//
//
// PlayerTH2 Version 0.1
// PlayerTH2 is used to analyse a 2D histogram by putting gates interactively.
//
// A PlayerTH2 object must be connected to a TCanvas and a TH2D histogram. It is done by using SetCanvas & SetMatrix.
// Once a canvas and a matrix has been linked, any key pressed on the defined canvas gives a gate (TBox) that can be
// modify graphically. The gates on the current canvas are use to  
//
//	Olivier Stezowski - 
//
// 
//

//#ifndef __CINT__

#include "RTypes.h"

#ifndef ROOT_TROOT
#include "TROOT.h"
#endif

/*
#ifndef ROOT_TROOT
#include <TFrame.h>
#endif*/

#ifndef ROOT_TApplication
#include "TApplication.h"
#endif

#ifndef ROOT_TBox
#include "TBox.h"
#endif

#ifndef ROOT_TString
#include "TString.h"
#endif

#ifndef ROOT_TH1
#include "TH1.h"
#endif

#ifndef ROOT_TH2
#include "TH2.h"
#endif

#ifndef ROOT_TStyle
#include "TStyle.h"
#endif

#ifndef ROOT_TSpectrum
#include "TSpectrum.h"
#endif

#ifndef ROOT_TFormula
#include "TFormula.h"
#endif

#ifndef ROOT_TCanvas
#include "TCanvas.h"
#endif

#ifndef ROOT_TFrame
#include "TFrame.h"
#endif

#include <iostream> 
using namespace std;

#include <RQ_OBJECT.h>

//#endif

/**
	PlayerTH2 is a tool to play with a TH2 histogram. It is mainly designed in the spirit of gamma-gamma matrices but could be used
	also for general purposes.

	


	@author Olivier Stezowski
 */
class PlayerTH2 
{
	RQ_OBJECT("PlayerTH2")

private:
	//! Box characteristics 
	TBox    fBoxX;	// Use to draw gates on the current pad
	TBox    fBoxY;	// Use to draw gates on the current pad
	TBox    fBoxBX;	// Use to draw gates on the current pad	
	TBox    fBoxBY;	// Use to draw gates on the current pad	
	
	//! Default with for gate X and Y
	pair<TFormula *, TFormula *> fWidthGate; 
   	
	//! Default stepping for gate X and Y in case of 
	pair<Float_t,Float_t> fStep; 	
   
	TList   fGatesX;	// to keep a current copy the gates list
	TList   fGatesY;	// to keep a current copy the gates list
	TList   fGatesBX;	// to keep a current copy the gates list
	TList   fGatesBY;	// to keep a current copy the gates list
	
   
private:
	static void SubstractBG_RAD(TH2 *);
	static void SubstractBG_SNIP(TH2 *);
   
   	void Clear();
   	void SetMatrix(TH2 *);	// Matrix to be analysed

private:
	TH2       *fMat;     	//! 2D matrix to work with
   
	TH1D      *fGx;		//! Current projection on X
	TH1D      *fGy;		//! Current protection on Y
	TH1D      *fBx;		//! Current background on X
	TH1D      *fBy;		//! Current background on Y
	TH1D      *fTotx;	//! Current total protection on X
	TH1D      *fToty;	//! Current total protection on Y
public:
	PlayerTH2();
	virtual ~PlayerTH2();
   
	//! 
	static void SubstractBG(const char *, Option_t *);
  
	void SetCanvas();		// set current canvas as the mother of the current pad
	void SetMatrix(const char*);	// Matrix to be analysed
   
	void Reset();         			// to reset the conditions of gating in the current pad
	void ShowGates(Option_t *o = "BGx");	// to show the gates used for the last gating  
	
	//! set a gate at position pos (on the horizontal axis) 
	/*
		Three options are available:
		- Gx: gate on the first axis of the matrix
		- Gy: gate on the second axis of the matrix
		- BX, BY:  background gate on the first and second axis of the matrix
	*/
	void SetGate(Double_t pos, Option_t *opt = "Gx");		// to set a gate a position x
//	void SetGate2D(Double_t, Double_t, Option_t *);	// to set a gate a position x
	
	//! get width for gate at position pos
	Double_t GetWidthGate(Double_t pos, Option_t *opt = "Gx");
	//! set default width for gates
	void SetWidthGate(Float_t wx = 2.5, Float_t wy = 2.5) { fWidthGate.first->SetParameter(0,wx); fWidthGate.second->SetParameter(0,wy); } 
	
	Int_t InitGating(const char *source = "gPad");
  
	//! to project
	void Projection(Option_t *option = "Gx");
     
  	// slots
   	void XEventAction(Int_t event, Int_t px, Int_t py, TObject *obj);
};


PlayerTH2::PlayerTH2()
{
// constructor   
	fMat = NULL; fGx = fGy = fTotx = fToty = fBx = fBy = NULL;
   
	fGatesX.SetOwner(kTRUE); fGatesY.SetOwner(kTRUE); fGatesBX.SetOwner(kTRUE); fGatesBY.SetOwner(kTRUE);
	
	gStyle->SetOptStat(0);
  
	fStep.first = fStep.second = 1.0; 
	
	fWidthGate.first  = new TFormula("PlayerTH2_WX","pol0"); fWidthGate.first->SetParameter(0,2.5);
	fWidthGate.second = new TFormula("PlayerTH2_WY","pol0"); fWidthGate.second->SetParameter(0,2.5); 
  
	fBoxX.SetFillColor(2);  fBoxX.SetFillStyle(3002);
	fBoxY.SetFillColor(4);  fBoxY.SetFillStyle(3013);
	fBoxBX.SetFillColor(16); fBoxBX.SetFillStyle(3016); fBoxBY.SetFillColor(16); fBoxBY.SetFillStyle(3020);
}

PlayerTH2::~PlayerTH2()
{
// destructor
	Clear();
	if ( fWidthGate.first ) delete fWidthGate.first; if (fWidthGate.second ) delete fWidthGate.second;
}   

void PlayerTH2::Clear()
{
	if ( fTotx ) delete fTotx; fTotx = NULL; if ( fToty ) delete fToty; fToty = NULL;
	
	if ( fGx ) delete fGx; fGx = NULL; if ( fGy ) delete fGy; fGy = NULL;
	if ( fBx ) delete fBx; fBx = NULL; if ( fBy ) delete fBy; fBy = NULL;
}

void PlayerTH2::SetCanvas()
{
// to connect the current canvas to the interactive gating 
	TVirtualPad *pad = TVirtualPad::Pad(); TCanvas *c = NULL;
	
	if ( pad ) 
		c = pad->GetCanvas();
	else {
		if ( (c = dynamic_cast<TCanvas *>(gROOT->GetListOfCanvases()->FindObject("Canvas_PlayerTH2"))) == NULL ) 
			c = new TCanvas("Canvas_PlayerTH2","Canvas to interact with PlayerTH2",1);
	}
				
  	c->Connect("ProcessedEvent(Int_t,Int_t,Int_t,TObject*)",
		   "PlayerTH2",this,"XEventAction(Int_t,Int_t,Int_t,TObject*)"); 	
	
	cout << " - Info - Interactive gating (by pressing any key) is now possible on any pad of the canvas " << c->GetName();
}

void PlayerTH2::SetMatrix(const char *matname)
{
	TObject *obj = gDirectory->Get(matname);
	if ( obj == NULL ) {
		cout << "Cannot find matrix named " << matname << " in the current TDirectory " << endl; 
	} else {	
		if ( obj->InheritsFrom("TH2") ) 
			SetMatrix(dynamic_cast<TH2 *>(obj));
		else 
			cout << "Object named " << matname << " is not a matrix !! " << endl;
	}
}

void PlayerTH2::SetMatrix(TH2 *mat)
{
	TString hname; TH1D *Gx, *Gy, *Bx, *By, *Totx, *Toty;
			
// nothing checked to avoid wrong pointers for mat .. except that this function is private
	if ( fMat == mat ) return;
	
	fMat = mat;
	Totx = Toty = Gx = Gy = Bx = By = NULL;
	
// it creates necessary histograms (or checks if they already exist) from mat on the same location (TDirectory) than the matrix
	// histograms to store total projection
	if ( mat->GetDirectory() ) { // mat belongs to a TDirectory .. so check if the histograms have already been created
		hname = mat->GetName(); hname += "_tx" ;
		Totx  = dynamic_cast<TH1D *>( mat->GetDirectory()->Get(hname.Data()) );
		hname = mat->GetName(); hname += "_ty" ;
		Toty  = dynamic_cast<TH1D *>( mat->GetDirectory()->Get(hname.Data()) );
		hname = mat->GetName(); hname += "_gx" ;
		Gx    = dynamic_cast<TH1D *>( mat->GetDirectory()->Get(hname.Data()) );
		hname = mat->GetName(); hname += "_gy" ;
		Gy    = dynamic_cast<TH1D *>( mat->GetDirectory()->Get(hname.Data()) );
		hname = mat->GetName(); hname += "_bx" ;		
		Bx    = dynamic_cast<TH1D *>( mat->GetDirectory()->Get(hname.Data()) );
		hname = mat->GetName(); hname += "_by" ;
		By    = dynamic_cast<TH1D *>( mat->GetDirectory()->Get(hname.Data()) );		
	}
	// histograms to store full projection
	if ( Totx == NULL ) { hname = mat->GetName(); hname += "_tx"; Totx = mat->ProjectionX(hname.Data()); }
	if ( Toty == NULL ) { hname = mat->GetName(); hname += "_ty"; Toty = mat->ProjectionY(hname.Data()); }
		
	// histograms to store gated spectra
	if ( Gx == NULL ) { hname = mat->GetName(); hname += "_gx"; Gx = (TH1D *)Totx->Clone(hname.Data()); }
	if ( Gy == NULL ) { hname = mat->GetName(); hname += "_gy"; Gy = (TH1D *)Toty->Clone(hname.Data()); }
	
	// histograms to store background spectra
	if ( Bx == NULL ) hname = mat->GetName(); hname += "_bx"; Bx = (TH1D *)Totx->Clone(hname.Data()); Bx->Reset("ICE"); 
	if ( By == NULL ) hname = mat->GetName(); hname += "_by"; By = (TH1D *)Toty->Clone(hname.Data()); By->Reset("ICE");	
	
// now keep a local copy protected from delete
	// remove previous projection
	Clear();
	
	// clone local histograms and set directory to NULL	
	fGx = dynamic_cast<TH1D *>( Gx->Clone("PlayerTH2_gx") ); fGx->SetDirectory(0); 
	fGy = dynamic_cast<TH1D *>( Gy->Clone("PlayerTH2_gy") ); fGy->SetDirectory(0);
	fBx = dynamic_cast<TH1D *>( Bx->Clone("PlayerTH2_bx") ); fBx->SetDirectory(0); 
	fBy = dynamic_cast<TH1D *>( By->Clone("PlayerTH2_by") ); fBy->SetDirectory(0);
	
	fTotx = dynamic_cast<TH1D *>( Totx->Clone("PlayerTH2_tx") ); fTotx->SetDirectory(0); 
	fToty = dynamic_cast<TH1D *>( Toty->Clone("PlayerTH2_ty") ); fToty->SetDirectory(0);
	
	Totx->Draw();
}

Double_t PlayerTH2::GetWidthGate(Double_t pos, Option_t *o)
{
	TString opt = o;
	if ( opt.Contains("Gx") ) return fWidthGate.first->Eval(pos);  
	else return fWidthGate.second->Eval(pos); 
}

// TODO
void PlayerTH2::SetGate(Double_t pos, Option_t *o)
{
// add a gate to the current pad 
	TH1 *hist; TObject *obj; TFrame *frame; 

	// check for a valid TPad
	TVirtualPad *pad = TVirtualPad::Pad(); if ( pad == NULL ) return;

	TIter next(pad->GetListOfPrimitives()); // look for an histogram
	frame = pad->GetFrame(); 
	hist = NULL;
	while ( (obj = next()) ) { 
		if ( obj->InheritsFrom("TH1") ) { hist = dynamic_cast<TH1 *>(obj); break; }
	}
	if ( hist == NULL || frame == NULL ) { cout << "Cannot add a gate, no spectrum selected " << endl; return; } 
	
	Double_t w, x1, x2, y1, y2; TString opt = o;  
	
	TBox *box = &fBoxX; w = GetWidthGate(pos,"Gx");
	if ( opt.Contains("Gy") ) { box = &fBoxY; w = GetWidthGate(pos,"Gy"); }
	if ( opt.Contains("Bx")  ) { box = &fBoxBX; w = GetWidthGate(pos,"Gx"); }
	
	// Add one gates
	x1 = pos-w/2.0; y1 = frame->GetY1(); x2 = pos+w/2.0; y2 = frame->GetY2(); box->DrawBox(x1,y1,x2,y2); pad->Update();
	
	cout << " Add gate at position " << pos << " <" << pos-w/2.0 << "," << pos+w/2.0 << ">"<< endl;
}

// TODO
Int_t PlayerTH2::InitGating()
{
// Init gating - It looks in the current pad for gates (boxes) with the good characteristics (color and fill style).
// It returns the number of gates found

	if ( TVirtualPad::Pad() == NULL ) return 0;

	TBox *box; 
	fGatesX.Delete(); fGatesY.Delete(); fGatesBX.Delete(); fGatesBY.Delete();

	TIter next(TVirtualPad::Pad()->GetListOfPrimitives());
	while ( (box = (TBox *)next()) ) { // look for a box and keep a copy
		if ( box->InheritsFrom("TBox") ) {
			if ( box->GetFillStyle() == fBoxX.GetFillStyle()  && box->GetFillColor() == fBoxX.GetFillColor() ) 
				fGatesX.Add(box->Clone());
			if ( box->GetFillStyle() == fBoxY.GetFillStyle()  && box->GetFillColor() == fBoxY.GetFillColor() ) 
				fGatesY.Add(box->Clone());
			if ( box->GetFillStyle() == fBoxBX.GetFillStyle() && box->GetFillColor() == fBoxBX.GetFillColor() ) 
				fGatesBX.Add(box->Clone());
			if ( box->GetFillStyle() == fBoxBY.GetFillStyle() && box->GetFillColor() == fBoxBY.GetFillColor() ) 
				fGatesBY.Add(box->Clone());	
		} // if
	} // while   
 
	cout << " Nb Gates found in the current pad : " << fGatesX.GetSize() << endl; return fGatesX.GetSize(); 
}

// TODO
void PlayerTH2::ShowGates(Option_t *o)
{
//
// look for an histogram in the current pad. If one is found, it displays the boxes saved in fGatesX
//
	TString opt = o;
	TObject *obj; TH1 *hist; TBox *box;

	if ( TVirtualPad::Pad() == NULL ) return;

	// look for an histogram
	hist = NULL;
	TIter next(TVirtualPad::Pad()->GetListOfPrimitives());
	while ( (obj = next()) ) { 
		if ( obj->InheritsFrom("TH1") ) { hist = dynamic_cast<TH1 *>(obj); break; }
	}
	if ( hist == NULL ) return;

	// draw gates
		
	TList *list = &fGatesX; 
	if ( opt.Contains("Gy") )  { list = &fGatesY; }
	if ( opt.Contains("Bx")  ) { list = &fGatesBX; }

	TIter next2(list);
		while ( (box = (TBox *)next2()) ) { // add all lines to the current pad
		box->DrawBox(box->GetX1(),0,box->GetX2(),hist->GetMaximum() / 2);
	}
}

void PlayerTH2::Reset()
{
// to reset the conditions of gating
	TObject *obj; TList toremove;

	if ( TVirtualPad::Pad() == NULL ) return;

	TIter next(TVirtualPad::Pad()->GetListOfPrimitives());
	while ( (obj = next()) ) { // look for a box and keep a reference
		if ( obj->InheritsFrom("TBox") ) { toremove.Add(obj); }
	}
	TIter remove(&toremove);
	while ( (obj = remove()) ) { // remove lines
			TVirtualPad::Pad()->GetListOfPrimitives()->Remove(obj);
	}
	toremove.Delete();

	TVirtualPad::Pad()->Update();
}

// TODO
void PlayerTH2::Projection(Option_t *option)
{
// Gating on a gamma-gamma matrix
	TString opt = option;
  
	if ( fMat == NULL ) {
		cout << "Cannot Project in PlayerTH2::Projection, no matrix selected ! " << endl; return;
	}
	 
// init gating from boxes drawn in the current pad
	if ( InitGating() == 0 ) return;
  
	TH1D *h, *hb, *htmp; TSpectrum worker; Float_t *bgi;
	
	if ( opt.Contains("Gy") ) { h = fGy; hb = fBy; }
	else { h = fBx; hb = fBy;}
  
	h->Reset("ICE"); hb->Reset("ICE");
  
	TBox *box; TIter next(&fGatesX);
	while ( (box = (TBox *)next()) ) { // add all lines to the current pad
   
		if (box->GetFillColor() ==  fBoxX. GetFillColor() )  
			h->Add( htmp = fMat->ProjectionX("PlayerTH2_tmp_gx",h->FindBin(box->GetX1()),h->FindBin(box->GetX2()) ) );
		else 
 			h->Add( htmp = fMat->ProjectionY("PlayerTH2_tmp_gx",h->FindBin(box->GetX1()),h->FindBin(box->GetX2()) ) );  
	 
		delete htmp; 
	}  
  
	if (opt.Contains("DoBg")) {
		bgi = new Float_t[fMat->GetNbinsX()];
		for ( Int_t i = 0; i < fMat->GetNbinsX(); i++ ) { bgi[i] = h->GetBinContent(i+1); }
		worker.Background1(bgi,fMat->GetNbinsX(),40);
		for ( Int_t i = 0; i < fMat->GetNbinsX(); i++ ) { h->SetBinContent(i+1,h->GetBinContent(i+1)-bgi[i]); } 
	}
	h->Draw();
  
	if ( opt.Contains("Gy") ) cout << " Projection on Y DONE \n ";
	else cout << " Projection on X DONE \n ";
}


void PlayerTH2::XEventAction(Int_t event, Int_t px, Int_t py, TObject *sel)
{
// Do action depending on X actions
	static Int_t fLastX = 0, fLastY = 0;
	
	TVirtualPad *pad = gROOT->GetSelectedPad(); if ( pad == NULL ) return;

	switch (event) {
		case kMouseMotion:
			fLastX = px; fLastY = py;
 			break;

		case kKeyPress:
		{
			Int_t key = px, keysym = py;
			switch ((EKeySym)keysym) {
				case kKey_Left:
					cout<<"left arrow"<<endl;
					break;
				case kKey_Right:
					cout<<"rightarrow"<<endl;
					break;
				case kKey_Down:
					cout<<"down arrow"<<endl;
					break;
				case kKey_Up:
					cout<<"up arrow"<<endl;
					break;
				case kKey_x: 
					//cout << " echo " << hex << px << " " << py << " " << kKey_Control << endl;
					if ( px != py  ) 
						{ /*cout << " BINGO " << endl;*/ }
					else 
						{ SetGate(pad->PadtoX(pad->AbsPixeltoX(fLastX)),"Gx");}
					break;
				case kKey_X: 
					cout << " ECHO " << hex << px << " " << py << " " << kKey_Control << endl;
					if ( px != py  ) 
						{ /*cout << " BINGO " << endl;*/ }
					else 
						{ SetGate(pad->PadtoX(pad->AbsPixeltoX(fLastX)),"Gx");}
					break;					
				case kKey_y:
					SetGate(pad->PadtoX(pad->AbsPixeltoX(fLastX)),"Gy");
					break;
				case kKey_b:
					SetGate(pad->PadtoX(pad->AbsPixeltoX(fLastX)),"Bx"); 
					break;
				case kKey_Y:
					cout<<"gate Y"<<endl;
					break;
				case kKey_B:
					cout<<"gate B"<<endl;
					break;					
			}			
		}
		break;
		
		default :
		break;
	} // switch
	// cout << " Px " << fLastX << " Py " << fLastY << endl;
}

void PlayerTH2::SubstractBG(const char *matname, Option_t *option)
{
	TString opt = option;
	
	// Test Option 
	if ( (!opt.Contains("RAD")) && (!opt.Contains("SNIP")) ) {
		cout << "Bad options - Choose 'RAD' or 'SNIP' " << endl; return;
	}
	
	TObject *obj = gROOT->FindObject(matname);
	if ( obj == NULL ) {
		cout << "Cannot find matrix named " << matname << endl;
	} else {	
		if ( obj->InheritsFrom("TH2") ) {
			if ( opt.Contains("RAD") ) SubstractBG_RAD((TH2 *)obj);
			if ( opt.Contains("SNIP")) SubstractBG_SNIP((TH2 *)obj);
		}
		else cout << "Object named "<< matname << " is not a matrix !! ";
	}
}

void PlayerTH2::SubstractBG_RAD(TH2 *mat)
{
	Int_t i, j; TSpectrum worker; // use to substract a background on the x and y projection
	
   	TH1D *proji, *projj;   
  	Float_t *ptoti, *ptotj, *bgi, *bgj; Stat_t T;
	
// to get the projection on X and Y
	proji = mat->ProjectionX("tmp_rad_x"); projj = mat->ProjectionY("tmp_rad_y");
	
//Smooth projection
	proji->Smooth(2); projj->Smooth(2); 
	
// to normalize with the total number of counts in the matrix.
	T = mat->Integral();
   	cout << " Number of counts in the gamma-gamma matrix: " << T << endl;
	
	ptoti = new Float_t[mat->GetNbinsX()]; bgi = new Float_t[mat->GetNbinsX()];
	ptotj = new Float_t[mat->GetNbinsY()]; bgj = new Float_t[mat->GetNbinsY()];
	for ( i = 0; i < mat->GetNbinsX(); i++ ) { ptoti[i] = bgi[i] = proji->GetBinContent(i+1); } 
	for ( j = 0; j < mat->GetNbinsY(); j++ ) { ptotj[j] = bgj[j] = projj->GetBinContent(j+1); }
	
	delete proji; delete projj;
		
   	worker.Background1(bgi,mat->GetNbinsX(),40);worker.Background1(bgj,mat->GetNbinsY(),40);  	
	//worker.Background1General(bgi,mat->GetNbinsX(),20,BACK1_INCREASING_WINDOW,BACK1_ORDER2,BACK1_INCLUDE_COMPTON);
	//worker.Background1General(bgj,mat->GetNbinsY(),20,BACK1_INCREASING_WINDOW,BACK1_ORDER2,BACK1_INCLUDE_COMPTON);
	
// now do the background substraction
	cout << " Background substraction on " << mat->GetName() << " by using the RADFORD procedure ..." << endl;
	
	Stat_t c, cmin = 0;
   	for ( i = 0; i < mat->GetNbinsX(); i++ ) {
		for ( j = 0; j < mat->GetNbinsY(); j++ ) {
		
			cout << " [" << i << "," << j << "] \r" ;
			// printf(" [%d,%d] \r",i,j); 
			
			c = mat->GetBinContent(i+1,j+1);
			c = c  - (ptoti[i]*ptotj[j] - (ptoti[i]-bgi[i])*(ptotj[j]-bgj[j]))/T;
			if ( c < 0 && c < cmin ) { cmin = c; }
							  
			mat->SetBinContent(i+1,j+1,c);
		} // j
   	} // i	
	cout << " ... DONE. Lowest value in the background substracted matrix " << cmin << endl;
	
	delete [] ptoti; delete [] ptotj; delete [] bgi; delete [] bgj;
}

void PlayerTH2::SubstractBG_SNIP(TH2 *mat)
{
// 2D Background substraction
// Based on the implementation given by M.Morhac NIM A,401 (1997) 113-132
// Based on SNIP procedure (Sensitive Nonlinear Iterative Peak)
	
	Int_t param = 3;
	
	cout << "*********************************************\n";
	cout << "* Background Substraction with SNIP Procedure\n";
	cout << "* Number of iterations : " << param << endl;
	cout << "*********************************************\n";
	cout << "* TH2 name : " << mat->GetName() << endl; 
	
	Int_t dimX, dimY;
	Float_t **v, **w, **y;
	
	Double_t P1, P2, P3, P4, S1, S2, S3, S4, a1, a2;
	
	
	dimX = mat->GetNbinsX();  dimY = mat->GetNbinsY();
	if (dimX <= 0) {cout << "Error, dimX <= 0\n"; return;} 
	if (dimY <= 0) {cout << "Error, dimY <= 0\n"; return;} 	
	 
	v = new Float_t*[dimX]; w = new Float_t*[dimX], y = new Float_t*[dimX];
	for (Int_t i=0 ; i<dimX ; i++){
		v[i] = new Float_t[dimY];
		w[i] = new Float_t[dimY];
		y[i] = new Float_t[dimY];
	}
	
	// Fill v
	for (Int_t i=0 ; i<dimX ; i++){
		for (Int_t j=0 ; j<dimY ; j++){
			y[i][j] = mat->GetBinContent(i+1,j+1);
			v[i][j] = mat->GetBinContent(i+1,j+1);
		}
	}
	
	
	// LLS operator
	for (Int_t i=0 ; i<dimX ; i++){
		for (Int_t j=0 ; j<dimY ; j++){
			v[i][j] = TMath::Log( TMath::Log( TMath::Sqrt(y[i][j] + 1) + 1) + 1);
			
		}
	}
	
	//SNIP Algorithm
	for (Int_t p=1 ; p<=param ; p++){
		for (Int_t i=p ; i<dimX-p ; i++){
			for (Int_t j=p ; j<dimY-p ; j++){
				P1 = v[i+p][j+p];
				P2 = v[i-p][j+p];
				P3 = v[i+p][j-p];
				P4 = v[i-p][j-p];
				
				S1 = v[i+p][j];
				S2 = v[i][j+p];
				S3 = v[i][j-p];
				S4 = v[i-p][j-p];
				
				S1 = TMath::Max(S1,(P1+P3)/2) - (P1+P3)/2;
				S2 = TMath::Max(S2,(P1+P2)/2) - (P1+P2)/2;
				S3 = TMath::Max(S3,(P3+P4)/2) - (P3+P4)/2;
				S4 = TMath::Max(S4,(P2+P4)/2) - (P2+P4)/2;
				
				a1 = v[i][j];
				a2 = (S1+S4)/2 + (S2+S3)/2 + (P1+P2+P3+P4)/4;
				w[i][j] = TMath::Min(a1,a2);
			} //j
		} //i
		for (Int_t i=p ; i<dimX-p ; i++){
			for (Int_t j=p ; j<dimY-p ; j++){
				v[i][j] = w[i][j];	
			}
		}
		// printf ("[ %i ]\n",p);
		
		//TEST
		/*for (Int_t i=0 ; i<dimX ; i++){
			for (Int_t j=0 ; j<dimY ; j++){
				mat->SetBinContent(i+1,j+1,y[i][j]-v[i][j]);
			} 
		}*/
		//ENDOFTEST
		
		//if (p==1) mat->ProjectionX("p=1")->Draw();
		//else mat->ProjectionX("p=1")->DrawClone("SAME");
		
		
	} //p
	
	
	// Inverse LLS operator
	for (Int_t i=0 ; i<dimX ; i++){
		for (Int_t j=0 ; j<dimY ; j++){
			v[i][j] = TMath::Exp( TMath::Exp( v[i][j] - 1) -1 ) * TMath::Exp( TMath::Exp( v[i][j]-1) - 1) -1;
			
		}
	}
	
	//fill mat
	for (Int_t i=0 ; i<dimX ; i++){
		for (Int_t j=0 ; j<dimY ; j++){
			mat->SetBinContent(i+1,j+1,y[i][j]-v[i][j]);
		} 
	}  
	
	//Delete useless pointers
	for (Int_t i=0 ; i<dimX ; i++){
		delete [] v[i];
		delete [] w[i];
		delete [] y[i];
	}
	delete [] v;
	delete [] w;
	delete [] y;
	
}
