/** \file FirstGEM.C to illustrate how to use GEM.

	the two libraries must be loaded first with  \n
	<em>gSystem->Load("libGWPHYSICS.so"); gSystem->Load("libGWGEM.so");</em>
	
	Another possibility is to use a rootlogon.C file to do the job. The are examples of such 
	a file in demos directory.
	
	The GEM class adds some functionnalities to BaseGEM by adding a feeding mechanism.
	It generates gamma-rays from a continuum. The system is designed to allow a parametrisation of the feeding. 
	
	It also shows how to 'degrade' the real cascade with a filter to generate gamma-rays 
	as expected after the tracking and to store them in a ROOT TTree.
	
 \author Olivier Stezowski
*/


#ifndef __CINT__
#include <Link.h>
#include <XGammaLink.h>
#include <GEM.h>
#include <GammaFilter.h>

#include <TStopwatch.h>
#include <TH1.h>
#include <TTree.h>

#include <iostream>
using namespace std;
#endif

using namespace Gw;

#define MAX_STRING_LENGTH 50

/** Function to generate discret gamma-rays between several level schemes

  Usage: 
	- root
	- root [0] .L FirstGEM.C
	- root [1] testBaseGEM("/path/to/levelscheme/files")


\author Olivier Stezowski - 01/2005

PS: libGWPHYSICS and libGWGEM must be loaded before running this macro (see rootlogon.C)
*/
void testBaseGEM(const char *pathtodata, Int_t nbcas = 100)
{
	TStopwatch watch; 
	
	watch.Start();
	
	// Experimental conditions (from PACE like)
	const Int_t nb_residu = 6;	// 151DY, 152DY, 153DY, 151TB, 149GD, 148GD
	
	BaseGEM nucleus[nb_residu];     // gem related
	
	TString filename[nb_residu];      // database related	
	TString nucleusname[nb_residu];   // database related	
	
	Float_t sigma[nb_residu]    = { 29.7, 47.1, 19.3, 19.3, 0.4, 1.31 };

	filename[0] = pathtodata; filename[0] += "DY_ENSDF.ens"; nucleusname[0] = "151DY";
	filename[1] = pathtodata; filename[1] += "152DY.ags";    nucleusname[1] = "152DY";
	filename[2] = pathtodata; filename[2] += "DY_ENSDF.ens"; nucleusname[2] = "153DY";	
	filename[3] = pathtodata; filename[3] += "TB_ENSDF.ens"; nucleusname[3] = "151TB";
	filename[4] = pathtodata; filename[4] += "GD_ENSDF.ens"; nucleusname[4] = "149GD";
	filename[5] = pathtodata; filename[5] += "148GD.ags";    nucleusname[5] = "148GD";	
		
	// same prototype function to init all nucleus
	Int_t ok = 0;
	for (int i = 0; i < nb_residu; i++ ) 
		{ ok += nucleus[i].Import(filename[i],nucleusname[i]); }

	if ( ok != nb_residu ) {
		cout << " A Level Scheme is missing !! " << endl; return;
	}
	// create the RandObj to select the nb_residu nuclei.
	RandObj channel;
	for (int i = 0; i < nb_residu; i++ ) channel.Add(nucleus+i,sigma[i]); 	

	// some spectra to be filled
	TH1F *K = new TH1F("BaseGEM_K","K",100,0,100);
	TH1F *projtot = new TH1F("BaseGEM","BaseGEM",8192,0,4096); 

	watch.Stop(); 
	cout << "Initialisation time " << endl; watch.Print();
	// start the Monte-Carlo
	
	watch.Reset(); watch.Start();	
	Cascade cas; BaseGEM *gem; GammaLink *gam; Int_t which;
	for (Int_t i = 0; i < nbcas; i++ ) {

		gem = (BaseGEM *)channel.Rand(which);	// select one nucleus
		gem->DoCascade(cas);	   		// determine a cascade on this nucleus

		K->Fill(cas.GetSize()) ;
		for (Int_t j = 0; j < cas.GetSize(); j++ ) {
		  gam = (GammaLink *)cas.At(j); projtot->Fill(gam->GetEnergy().Get());
		} // j		
	} //i
	watch.Stop(); 	
	cout << "Monte-Carlo time " << endl; watch.Print();
	
	projtot->Draw(); 
}

/** Function to generate discret gamma-rays between several level schemes ...

  it also adds to all the cascades some gamma-rays coming from a continuous feeding
   
  Usage: 
	- root
	- root [0] .L FirstGEM.C
	- root [1] testGEM("/path/to/levelscheme/files")


\author Olivier Stezowski - 01/2005

PS: libGWPHYSICS and libGWGEM must be loaded before running this macro (see rootlogon.C)
*/
void testGEM(const char *pathtodata, Int_t nbcas = 100)
{
	TStopwatch watch; 
	
	watch.Start();
	
	// Experimental conditions (from PACE like)
	const Int_t nb_residu = 6;	  // 151DY, 152DY, 153DY, 151TB, 149GD, 148GD
	
	GEM nucleus[nb_residu];           // gem related
	
	TString filename[nb_residu];      // database related	
	TString nucleusname[nb_residu];   // database related	
	
	Float_t sigma[nb_residu]    = { 29.7, 47.1, 19.3, 19.3, 0.4, 1.31 };

	filename[0] = pathtodata; filename[0] += "DY_ENSDF.ens"; nucleusname[0] = "151DY";
	filename[1] = pathtodata; filename[1] += "152DY.ags";    nucleusname[1] = "152DY";
	filename[2] = pathtodata; filename[2] += "DY_ENSDF.ens"; nucleusname[2] = "153DY";
	filename[3] = pathtodata; filename[3] += "TB_ENSDF.ens"; nucleusname[3] = "151TB";
	filename[4] = pathtodata; filename[4] += "GD_ENSDF.ens"; nucleusname[4] = "149GD";
	filename[5] = pathtodata; filename[5] += "148GD.ags";    nucleusname[5] = "148GD";
				
	// same prototype function to init all nucleus
	Int_t ok = 0;
	for (int i = 0; i < nb_residu; i++ ) { 
		ok += nucleus[i].Import(filename[i],nucleusname[i]); 
		// entry point 80 MeV, spin 60		
		nucleus[i].SetFeeding("TestFeeding"); nucleus[i].SetExI(80000,60); 
	}

	if ( ok != nb_residu ) {
		cout << " A Level Scheme is missing !! " << endl; return;
	}
	// create the RandObj to select the nb_residu nuclei.
	RandObj channel;
	for (int i = 0; i < nb_residu; i++ ) channel.Add(nucleus+i,sigma[i]); 	

	// some spectra to be filled
	TH1F *K = new TH1F("GEM_K","K",100,0,100);
	TH1F *projtot = new TH1F("GEM","GEM",8192,0,4096); 

	watch.Stop(); 
	cout << "Initialisation time " << endl; watch.Print();
	// start the Monte-Carlo
	
	watch.Reset(); watch.Start();	
	Cascade cas; BaseGEM *gem; GammaLink *gam; Int_t which;
	for (Int_t i = 0; i < nbcas; i++ ) {

		gem = (BaseGEM *)channel.Rand(which);	// select one nucleus
		gem->DoCascade(cas);	   		// determine a cascade on this nucleus

		K->Fill(cas.GetSize()) ;
		for (Int_t j = 0; j < cas.GetSize(); j++ ) {
		  gam = (GammaLink *)cas.At(j); projtot->Fill(gam->GetEnergy().Get());
		} // j		
	} //i
	watch.Stop(); 	
	cout << "Monte-Carlo time " << endl; watch.Print();
	
	projtot->Draw(); 
}

/** Function to generate tracked like events

  With the previous functions, it is possible to generate cascades of real gammas. Here an filter
  is apply to simulate the 'real' events as they could be obtained after the tracking.
  
  To realize this, some histograms are needed; they are stored in root files. They are available
  in the demos/gem directory and correpond to AGATA and EUROBALL like filter.
   
  Usage: 
	- root
	- root [0] .L FirstGEM.C
	- root [1] testGammaFilter1("/path/to/levelscheme/files")


\author Olivier Stezowski - 01/2005

PS: libGWPHYSICS and libGWGEM must be loaded before running this macro (see rootlogon.C)
*/
void testGammaFilter1(const char *pathtodata, Int_t nbcas = 100, const char *filtername = "AGATA_FILTER.root")
{
	TStopwatch watch; 
	
	watch.Start();
	
	// Experimental conditions (from PACE like)
	const Int_t nb_residu = 6;	// 151DY, 152DY, 153DY, 151TB, 149GD, 148GD
	
	BaseGEM nucleus[nb_residu];     // gem related
	
	TString filename[nb_residu];      // database related	
	TString nucleusname[nb_residu];   // database related	
	
	Float_t sigma[nb_residu]    = { 29.7, 47.1, 19.3, 19.3, 0.4, 1.31 };
	
	filename[0] = pathtodata; filename[0] += "DY_ENSDF.ens"; nucleusname[0] = "151DY";
	filename[1] = pathtodata; filename[1] += "152DY.ags";    nucleusname[1] = "152DY";
	filename[2] = pathtodata; filename[2] += "DY_ENSDF.ens"; nucleusname[2] = "153DY";
	filename[3] = pathtodata; filename[3] += "TB_ENSDF.ens"; nucleusname[3] = "151TB";
	filename[4] = pathtodata; filename[4] += "GD_ENSDF.ens"; nucleusname[4] = "149GD";
	filename[5] = pathtodata; filename[5] += "148GD.ags";    nucleusname[5] = "148GD";
						
	// same prototype function to init all nucleus
	Int_t ok = 0;
	for (int i = 0; i < nb_residu; i++ ) 
		{ ok += nucleus[i].Import(filename[i],nucleusname[i]); }

	if ( ok != nb_residu ) {
		cout << " A Level Scheme is missing !! " << endl; return;
	}
	// create the RandObj to select the nb_residu nuclei.
	RandObj channel;
	for (int i = 0; i < nb_residu; i++ ) channel.Add(nucleus+i,sigma[i]); 	

	// initiate the GammaFilter
	Float_t e[1000]; // up to 1000 energies .. should be ok for the moment
	GammaFilter filter; filter.InitFilter(filtername);
	if ( filter.IsEffective() == false ) return ;
	
	// some spectra to be filled
	TH1F *K = new TH1F("GammaFilter1_K","K",100,0,100);
	TH1F *projtot = new TH1F("GammaFilter1","GammaFilter1",8192,0,4096); 

	watch.Stop(); 
	cout << "Initialisation time " << endl; watch.Print();
	// start the Monte-Carlo
	
	watch.Reset(); watch.Start();	
	Cascade cas; BaseGEM *gem; GammaLink *gam; Int_t which;
	for (Int_t i = 0; i < nbcas; i++ ) {

		gem = (BaseGEM *)channel.Rand(which);	// select one nucleus
		gem->DoCascade(cas);	   		// determine a cascade on this nucleus

		// apply experimental filter on this cascade.
		Int_t nbgamma = filter.ApplyE(cas,e);
		
		K->Fill(nbgamma) ;
		for (Int_t j = 0; j < nbgamma; j++ ) {
		  projtot->Fill(e[j]);
		} // j		
	} //i
	watch.Stop(); 	
	cout << "Monte-Carlo time " << endl; watch.Print();
	
	projtot->Draw(); 
}

/** Function to generate tracked like events and store them in a ROOT TTree

  With the previous functions, it is possible to generate cascades of real gammas. Here an filter
  is apply to simulate the 'real' events as they could be obtained after the tracking.
  
  To realize this, some histograms are needed; they are stored in root files. They are available
  in the demos/gem directory and correpond to AGATA and EUROBALL like filter. The generated events
  are stored in a ROOT TTree.
   
  Usage: 
	- root
	- root [0] .L FirstGEM.C
	- root [1] testGammaFilter2("/path/to/levelscheme/files")


\author Olivier Stezowski - 01/2005

PS: libGWPHYSICS and libGWGEM must be loaded before running this macro (see rootlogon.C)
*/
void testGammaFilter2(const char *pathtodata, Int_t nbcas = 100, const char *filtername = "AGATA_FILTER.root")
{
	TStopwatch watch; 
	
	watch.Start();
	
	// Experimental conditions (from PACE like)
	const Int_t nb_residu = 6;	// 151DY, 152DY, 153DY, 151TB, 149GD, 148GD
	
	GEM nucleus[nb_residu];     // gem related
	
	TString filename[nb_residu];      // database related	
	TString nucleusname[nb_residu];   // database related	
	
	Float_t sigma[nb_residu]    = { 29.7, 47.1, 19.3, 19.3, 0.4, 1.31 };
	
	filename[0] = pathtodata; filename[0] += "DY_ENSDF.ens"; nucleusname[0] = "151DY";
	filename[1] = pathtodata; filename[1] += "152DY.ags";    nucleusname[1] = "152DY";
	filename[2] = pathtodata; filename[2] += "DY_ENSDF.ens"; nucleusname[2] = "153DY";
	filename[3] = pathtodata; filename[3] += "TB_ENSDF.ens"; nucleusname[3] = "151TB";
	filename[4] = pathtodata; filename[4] += "GD_ENSDF.ens"; nucleusname[4] = "149GD";
	filename[5] = pathtodata; filename[5] += "148GD.ags";    nucleusname[5] = "148GD";
						
	// same prototype function to init all nucleus
	Int_t ok = 0;
	for (int i = 0; i < nb_residu; i++ ) { 
		ok += nucleus[i].Import(filename[i],nucleusname[i]); 
		// entry point 80 MeV, spin 60		
		nucleus[i].SetFeeding("TestFeeding"); nucleus[i].SetExI(80000,60); 
	}
	
	if ( ok != nb_residu ) {
		cout << " A Level Scheme is missing !! " << endl; return;
	}
	// create the RandObj to select the nb_residu nuclei.
	RandObj channel;
	for (int i = 0; i < nb_residu; i++ ) channel.Add(nucleus+i,sigma[i]);

	// initiate the GammaFilter
	Int_t nbgamma; Float_t e[1000]; // up to 1000 energies ..should be ok for the moment
	GammaFilter filter; filter.InitFilter(filtername);
	if ( filter.IsEffective() == false ) return ;
		
	// some spectra to be filled
	TH1F *K = new TH1F("GammaFilter2_K","K",100,0,100);
	TH1F *projtot = new TH1F("GammaFilter2","GammaFilter2",8192,0,4096); 
	
	// open a root file to store the produced events in a ROOT TTree
	TFile fileout("SimuDY.root","recreate");
	
	TTree *treeout; 
	treeout = (TTree *)fileout.Get("SimuDY");
	if ( treeout == NULL ) {
		treeout = new TTree("SimuDY","FromGammaFilter2");
		treeout->Branch("mult",&nbgamma,"nbgamma/I"); treeout->Branch("e",e,"e[nbgamma]/F");
	} 
//	else {
//   		treeout->SetBranchAddress("nbgamma",&nbgamma);
//	
//	}

	watch.Stop(); 
	cout << "Initialisation time " << endl; watch.Print();
	// start the Monte-Carlo
	
	watch.Reset(); watch.Start();	
	Cascade cas; BaseGEM *gem; GammaLink *gam; Int_t which;
	for (Int_t i = 0; i < nbcas; i++ ) {

		gem = (BaseGEM *)channel.Rand(which);	// select one nucleus
		gem->DoCascade(cas);	   		// determine a cascade on this nucleus

		// apply experimental filter on this cascade.
		nbgamma = filter.ApplyE(cas,e);
		
		K->Fill(nbgamma) ;
		for (Int_t j = 0; j < nbgamma; j++ ) {
		  projtot->Fill(e[j]);
		} // j	
		
		treeout->Fill();	
	} //i
	
	treeout->Write();
	
	watch.Stop(); 	
	cout << "Monte-Carlo time " << endl; watch.Print();
	
	projtot->Draw(); 
}


