#ifndef __CINT__

#ifndef ROOT_TObjArray
#include "TObjArray.h"
#endif
#ifndef ROOT_TH1
#include "TH1.h"
#endif
#ifndef ROOT_TH2
#include "TH2.h"
#endif
#ifndef ROOT_TMultiGraph
#include "TMultiGraph.h"
#endif
#ifndef ROOT_TGraphErrors
#include "TGraphErrors.h"
#endif
#ifndef ROOT_TLegend
#include "TLegend.h"
#endif
#ifndef ROOT_TLeaf
#include "TLeaf.h"
#endif
#ifndef ROOT_TLine
#include "TLine.h"
#endif
#ifndef ROOT_TPad
#include "TPad.h"
#endif
#ifndef ROOT_TMarker
#include "TMarker.h"
#endif
#ifndef GW_EVAPLMOF_H
#include "EvapLMOF.h"
#endif

#endif

#include "EvapLMOF.h"

const Int_t MAXZ = 120; const Int_t MAXN = 200; // upper limit for the number of protons and neutrons for one nucleus

using namespace Gw;

/** Function to scan a pax file and display the content on the standard ouput

	It displays on the standard output all the cascacades between event # 0 and event # 100. Also 
	illustrated the idonly flag. If activated (with pax << idonly), only the header part of the cascacades is read. 
	
	fname is the file name of the pax file to be scanned
*/
void Scan(const char *fname)
{
	EvapLMOF pax; 
	pax.AddFile(fname); pax.VERBOSE = 1; pax << idonly; pax.Scan(0,100); pax.ShowCurrentConditions();
}

/** Function to plot a 2D plot of all the produced residus 
 	
	It displays on a 2D plot (Z,N) the cross section of the different residus produced and the total cross section for
	the compound nucleus. In case fis is set to true, it gives the fissioning nuclei.
	
	- hname is the name of the histogram
	- fname is the base file name of the pax files to be scanned
	- pathtofile the directory where are stored the pax files to be chained
*/
TH2D *Production2D(const char *hname, const char *fname, const char *pathtofile = "./",bool fis = false)
{
	// allocate 2D plot 
	TH2D *h = new TH2D(hname,"Cross Section",MAXN,0,MAXN,MAXZ,0,MAXZ);
	
	// open chain of files
	EvapLMOF pax; pax.AddFile(fname); pax.SetPath(pathtofile); pax << idonly;
		
	while( pax.NextEvent() ) { // loop on events in the generated pax file
		if ( pax.GetId().FIS == fis ) {
			h->Fill(pax.GetId().Z,pax.GetId().N);
		}
	} // while reader
	
	h->Fill(pax.GetGlobal().ZC,pax.GetGlobal().AC-pax.GetGlobal().ZC,pax.GetGlobal().NCASC);
	
	// now scale to get the cross sections
	h->Scale(pax.GetGlobal().SIGMA/pax.GetGlobal().NCASC);
	
	pax.ShowCurrentConditions();
	
	h->Draw("col"); return h;
}

/** Function to plot Excitation function
 
	With evapOR, a list of pax files could be generated for different energies (see the evloop script). 
	If in your input file for evapor, the following line has been added "*NAMES XXX INDEX", the pax file are named XXX001.pax, XXX002.pax ...	
	This function chaines from the directory pathtofile all the files with the name constructed from the filebasename and the integers from istart to iend (there is no error if an non-existing file is given).
	It run through the files to plot the excitation function (one graph per residu) with different styles (color and markers).
	
	It keeps residu production if the production cross section is greater than threshold
	
	- filebasename is the base file name of the pax files to be scanned
	- pathtofile the directory where are stored the pax files to be chained	
*/
TMultiGraph *Excitation(const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10, Float_t threshold = 0.01)
{
	// style for each graph ... colmarq_Z color for an isotope, stymarq_N style for isotone
	// firstZ firstN indicate the starting point in (Z,N) .. could be the compound nucleus
	// so the indice #i in colmarq_Z (stymarq_N) corresponds to the nucleux with firstZ-i (firstN-i) protons (neutrons)
	// Fission have a different style as well as nuclei 'out of range' 
	const Short_t NBSTYLE = 11; 
	Short_t firstZ = 64;
	Short_t firstN = 88;
	const Short_t colmarq_Z[NBSTYLE] = { 2, 3, 4, 6, 7, 8, 9,20,30,40,50}; const Short_t colmarq_FISSION = 1; 
	const Short_t stymarq_N[NBSTYLE] = {20,21,22,23,24,25,26,27,28,29,30}; const Short_t stymarq_FISSION = 5; 
		
	const Int_t nbin = 2000, min = 0, max = 200; // histograms definitions
	
	// array of histograms that are transformed to TGraph at the end.
	TObjArray ar(MAXN*MAXZ); ar.SetOwner(); TH1F *h; TString tmp;
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << idonly; 
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	while( pax.NextEvent() ) { // loop on events in the generated pax file		
		if ( pax.GetId().FIS == kTRUE ) { // slot 0 of the TObjArray to store fission like events
			if ( ar[0] == NULL ) { 
				tmp ="Fission"; h = new TH1F(tmp.Data(),filebasename,nbin,min,max); ar.AddAt(h,0); 
				
				// set attributes
				h->SetMarkerColor(colmarq_FISSION); h->SetMarkerStyle(stymarq_FISSION); h->SetLineColor(colmarq_FISSION);
			}
			h = (TH1F *)(ar.At(0));
			if ( h ) h->Fill(pax.GetGlobal().ENER,pax.GetGlobal().SIGMA/pax.GetGlobal().NCASC);
		}
		else { // residus
			if ( pax.GetId().N > 0 && pax.GetId().Z > 0 ) {
				if ( ar[pax.GetId().Z * MAXN + pax.GetId().N] == NULL ) {
					tmp = pax.GetId().Z; tmp += "-"; tmp += pax.GetId().N;
					h = new TH1F(tmp.Data(),filebasename,nbin,min,max); ar.AddAt(h,pax.GetId().Z * MAXN + pax.GetId().N); 
					
					// set histo style color and marker
					if ( (firstN - pax.GetId().N) > -1 && (firstN - pax.GetId().N) < NBSTYLE 
						&& (firstZ - pax.GetId().Z) > -1 && (firstZ - pax.GetId().Z) < NBSTYLE  ) { 
						h->SetMarkerColor(colmarq_Z[firstZ - pax.GetId().Z]); 
						h->SetMarkerStyle(stymarq_N[firstN - pax.GetId().N]); 
						h->SetLineColor(colmarq_Z[firstZ - pax.GetId().Z]);	
					}
					else {
						h->SetMarkerColor(14); 
						h->SetMarkerStyle(2); 
						h->SetLineColor(14);
						h->SetLineStyle(2);	
					}
				}
				h = (TH1F *)(ar.At(pax.GetId().Z * MAXN + pax.GetId().N));
				if ( h ) h->Fill(pax.GetGlobal().ENER,pax.GetGlobal().SIGMA/pax.GetGlobal().NCASC);	
			}
		}
	} // while reader	
	pax.ShowCurrentConditions();
		
	// to put all histograms in graphs
	TMultiGraph *stack = new TMultiGraph(filebasename,filebasename); 
	
	TLegend *leg = new TLegend(0.8,0.8,0.99,0.99);	
	
	Short_t ihisto = 0;
	if ( ar.GetEntries() > 0 ) { // at least one histo exists
		for (Int_t i = 0; i < MAXN*MAXZ; i++  ) {
			if ( ar[i] ) { // an histogram exists at this slot
			
				h = (TH1F *)ar[i]; if ( h->GetMaximum() < threshold ) continue;
			
				// new graph added to the collection
				TGraphErrors *g = new TGraphErrors(); 
				stack->Add(g); 
				
				// to set different style for each graph and the legend
				g->SetMarkerColor(h->GetMarkerColor()); g->SetMarkerStyle(h->GetMarkerStyle());
				g->SetLineColor(h->GetLineColor());
				g->SetLineStyle(h->GetLineStyle());
					
				g->SetNameTitle(h->GetName(),h->GetTitle()); // to get the histogram that is used to build the TGraph
				
				TString tmp;
				if ( i == 0 ) tmp = "Fission";
				else {
					tmp ="X ^{"; tmp +=  (i / MAXN + i % MAXN) ; tmp +="}_{"; tmp += i / MAXN; tmp +="}";
				}
				leg->AddEntry(g,tmp.Data(),"L P");
				
				// fill the graph
				Int_t ipoint = 0;
				for (Int_t ibin = 1; ibin < nbin+1; ibin++ ) { // keep bincontent > threshold
					if ( h->GetBinContent(ibin) > threshold ) { 
						g->SetPoint(ipoint,h->GetBinCenter(ibin),h->GetBinContent(ibin));
						g->SetPointError(ipoint,h->GetBinWidth(ibin),h->GetBinError(ibin));
						ipoint++; 
					}
				}
				ihisto++; 
			} 
		}	
		stack->Draw("ALP");
		stack->GetXaxis()->SetTitle("Excitation energy (MeV)"); stack->GetYaxis()->SetTitle("Cross section (mbarns)");
		leg->Draw();
	}
	return stack;
}

/** Function to draw a single distribution
	
	Just allocate the histogram with a title corresponding to the distribution you want to draw. For instance:
	
	TH1F h("MyHisto","MG",20,0,20) 
	
	then, call Distribution(h,...) and the distribution will be drawn.
	
	- filebasename is the base file name of the pax files to be scanned
	- pathtofile the directory where are stored the pax files to be chained			
*/
void Distribution(TH1F &histo, const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10)
{
	TString tmp; 
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << full;  // pax << idonly; pax.VERBOSE = 1;
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	TLeaf *x = pax.GetLeaf(histo.GetTitle()), *w = pax.GetLeaf("WT");
	if ( x == NULL ) 
		{ cout << " Unknown distribution " <<  tmp.Data() << endl; return; }
	
	// fill the histogram and draw it
	while( pax.NextEvent() ) { // loop on events in the generated pax file	
		histo.Fill(x->GetValue(),w->GetValue());
	}	
	pax.ShowCurrentConditions(); histo.Draw(); delete x; delete w;
}

/** Function to draw a single distribution

	To follow the evolution of a variable as a function of the event number.
*/
TGraph *Evolution1D(const char *var, const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10, Int_t which = 0)
{
	TString tmp; 
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << full;  
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	// to get the variable to be inspected
	TLeaf *x = pax.GetLeaf(var);
	if ( x == NULL ) 
		{ cout << " Unknown variable " <<  tmp.Data() << endl; return NULL; }
	
	// fill the graph and draw it
	TGraph *gr = new TGraph(); 
	Int_t i = 0;
	while( pax.NextEvent() ) { // loop on events in the generated pax file	
		/* if ( x->GetValue(which) > 1 ) pax.ShowCurrentEvent() ; */ gr->SetPoint(i,i,x->GetValue()); i++;
	}	
	pax.ShowCurrentConditions(); gr->Draw("A*"); delete x; return gr;
}


/** Function to draw a 2D distribution
		
	As for Distribution, just allocate your 2D histogram with a title corresponding to what you would like to draw.
	
	TH1F h("MyHisto","JE:EE",20,0,20,20,0,20); Distribution2D(h, ... ); 
	
	will draw the entry point distribution in the (Spin,Excitation Energy) space.
*/
void Distribution2D(TH2 &histo, const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10)
{
	TString tmp; 
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << full;  // pax << idonly; pax.VERBOSE = 1;
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	// to get x, y from the histogram's name	
	TLeaf *x = NULL, *y = NULL, *w = pax.GetLeaf("WT"); 
	
	tmp = histo.GetTitle(); TObjArray *arr = tmp.Tokenize(":"); 
	if ( arr && arr->GetEntries() == 2 ) { // look for what is to be drawn
		x = pax.GetLeaf(arr->At(0)->GetName());
		y = pax.GetLeaf(arr->At(1)->GetName());	
		
		delete arr;
	}

	if ( x == NULL || y == NULL ) 
		{ cout << " Unknown distribution " <<  tmp.Data() << endl; return; }
	
	// fill the histogram and draw it
	while( pax.NextEvent() ) { // loop on events in the generated pax file	
//		if ( pax.GetId().FIS == false ) histo.Fill(x->GetValue(),y->GetValue(),w->GetValue());
		histo.Fill(x->GetValue(),y->GetValue(),w->GetValue());		
	}	
	pax.ShowCurrentConditions(); histo.Draw(); delete x; delete y; delete w;
}

/** Function to draw the Energy of the particule as a function of the theta angle
	
*/
void EPxTHETA(TH2 &histo, const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10, Int_t mode = 1)
{
	TString tmp; 
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << full;  // pax << idonly; pax.VERBOSE = 1;
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	// to get x, y from the histogram's name	
	TLeaf *x = pax.GetLeaf("THETA"), *y = pax.GetLeaf("EEJ"), *w = pax.GetLeaf("WT"); 

	
	// fill the histogram and draw it
	while( pax.NextEvent() ) { // loop on events in the generated pax file	
		for( Short_t i = 0; i < pax.GetId().NUM ; i++){
			if ( pax.GetData().MODE[i] == mode ) histo.Fill(x->GetValue(i)*180/TMath::Pi(),y->GetValue(i),w->GetValue());
		}
	}	
	pax.ShowCurrentConditions(); histo.Draw(); delete x; delete y; delete w;
}

/** Function to show cascades in the E x I plan
	
	Press a keyboard to run without interruption, click to stop, double click to run step by step
	Crtl-C to stop the display.
*/
void ShowCascades(const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10)
{
	TString tmp; TLine line; line.SetLineWidth(2); TMarker mark; mark.SetMarkerStyle(23); 
	
	Short_t colors[11] = {0,1,30,28,2,4,6,8,9,10,11};
	Short_t styles[11] = {0,2,1,1,1,1,1,8,9,10,11};
	
	if ( gPad == NULL ) return;
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << full;  // pax << idonly; pax.VERBOSE = 1;
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	Int_t refresh = 0;
	while( pax.NextEvent() ) { // loop on events in the generated pax file	
		
		if ( pax.GetId().FIS == kTRUE ) continue;
		
		if ( refresh % 3 == 0 ) 
			{ refresh = 0 ; gPad->Clear(); gPad->DrawFrame(0,0,TMath::Max(1.1*pax.GetId().L+5,10.),1.1*pax.GetGlobal().ENER); }
		refresh++;
		
		mark.DrawMarker(pax.GetId().JE,pax.GetId().EE); // entry point as given by the id
		
		for( Int_t i = 0; i < pax.GetId().NUM  ; i++){ // to draw the full cascade
			line.SetLineColor(colors[pax.GetData().MODE[i]]); 
			line.SetLineStyle(styles[pax.GetData().MODE[i]]); 	
			if ( i == pax.GetId().NUM - 1 ) 
				line.DrawLine(pax.GetData().JI[i],pax.GetData().EI[i],0,0);
			else 
				line.DrawLine(pax.GetData().JI[i],pax.GetData().EI[i],pax.GetKin().JF[i],pax.GetData().EI[i+1]);	
		}
		gPad->Update(); TObject *obj = gPad->WaitPrimitive(); 
					
		// pax.ShowCurrentEvent(); 
		if ( gROOT->IsInterrupted() ) { cout << "loop stopped " << endl; break; }
	}
}

/** 

*/
void Distribution(const char *filebasename, const char *pathtofile = "./", Int_t istart = 1, Int_t iend = 10)
{
	TH1F *histo = new TH1F("","",200,-100,100);
	
	TString tmp; 
	
	// open chain of files that are in the pathtofile directory
	EvapLMOF pax; pax.SetPath(pathtofile); pax << full;  // pax << idonly; pax.VERBOSE = 1;
		
	// add files to the pax reader
	for (Int_t i = istart; i < iend; i++  ) {
		tmp = filebasename; 
		if ( i < 10 ) 
			tmp += "00";
		if ( 10 <= i && i < 100 ) 
			tmp += "0";	
	
		tmp += i; tmp += ".pax";		
		if ( pax.VERBOSE ) cout << "Adding " << tmp.Data() << endl; pax.AddFile(tmp.Data());
	} 
	
	// fill the histogram and draw it
	while( pax.NextEvent() ) { // loop on events in the generated pax file	
		// if ( pax.GetId().L-pax.GetData().JI[0] > -1 ) { 
			histo->Fill(pax.GetId().L-pax.GetData().JI[0]); 
		// pax.ShowCurrentEvent();}
	}	
	pax.ShowCurrentConditions(); histo->Draw();
}


