/** \file ensdfdemo.C to illustrate how to use the new Import method (GammaWare 0.2)

   This macro re-defines the oneLS() funtion defines in firstdemo.C and shows how to use
   the new LevelScheme::Import method to build a level scheme with ags or ensdf files.
   
 \author Olivier Stezowski 
*/

using namespace Gw;

/** Function to make a monte carlo on any AGS/or ENSDF file 

  Usage: 
	- root
	- root [0] .L ensdfdemo.C
	- root [1] oneLS("151TB.ags","151TB") or root [1] oneLS("TB_ENSDF.ens","151TB")

filename is the name of the AGS or ENSDF file used to produce the random cascades 
nucleus  is a the nucleus to which belongs the levelscheme (needed for ENSDF files - see LevelScheme::Import())
nbcas corresponds to the number of cascades to be simulated

It produces two histograms in the current directory
             - the multiplicity distribution of the simlated cascades
             - the gamma ray spectrum of all gammas

Olivier Stezowski - 07/2004

PS: libGWPHYSICS and libGWGEM must be loaded before running this macro (see rootlogon.C)
*/
void oneLS(const Char_t *filename, const Char_t *nucleus, Int_t nbcas = 100)
{
	// load the AGS file
	BaseGEM random;
	if ( random.Import(filename,nucleus) == 0 ) {
		cout << " Level Scheme not correctly loaded " << endl; return;
	}

	// some spectra to be filled
	TH1F *K = new TH1F("K","K",100,0,100);
	TH1F *h = new TH1F("ProjTot","ProjTot",8192,0,4096);

	// start the Monte-Carlo
	Cascade cas; GammaLink *gam;
	for (Int_t i = 0; i < nbcas; i++ ) {

		random.DoCascade(cas);

		K->Fill(cas.GetSize()) ;
		for (Int_t j = 0; j < cas.GetSize(); j++ ) {
			gam = (GammaLink *)cas.At(j); h->Fill(gam->GetEnergy().Get());
		} // j
	} //i
	h->Draw();
}

/** Function to make a monte carlo on any AGS/or ENSDF file 

Olivier Stezowski - 07/2004

PS: libGWPHYSICS and libGWGEM must be loaded before running this macro (see rootlogon.C)
*/
void testfeeding(const Char_t *filename, const Char_t *nucleus, Int_t nbcas = 100)
{
	// load the level scheme from an external file
	GEM random;
	if ( random.Import(filename,nucleus) == 0 ) {
		cout << " Level Scheme not correctly loaded " << endl; return;
	}
	// Adds a feeding to the GEM process
	random.SetFeeding("TestFeeding"); random.SetExI(80000,60); // entry point 50 MeV, spin 60
	
	// some spectra to be filled
	TH1F *K = new TH1F("K","K",100,0,100);
	TH1F *h = new TH1F("ProjTot","ProjTot",8192,0,4096);

	// start the Monte-Carlo
	Cascade cas; GammaLink *gam;
	for (Int_t i = 0; i < nbcas; i++ ) {

		random.DoCascade(cas);

		K->Fill(cas.GetSize()) ;
		for (Int_t j = 0; j < cas.GetSize(); j++ ) {
			gam = (GammaLink *)cas.At(j); h->Fill(gam->GetEnergy().Get());
		} // j
	} //i
	h->Draw();
}

