/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file AgataKeyFactory.h header file for AgataKeyFactory.cpp */

#ifndef ADF_AgataKeyFactory
#define ADF_AgataKeyFactory
#define ADF_AGATAKEYFACTORY_H ADF_AgataKeyFactory

#ifndef ADF_KeyFactory
#include "KeyFactory.h"
#endif
#ifndef ADF_Key
#include "Key.h"
#endif
#ifndef ADF_DotConf
#include "DotConf.h"
#endif

#include <iostream>
#include <string>
#include <map>
#include <utility>

namespace ADF
{ 
//! Interface to AgataKey
/*!
	All AgataKeys inherit from this class that defined the interface for an agata key i.e. the type of
	key that structure the data flow in Narval.
	
	As a Key, it defines the length of the frame, a message type, an event number and a timestamp : [length][message][event#][TS]
 
	All this fields are not necessarely written on the data flow. It depends of which version of the Agatakey
	you are working with. If a field is not coded, for instance the timestamp, the value returns by the
	corresponding  GetXX method is the default value, mainly 0. As well, the Set method does not write anything but instead set the 
	flag of the key to kCorrupt.
	
	AgataKeys are allocated by the AgataKeyFactory. A factory name, a type of key and a version number (FactoryItem) are then required.
	
	The comparison of two agata key is done by comparing the Message type which has to be identical.
*/
class AgataKey : public Key
{
public:
  	friend class AgataKeyFactory;
    
protected:
	UInt_t kKeyLength     ;
	UInt_t kMaxDataLength ;   
    
protected:
	UInt_t fMessage, fMask;
    
protected:
	//! To set the message type 
	virtual void SetMessage(UInt_t)     = 0;  
       
public:
	AgataKey(BaseBuffer::EEndian e, UInt_t s = 4u*aByte) : 
		Key(e,s), 
		fMessage(0x0A000000), 
		fMask(0x0F000000) 
	{ 
		kKeyLength = s; 
		kMaxDataLength = ADF::kMaxUInt_t - kKeyLength; 
	}
	virtual ~AgataKey() 
		{;}    
	
	//! To get the message type encoded
	virtual UInt_t GetMessage()  const   = 0;        
	//! To get the event number encoded
	virtual UInt_t GetEventNumber() const = 0;     
	
	//! To set the event number     
	virtual void SetEventNumber(UInt_t) = 0;  
	
	//! To set the timestamp
	virtual ULong64_t GetTimeStamp() const = 0;  
	virtual void SetTimeStamp(ULong64_t)   = 0;       
	
	//! stream in buffer the data length
	/*!
		An AgataKey books 4 bytes for the frame length
	*/   
	virtual void SetDataLength(UShort_t s) 
	{ 
		fBuffer->SetOffset(); 
		(*fBuffer) << (UInt_t(s)+kKeyLength); 
	} 
	
	//! stream in buffer the data length
	virtual void SetDataLength(UInt_t i) 
	{ 
		UInt_t ri = kKeyLength; 
		if ( i < kMaxDataLength ) 
				ri += i; 
		fBuffer->SetOffset(); 
		(*fBuffer) << ri; 
	}     
		
	virtual UInt_t GetDataLength() const 
	{ 
		UInt_t i = 0u; 
		fBuffer->SetOffset(); 
		(*fBuffer) >> i; 
		if ( i >= kKeyLength ) 
			i = i - kKeyLength ; 
		
		return i;
	}	 
		
	virtual UInt_t GetFrameLength() const 
	{ 
		UInt_t i = kKeyLength; 
		fBuffer->SetOffset(); 
		(*fBuffer) >> i; 
		return i; 
	}
		
	//! Compare it to another key.     
	/*!
		The two keys have to have the same signature and the same message type to return true
	*/       
	virtual Bool_t IsAKey(const Key *akey) const
	{ 
		Bool_t ok = false;
		if ( akey->GetSignature().GetFactoryName() == GetSignature().GetFactoryName() ) { // be sure an agata factory has allocated akey
			const AgataKey *ckey = dynamic_cast<const AgataKey *> (akey);
			if ( ckey ) {
				ok = (fMessage == (ckey->GetMessage() & fMask));				
			}
		}
		return ok;
	} 
		
	//! to modify the Agata type for this key 
	virtual void ChangeType(UInt_t messtype, UInt_t messmask) 
	{ 
		fMessage = messtype; 
		fMask = messmask; 
	}     
	
	virtual void Print(std::ostream &out = std::cout) const;  
	
	//! convert a key to another key ... not an easy task ... use it carefully
	/*!
		By default, converting a key means just copy the data length
		
			\todo timestamp implementation
	*/
	virtual Bool_t Convert(const Key *akey); 	
};


//! Base class for a AgataKeyFactory
/*!
	The factory delivers AGATA keys and thus keeps a list (std::map) of strings associated to unique addresses. 
	It is initiated by reading ADF.conf in the distribution and/or in the environnment variable ADF_CONF_PATH.
 
 Depending of the option given in the header in the ADF.conf file found under $ADF_CONF_PATH, the data flow agent is reset or not :
 - Init : reset
 - +,== : add/replace mode. execpt some hard coded key, it allows to modified the key content to be able to read old produced files and to add new definitions (event-like for instance or to deal with a new frame factory)
 
 #////ADF::AgataKeyFactory_beg//// - -1 Init 

 The format for the map is given there :
 \code
 # Format: KeyNAME message mask 
 #
 # KeyNAME is the string to be passed to the AgataKeyFactory. It should be unique !
 # message is the message part as coded in the key
 # mask    is to be used to extract part of the message when comparing keys
 #
 # with 0xWAXXYYZZ
 #
 #	W = F for simple frame, C for composite frame - A means used AgataKey factory
 #
 #	XX = Sub-detector number (associated to a given factories)
 #		01 for Agata (Agata Frame factory), pure Agata
 #		02 for Agata (Agata Frame factory), raw ancillary for Agata
 #               > 02 for other factories
 #	YY = Frame type
 #		01 data  (produce/consume by algorithm)
 #		11 meta  (general, does not depend on an algorithm)
 #		02 conf  (conf associated to a data/meta frame)
 #		03 algo  (conf specific for an algorithm)
 #		04 info  (conf like, just some comments)
 #		05 rmess (message, raw frame)
 #		06 cmess (message, conf frame)
 #
 #	ZZ = ALGO ID, should be unique in the data flow
 \endcode
	
	data are exchanged between different algorithms. they are input for filters and ouptut of filters. 
	meta are also data BUT general i.e. not proper to a specific algorithm
	conf are configuration frames that are use to encode/decode the corresponding data frame. for instance it could contain digitizer/calibration coefficients (inner ADF usage)
	algo are configuration frames needed to reconfigure a specific algorithm. It could be a modification of some calibration coefficients, parameters (usage : for the user)
	info are configuration frames just use to pass some comments/ general informations (usage : for the user)
	rmess are raw data frame that could be used by the user for passing informations between two parts of the system
	cmess are configuration frame that could be used by the user for passing informations between two parts of the system

	all the previous frame are for simple frame (0xFA... == Frame Agata) like keys. Composite frames start with 0xCA (0xCA... == CompositeFrame Agata) 
 
 NOTE : in the very first version of agata key, frame out of the event builder and the merger were labelled 0xFA010103 / 0xFA010104.
 thus breaking the rule for composite frames. It should be 0xCA010102 / 0xCA000100. 
 To allow reading back all the adf files from the beginning, two keys have been introduced : \n
 data:eb / data:merger \n
 which are at some level consitent since these are the kind of data procuced by eb and merger.
 
 Doing that the system is backward compatible. Since conf:global frame are written in .adf files the key content are modified @ reconfiguration to properly read the data already produced.
 There is only one consequence : in any .adf file produced based on old .adf, the composite frames for event:data and event:data:psa would be with the older definition i.e. 0xFA010103 / 0xFA010104.
 To allow this compatibility, a version # has been introduced for the factory: \n
 - For version -1, the general convention is not used for event:data and event:data:pas keys. the extention of the system is restricted\n
 - For version 0, the general convention is used. the user can extend the system by adding new keys / building any kind of event
  
 To move to version 0, jsut change -1 into 0 in the header part:
 
 #////ADF::AgataKeyFactory_beg//// - 0 Init 

 Of course, this is transparent for the user as soon as she/he uses adf/gw to read the adf data.
 
*/
class AgataKeyFactory : public AKeyFactory<AgataKeyFactory>
{
	friend class AKeyFactory<AgataKeyFactory>;
    
private:
	// correspondance between a string (human readable) and a message type (UInt_t)
	std::map<std::string,std::pair<UInt_t,UInt_t> > fMessageTable; 
    
private:
	//! hardcoded frames : could not be changed by ADF.conf in any cases
	std::map<std::string,std::pair<UInt_t,UInt_t> > fHardcodedMessageTable; 
	
private:
	//! a patch to go around a non-general definition of event key in the very first version of this
	Bool_t fIsPatchEvent; 
	
private:
	AgataKeyFactory();
	virtual ~AgataKeyFactory () 
		{;}

protected: 
	//! Ask this factory for a new AgataKey
	/*!
		it returns 0x0 if the key is not know by the system or if the request is for a factory version greater that the current definition of the factory
	*/   
	virtual Key *DoNew(const FactoryItem &);
	
	//! During configuration, process a single line    
	virtual Bool_t ProcessLine(const std::string &); 
	//! Reset the table            
	virtual void Reset(const Char_t * = ""); 
	
public:
	//! Agata files are always little endian !!
	virtual void SetEndian(BaseBuffer::EEndian) 
		{ KeyFactory::SetEndian(BaseBuffer::kLittle); }

	//! to just browse the list of known keys
	const std::map<std::string,std::pair<UInt_t,UInt_t> > & GetKnownKeys() const
	{
		return fMessageTable;
	}

	//! to check if a given keyname is in the list of known keys (return true in this case, false otherwise) and to access the corresponding message type, mask
	Bool_t GetMessage(std::string keyName, UInt_t &message, UInt_t &mask);
	Bool_t GetMessage(std::string keyName, UInt_t &message)
	{
		UInt_t mask;
		return GetMessage(keyName,message,mask);
	}

	//! to get the message part as a string 
	const Char_t *GetMessage(UInt_t);
	
	//! write the message table
	virtual Bool_t DoConfigure(std::ostream &) ;
	virtual Bool_t DoConfigure(std::istream &in, Bool_t allow_init) 
		{ return DotConf::DoConfigure(in, allow_init); }
}; 
// AgataKeyFactory inline members //////////////////////////////////////////////////////


} // namespace ADF
#endif





