/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ADF_BaseBuffer
#define ADF_BaseBuffer
#define ADF_BASEBUFFER_H ADF_BaseBuffer

#ifndef ADF_ADFConfig
#include "ADFConfig.h"
#endif

#include <cstring>

namespace ADF
{ 

const UInt_t  aByte = 1u;
const UInt_t aKByte = 1024u;
const UInt_t aMByte = aKByte*aKByte;
const UInt_t aGByte = aKByte*aMByte;
    
//! Base class for a buffer. 
/*!
	It defines a set of useful functions to deal with buffers of basic types. It also determined the endian
	type of the system you are working with. 
	This class is used by Buffer and EndianBuffer to get buffers that can be read whatever their origin 
	as soon as the endian type of the source of data is known. 
	
		 \todo - the copy constructor must ne written as fast as possible !
*/
class BaseBuffer
{
public:
    //! Status for this buffer
    /*!
    	 - kGood:	everything is all right
	 - kFail:	the last operation failed, not enough place to in the buffer
	 - kCorrupt:		 
    */
    enum EStatus { kGood, kFail, kCorrupt } ;    
    
    //!
    /*!
       The adjectives big-endian and little-endian refer to which bytes are most significant in multi-byte data 
       types and describe the order in which a sequence of bytes is stored in a computer’s memory.
       In a big-endian system, the most significant value in the sequence is stored at the lowest storage address (i.e., first). 
       In a little-endian system, the least significant value in the sequence is stored first. 
       
       \todo check byte swapping for Long_t, 
    */
    enum EEndian { kLittle = 0, kBig = 1 } ;
    
    // enum ESeek { kBeg, kCur, kEnd} ;
          
private:
    //! Agent that determines the endian type at starting time
    /*!
    	The endian type is determined in the constructor (many thanks to Fons for the tip) 
	which is called at least once when allocating gAgent
    */
    class Agent {
    	EEndian fEndian; 
    public:
    	Agent();
	EEndian GetEndian() { return fEndian; }
    };
    static Agent gAgent; // static Agent *gAgent;
    
public:
    BaseBuffer();
    virtual ~BaseBuffer();
    
    //! check out the endian type of the running system
    /*!
    	The request is treated by the Agent
    */
    static bool IsSysEndian(BaseBuffer::EEndian e) 
    	{ return gAgent.GetEndian() == e; }
    static EEndian SysEndian() 
    	{ return gAgent.GetEndian() ; }
       
    static Char_t   *New(Char_t   *p, Int_t nb = 32*aKByte);
    static Short_t  *New(Short_t  *p, Int_t nb = 32*aKByte);
    static Int_t    *New(Int_t    *p, Int_t nb = 32*aKByte);
    static UChar_t  *New(UChar_t  *p, Int_t nb = 32*aKByte);
    static UShort_t *New(UShort_t *p, Int_t nb = 32*aKByte);
    static UInt_t   *New(UInt_t   *p, Int_t nb = 32*aKByte);
    static Float_t  *New(Float_t  *p, Int_t nb = 32*aKByte);
    static Double_t *New(Double_t *p, Int_t nb = 32*aKByte);    
    
    static void Zero(Char_t   *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,nb); } 
    static void Zero(Short_t  *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Short_t)*nb); }
    static void Zero(Int_t    *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Int_t)*nb);   }
    static void Zero(UChar_t  *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(UShort_t *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(UShort_t)*nb); }
    static void Zero(UInt_t   *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(UInt_t)*nb);   }
    static void Zero(Float_t  *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Float_t)*nb); }
    static void Zero(Double_t *p, Int_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Double_t)*nb); }
    
    static Char_t   *New(Char_t   *p, UInt_t nb );
    static Short_t  *New(Short_t  *p, UInt_t nb );
    static Int_t    *New(Int_t    *p, UInt_t nb );
    static UChar_t  *New(UChar_t  *p, UInt_t nb );
    static UShort_t *New(UShort_t *p, UInt_t nb );
    static UInt_t   *New(UInt_t   *p, UInt_t nb );
    static Float_t  *New(Float_t  *p, UInt_t nb );
    static Double_t *New(Double_t *p, UInt_t nb );     
    
    static void Zero(Char_t   *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(Short_t  *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Short_t)*nb); }
    static void Zero(Int_t    *p, UInt_t nb)
    	 { if ( nb < 1 ) return; ::memset(p,0,sizeof(Int_t)*nb);   }
    static void Zero(UChar_t  *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(UShort_t *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(UShort_t)*nb); }
    static void Zero(UInt_t   *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(UInt_t)*nb);   }
    static void Zero(Float_t  *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Float_t)*nb); }
    static void Zero(Double_t *p, UInt_t nb) 
    	{ if ( nb < 1 ) return; ::memset(p,0,sizeof(Double_t)*nb); }     
    
    static void Delete(Char_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(Short_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(Int_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(UChar_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(UShort_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(UInt_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(Float_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }
    static void Delete(Double_t *p)	
    	{ if ( p ) { delete [] p; p = NULL;} }

    //! copy the content from -> to
    /*!
    	
    */
    static void Copy(const void *from, void *to, Int_t size)  
    	{ if ( size > 0 ) ::memcpy(to,from,size); }
    static void Copy(const void *from, void *to, UInt_t size) 
    	{ if ( size > 0 ) ::memcpy(to,from,size); }

    //! the Swap members
    /*!
    
    	Swap(T, Char_t *buf) writes type T in the buffer pointed by buf swapping the bytes
	 
	This members used, if defined, the ROOT macros defined in Bytes.h which take into account byte swapping optimizations from the some systems.
    */
    static void Swap(Short_t, Char_t *);
    static void Swap(UShort_t, Char_t *); 
    static void Swap(Int_t, Char_t *);
    static void Swap(UInt_t, Char_t *);
    static void Swap(Float_t, Char_t *);
    static void Swap(Double_t, Char_t *);  
    static void Swap(Long64_t, Char_t *);
    static void Swap(ULong64_t, Char_t *); 
    static void Swap(Long_t, Char_t *);
    static void Swap(ULong_t, Char_t *);   
          
    static void Swap(const Char_t *,Short_t &);
    static void Swap(const Char_t *,UShort_t &); 
    static void Swap(const Char_t *,Int_t &);
    static void Swap(const Char_t *,UInt_t &);   
    static void Swap(const Char_t *,Float_t &);
    static void Swap(const Char_t *,Double_t &);      
    static void Swap(const Char_t *,Long64_t &);
    static void Swap(const Char_t *,ULong64_t &); 
    static void Swap(const Char_t *,Long_t &);
    static void Swap(const Char_t *,ULong_t &);         
//    void Test();
};

// BaseBuffer inline members //////////////////////////////////////////////////////
inline Char_t   * BaseBuffer::New(Char_t   *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Char_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Short_t  * BaseBuffer::New(Short_t  *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Short_t[nb]; if ( p ) Zero(p,nb); return p; }
inline Int_t    * BaseBuffer::New(Int_t    *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Int_t[nb];   if ( p ) Zero(p,nb); return p; }
inline UChar_t  * BaseBuffer::New(UChar_t  *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new UChar_t[nb]; if ( p ) Zero(p,nb); return p; }
inline UShort_t * BaseBuffer::New(UShort_t *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new UShort_t[nb];if ( p ) Zero(p,nb); return p; }
inline UInt_t   * BaseBuffer::New(UInt_t   *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new UInt_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Float_t  * BaseBuffer::New(Float_t  *p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Float_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Double_t * BaseBuffer::New (Double_t * p,Int_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Double_t[nb]; if ( p ) Zero(p,nb); return p;}

inline Char_t   * BaseBuffer::New(Char_t   *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Char_t[nb];   if ( p ) Zero(p,nb); return p; }
inline Short_t  * BaseBuffer::New(Short_t  *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Short_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Int_t    * BaseBuffer::New(Int_t    *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Int_t[nb];    if ( p ) Zero(p,nb); return p; }
inline UChar_t  * BaseBuffer::New(UChar_t  *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new UChar_t[nb];  if ( p ) Zero(p,nb); return p; }
inline UShort_t * BaseBuffer::New(UShort_t *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new UShort_t[nb]; if ( p ) Zero(p,nb); return p; }
inline UInt_t   * BaseBuffer::New(UInt_t   *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new UInt_t[nb];   if ( p ) Zero(p,nb); return p; }
inline Float_t  * BaseBuffer::New(Float_t  *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Float_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Double_t * BaseBuffer::New (Double_t *p,UInt_t nb) 
	{ if ( nb < 1 ) return NULL; p = new Double_t[nb]; if ( p ) Zero(p,nb); return p; }


inline void BaseBuffer::Swap(UShort_t x, Char_t *buf)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   *((UShort_t *)buf) = Rbswap_16(x); 
# else
   // To work around a stupid optimization bug in MSVC++ 6.0
   const UShort_t *intermediary = &x;
   char *sw = (char *) intermediary;
   buf[0] = sw[1];
   buf[1] = sw[0]; 
# endif
}
inline void BaseBuffer::Swap(Short_t x, Char_t *buf) 
	{ BaseBuffer::Swap((UShort_t) x,buf) ;}

inline void BaseBuffer::Swap(UInt_t x, Char_t *buf)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   *((UInt_t *)buf) = Rbswap_32(x);
# else
   // To work around a stupid optimization bug in MSVC++ 6.0
   const UInt_t *intermediary = &x;
   char *sw = (char *)intermediary;
   buf[0] = sw[3];
   buf[1] = sw[2];
   buf[2] = sw[1];
   buf[3] = sw[0];
# endif
}
inline void BaseBuffer::Swap(Int_t x, Char_t *buf) 
	{ BaseBuffer::Swap((UInt_t) x,buf);}

inline void BaseBuffer::Swap(Float_t x, Char_t *buf)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   union {
      volatile UInt_t  i;
      volatile Float_t f;
   } u;
   u.f = x;
   *((UInt_t *)buf) = Rbswap_32(u.i);   
# else
   union {
     volatile char    c[4];
     volatile Float_t f;
   } u;
   u.f = x;
   buf[0] = u.c[3];
   buf[1] = u.c[2];
   buf[2] = u.c[1];
   buf[3] = u.c[0];
# endif
}

inline void BaseBuffer::Swap(Double_t x, Char_t *buf)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   union {
      volatile ULong64_t l;
      volatile Double_t  d;
   } u;
   u.d = x;
   *((ULong64_t *)buf) = Rbswap_64(u.l);   
# else
   union {
     volatile char     c[8];
     volatile Double_t d;
   } u;
   u.d = x;
   buf[0] = u.c[7];
   buf[1] = u.c[6];
   buf[2] = u.c[5];
   buf[3] = u.c[4];
   buf[4] = u.c[3];
   buf[5] = u.c[2];
   buf[6] = u.c[1];
   buf[7] = u.c[0];
# endif
}

inline void BaseBuffer::Swap(ULong64_t x, Char_t *buf)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   *((ULong64_t *)buf) = Rbswap_64(x); 
# else
   // To work around a stupid optimization bug in MSVC++ 6.0
   const ULong64_t *intermediary = &x;
   char *sw = (char *)intermediary;
   buf[0] = sw[7];
   buf[1] = sw[6];
   buf[2] = sw[5];
   buf[3] = sw[4];
   buf[4] = sw[3];
   buf[5] = sw[2];
   buf[6] = sw[1];
   buf[7] = sw[0];
# endif
}
inline void BaseBuffer::Swap(Long64_t x, Char_t *buf) 
	{ BaseBuffer::Swap((ULong64_t) x,buf) ;}
	
inline void BaseBuffer::Swap(Long_t x, Char_t *buf)
{
   const Long_t *intermediary = &x;
   char *sw = (char *)intermediary;
   if (sizeof(Long_t) == 8) {
      buf[0] = sw[7];
      buf[1] = sw[6];
      buf[2] = sw[5];
      buf[3] = sw[4];
      buf[4] = sw[3];
      buf[5] = sw[2];
      buf[6] = sw[1];
      buf[7] = sw[0];
   } else {
      if (x < 0) {
         buf[0] = (char) -1;
         buf[1] = (char) -1;
         buf[2] = (char) -1;
         buf[3] = (char) -1;
      } else {
         buf[0] = 0;
         buf[1] = 0;
         buf[2] = 0;
         buf[3] = 0;
      }
      buf[4] = sw[3];
      buf[5] = sw[2];
      buf[6] = sw[1];
      buf[7] = sw[0];
   }
}

inline void BaseBuffer::Swap(ULong_t x, Char_t *buf)
{
   const ULong_t *intermediary = &x;
   char *sw = (char *)intermediary;
   if (sizeof(ULong_t) == 8) {
      buf[0] = sw[7];
      buf[1] = sw[6];
      buf[2] = sw[5];
      buf[3] = sw[4];
      buf[4] = sw[3];
      buf[5] = sw[2];
      buf[6] = sw[1];
      buf[7] = sw[0];
   } else {
      buf[0] = 0;
      buf[1] = 0;
      buf[2] = 0;
      buf[3] = 0;
      buf[4] = sw[3];
      buf[5] = sw[2];
      buf[6] = sw[1];
      buf[7] = sw[0];
   }
}

inline void BaseBuffer::Swap(const Char_t *buf, UShort_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   x = Rbswap_16(*((UShort_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[1];
   sw[1] = buf[0];
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf, Short_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   x = Rbswap_16(*((Short_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[1];
   sw[1] = buf[0];
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf, UInt_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   x = Rbswap_32(*((UInt_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[3];
   sw[1] = buf[2];
   sw[2] = buf[1];
   sw[3] = buf[0];
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf, Int_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   x = Rbswap_32(*((Int_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[3];
   sw[1] = buf[2];
   sw[2] = buf[1];
   sw[3] = buf[0];
# endif
}
inline void BaseBuffer::Swap(const Char_t *buf, Float_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   // Use a union to allow strict-aliasing
   union {
      volatile UInt_t  i;
      volatile Float_t f;
   } u;
   u.i = Rbswap_32(*((UInt_t *)buf));
   x = u.f;
# else
   union {
      volatile char    c[4];
      volatile Float_t f;
   } u;
   u.c[0] = buf[3];
   u.c[1] = buf[2];
   u.c[2] = buf[1];
   u.c[3] = buf[0];
   x = u.f;
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf, Double_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   // Use a union to allow strict-aliasing
   union {
      volatile ULong64_t l;
      volatile Double_t  d;
   } u;
   u.l = Rbswap_64(*((ULong64_t *)buf));
   x = u.d;
# else
   // Use an union to prevent over-zealous optimization by KCC
   // related to aliasing double.
   // + Use a volatile here to work around error in KCC optimizer
   union {
      volatile char     c[8];
      volatile Double_t d;
   } u;
   u.c[0] = buf[7];
   u.c[1] = buf[6];
   u.c[2] = buf[5];
   u.c[3] = buf[4];
   u.c[4] = buf[3];
   u.c[5] = buf[2];
   u.c[6] = buf[1];
   u.c[7] = buf[0];
   x = u.d;
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf, ULong64_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   x = Rbswap_64(*((ULong64_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[7];
   sw[1] = buf[6];
   sw[2] = buf[5];
   sw[3] = buf[4];
   sw[4] = buf[3];
   sw[5] = buf[2];
   sw[6] = buf[1];
   sw[7] = buf[0];
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf,Long64_t &x)
{
# if ADF_STANDALONE == 0 && defined(R__USEASMSWAP)
   x = Rbswap_64(*((Long64_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[7];
   sw[1] = buf[6];
   sw[2] = buf[5];
   sw[3] = buf[4];
   sw[4] = buf[3];
   sw[5] = buf[2];
   sw[6] = buf[1];
   sw[7] = buf[0];
# endif
}

inline void BaseBuffer::Swap(const Char_t *buf,ULong_t &x)
{
   char *sw = (char *)&x;
   if (sizeof(ULong_t) == 8) {
      sw[0] = buf[7];
      sw[1] = buf[6];
      sw[2] = buf[5];
      sw[3] = buf[4];
      sw[4] = buf[3];
      sw[5] = buf[2];
      sw[6] = buf[1];
      sw[7] = buf[0];
   } else {
      sw[0] = buf[7];
      sw[1] = buf[6];
      sw[2] = buf[5];
      sw[3] = buf[4];
   }
}

inline void BaseBuffer::Swap(const Char_t *buf,Long_t &x)
{
   char *sw = (char *)&x;
   if (sizeof(Long_t) == 8) {
      sw[0] = buf[7];
      sw[1] = buf[6];
      sw[2] = buf[5];
      sw[3] = buf[4];
      sw[4] = buf[3];
      sw[5] = buf[2];
      sw[6] = buf[1];
      sw[7] = buf[0];
   } else {
      sw[0] = buf[7];
      sw[1] = buf[6];
      sw[2] = buf[5];
      sw[3] = buf[4];
   }
}



} // namespace ADF

#endif





