/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file BasicAFC.cpp compiled in libADF.so */

#ifndef ADF_BasicAFC
#include "BasicAFC.h"
#endif

#include "RunAgent.h"
#include "MetaFrame.h"

#include <iostream>
#include <iomanip>
#include <fstream>

using namespace ADF; 

BasicAFC::BasicAFC(Int_t id) : 
NarvalConsumer(), 
fEndOfFrames(aMByte,ConfAgent::kWrite),
fPath("./"), 
fBaseForName("AFC_"),
fCurrentFile(0x0),
fCurrentFileNumber(0u),
fMaxSize(kMaxInt_t),
fMyID(id),
fVertex(0x0)
{
	// log system
	Log.GetProcessName() = "BasicAFC"; 
	Log.SetPID(GetPID()); 
	
	// frameio related
	GetFrameIO().SetName("BasicAFC");
	// to be sure unknown frames are written in the ouput
	GetFrameIO().SetModel(ConfAgent::kSafe); 
	// to protect readings in the source of frames
	fEndOfFrames.SetModeIO(ConfAgent::kWrite);
	
	GetFrameIO().Attach(0x0,&fEndOfFrames);
}

BasicAFC::~BasicAFC() 
{  
	// in principle not needed ... just in case reset it has not been called by narval
	UInt_t error = 0u; process_reset(&error) ; 
} 

void BasicAFC::process_config (const Char_t *directory_path, UInt_t *error_code)
{
	// first init narval stuff
	NarvalInterface::process_config(directory_path,error_code);
	
	if ( (*error_code) == 0u ) {
		// now init your stuff if required
	}
}

Bool_t  BasicAFC::NewFile()
{
	Log.SetProcessMethod("NewFile"); 
	
	Bool_t ok = false;
	
	// close if already open
	if ( fCurrentFile ) {
		::fclose(fCurrentFile);
		fCurrentFile = 0x0;
		std::cout << "  -BasicAFP: the input file has just been closed " << fCurrentFileName << std::endl;
	}
	
	// open the file 
	std::string tmp = fPath;
	if ( tmp.size() > 0 && !(tmp.at(tmp.size()-1) == '/') ) 
		tmp += '/';
	
	std::ostringstream filename;
	filename	<< tmp << fBaseForName 
	<< std::setfill('0') << std::setw(4) << fCurrentFileNumber++ << std::setfill(' ')
	<< ".adf";
	fCurrentFileName = filename.str();
	
	fCurrentFile = ::fopen(fCurrentFileName.c_str(),"wb");
	if ( fCurrentFile != 0x0 ) {
		Log << info << "A new output file has just been open " 
		<< fCurrentFileName << nline;
		
		fEndOfFrames.SetFile(fCurrentFile,fMaxSize);
		GetFrameIO().SetStatus(BaseFrameIO::kIdle);
		
		size_t len = fCurrentFileName.length();
		size_t lpt = fCurrentFileName.find_last_of('/');
		fCurrentName = fCurrentFileName.substr(lpt+1, len-lpt-1);
		
		// add conf frame at the beginning of the file and a vertex frame.
		GetFrameIO().RecordGlobalConfiguration(); 
		
		ok = true;
	}
	else {
		Log << warning << "Cannot open file " << fCurrentFileName << nline;
		GetFrameIO().SetStatus(BaseFrameIO::kFinished); 
		ok = false;
	}
	
	Log << dolog;
	return ok;
}

UInt_t BasicAFC::ProcessBlock (ADF::FrameBlock &in)
{	
	Log.ClearMessage(); Log.SetProcessMethod("ProcessBlock"); 
	
	// attach the FrameBlocks
	GetFrameIO().Attach(&in,0x0);
	
	UInt_t nevtsIn = GetFrameIO().GetFramesWritten();
	while ( 1 ) {
		// fill the output block with the frames from the input block
		while ( GetFrameIO().Notify() ) { 	
		}
		if ( GetFrameIO().GetCurrentBlockOUT()->IsEoB() ) { // the output file is full 	
			NewFile();
			continue;
		}
		// interruption: means end of file or out is full
		if ( GetFrameIO().GetCurrentBlockIN()->IsEoB() )  // the input block is treated
			break;
	}
	UInt_t nevtsOut = GetFrameIO().GetFramesWritten();
	// 
	GetFrameIO().Detach(&in,0x0);	
	
	if(fMyID >= 0) {
		std::cout << std::setw(2) << fMyID;
		std::cout << std::left << std::setw(24) << "-BasicAFC:" << std::right
      << " " << std::setw(5) << nevtsOut-nevtsIn
      << " evts (" << std::setw(8) << in.GetSize() << ")"
      << " Tot = " << std::setw(8) << nevtsOut
      << "           " << fCurrentName << std::endl;
	}
	return 0u;
}


void BasicAFC::process_initialise (UInt_t *error_code)
{
	*error_code = 0;
	Log.ClearMessage(); Log.SetProcessMethod("process_initialise"); 
	
	// creates a trigger to register at the beginning of the file a global vertex frame
	// that is initiated from ADF.conf
	fVertex = AgataFrameTrigger::Build("GVertex","","meta:vertex");
	if ( fVertex )
		GetFrameIO().Register(fVertex);
	
	// read an input file to get the base filename and the path  
	std::string conffile = GetConfPath() + "BasicAFC.conf";
	
	std::ifstream filein(conffile.data());
	if ( filein.is_open() == true ) { 
		
		std::string pathforfiles, basename; 
		UInt_t starting_number;
		filein >> pathforfiles >> basename >> starting_number;
		
		if ( filein.good() ) {
			fPath = pathforfiles;
			fBaseForName = basename;
			fCurrentFileNumber = starting_number;
			//	GetFrameIO().GetConfAgent()->GetRunAgent()->SetSubRun(starting_number);
		}
		filein.close();
	}
	
	// open the first file
	if ( !NewFile() )
		*error_code = 1;
	
	Log << dolog;		
}

void BasicAFC::process_reset (UInt_t *error_code)
{
	*error_code = 0;
	Log.ClearMessage(); Log.SetProcessMethod("process_reset");
	
	if ( fCurrentFile )
		::fclose(fCurrentFile);
	
	GetFrameIO().Print( Log() );
	
	Log << dolog;		
}

/*
 void BasicAFC::process_unload (UInt_terror_code)
 {
 std::cout << "process_unload called with GetPID()" << GetPID() << std::endl;
 *error_code = 0;
 } */





