/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file BufferIO.h header file for BufferIO.cpp */

#ifndef ADF_BufferIO
#define ADF_BufferIO
#define ADF_BUFFERIO_H ADF_BufferIO

#ifndef ADF_BaseBuffer
#include "BaseBuffer.h"
#endif

#include <stdio.h>
#include <iostream>
#include <string>

namespace ADF {

//! A buffer is used to read/write basic data from/to buffers. With minimal requirements, it takes also into account byte swapping
/*!
	A BufferIO wraps a 
	
	Copy methods ==> similar behavior that Link  
	
	It is the charge of the user to reset the status flag if it goes to kFail.
	
	@author Olivier Stezowski
*/

class CompositeFrame; 

class BufferIO : protected BaseBuffer
{
public:
   	friend class CompositeFrame;    
    
protected:    
	//! current status
	mutable EStatus fStatus;
	//! current endian    
	EEndian fEndian;  
      
private:
	//! internal buffer owned by this and thus deleted
	Char_t * fInner;
	//! external buffer in case the attach method is used. NOT DELETED
	Char_t * fOuter;
	//! Real size for the inner buffer
	UInt_t fRealInnerSize;	
	//! Real size for the outer buffer  
	UInt_t fRealOuterSize;	 
             
protected:    
	//!  current buffer in which data are read/written
	Char_t * fBufferIO;
	//! effective size of the current buffer	
	UInt_t fSize;
	//! Real size of the current buffer 	
	UInt_t fRealSize;	
	
	//! current position in the buffer (mutable so that reading are allowed for constant)
	mutable UInt_t fCurrent; 
    
protected:            
	void AllocInner(UInt_t);    
	void SetEndian(BaseBuffer::EEndian e) 
		{ fEndian = e; }
		
	//! to switch between inner and outer   
	void SetCurrent(const char = 'i');  
        
public:
	//! Default size for a BufferIO is one byte just because it could be used only with an external buffer
	/*!
	*/
	explicit BufferIO(UInt_t s = aByte);
	virtual ~BufferIO();
	
	//! for classes that needs it to write directly data into it. 
	Char_t *Address() 
		{ return fBufferIO; }    
	//! for classes that needs it to write directly data into it. 
	Char_t *CurrentAddress() 
		{ return fBufferIO + Offset(); }  
	
	Bool_t IsStatus(BufferIO::EStatus s) const 
		{ return fStatus == s; }
	void  SetStatus(BufferIO::EStatus s = BaseBuffer::kGood) 
		{ fStatus = s;}
	
	BufferIO::EStatus GetStatus() const 
		{ return fStatus; }
	
	BaseBuffer::EEndian GetEndian() const 
		{ return fEndian; }
	Bool_t IsEndian(BaseBuffer::EEndian e) const 
		{ return fEndian == e ; }
	
	//! Link an external buffer to this class. 
	/*!
		An external buffer is linked to used the services of BufferIO i.e. reading/writing.
		In case of success, it returns true
		In case the pointer in NULL, it is not linked and the current buffer is not changed 
		(returns false).
		
		Offset is set to 0 (beginning of the buffer)
		
		\warning
		It is the charge of the users to pass the right size of the external buffer!
		
		The linked buffer is not owned by BufferIO, it will NOT be deleted with BufferIO
	*/    
	virtual Bool_t Link(const Char_t *from, UInt_t from_size);
	
	//! copy an external buffer to this. 
	/*!
		It works the same way than Link and thus propose a safer method
		(a external buffer may be deleted or modified during the processing of that one)
		
		As a consequence:
		- switch to inner buffer (in case you don't want this, use Import/Export methods instead)
		- It starts the copy from the beginning in the destination buffer 
		- Try to expand if the current size in too small.
		- Offset is set to 0 (beginning of the buffer)
	*/    
	virtual Bool_t Copy(const Char_t *from, UInt_t from_size);
	
	//! Link a buffer to this buffer
	/*!
	*/    
	virtual Bool_t Link(const BufferIO &buf, UInt_t pos, UInt_t size);
	
	//! Unlink the external buffer to this class. 
	/*!   
		It returns the external buffer that has been linked using the Link method.
		It returns 0 if no external buffer is Linked. The inner buffer becomes the current buffer and the 
		current position is set to O.
	*/
	virtual Char_t *Unlink();   
	
	//! check if this is linked or not 
	virtual Bool_t IsLinked() 
		{ return fBufferIO != fInner; }
	
	//! copy from a buffer to another one
	/*!
		Conventions: copy n bytes from the first buffer (from offset off_from) to the second one 
		(current position). \n
		Returns false if something goes wrong. 
		
		No Expand called, the user must do it himself. 
		The offset in the destination buffer is moved to the new position.
	*/
	static Bool_t Copy(const BufferIO &from, BufferIO &to, UInt_t size, UInt_t off_from = 0u) 
	{ 
		if ( off_from + size > from.Size() ) 
			return false;
		if ( size > to.FreeSize() ) 
			return false ; 
		BaseBuffer::Copy(from.GetAddress()+off_from,to.CurrentAddress(),size); 
		to.SetOffset(to.Offset()+size); 
		return true; 
	}
		
	//! copy a buffer to another one
	/*!
		Conventions: copy n bytes from the first buffer (from the beginning) 
		to the second one (current position). \n
		Returns false if the destination buffer is to small. No Expand called, the user must do it himself
	*/
	/* static Bool_t Copy(const BufferIO &from, BufferIO &to, UInt_t size) 
		{ if ( size > to.FreeSize() ) 
			return false ; 
		BaseBuffer::Copy(from.GetAddress(),to.GetCurrentAddress(),size); 
		to.SetOffset(to.Offset()+size); 
		return true; 
		} 
	*/
			
	//! copy a buffer to another one
	/*!
		Conventions: copy all bytes from the first buffer (from the beginning) 
		to the second one (current position). \n
		Returns false if the destination buffer is to small. No Expand called, the user must do it himself
	*/
	static Bool_t Copy(const BufferIO &from, BufferIO &to) 
		{ return Copy(from,to,from.Size()); }    
		
	//! return the right buffer taking Into account the endian argument compared to the underlying endian system type. 
	/*!
	
	*/    
	static BufferIO *New(BaseBuffer::EEndian e, UInt_t s = aByte);
	
	//! Pointer to the current underlying array of bytes.
	/*!
	*/
	const Char_t *GetAddress() const 
		{ return fBufferIO; }
	
	//! it returns the maximum number of bytes in this buffer   
	UInt_t Size() const 
		{ return fSize; }
	
	//! in case the buffer is partly filled and you would like to read it again
	/*!
		If the given size is higher than the current size, it does not affect
		the size of the buffer. Otherwise, the size of the buffer is set to the given value.
		
		It returns the size of the buffer after application of this method.
		
		To come back with a situation with Real and Effective size equal, 
		just call SetEffectiveSize();
	*/           
	UInt_t SetEffectiveSize(UInt_t size = kMaxUInt_t);   
		
	//! it returns the number of free bytes to the end
	UInt_t FreeSize() const 
		{ return fSize - fCurrent; }  
	
	//! it returns the current position in the buffer   
	UInt_t Offset() const 
		{ return fCurrent ; }
	
	//! change the current position.
	/*!
		If the given offset is greater than the size of the buffer, it sets the current offset 
		at the end of the buffer.
		
		It returns the current position in the buffer
	*/ 
	UInt_t SetOffset(UInt_t off = 0u) const;
	
	//! change the current position.   
	/*!
		if too small, set to 0. if too high, set to max size
	*/ 
	UInt_t SetOffset(const Char_t *off) const;   
	
	//!	check if this address belongs to this buffer
	/*!
		if true, pos is set to the corresponding position in the buffer
	*/
	Bool_t IsAnOffset(const Char_t *off, UInt_t &pos) const;
	
	//! Reset means set all elements to 0 and the current position is 0
	/*!	
	*/
	virtual void Reset(); 
	virtual void FastReset();
	
	//! Expand the size of the current buffer
	/*!
		This method expands the size of the current buffer. \n
		if the real size of the buffer is enough, this method is applied the same 
		way for the inner part or the outer part. 
		
		In case the size is too small and a real reallocation is required to expand
		 it is ONLY applied if the current buffer is the inner one.
		
		The argument corresponds to the new-size of the buffer \n
		It returns false if:
		- one try to Expand the outer buffer
		- it fails to expand the buffer (the final size is to high)
	*/   
	virtual Bool_t Expand(UInt_t new_size);

	//! to reserve a given size (from the current position) to be sure one are able to write something
	virtual Bool_t Reserve(UInt_t size);
		
	virtual BufferIO &operator << (Char_t);
	virtual BufferIO &operator << (UChar_t);
	virtual BufferIO &operator << (Short_t);
	virtual BufferIO &operator << (UShort_t);
	virtual BufferIO &operator << (Int_t);
	virtual BufferIO &operator << (UInt_t);
	virtual BufferIO &operator << (Float_t);
	virtual BufferIO &operator << (Double_t);
	virtual BufferIO &operator << (Long_t);
	virtual BufferIO &operator << (ULong_t);        
 	virtual BufferIO &operator << (Long64_t);
	virtual BufferIO &operator << (ULong64_t);
	
	//! string can be handled BUT only with size less that 2^16 chars. Use Import/Export method otherwise
	virtual BufferIO &operator << (std::string);
	
	virtual const BufferIO &operator>> (Char_t &)  const;
	virtual const BufferIO &operator>> (UChar_t &) const;
	virtual const BufferIO &operator>> (Short_t &) const;
	virtual const BufferIO &operator>> (UShort_t &) const;
	virtual const BufferIO &operator>> (Int_t &) const;
	virtual const BufferIO &operator>> (UInt_t &) const;
	virtual const BufferIO &operator>> (Float_t &) const;
	virtual const BufferIO &operator>> (Double_t &) const; 
	virtual const BufferIO &operator>> (Long_t &) const;
	virtual const BufferIO &operator>> (ULong_t &) const;
	virtual const BufferIO &operator>> (Long64_t &) const;
	virtual const BufferIO &operator>> (ULong64_t &) const;
	
	//! string can be handled BUT only with size less that 2^16 chars. Use Import/Export method otherwise
	virtual const BufferIO &operator>> (std::string &) const;

	
	//! add the same character how_many times from the current position
	virtual UInt_t FillWith(Char_t c, UInt_t how_many);

	//! Export this buffer to an array 
	/*!
		The second arguments is the number of elements to be exported.
		It returns the number of BYTES effectively extracted from the buffer
	*/
	virtual UInt_t Export(Char_t *, UInt_t) const;	
	virtual UInt_t Export(UShort_t *, UInt_t) const;
	virtual UInt_t Export(Short_t *, UInt_t) const;
	virtual UInt_t Export(UInt_t *, UInt_t) const;
	virtual UInt_t Export(Int_t *, UInt_t) const;
	virtual UInt_t Export(Float_t *, UInt_t) const;
		
	//! Export this buffer to an array of doubles
	/*!
		It returns the number of bytes effectively extracted from the buffer
	*/
	virtual UInt_t Export(Double_t *, UInt_t) const;
	

	//! Import the given array in this buffer
	/*!
		It returns the number of bytes effectively written in the buffer
	*/    
	UInt_t Import(const Char_t *from, UInt_t size_ext_buf);
	
	//! Import the given array in this buffer
	/*!
		It returns the number of bytes effectively written in the buffer
	*/
	virtual UInt_t Import(const UShort_t *, UInt_t);
	virtual UInt_t Import(const Short_t *, UInt_t);
	virtual UInt_t Import(const UInt_t *, UInt_t);
	virtual UInt_t Import(const Int_t *, UInt_t);	
	virtual UInt_t Import(const Float_t *, UInt_t);
	//! Import the given array in this buffer
	/*!
		It returns the number of bytes effectively written in the buffer
	*/	
	virtual UInt_t Import(const Double_t *, UInt_t);
	
	//! Import in this buffer the content of another buffer
	/*!
		It returns the number of bytes written in the buffer: 
	*/    
	virtual UInt_t Import(const BufferIO &, UInt_t);
		
	//! Import in the buffer data coming from a C file
	/*!
		It reads n bytes into a C file added from the current position to the buffer. 
		So if you need to start from the beginning, do SetOffset() before calling this method.
		
		It returns the number of bytes written in the buffer
	*/    
	virtual UInt_t Import(FILE *, UInt_t);  
	
	//! Export this buffer to a C files
	/*!
		It writes n bytes into a C file from the beginning of the buffer. 
		It returns the number of bytes effectively written in the file
	*/
	virtual UInt_t Export(FILE *, UInt_t) const;   
	
	//! Import in the buffer data coming from a C++ stream     
	virtual UInt_t Import(std::istream &, UInt_t);
	
	//! Export this buffer          
	virtual UInt_t Export(std::ostream &, UInt_t) const ;       
};
// inline BufferIO members
inline BufferIO &BufferIO::operator << (Char_t data) 
{ 
	if (FreeSize()>=sizeof(Char_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(Char_t)); 
		fCurrent += sizeof(Char_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &BufferIO::operator << (UChar_t data)
{ 
	if (FreeSize()>=sizeof(Char_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(UChar_t)); 
		fCurrent += sizeof(Char_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);
}
inline BufferIO &BufferIO::operator << (Short_t data) 
{
	if (FreeSize()>=sizeof(Short_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(Short_t)); 
		fCurrent += sizeof(Short_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &BufferIO::operator << (UShort_t data)
{
	if (FreeSize()>=sizeof(Short_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(UShort_t)); 
		fCurrent += sizeof(UShort_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &BufferIO::operator << (Int_t data) 
{
	if (FreeSize()>=sizeof(Int_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(Int_t)); 
		fCurrent += sizeof(Int_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &BufferIO::operator << (UInt_t data)
{
	if (FreeSize()>=sizeof(Int_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(UInt_t)); 
		fCurrent += sizeof(UInt_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &BufferIO::operator << (Float_t data) 
{ 
	if (FreeSize()>=sizeof(Float_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(Float_t)); 
		fCurrent += sizeof(Float_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &BufferIO::operator << (Double_t data) 
{
	if (FreeSize()>=sizeof(Double_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(Double_t));
		fCurrent += sizeof(Double_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}

inline BufferIO &BufferIO::operator << (Long_t data) 
{
	// long are written on 64 bits whatever the system
	if (FreeSize()>=sizeof(Long64_t)) { 
		if ( sizeof(Long_t) == 8 ) {
			::memcpy((fBufferIO+fCurrent),&data,sizeof(Long_t)); 
		}
		else {
			if (data < 0) {
				fBufferIO[fCurrent+0] = (char) -1;
				fBufferIO[fCurrent+1] = (char) -1;
				fBufferIO[fCurrent+2] = (char) -1;
				fBufferIO[fCurrent+3] = (char) -1;
			} else {
				fBufferIO[fCurrent+0] = 0;
				fBufferIO[fCurrent+1] = 0;
				fBufferIO[fCurrent+2] = 0;
				fBufferIO[fCurrent+3] = 0;
			}
			::memcpy((fBufferIO+fCurrent+4), &data, 4);
		}
		fCurrent += sizeof(Long64_t);	
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}

inline BufferIO &BufferIO::operator << (ULong_t data) 
{
	if (FreeSize()>=sizeof(ULong64_t)){ 
		if ( sizeof(ULong_t) == 8 ) {
			::memcpy((fBufferIO+fCurrent),&data,sizeof(ULong_t)); 
		}
		else {
//			if (data < 0) {
//				fBufferIO[fCurrent+0] = (char) -1;
//				fBufferIO[fCurrent+1] = (char) -1;
//				fBufferIO[fCurrent+2] = (char) -1;
//				fBufferIO[fCurrent+3] = (char) -1;
//			} else {
				fBufferIO[fCurrent+0] = 0;
				fBufferIO[fCurrent+1] = 0;
				fBufferIO[fCurrent+2] = 0;
				fBufferIO[fCurrent+3] = 0;
//			}
			::memcpy((fBufferIO+fCurrent+4), &data, 4);
		}
		fCurrent += sizeof(ULong64_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}

inline BufferIO &BufferIO::operator << (Long64_t data) 
{
	if (FreeSize()>=sizeof(Long64_t)) { 
		::memcpy((fBufferIO+fCurrent),&data,sizeof(Long64_t)); 
		fCurrent += sizeof(Long64_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &BufferIO::operator << (ULong64_t data) 
{
	if (FreeSize()>=sizeof(ULong64_t)) {
		 ::memcpy((fBufferIO+fCurrent),&data,sizeof(ULong64_t)); 
		 fCurrent += sizeof(ULong64_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &BufferIO::operator << (std::string data) 
{
	UShort_t size_string = (UShort_t)data.size();
	UShort_t size_needed = size_string + sizeof(size_needed);
	
	if (FreeSize()>=size_needed) {
		::memcpy((fBufferIO+fCurrent),&size_string,sizeof(size_needed)); 
		::memcpy((fBufferIO+fCurrent+sizeof(size_needed)),data.data(),size_string); 

		fCurrent += size_needed;
	}	
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline const BufferIO &BufferIO::operator >> (Char_t &data) const 
{
	if (FreeSize()>=sizeof(Char_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(Char_t));
		 // data = fBufferIO[fCurrent];	
		fCurrent += sizeof(Char_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &BufferIO::operator >> (UChar_t &data) const
{
	if (FreeSize()>=sizeof(Char_t)) {
		 ::memcpy(&data,fBufferIO+fCurrent,sizeof(UChar_t));
		 // data = (*((UChar_t *)(fBufferIO+fCurrent)));
		 fCurrent += sizeof(UChar_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &BufferIO::operator >> (Short_t &data) const
{
	if (FreeSize()>=sizeof(Short_t)) { 	
		::memcpy(&data,fBufferIO+fCurrent,sizeof(Short_t));
		// data = (*((Short_t *)(fBufferIO+fCurrent)));		
		fCurrent += sizeof(Short_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &BufferIO::operator >> (UShort_t &data) const
{
	if (FreeSize()>=sizeof(Short_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(UShort_t));
	//	data = (*((UShort_t *)(fBufferIO+fCurrent)));
		fCurrent += sizeof(UShort_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO & BufferIO::operator >> (Int_t &data) const
{
	if (FreeSize()>=sizeof(Int_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(Int_t));
		// data = (*((Int_t *)(fBufferIO+fCurrent)));		
		fCurrent += sizeof(Int_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this) ;
}
inline const BufferIO &BufferIO::operator >> (UInt_t &data) const
{
	if (FreeSize()>=sizeof(Int_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(UInt_t));
		// data = (*((UInt_t *)(fBufferIO+fCurrent)));
		fCurrent += sizeof(UInt_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &BufferIO::operator >> (Float_t &data) const
{
	if (FreeSize()>=sizeof(Float_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(Float_t));
		// data = (*((Float_t *)(fBufferIO+fCurrent)));
		fCurrent += sizeof(Float_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &BufferIO::operator >> (Double_t &data) const
{
	if (FreeSize()>=sizeof(Double_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(Double_t));
		// data = (*((Double_t *)(fBufferIO+fCurrent)));		
		fCurrent += sizeof(Double_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO & BufferIO::operator >> (Long_t &data) const
{
	if (FreeSize()>=sizeof(Long64_t)) {
   		if (sizeof(Long_t) == 8) 	
			::memcpy(&data,fBufferIO+fCurrent,sizeof(Long_t));
		else 
      			::memcpy(&data,fBufferIO+fCurrent+4,4);	
		
		fCurrent += sizeof(Long64_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this) ;
}
inline const BufferIO & BufferIO::operator >> (ULong_t &data) const
{
	if (FreeSize()>=sizeof(ULong64_t)) {
   		if (sizeof(ULong_t) == 8) 	
			::memcpy(&data,fBufferIO+fCurrent,sizeof(ULong_t));
		else 
      			::memcpy(&data,fBufferIO+fCurrent+4,4);	
		
		fCurrent += sizeof(ULong64_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this) ;
}

inline const BufferIO & BufferIO::operator >> (Long64_t &data) const
{
	if (FreeSize()>=sizeof(Long64_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(Long64_t));
		// data = (*((Long64_t *)(fBufferIO+fCurrent)));		
		fCurrent += sizeof(Long64_t); 
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this) ;
}
inline const BufferIO & BufferIO::operator >> (ULong64_t &data) const
{
	if (FreeSize()>=sizeof(ULong64_t)) { 
		::memcpy(&data,fBufferIO+fCurrent,sizeof(ULong64_t));
		// data = (*((ULong64_t *)(fBufferIO+fCurrent)));		
		fCurrent += sizeof(ULong64_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this) ;
}
inline const BufferIO &BufferIO::operator >> (std::string &data) const
{
	UShort_t size_string;
	if ( FreeSize()>=sizeof(size_string) ){ 
		::memcpy(&size_string,fBufferIO+fCurrent,sizeof(size_string));
		if (FreeSize()>=size_string) 
		{
			data.assign(fBufferIO+fCurrent+sizeof(size_string),size_string);
			fCurrent += ( size_string + sizeof(size_string) );
		}
		else fStatus = BaseBuffer::kFail ;
	}
	else fStatus = BaseBuffer::kFail ;			
	return (*this);
}
inline UInt_t BufferIO::Export(Char_t *output, UInt_t size) const
{
	UInt_t bread = size;
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::Export(UShort_t *output, UInt_t size) const
{
	UInt_t bread = size*sizeof(UShort_t);
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::Export(Short_t *output, UInt_t size) const
{
	UInt_t bread = size*sizeof(Short_t);
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::Export(UInt_t *output, UInt_t size) const
{
	UInt_t bread = size*sizeof(UInt_t);
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::Export(Int_t *output, UInt_t size) const
{
	UInt_t bread = size*sizeof(Int_t);
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::Export(Float_t *output, UInt_t size) const
{
	UInt_t bread = size*sizeof(Float_t);
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::Export(Double_t *output, UInt_t size) const
{
	UInt_t bread = size*sizeof(Double_t);
	
	if (FreeSize()>=bread) 
		{ BaseBuffer::Copy(fBufferIO+fCurrent,output,bread); fCurrent += bread; }
	else { fStatus = BaseBuffer::kFail ; bread = 0u ; }
	
	return bread;
}
inline UInt_t BufferIO::FillWith(Char_t c, UInt_t how_many)
{
	UInt_t bwritten;
	how_many > FreeSize() ? bwritten = FreeSize() : bwritten = how_many ;

	::memset(fBufferIO+fCurrent,c,bwritten); 
	fCurrent += bwritten; 
	
	return bwritten;
}
inline UInt_t BufferIO::Import(const UShort_t *input, UInt_t size)
{
	UInt_t bwritten = size*sizeof(UShort_t);
	
	if (FreeSize()>=bwritten) 
		{ BaseBuffer::Copy(input,fBufferIO+fCurrent,bwritten); fCurrent += bwritten; }
	else { fStatus = BaseBuffer::kFail ; bwritten = 0u ; }
	
	return bwritten;
}
inline UInt_t BufferIO::Import(const Short_t *input, UInt_t size)
{
	UInt_t bwritten = size*sizeof(Short_t);
	
	if (FreeSize()>=bwritten) 
		{ BaseBuffer::Copy(input,fBufferIO+fCurrent,bwritten); fCurrent += bwritten; }
	else { fStatus = BaseBuffer::kFail ; bwritten = 0u ; }
	
	return bwritten;
}  
inline UInt_t BufferIO::Import(const UInt_t *input, UInt_t size)
{
	UInt_t bwritten = size*sizeof(UInt_t);
	
	if (FreeSize()>=bwritten) 
		{ BaseBuffer::Copy(input,fBufferIO+fCurrent,bwritten); fCurrent += bwritten; }
	else { fStatus = BaseBuffer::kFail ; bwritten = 0u ; }
	
	return bwritten;
}
inline UInt_t BufferIO::Import(const Int_t *input, UInt_t size)
{
	UInt_t bwritten = size*sizeof(Int_t);
	
	if (FreeSize()>=bwritten) 
		{ BaseBuffer::Copy(input,fBufferIO+fCurrent,bwritten); fCurrent += bwritten; }
	else { fStatus = BaseBuffer::kFail ; bwritten = 0u ; }
	
	return bwritten;
}
inline UInt_t BufferIO::Import(const Float_t *input, UInt_t size)
{
	UInt_t bwritten = size*sizeof(Float_t);
	
	if (FreeSize()>=bwritten) 
		{ BaseBuffer::Copy(input,fBufferIO+fCurrent,bwritten); fCurrent += bwritten; }
	else { fStatus = BaseBuffer::kFail ; bwritten = 0u ; }
	
	return bwritten;
} 
inline UInt_t BufferIO::Import(const Double_t *input, UInt_t size)
{
	UInt_t bwritten = size*sizeof(Double_t);
	
	if (FreeSize()>=bwritten) 
		{ BaseBuffer::Copy(input,fBufferIO+fCurrent,bwritten); fCurrent += bwritten; }
	else { fStatus = BaseBuffer::kFail ; bwritten = 0u ; }
	
	return bwritten;
}   

// inline BufferIO members //////////////////////////

//! A Endianbuffer is used to read/write raw data buffers from/on files
/*!
	Should never be used directly. Use BufferIO::New instead
	
	@author Olivier Stezowski
*/
class EndianBufferIO : public BufferIO
{  
public:
	EndianBufferIO(UInt_t s = aKByte);
	virtual ~EndianBufferIO();

	virtual BufferIO &operator << (Char_t) ;
	virtual BufferIO &operator << (UChar_t);
	virtual BufferIO &operator << (Short_t);
	virtual BufferIO &operator << (UShort_t);
	virtual BufferIO &operator << (Int_t);
	virtual BufferIO &operator << (UInt_t);
	virtual BufferIO &operator << (Float_t);
	virtual BufferIO &operator << (Double_t);
	virtual BufferIO &operator << (Long64_t);
	virtual BufferIO &operator << (ULong64_t);
	virtual BufferIO &operator << (Long_t);
	virtual BufferIO &operator << (ULong_t);

	virtual BufferIO &operator << (std::string);

	virtual const BufferIO &operator >> (Char_t &) const;
	virtual const BufferIO &operator >> (UChar_t &) const;    
	virtual const BufferIO &operator >> (Short_t &) const;
	virtual const BufferIO &operator >> (UShort_t &) const;
	virtual const BufferIO &operator >> (Int_t &) const;
	virtual const BufferIO &operator >> (UInt_t &) const;
	virtual const BufferIO &operator >> (Float_t &) const;
	virtual const BufferIO &operator >> (Double_t &) const;
	virtual const BufferIO &operator >> (Long64_t &) const;
	virtual const BufferIO &operator >> (ULong64_t &) const;
	virtual const BufferIO &operator >> (Long_t &) const;
	virtual const BufferIO &operator >> (ULong_t &) const;	
	
	virtual const BufferIO &operator>> (std::string &) const;

};
// inline EndianBufferIO members
inline BufferIO &EndianBufferIO::operator << (Char_t data)  
	{ return BufferIO::operator << (data) ;}
inline BufferIO &EndianBufferIO::operator << (UChar_t data) 
	{ return BufferIO::operator << (data); }    
inline BufferIO &EndianBufferIO::operator << (Short_t data) 
{
	if (FreeSize()>=sizeof(Short_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(Short_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &EndianBufferIO::operator << (UShort_t data)
{
	if (FreeSize()>=sizeof(UShort_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent); 
		fCurrent += sizeof(UShort_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &EndianBufferIO::operator << (Int_t data)
{
	if (FreeSize()>=sizeof(Int_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(Int_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline BufferIO &EndianBufferIO::operator << (UInt_t data)
{
	if (FreeSize()>=sizeof(UInt_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(UInt_t); 
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &EndianBufferIO::operator << (Float_t data)
{ 
	if (FreeSize()>=sizeof(Float_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent); 
		fCurrent += sizeof(Float_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &EndianBufferIO::operator << (Double_t data) 
{
	if (FreeSize()>=sizeof(Double_t)){  
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(Double_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline BufferIO &EndianBufferIO::operator << (Long64_t data)
{ 
	if (FreeSize()>=sizeof(Long64_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(Long64_t); 
	}
	else fStatus = BaseBuffer::kFail ; 
	return (*this);	
}
inline BufferIO &EndianBufferIO::operator << (ULong64_t data) 
{
	if (FreeSize()>=sizeof(ULong64_t)){ 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(ULong64_t); 
	}
	else fStatus = BaseBuffer::kFail ; 
	return (*this);
}
inline BufferIO &EndianBufferIO::operator << (Long_t data)
{ 
	if (FreeSize()>=sizeof(Long64_t)) { 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(Long64_t); 
	}
	else fStatus = BaseBuffer::kFail ; 
	return (*this);	
}
inline BufferIO &EndianBufferIO::operator << (ULong_t data) 
{
	if (FreeSize()>=sizeof(ULong64_t)){ 
		BaseBuffer::Swap(data,fBufferIO+fCurrent);
		fCurrent += sizeof(ULong64_t); 
	}
	else fStatus = BaseBuffer::kFail ; 
	return (*this);
}
inline BufferIO &EndianBufferIO::operator << (std::string data) 
{
	UShort_t size_string = (UShort_t)data.size();
	UShort_t size_needed = size_string + sizeof(size_needed);
	
	if (FreeSize()>=size_needed) {
		BaseBuffer::Swap(size_string,fBufferIO+fCurrent);			
		::memcpy((fBufferIO+fCurrent+sizeof(size_needed)),data.data(),size_string); 
		
		fCurrent += size_needed;
	}	
	else fStatus = BaseBuffer::kFail ;
	return (*this);
}
inline const BufferIO &EndianBufferIO::operator >> (Char_t &data)  const
{ 
	return BufferIO::operator >> (data) ;
}
inline const BufferIO &EndianBufferIO::operator >> (UChar_t &data)  const
{ 
	return BufferIO::operator >> (data); 
}    
inline const BufferIO &EndianBufferIO::operator >> (Short_t &data) const
{
	if (FreeSize()>=sizeof(Short_t)){ 
		BaseBuffer::Swap(fBufferIO+fCurrent,data);
		fCurrent += sizeof(Short_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (UShort_t &data) const
{
	if (FreeSize()>=sizeof(UShort_t)){ 
		BaseBuffer::Swap(fBufferIO+fCurrent,data);
		fCurrent += sizeof(UShort_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (Int_t &data) const
{
	if (FreeSize()>=sizeof(Int_t)){ 
		BaseBuffer::Swap(fBufferIO+fCurrent,data);  
		fCurrent += sizeof(Int_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (UInt_t &data) const
{
	if (FreeSize()>=sizeof(UInt_t)){ 
		BaseBuffer::Swap(fBufferIO+fCurrent,data);
		fCurrent += sizeof(UInt_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}

inline const BufferIO &EndianBufferIO::operator >> (Long64_t &data) const
{
	if (FreeSize()>=sizeof(Long64_t)){  
		BaseBuffer::Swap(fBufferIO+fCurrent,data); 
		fCurrent += sizeof(Long64_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (ULong64_t &data) const
{
	if (FreeSize()>=sizeof(ULong64_t)){
		BaseBuffer::Swap(fBufferIO+fCurrent,data); 
		fCurrent += sizeof(ULong64_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}

inline const BufferIO &EndianBufferIO::operator >> (Float_t &data) const
{
	if (FreeSize()>=sizeof(Float_t)){
		BaseBuffer::Swap(fBufferIO+fCurrent,data); 
		fCurrent += sizeof(Float_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (Double_t &data) const
{
	if (FreeSize()>=sizeof(Double_t)){
		BaseBuffer::Swap(fBufferIO+fCurrent,data);
		fCurrent += sizeof(Double_t);
	}
	else fStatus = BaseBuffer::kFail ;	
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (Long_t &data) const
{
	if (FreeSize()>=sizeof(Long64_t)){  
		BaseBuffer::Swap(fBufferIO+fCurrent,data); 
		fCurrent += sizeof(Long64_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator >> (ULong_t &data) const
{
	if (FreeSize()>=sizeof(ULong64_t)){
		BaseBuffer::Swap(fBufferIO+fCurrent,data);
		fCurrent += sizeof(ULong64_t);
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}
inline const BufferIO &EndianBufferIO::operator>> (std::string &data) const
{
	UShort_t size_string;
	
	if (FreeSize()>=sizeof(size_string)){
		BaseBuffer::Swap(fBufferIO+fCurrent,size_string);
		if (FreeSize()>=size_string) 
		{
			data.assign(fBufferIO+fCurrent+sizeof(size_string),size_string);
			fCurrent += ( size_string + sizeof(size_string) );
		}
		else fStatus = BaseBuffer::kFail ;		
	}
	else fStatus = BaseBuffer::kFail ;
	return (*this);	
}

}; // ADF namespace

#endif
