/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file Frame.h header file for Frame.cpp */

#ifndef ADF_CompositeFrame
#define ADF_CompositeFrame

#ifndef ADF_Frame
#include "Frame.h"
#endif

#ifndef ADF_PtrStack
#include "PtrStack.h"
#endif

namespace ADF
{ 

//! Base class for a CompositeFrame
/*!
	A composite Frame is a frame composed of sub-frames. Thus, this interface proposes 
	methods to manipulate such sub-frames (access, extraction, add etc ..)
	
	As a Frame, a composite frame starts with a Key. A second Key characterises a composite
	frame since all sub-frames in it start with this second key type. Of course, the same kind
	of key could be used in both cases. That is why two constructors are provided 
	for classes that inherits from CompositeFrame:
	
	- CompositeFrame(const Key *)
	- CompositeFrame(const Key *, FactoryItem &)
		
	If only one key is given in the constructor, 
*/
class CompositeFrame : public ConcreteFrame
{   
protected:
	//! Key used to scan the main frame (link to parts of the main frame)
	Key *fSubKey;      
	//! ConstKey remains unchanged (unlinked) to be sure the comparison IsAKey is always good
	Key *fSubKeyConst;     
    
private:    
	//! list of current keys composing this frame
	PtrStack<Key> fCurrentKeys; 
	//! stack of created keys to avoid too many new/delete cycles
	PtrStack<Key> fStackOfKeys;  
    
protected:    
	//!     
	CompositeFrame(Key *);    
	CompositeFrame(Key *,const FactoryItem &);   
	
	//! to get a new subkey  
	Key *NewSubKey();
		    
public:      
	virtual ~CompositeFrame();
		
	//! tells if this frame is a composite frame i.e. if it is composed of sub-frames
	/*!
	*/   
	virtual Bool_t IsComposite() const 
		{ return true; }
		
	//! Reads from the CompositeFrame the content into dedicated structures
	/*!
		return the number of bytes read
	*/    
	virtual UInt_t Read() 
		{ return Scan(); }
		
	virtual UInt_t Write() 
		{ return Scan(); }
	
	//! Scan this CompositeFrame. If it is a composite frame, it looks for the keys of sub-frames
	/*!
	*/   
	virtual UInt_t Scan();   
	
	//! to get one of the key composing the buffer. if out of range, return the main key
	virtual const Key *GetSubKey(UInt_t which) const { 
		Key *key = fCurrentKeys.At(which); 
		if ( key ) 
			return key; 
		return fKey; 
	}
	
	//! Returns the number of sub-frames composing this frame. Scan have to be called first
	/*!
	*/   
	virtual UInt_t GetNbSubFrame() const 
		{ return fCurrentKeys.GetSize(); }    
		
	//! to check if a subframe could be added to the composite frame   
	/*!
	
	*/
	virtual Bool_t IsASubFrame(const Frame &);	
		
	//! Add a subframe to that frame (only if composite)     
	/*!
	
	*/
	virtual Bool_t AddSubFrame(const Frame &);
	
	//! Attach the sub-frame corresponding to the given subkey to the Frame given in the second argument
	/*!
		The data part of the sub-frame is linked to the frame given as an argument.
		Its key is modified just to set the correct length using GetDataLength. The Key part is not linked
		because one may need to get a sub-frame 
	*/   
	virtual Bool_t LinkSubFrame(UInt_t , Frame *) ;      
		
	//! copy the sub-frame #i to the CompositeFrame given in the second argument
	/*!
	*/   
	//virtual Bool_t CopySubFrame(UInt_t, Frame *) { return false; }   
	
	//! transfert subframe whichsub from from_frame to this 
	virtual Bool_t TransfertSubFrame(const Frame &from, UInt_t whichsub);
			
	//! Reset the current frame
	/*!
		remove all keys from the list of sub-keys, set the full buffer to 0 and offset to 0
		Set length in key to 0.
		
		If you call the method often, setting the full buffer to 0 may take time. If you don't need it, 
		better to re-implement this method in your frame and to use FastReset() 
	*/   
	virtual void Reset() 
		{ ConcreteFrame::Reset(); fCurrentKeys.Clear(); }   
	
	//! FastReset the current frame, means the data part keep the previous values 
	virtual void FastReset() 
		{ ConcreteFrame::FastReset(); fCurrentKeys.Clear(); } 
}; 

// CompositeFrame inline members //////////////////////////////////////////////////////

} // namespace ADF
#endif




