/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file ConfAgent.cpp compiled in libGWADF.so */

#ifndef ADF_ConfAgent
#include "ConfAgent.h"
#endif
#ifndef ADF_ConfigurationFrame
#include "ConfigurationFrame.h"
#endif
#ifndef ADF_RunAgent
#include "RunAgent.h"
#endif 
#ifndef ADF_DFAgent
#include "DFAgent.h"
#endif

#include <stdlib.h>

ADF::ConfAgent *ADF::ConfAgent::gTheGlobalAgent = 0x0;

using namespace ADF;
using namespace std;

namespace { UInt_t gAllocated = 0u; }

ConfAgent::ConfAgent() : DotConf("ADF::ConfAgent")
{ 
	gAllocated++;
	SetWithRecord(true);
} 

ConfAgent::~ConfAgent() 
{
	gAllocated--;
	
	if ( gAllocated == 1 ) {
#ifdef DEBUGSTATIC			
		std::clog << "[[static]] ConfAgent::~ConfAgent() ]\n[...\n" ;
#endif
		KilltheGlobalAgent();
#ifdef DEBUGSTATIC			
		std::clog << " done\n...]\n";	
#endif
	}
}

ConfAgent::EModel ConfAgent::GetModel(std::string model)
{
	ConfAgent::EModel m = ConfAgent::kSafe; 
	
	if ( model == "kStrict" )
		m = ConfAgent::kStrict;
	else if ( model == "kGrowing" )
		m = ConfAgent::kGrowing;
	
	return m;
}

const Char_t *ConfAgent::GetStringModel(ConfAgent::EModel m)
{
	const char *model = 0x0;
	switch (m) {
		case kStrict:
			model = "kStrict";
			break;	
		case kSafe:
			model = "kSafe";			
			break;	
		case kGrowing:
			model = "kGrowing";
			break;
	}
	return model; 
}

void ConfAgent::KilltheGlobalAgent()
{
	if ( gTheGlobalAgent && gAllocated == 1 ) {
#ifdef DEBUGSTATIC					
		std::clog << "[[static]] KilltheGlobalAgent() ]\n[...\n" ;
		std::clog << " Destroying the global agent ... " << std::endl;
#endif		
		delete gTheGlobalAgent;
#ifdef DEBUGSTATIC			
		std::clog << " done\n...]\n";	
#endif		
		gTheGlobalAgent = 0x0;	
	}
}


namespace ADF {
	
	//! The Agata Data Flow agent
	/*!
	 this contents all the information concerning the agata data flow @ a given point. A global, static agent is loaded @ the beginning.
	 It is initialised by the ADF.conf file. One is installed in the distribution, the other one is in the user space in the directory defined under ADF_CONF_PATH.
	 
	 Any sub-system (actor) has its own agent (local) which is a Clone of the global one so that @ the beginning all the system is synchronized.
	 Local agents could be re-configured locally (for instance by configuration frames) independently to adapt themself to the local data flow without breaking the whole system.
	 
	 As a agata data flow agent, it defines the factories needed to allocate keys/frames. The global agent has references to global (singleton) factories.
	 Local agent have clones of the global ones. 
	 The agent also defines the endianness of the data flow, the factories it really needs to work completely (it should be included in the list of loaded ones),
	 as well as the definition of the frames going one.
	 
	 The Primarykey is the minimum information to be able to map the flow. Indeed, one may have the length of the frames coded on 2 or 4 bytes.
	 The AutoConf frame is used by the inner ADF system to reconfigure itself the agent. Comment the line to switch off this feature.
	 
	 \code
	 # This is a begin of the ADF configuration record 
	 # Option is Init. It means the Agent is reset before being initialised 
	 # 
	 #////ADF::ConfAgent_beg//// Agata 1 Init 
	 #  
	 #
	 Endian: kLittle
	 #
	 # List of factories required 
	 # 
	 # Factories for Keys 
	 #
	 KeyFactory: Default 
	 KeyFactory: Agata 
	 #
	 # Factories for Frames
	 #
	 FrameFactory: Agata 
	 # 
	 # 
	 # Version of the primary key that structures the data flow 
	 # and the frame expected to reconfigure it
	 # 
	 PrimaryKey:	Default FS 0 1 
	 #
	 # comment/uncomment to switch off/on auto-reconf
	 AutoConf:	Agata conf:global 2 0 Agata conf:global 0 0
	 #
	 #
	 # Current Frames going through the data flow
	 #
	 Frame:  Agata data:crystal 4 0  Agata data:crystal 65000 0  
	 Frame:  Agata data:ccrystal 4 0  Agata data:ccrystal 65000 0  
	 Frame:  Agata data:ranc0 4 0  Agata data:ranc0 65000 0  
	 Frame:  Agata data:ranc1 4 0  Agata data:ranc1 65000 0  
	 Frame:  Agata data:psa 4 0  Agata data:psa 65000 1  
	 Frame:  Agata data:tracked 4 0  Agata data:tracked 65000 0  
	 Frame:  Agata event:data 4 0  Agata event:data 4 0  
	 Frame:  Agata event:data:psa 4 0  Agata event:data:psa 4 0  
	 #
	 # end of Data Flow Defaults  
	 #
	 #////ADF::ConfAgent_end//// 
	 \endcode
	 
	 To work with a factory version different that -1 (which is the default), just set
	 
	 \code
	 KeyFactory: Agata 0
	 \endcode
	 
	 
	 */
	class AgataConfAgent : public ConfAgent
	{	
	private:
		//! global AgataConf Agent
		static AgataConfAgent *gTheGlobalAgent;
		
	private:
		//! Agent in charge of the structrure of the data flow
		DFAgent *fDFAgent;	
		//! Agent in charge of the informations relatives to the run
		RunAgent *fRunAgent;
		
	public:
		//! used to set global agent
		AgataConfAgent(); 
		//! used to set local agent i.e. clones
		AgataConfAgent(const char *);
		//! destructor
		virtual ~AgataConfAgent(); 
		
		//! to get the Data Flow Agent in charge of knowing the structure of the DF
		virtual DFAgent *GetDFAgent()
		{ return fDFAgent ;}
		//! to get the run Agent in charge of knowing the conditions of the DF being processed
		virtual RunAgent *GetRunAgent()
		{ return fRunAgent ;}
		virtual const DFAgent *GetConstDFAgent() const
		{ return const_cast<const DFAgent *> (fDFAgent) ;}
		virtual const RunAgent *GetConstRunAgent() const
		{ return const_cast<const RunAgent *> (fRunAgent) ;}
		
		ConfAgent *Clone(const char *comment = "Clone of the global ConfAgent");
		
		// to get the global AGATA agent
		static AgataConfAgent *theGlobalAgent(); 
		
		//! to avoid some warning of hidden methods 
		virtual Bool_t Configure(const Char_t *name,  const char *option = "in", Bool_t allow_init = true)
		{ return DotConf::Configure(name,option,allow_init); }
		virtual Bool_t Configure(ConfigurationFrame *,  const char *option = "in", Bool_t allow_init = false);
	}; 	
} // namespace

ADF::AgataConfAgent *ADF::AgataConfAgent::gTheGlobalAgent = 0x0;

AgataConfAgent::AgataConfAgent() : 
ConfAgent(), 
fDFAgent(new DFAgent("ADF::ConfAgent","Agata",true,true)), 
fRunAgent(new RunAgent("ADF::ConfAgent","Run"))
{ 	
	fLog.GetProcessName() = "ADF::AgataConfAgent"; 
	SetLastModif();	
	
	if ( fDFAgent )
		AddConf(fDFAgent);
	if ( fRunAgent )
		AddConf(fRunAgent);	
	
	// configuration from the input file of the distribution
	// then the one pointed by ADF_CONF_PATH. if none of these two are found, try the current directory
	// if not found in the install path (make install not done properly)
	// try in the current directory
	std::string tmp; 
	
	// first the one in the distribution
	
	tmp = GetPathToConfFile();
	if ( tmp.size() > 0 && !(tmp.at(tmp.size()-1) == '/') ) 
		tmp += '/';
	tmp += "ADF.conf";
	
	if ( Configure(tmp.data(),"in") ) 
		fLog << info << "Init Global Conf Agent from distribution " << tmp << nline;
	else
		fLog << error << "Failed to init Global Conf Agent from distribution " << tmp << nline;
	
	// send snapshot of the agent in the log system
	if ( fLog.IsDebug(2) ) {
		fLog.SetLevel(LogMessage::kInfo,1);
		DoConfigure( fLog() );
		fLog.SetLevel(LogMessage::kInfo,0);
	}
	// second the one in the ADF_CONF_PATH path	or if not defined in the current directory
	if ( ::getenv("ADF_CONF_PATH") ) {
		
		tmp = ::getenv("ADF_CONF_PATH");
		
		if ( tmp.size() > 0 && !(tmp.at(tmp.size()-1) == '/') ) 
			tmp += '/';
		tmp += "ADF.conf";
		if ( Configure(tmp.data(),"in") ) 
			fLog << info << "Init Global Conf Agent from ADF_CONF_PATH " << tmp << nline;
		else 
			fLog << warning << "Failed to init Global Conf Agent from ADF_CONF_PATH " << tmp << nline;
	}	
	else {
		fLog << warning << "ADF_CONF_PATH NOT DEFINED !, try init Global Conf Agent from current dir ..." << nline;
		if ( Configure("./ADF.conf","in") ) 
			fLog << warning << "... DONE" << nline;
		else 
			fLog << warning << "... FAILED" << nline;
	}
	
	// send snapshot of the agent in the log system
	if ( fLog.IsDebug(2) ) {
		fLog.SetLevel(LogMessage::kInfo,1);
		DoConfigure( fLog() );
		fLog.SetLevel(LogMessage::kInfo,0);
	}
	fLog << dolog;
} 

AgataConfAgent::AgataConfAgent(const char *comment) : 
ConfAgent(), 
fDFAgent(new DFAgent("ADF::ConfAgent","Agata",true,false)), 
fRunAgent(new RunAgent("ADF::ConfAgent","Run"))
{ 	
	fLog.GetProcessName() = "ADF::AgataConfAgent"; 
	SetComment(comment);
	
	if ( fDFAgent )
		AddConf(fDFAgent);
	if ( fRunAgent )
		AddConf(fRunAgent);	
} 

AgataConfAgent::~AgataConfAgent() 
{
	if ( fDFAgent ) {
		delete fDFAgent; fDFAgent = 0x0;
	}
	if ( fRunAgent ) {
		delete fRunAgent; fRunAgent = 0x0;
	}	
}

AgataConfAgent *AgataConfAgent::theGlobalAgent()
{
	if ( gTheGlobalAgent == 0x0 ) {
#ifdef DEBUGSTATIC			
		std::clog << "[[static]]  Creating the global ConfAgent ]\n[...\n";
#endif
		// be sure factories global main factories are there  
		// the two main factories are added to the main factory in case it has not yet been done
		// key factory	
		if ( !MainKeyFactory::theMainFactory().IsKnown("Agata") )
			MainKeyFactory::theMainFactory().Register("Agata"); 	
		// Frame factory
		if ( !MainFrameFactory::theMainFactory().IsKnown("Agata") )	
			MainFrameFactory::theMainFactory().Register("Agata"); 
		
		gTheGlobalAgent = new AgataConfAgent();	
		
#ifdef DEBUGSTATIC			
		if ( gTheGlobalAgent ) 
			std::clog << " done\n...]\n";
		else 
		{ std::clog << " *** ERROR *** \n"; std::clog << "...]\n"; }
#endif		
	}
	return gTheGlobalAgent; 
	
}

ConfAgent *AgataConfAgent::Clone(const char *comment)
{	
	// check the global agent exists
	AgataConfAgent::theGlobalAgent();
	
	// first creats a new agent
	AgataConfAgent *aconf = new AgataConfAgent(comment);
	if ( aconf == 0x0 )
		return 0x0;
	
	// now copy the content of the global agent to it using Configure
	std::ostringstream o;  
	AgataConfAgent::theGlobalAgent()->DoConfigure(o);
	std::string tmp;
	tmp = o.str();
	std::istringstream i(tmp);
	aconf->DoConfigure(i,true); // the new agent is cloned, so reset it first ... even if aconf has just been created ... needed ?
	
	return aconf;
}

Bool_t AgataConfAgent::Configure(ConfigurationFrame *frame,  const char *option, Bool_t allow_init)
{
	// set reference to this method for log system
	fLog.SetProcessMethod("Configure(ConfigurationFrame *name)");
	
	Bool_t ok; std::string opt = option;
	
	if ( opt == "in" ) {
		// load the content of the frame in the string
		frame->Read();
		// use stream facility 
		std::istringstream in;
		in.clear();
		in.str((frame->String()));
		ok = DoConfigure(in, allow_init);
	}
	else {
		// use stream facility 
		std::ostringstream out;
		
		// the global conf frame starts an adf file. if the global conf file is not set properly (too high !)
		// the reading of the conf frame is cut and this could cause a wrong initialisation of the Agent.
		// to be safe, add some blank space 
		out << ends << 	ends << ends << ends << ends << ends << ends << ends << ends << ends << ends << ends;
		out << ends << 	ends << ends << ends << ends << ends << ends << ends << "\n";	
		ok = DoConfigure(out);
		frame->String() = out.str();
		frame->Write();
	}
	fLog << dolog;
	
	return ok;
}	

ConfAgent *ConfAgent::theGlobalAgent(std::string which_agent)
{ 
	if ( gTheGlobalAgent == 0x0 ) {
#ifdef DEBUGSTATIC			
		std::clog << "[[static]]  Creating the global ConfAgent ]\n[...\n";
#endif
		if ( which_agent == "Agata" ) {
			gTheGlobalAgent = AgataConfAgent::theGlobalAgent();
		}
#ifdef DEBUGSTATIC			
		if ( gTheGlobalAgent ) 
			std::clog << " done\n...]\n";
		else 
		{ std::clog << " *** ERROR *** \n"; std::clog << "...]\n"; }
#endif		
	}
	return gTheGlobalAgent; 
}
