/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file ConfAgent.h header file for ConfAgent.cpp */

#ifndef ADF_ConfAgent
#define ADF_ConfAgent

#ifndef ADF_DotConf
#include "DotConf.h"
#endif

namespace ADF
{

class DFAgent;
class RunAgent;
	
//! ConfAgent
/*!
	This class contains all the definitions relative to the data flow being read/written.
	
	In particular it contains a DFAgent, in charge of the structure od the data flow
	and a RunAgent that contains global informations concerning the current conditions.
 
	A global ConfAgent is allocated using theGlobalAgent(). 
	Local agent could be built using the Clone method. Once built they can be re-initialised 
	independently of the Global Agent, from, for instance, a Configuration Frame
 
	The data flow model could be kSafe, kStrict, kGrowing
	- kSafe : do exactly what is asked by the user for the data flow part he is working on but do nothing for the other part
	- kStrict : strictly take into account the user's wishes 
	- kGrowing : nothing is lost, the data flow is growing at each stage of the processing
	
	Extention
 
	This class should be the central point whatever the data flow processed. For the moment a single class implements such a ConfAgent : AgataConfAgent
 
	In order to possibly extends the system, one may add to the class a list of static 
	corresponding to the possible ConfAgent. To do this:
	-one should add to the constructor a Register mehod so that any class that inherits from this could Register itself
	-one must built an class that inherits from this
	-one should modify the method theGlobalAgent to deliver the proper Agent depending of an input string ...
 
*/
class ConfAgent : public DotConf
{	
public:
	//! Operating mode for frame IO
	enum EMode { kRead, kWrite } ;	
	//! Model for FrameIO   
	enum EModel { kStrict, kSafe, kGrowing } ;
	
private:
	//! the global agent
	static ConfAgent *gTheGlobalAgent;
	
protected:
	ConfAgent(); 
	
public:
	virtual ~ConfAgent(); 
	
	//! to get the Data Flow Agent in charge of knowing the structure of the DF
	virtual DFAgent *GetDFAgent() = 0;
	virtual const DFAgent *GetConstDFAgent() const = 0;
	
	//! to get the run Agent in charge of knowing the conditions of the DF being processed
	virtual RunAgent *GetRunAgent() = 0;
	virtual const RunAgent *GetConstRunAgent() const = 0;		
	
	static ConfAgent::EModel GetModel(std::string);
	static const Char_t *GetStringModel(ConfAgent::EModel);

	//! Clone of the main (global) Agent
	virtual ConfAgent *Clone(const char *comment = "Clone of the global ConfAgent") = 0;
	
	//! to get the global agent
	static ConfAgent *theGlobalAgent(std::string = "Agata"); 
	
	//! to kill the global agent	
	static void KilltheGlobalAgent();
}; 

} // namespace ADF

#endif








