/***************************************************************************
 *   Copyright (C) 2010 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file DFAgent.h header file for DFAgent.cpp */

#ifndef ADF_DFAgent
#define ADF_DFAgent

#ifndef ADF_FactoryItem
#include "FactoryItem.h"
#endif
#ifndef ADF_ConfAgent
#include "ConfAgent.h"
#endif
#ifndef ADF_BaseBuffer
#include "BaseBuffer.h"
#endif

#include <list>
#include <utility>

namespace ADF
{
class KeyFactory;
class FrameFactory;
class Frame;
class ConfigurationFrame;

//! A DFAgent contains all the information concerning the structure of the data flow
/*!
 The DFAgent is in charge of keeping all the informations concerning the structrure of the data flow at a given point
 
 For instance it is the place to store the endian type of the data flow. 
 It could be also the default size of a block in case the DF in composed of fixed blocks.
 
 The model corresponds to the way the data flow is manipulated, for instance if the input data are consumed or not.

 The primary key is used to know the number of bytes (2 bytes, short or 4 bytes, integer) used to read the length of the frames.
 
 The other part are more specific to the data flow itsel. It corresponds to the full definition of the Frame that
 reconfigures completely the data flow and the list of Frame, with their version number, expected on the data flow.
 
 It keeps also a list of factories. By default this is the list of the global factories found in the MainFactory once a DFAgent is created
 
 In a complex topology, the situation could be different at one point to the data flow compared to another one. It is then mandatory 
 to have at least one DFAgent per local point (per actor or data flow line). As well reconfiguration of one agent could occur from a Configuration Frame requiring a local agent.
 Having local agents helps also to be safer in muti-threading mode.  

 Local agents are at the beginning clones of the global agent. As for the agent it requires then to have a clone of the factories.
 This is done by the user, who knows which agent is local or global, by calling the constructor with the option global = false.
 The Cloned factories belong to the main factories so they are no deleted here. In case of reconfiguration, only the cloned factories are reconfigured 
 which helps for consitency and safety in case of multithreading.
*/
class DFAgent : public DotConf
{	
private:
	//! current IO model, not written in the dataflow
	ConfAgent::EModel fModel;    
private:
// Configurable part : value are modified in case of reconfiguration
	//! Endian type of the data flow
	BaseBuffer::EEndian fEndian;
	//! Defined the primary key for any input data block (mainly 2 or 4 bytes for frame length)
	FactoryItem fPrimaryKey;
	//! Defined the frame that trigs a global reconfiguration of the agent   
	std::pair <FactoryItem,FactoryItem> fAutoConf;
	//! list of the frame (keyitem,frameitem) going through the data flow         
	mutable std::list<std::pair <FactoryItem,FactoryItem> > fListOfFrames; 
	//! list of factories required to process correctly/completely the data flow (should be a subset of fListOfKeyFactories or fListOfFrameFactories)              
	mutable std::list< FactoryItem > fListOfRequiredKeyFactories;
	//!
	mutable std::list< FactoryItem > fListOfRequiredFrameFactories;
	
private:
// Not completely configurable part : the list is not modified BUT the configuration of the factories is called. Be sure they are local in case of complex topology/multi-threading
	//! list of key factories needed to have a full desciption of the data flow
	std::list< KeyFactory * > fListOfKeyFactories;
	//! list of frame factories needed to have a full desciption of the data flow
	std::list< FrameFactory * > fListOfFrameFactories;

private:
	//! keep a list of key and frame factories used to deal with the data flow
	/*!
	 The factories are not owned by the agent but by the main factories. if global is true (default)
	 it takes the pointers to the factories defined in the mainfactories. if wrong, it clones thoses factories.
	 In any case, the lsit is clear before filling it.
	 */
	void LinkFactories(Bool_t global = true);
	
protected:
	//! decode an input line to configure/reconfigure this DFAgent
	virtual Bool_t ProcessLine(const std::string &);  
			
	//! internal use to avoid adding a definition of a Frame that cannot be allocated
	void AddKnownFrame(const FactoryItem &item_key, const FactoryItem &item_frame, const Char_t opt = '+') ;
	
	//! look in the DFAgent list if this factory exist and with a version # greater that the given one
	KeyFactory *GetKeyFactory(const FactoryItem &item) ;
	//! look in the DFAgent list if this factory exist and with a version # greater that the given one
	FrameFactory *GetFrameFactory(const FactoryItem &item);	
	
public:
	//! in case global is true, it keeps a list of pointers to the global factories otherwise asked for Clones of them
	DFAgent(const Char_t *rid, const Char_t *srid, Bool_t is_record = true, Bool_t global = true); 
	virtual ~DFAgent(); 
	
	//! to just browse the list of known frame
	const std::list<std::pair <FactoryItem,FactoryItem> > & GetListOfKnownFrames() const
	{
		return fListOfFrames;
	}
			
	//! check if the factory is referenced by the agent. 
	/*!
		if it is referenced, it keeps a record of the factory, its version # and the address of the factory to deliver the requests.
	*/
	Bool_t SetKnownKeyFactory(const FactoryItem &factory_item);
	//! check if the factory is referenced by the agent. 
	/*!
	 if it is referenced, it keeps a record of the factory, its version # and the address of the factory to deliver the requests.
	*/
	Bool_t SetKnownFrameFactory(const FactoryItem &factory_item);
	
	//! Take the item and prepare it to be an item related to the is agent i.e. set the factory version # and pointer as referenced in the list of required factory
	Bool_t SetAsCompatibleKeyItem(FactoryItem &factory_item) const;
	//! Take the item and prepare it to be an item related to the is agent i.e. set the factory version # and pointer as referenced in the list of required factory
	Bool_t SetAsCompatibleFrameItem(FactoryItem &factory_item) const;
	//! change both item as the same time
	Bool_t SetAsCompatibleItems(FactoryItem &factory_key_item, FactoryItem &factory_frame_item) const
		{ return SetAsCompatibleKeyItem(factory_key_item) && SetAsCompatibleFrameItem(factory_frame_item); }

	// endianness of the data flow 
	BaseBuffer::EEndian GetEndian() const
		{ return fEndian; }
	ConfAgent::EModel GetModel() const
		{ return fModel; }
	
	// change th e endian. it is applied to all the factories referenced in this agent
	void SetEndian(BaseBuffer::EEndian); 
	// io model 
	void SetModel(ConfAgent::EModel model)
		{ fModel = model; }	
	
	//! Just to know if the size of the frame is coded on 2 bytes, 4 bytes ... or 8 bytes
	const FactoryItem &GetPrimaryKey() const;
	
	//! To get the definition of the Key and Frame associated to a global reconfiguration
	const FactoryItem &GetAutoConfKey() const
		{ return fAutoConf.first; }
	const FactoryItem &GetAutoConfFrame() const
		{ return fAutoConf.second; }

	//! to set the definition of the Keys for input
	/*!
		TODO check that the key exists before any changes
	*/	
	void SetPrimaryKey(const FactoryItem &item);	
	
	//! to add a known frame (used when building the list of known frames)
	/*!
	 \code
		r : the list is reset first 
		- : the item if found is removed from the list
		+ : added to the list
	 \endcode
	*/
	void AddKnownFrame(const Frame *frame, const Char_t opt = '+') ;
		
	//! To know if a Frame is known by the Agent under the conditions given 
	/*!
		It returns the number of defintion suitable
	*/
	UShort_t IsAKnownFrame 
		(const FactoryItem &kit, const FactoryItem &fit, PF_FactoryItemChange kch = ChangeOfVersion, PF_FactoryItemChange fch = ChangeOfVersion) const;
		 	
	//! To get the exact definition of a key and a Frame	
	FactoryItem WhichKnownFrame  
		(const FactoryItem &kit, const FactoryItem &fit, PF_FactoryItemChange kch = ChangeOfVersion, PF_FactoryItemChange fch = ChangeOfVersion) const;
	FactoryItem WhichKnownKey
		(const FactoryItem &kit, const FactoryItem &fit, PF_FactoryItemChange kch = ChangeOfVersion, PF_FactoryItemChange fch = ChangeOfVersion) const;
	
	//! 
	virtual void Reset(const Char_t *opt = "");

	//! configure this from an input stream
	virtual Bool_t DoConfigure(std::istream &in, Bool_t allow_init)	
		{ return DotConf::DoConfigure(in,allow_init); }
	//! write the configuration in an output stream	
	virtual Bool_t DoConfigure(std::ostream &);
}; 

} // namespace ADF

#endif








