/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file DataHandling.h header file for DataHandling.cpp */

#ifndef ADF_DataHandling
#define ADF_DataHandling
#define ADF_DATAHANDLING_H ADF_DataHandling

#ifndef ADF_ADFConfig
#include "ADFConfig.h"
#endif
#ifndef ADF_BufferIO
#include "BufferIO.h"
#endif

#include<iostream>
#include<string>
#include<cstdlib>

namespace ADF
{ 

const char *NamedItemTypeOf(Bool_t); 
const char *NamedItemTypeOf(UShort_t); 
const char *NamedItemTypeOf(Short_t); 
const char *NamedItemTypeOf(UInt_t); 
const char *NamedItemTypeOf(Int_t); 
const char *NamedItemTypeOf(ULong64_t);
const char *NamedItemTypeOf(Long64_t);
const char *NamedItemTypeOf(Double_t);
const char *NamedItemTypeOf(Float_t); 
const char *NamedItemTypeOf(std::string); 

//! A DataLink links too variables in memory 
/*!
	Two addresses of the same type of data are linked by a DataLink. 
	Once linked, the value of the first one (second one) is copied in the second one (first one) 
	each time the CopyA1_to_A2 (CopyA2_to_A1) method is called. 
	
	Both values could be set to a default value by calling the Reset method. The default value 
	are set in the constructor
	
	Be careful, it is the charge of the user to be sure a DalaLink is correctly 
	linked with two valid addresses.
	
	\code 
	UShort_t a1, a2; DataLinkImp<UShort_t> link; 
	
	a1 = 1; a2 = 2; 
	link(&a1,&a2); 
	link.CopyA1_to_A2(); ==> a1 = a2 = 1
	
	a2 = 2;
	link.CopyA2_to_A1(); ==> a1 = a2 = 2
	
	link.Reset(); ==> a1 = a2 = 0
	\endcode 
	With this, user's variables could be linked on the basis of a common name. 
	
	See for instance ANamedItem, NamedItem and GObject.
*/   
class DataLink 
{
public:
	friend class GObject;
	
protected:
	virtual void *A1()
		{ return 0x0; }	
	virtual void *A2()
		{ return 0x0; }

public:
	DataLink() 
		{;} 
	virtual ~DataLink()
		{;}

	//! test if the given address is one of the Link
// 	virtual UShort_t IsALink(void *) 
// 		{ return 0u; }
				
	//! Link a1 to a2
	virtual void Link(void *, void *) 
		{;}
	
	//! Copy a1 into a2
	virtual void CopyA1_to_A2() 
		{;}
	
	//! Copy a2 into a1
	virtual void CopyA2_to_A1() 
		{;}
	
	//! both linked data to default
	virtual void Reset() 
		{;}
	
	//! set A1 to default
	virtual void ResetA1() 
		{;}	
	
	//! set A2 to default
	virtual void ResetA2() 
		{;}	
};

//! Implementation of a concrete DataLink 
/*!
*/  
template<typename Data_T> class DataLinkImp : public DataLink
{
private:
	Data_T fDefault; Data_T *fA1, *fA2;

protected:
	virtual void *A1()
		{ return fA1; }
	virtual void *A2()
		{ return fA2; }
public:
	DataLinkImp(Data_T d) : 
		DataLink(), fDefault(d), fA1(0x0), fA2(0x0) {;}
	virtual ~DataLinkImp() 
		{ fA1 = fA2 = 0x0; }

	virtual void Link(void *a1, void *a2) 
		{ fA1 = (Data_T *)a1; fA2 = (Data_T *)a2; };
	
	virtual void CopyA1_to_A2() 
		{ (*fA2) = (*fA1) ;}
	virtual void CopyA2_to_A1() 
		{ (*fA1) = (*fA2) ;}
	
	virtual void ResetA1() 
		{ (*fA1) = fDefault; }	
	virtual void ResetA2() 
		{ (*fA2) = fDefault; }
		
	virtual void Reset() 
		{ (*fA2) = (*fA1) = fDefault; }	
};

//! ANamedItem 
/*!
	A ANamedItem is a variable with a name and a type. It could be linked with an external address (See for instance GObject). 
	
	Basic ADF types are supported :
            - S  : a 16 bit signed integer (Short_t)
            - US : a 16 bit unsigned integer (UShort_t)
            - I  : a 32 bit signed integer (Int_t)
            - UI : a 32 bit unsigned integer (UInt_t)
            - F  : a 32 bit floating point (Float_t)
            - D  : a 64 bit floating point (Double_t)
            - L  : a 64 bit signed integer (Long64_t)
            - UL : a 64 bit unsigned integer (ULong64_t)
            - B  : a boolean (Bool_t)	
	    
	An simple stucture could be a ANamedItem. In order to properly define it, 
	the following function has to be defined:
	 
	\code const char *ADF::NamedItemTypeOf(MyStruct_t) \endcode 
	
	were MyStruct_t is the type of the structure you would like to add.
	As well, the three operators << and >> (on a BufferIO) and = have to be properly defined.
	
	\code
	\endcode 	
*/   
class ANamedItem 
{
protected:
	std::string fName;
	std::string fType;
	
public:
	//! Constructor for a NamedItem
	ANamedItem(const char *name, const char *type) : 
		fName(name), fType(type) {;}
	virtual ~ANamedItem() 
		{;}
	
	//! clone that named item
	virtual ANamedItem *MakeAClone() const = 0;
		
	//! check is the passed DataLink is linked to the internal value of this NamaedItem
// 	virtual UShort_t IsLinked(DataLink *) = 0;
		
	//! ask this to create a suitable DataLink between the value of the NameItem and the external address
	virtual DataLink *MakeADataLink(void *) = 0;
	
	//! ask this to create a proxy for that item
	virtual ANamedItem *MakeAProxy()= 0;
	
	//! change value of this nmaed item using the passed one. it checks if asked, it check first if of the same type/name
	virtual Bool_t Set( const ANamedItem * ) = 0;
	//! change value of this nmaed item using the passed one. it checks if asked, it check first if of the same type/name	
	// virtual Bool_t Set( const ANamedItem *, Bool_t check_type = true, check_name = false ) = 0;

	//! Name of the Item	
	std::string GetName() const
		{ return fName; }
	//! Type of the Item	
	std::string GetType() const
		{ return fType; }	
	
	//! Name of the Item	
	std::string &Name() 
		{ return fName; }
	//! Type of the Item	
	std::string &Type() 
		{ return fType; }
		
	//! Check if this is of type 'type'
	Bool_t IsType(const char *type) 
		{ return fType.compare(type) == 0 ; }
			
	//! write the value of this item to the buffer	
	virtual void SetItem(BufferIO &) const 
		{;}	
	//! read the value of this item from the buffer	
	virtual void GetItem(const BufferIO &) 
		{;}
	
	//! set the item to the default value
	virtual void Reset() 
		{;}
		
	virtual void Print(std::ostream & = std::cout) const 
		{;}
};

template<typename Data_T> class NamedItem : public ANamedItem
{
private:
	Data_T fDefault; Data_T fData;
	
public:
	NamedItem(const char *name, Data_T d = 0): 
		ANamedItem(name,"NOTYPE"),fDefault(d),fData(d) { fType = NamedItemTypeOf(fData); }
	virtual ~NamedItem() 
		{;}

	//! clone that named item
	virtual ANamedItem *MakeAClone() const
	{ 
		NamedItem<Data_T> *clone = new NamedItem<Data_T>(GetName().data(),GetDefault()) ;
		clone->Set(Get());
		
		return clone;
	}
	//! change value of this named item using the passed one. It has to be of the same type ! returns false if not otherwise do the change
	virtual Bool_t Set( const ANamedItem *item )
	{
		Bool_t ok = false;
		if ( item->GetType() == GetType() ) {
			const NamedItem<Data_T> *t = dynamic_cast< const NamedItem<Data_T> *> (item);
			Set( t->Get() );
			ok = true;
		}		
		return ok;
	}
	//! ask this to create a proxy for it	
	virtual ANamedItem *MakeAProxy();
		
	//! ask this to create a suitable DataLink
	virtual DataLink *MakeADataLink(void *extern_add)
	{ 
		DataLinkImp<Data_T> *link = new DataLinkImp<Data_T>(fDefault); 
		if ( link ) 
			link->Link(&fData,extern_add);
		return link;
	}
	//! Set the data	
	virtual void Set(Data_T t) 
		{ fData = t; }
	//! Get the data	
	virtual Data_T Get() const
		{ return fData; }
		
	//! Set the data	
	virtual void SetDefault(Data_T t) 
		{ fDefault = t; }
	//! Get the data	
	virtual Data_T GetDefault() const
		{ return fDefault; }
		
	//! write the value of this item to the buffer
	virtual void SetItem(BufferIO &buf) const  
		{buf << fData;}	
	//! read the value of this item from the buffer	
	virtual void GetItem(const BufferIO &buf) 
		{buf >> fData;}	
		
	//! set the item to the default value
	virtual void Reset() 
		{ fData = fDefault; }
		
	virtual void Print(std::ostream & /*out */ = std::cout) const 
	{
//		out << " " << GetName() << " " << GetType() << " " << fData ;
//		out << " " << GetName() << " " << GetType() << " " ;		
	}
};

class AProxyNamedItem 
{
public:
	AProxyNamedItem() 
		{;}
	virtual ~AProxyNamedItem() 
		{;}
		
	//! the passed item becomes the real item
	virtual void SetProxy(ANamedItem * /*ritem*/) 
		{;}
};

template<typename Data_T> class ProxyNamedItem : public AProxyNamedItem, public NamedItem<Data_T>
{
private:
	NamedItem<Data_T> *fRealItem;

public:
	ProxyNamedItem(const char *name, Data_T d = 0): 
		NamedItem<Data_T>(name,d),fRealItem(0x0) { fRealItem = this; }
	virtual ~ProxyNamedItem() 
		{;}
	
	//! the passed item becomes the real item
	virtual void SetProxy(ANamedItem *ritem)
	{ 
		fRealItem = this;
		if ( ritem == 0x0 )
			return;
		if ( dynamic_cast< NamedItem<Data_T> *>(ritem) )
			fRealItem = dynamic_cast< NamedItem<Data_T> *>(ritem);
	} 

	//! ask this to create a proxy for that item
	/*!
		Already a proxy so just return this
	*/
	virtual ANamedItem *MakeAProxy()
		{ return this; }	
	
	//! ask this to create a suitable DataLink
	virtual DataLink *MakeADataLink(void *extern_add)
		{ return fRealItem->MakeADataLink(extern_add); }
	
	virtual Bool_t Set( const ANamedItem *t )
		{ return fRealItem->Set(t) ; }
	virtual void Set(Data_T t) 
		{ fRealItem->Set(t) ; }
	virtual Data_T Get() const
		{ return fRealItem->Get(); }
		
	virtual void SetDefault(Data_T t) 
		{ fRealItem->SetDefault(t) ; }
	virtual Data_T GetDefault() const
		{ return fRealItem->GetDefault(); }
		
	//! write the value of this item to the buffer
	virtual void SetItem(BufferIO &buf) const  
		{fRealItem->SetItem(buf);}	
	//! read the value of this item from the buffer	
	virtual void GetItem(const BufferIO &buf) 
		{fRealItem->GetItem(buf);}	
		
	//! set the item to the default value
	virtual void Reset() 
		{ fRealItem->Reset(); }	
};

template <typename Data_T> ANamedItem *NamedItem<Data_T>::MakeAProxy()
{
	ProxyNamedItem<Data_T> *ritem 
		= new ProxyNamedItem<Data_T>(Name().data(),GetDefault());
	ritem->SetProxy(this);
	
	return ritem;
}

//! Interface to digitise real values.
/*!

*/

class Digitizer
{
private:
	Double_t fXmin, fXmax, fDeltaX;
	
	UShort_t fNBins2;
	UInt_t   fNBins4;	 
	
public:
	Digitizer(Double_t xmin = 0., Double_t xmax = 1., UInt_t bins = 1u);
	virtual ~Digitizer() {;}
	
	virtual void Init(Double_t xmin = 0., Double_t xmax = 1., UInt_t bins = 1u);
	
	Double_t GetMin() const 
		{ return fXmin; }
	Double_t GetMax() const 
		{ return fXmax; }
	
	UShort_t GetBins2() const 
		{ return fNBins2; }
	UInt_t   GetBins4() const 
		{ return fNBins4; }
		
	//! returns (as UShort_t) the bin corresponding to x 
	/*!
	*/
	virtual UShort_t D2Bytes(Double_t x) const {
		if ( x <= fXmin ) return 0u;  
		if ( x > fXmax ) return fNBins2; return UShort_t(fNBins2*(x-fXmin)/fDeltaX) ; 
	}
	 
	//! returns (as UInt_t) the bin corresponding to x 
	/*!
	*/	 
	virtual UInt_t D4Bytes(Double_t x) const {  
		if ( x <= fXmin ) return 0u;
		if ( x > fXmax ) return fNBins4; return UInt_t(fNBins4*(x-fXmin)/fDeltaX) ;
	}	 
	 
	//! returns a randomized value corresponding to the given bin 
	/*!
	*/	
	virtual Double_t UndigitR(UInt_t i) const 
	  { if ( i > fNBins4 ) return fXmax; return fXmin + (i+Double_t(::rand())/RAND_MAX) * fDeltaX / fNBins4; }
	  
	//! returns a randomized value corresponding to the given bin 
	/*!
	*/			
	virtual Double_t UndigitR(UShort_t i) const
	  { if ( i > fNBins2 ) return fXmax; return fXmin + (i+Double_t(::rand())/RAND_MAX) * fDeltaX / fNBins2; }	
	  
	//! returns a non-randomized value corresponding to the given bin 
	/*!
	*/	
	virtual Double_t Undigit(UInt_t i) const 
	  { if ( i > fNBins4 ) return fXmax; return fXmin + (i) * fDeltaX / fNBins4; }
	  
	//! returns a non-randomized value corresponding to the given bin 
	/*!
	*/		
	virtual Double_t Undigit(UShort_t i) const
	  { if ( i > fNBins2 ) return fXmax; return fXmin + (i) * fDeltaX / fNBins2; }	
	  
	//! returns a non-randomized value corresponding to the given bin (center) 
	/*!
	*/	  
	virtual Double_t UndigitC(UInt_t i) const 
	  { if ( i > fNBins4 ) return fXmax; return fXmin + (i+0.5) * fDeltaX / fNBins4; }
	  
	//! returns a non-randomized value corresponding to the given bin (center) 
	/*!
	*/			
	virtual Double_t UndigitC(UShort_t i) const
	  { if ( i > fNBins2 ) return fXmax; return fXmin + (i+0.5) * fDeltaX / fNBins2; }		  
};

} // namespace ADF
#endif

