/***************************************************************************
 *   Copyright (C) 2010 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file DotConf.h header file for DotConf.cpp */

#ifndef ADF_DotConf
#define ADF_DotConf

#ifndef _ADFConfig
#include "ADFConfig.h"
#endif
#include "ADFLogMessage.h"

#include <list>

namespace ADF
{
	
class ConfigurationFrame;

//! DotConf : Utility for class configuration from ascii file or Configuration frames
/*!
 An object that inherits from this class could be intialised/modified using a simple ascii file. This file could be embedded inside a configuration Frame.
 
 The file/frame may be composed of records that start/end with a particular string. 
 In this case IsWithRecord is true and the BeginOfRecord and EndOfRecord look like this :
 
 #////ADF::ConfAgent_beg//// Agata 0 Init
...
 #////ADF::ConfAgent_end////
 
 with :
 - ADF::ConfAgent : it is, in principle the name of the class (RID Record ID) to wish is addressed the information between the begin and end record.
 - Agata : it is a additional word (SRID Sub Record ID) to distinguish eventually between several instances of a class (by name or type) 
 - 0 : a short value, to be able to deal internally with some kind of versionning 
 - Init : Option that changes the way the object is modified. In case of Init, Reset is called once reading the BeginOfRecord's record which is not the case for any other options 
  
 In principle a configuration frame should not Init completely the data members of the class that inherits from DotConf. 
 So if you design your class using this base class, think of the data members as belonging to two categories.
 The first one are changeable (or 'permanent') data members i.e., @ any time, a conf frame (or a conf file) could come with only a modification of one data member or to add some more info :
 Make sure this action is well managed by your reading (see for instance ProcessLine)
 
 The second one groups data members that cannot be changed : this could be a path to some directory. In such case such informations
 in the conf frame would cause errors since the path written is related to the system on which the data has been produced and is most likely not adapted to yours !
 
 By default Configure(ConfigurationFrame *,  const char *option = "in", Bool_t allow_init = false); is protected so that it is impossible to read a configuration frame in Init mode. 
 It may not suit your wishes. In such case, you should set the last argument of the method, allow_init, to true.
 
 The methods :
 \code 
 virtual Bool_t DoConfigure(std::istream &);
 virtual Bool_t DoConfigure(std::ostream &);
 \endcode
 do the work to read/write the stream used to initialise/change the data member of this object.
 \code 
 virtual Bool_t DoConfigure(std::istream &);
 \endcode 
 could be use as it is ... it just reads the stream line by line and calls the ProcessLine method
 which could be overwritten.
 the methods:
 \code 
 //! configuration from/to a configuration frame	
 virtual Bool_t Configure(ConfigurationFrame *,  const char *option = "in");
 //! configuration from/to a file
 virtual Bool_t Configure(const char *name, const char *option = "in");
 \endcode
 just call after some processing 
 \code 
 virtual Bool_t DoConfigure(std::istream &);
 virtual Bool_t DoConfigure(std::ostream &);
 \endcode
 
 In reading mode, false means an attempt to read something expected has rised an error.
 Otherwise if the record is not there or reading is ok, it returns true.
 
 \author O. Stezowski
 
*/
class DotConf
{	
protected:
	//! the log messenger ... to know the object hae been properly intitialised/modified
	LogMessage fLog;    

private: 
//! headers definition : Value don't change on Configure
	struct HeaderDef {
		HeaderDef(Char_t skip = '#', Bool_t is_record = false, std::string rid = "", std::string srid = "", Short_t version = -1) : fSkipLine(skip), fWithRecord(is_record), fRID(rid), fSRID(srid), fVersion(version) 
			{;}
		//! lines starting with this character are skipped
		Char_t fSkipLine;
		//! if true, a begin and end of record are used in reading and writing mode
		Bool_t fWithRecord;
		//! Record identificator 
		std::string fRID;	
		//! Sub Record ID   
		std::string fSRID;
		//! this could manage version up to this value
		Short_t fVersion;
	};
	HeaderDef fHeaderDef;
	
private:
//! value are read from conf file/frame
	struct Header {
		Header(Short_t version, Short_t pversion, std::string streamopt = "Init") :  fVersion(version), fPreviousVersion(pversion), fStreamOption(streamopt)
			{;}
		//! current version number : default -1 = unversionned, otherwise >= 0
		Short_t fVersion; 
		//! previous version number in case of reconfiguration/modification
		Short_t fPreviousVersion; 
		//! Stream Options : Init / Change 
		std::string fStreamOption;   		
	};
	Header fHeader;

private:
	//! list of DotConf : DoConfigure is called for all of them
	std::list< DotConf * > fListOfDotConf;
	
private:
//! additionnal informations: written out as comment/ not read.
	//! Keep track of the date of the last modification   
	std::string fLastModif;
	//! A comment about how/why it has been modified       
	std::string fComment;    
	
protected:
	//! string that gives a start of record for this object
	std::string BeginOfRecord();
	//! string that gives an end of record for this object
	std::string   EndOfRecord();
	//! Return true if this string is a begin of record and call Reset if Init is found
	/*!
		A record has the following format : $SkipLine////$RID_beg//// SRID Version OPT
	 
		RID and SRID are used to identify the agent in which the record is processed : they are not changed @ reading
		Version and OPT are extracted from the istream provided. 
		Previous version number of the .conf is kept so that it is possible to track back the evolution of the configuration from one call to the other one.
		
		If OPT is Init, Reset is called in this method. Set allow_int to false in case it is not the required feature.
	 
		In case your SRID is set to "", the presence of SRID is not mandatory in the header record.
		The presence of version is not mandatory also : in this case it is set to -1 meaning not versionning. OPT is mandatory
	*/
	Bool_t IsBeginOfRecord(const std::string &, Bool_t allow_init);
	//! Return true if this string is an end of record 
	Bool_t   IsEndOfRecord(const std::string &);	
	//! Add a begin or end of record, return true if done
	Bool_t AddRecord(std::ostream &, char t = 'b');	
	
protected:
	//! set in last modif string the current date.
	void SetLastModif();
	//!
	void SetVersion(Short_t v)
		{ fHeader.fVersion = v; }
	//!
	void SetMaxVersion(Short_t v)
		{ fHeaderDef.fVersion = v; }	
	
public:
	//! a helper template to do clones : it requires a default constructor. then it uses the Configure facilities to configure the new object
	/*!
		be careful, all data members may not be in the initialisation file/frame. it is the charge of the user to modifiy/set this data members in his Clone method
	*/
	/*
	template <class Data_t> Data_t *DoClone(Data_t *newobj)
	{
		newobj = new Data_t ();
		if ( newobj == 0x0 )
			return 0x0;
		
		// use DotConf facilities to clone
		Bool_t ok = true;
		//
		std::ostringstream o; 
		ok = ok &&
			this->DoConfigure(o);
		//
		// std::string tmp;
		// tmp = o.str();
		std::istringstream i(o.str());
		ok = ok &&
			newobj->DoConfigure(i,true); // 	
		
		if ( !ok )
			{ delete newobj; newobj = 0x0; }
		
		return newobj;
	}
	 */
	
protected:
	//! to be overwritten
	/*!
	  return false only in case something goes wrong at reading (wrong format)
	*/
	virtual Bool_t ProcessLine(const std::string &) 
		{ return true; }  	

public:
	DotConf(); 
	DotConf(const Char_t *rid, const Char_t *srid = "", Bool_t is_record = true);
	virtual ~DotConf(); 
	
// Get/Set method
public:
	//! to know if this is intialisased with a stream composed of records.
	Bool_t IsWithRecord() const
		{ return fHeaderDef.fWithRecord; }
	//! to change the behavior 
	void SetWithRecord(Bool_t is_record = true)
		{ fHeaderDef.fWithRecord = is_record; }
	
	void SetComment(const Char_t *comment)
		{ fComment = comment ; }
	const Char_t *GetComment() const
		{ return fComment.data(); }
	
	const Char_t *GetLastModif() const
		{ return fLastModif.data(); }
	
	//! this class could managed moddification of the versionned part up to MaxVersion : should be set in the constructor part of the inheriting class
	Short_t GetMaxVersion() const
		{return fHeaderDef.fVersion;}
	
	//! Reset : all values are set to default
	virtual void Reset(const Char_t *opt = "");
	
	//! to set the record id 
	virtual void SetRID(const Char_t *rid, const Char_t *srid ="") 
		{ fHeaderDef.fRID = rid; fHeaderDef.fSRID = srid; }	
	//! to get the record id 
	virtual const Char_t *GetRID() const
		{ return fHeaderDef.fRID.data(); }	
	//! to get the record id 
	virtual const Char_t *GetSRID() const
		{ return fHeaderDef.fSRID.data(); }	
	//! to get the curent version of this object
	virtual Short_t GetVersion() const
		{ return fHeader.fVersion; }	
	
	//! get/set option 
	virtual const Char_t *GetStreamOption() const
		{ return fHeader.fStreamOption.data(); }	
	virtual void SetStreamOption(const Char_t *opt) 
		{ fHeader.fStreamOption = opt; }	
	
	//! add a conf object to this. The added objects are not owned by this so not deleted. 
	/*!
		Each time a Configure method is called, it is called also for all the registered DotConf objects.
	 */
	virtual void AddConf ( DotConf * );
	// clear conf list
	virtual void RemoveConf ( DotConf *aconf = 0x0 )
	{
		if ( aconf == 0x0 )
			fListOfDotConf.clear();
		else
			fListOfDotConf.remove(aconf);
	}

// Interface for any DotConf object	
public:
	//! configure this from an input stream
	/*!
		Main method to be overwritten by daughter classes ... but Configure could also be overwritten
	 
	 allow_init is used to switch on/off Reset bfore streaming the content of istream
	 
		return false only in case something wrong goes on
	*/
	virtual Bool_t DoConfigure(std::istream &, Bool_t allow_init);
	//! write the configuration in an output stream	
	/*!
		Main method to be overwritten by daughter classes ... but Configure could also be overwritten
		return false only in case something wrong goes on
	 */	
	virtual Bool_t DoConfigure(std::ostream &);
	
	//! configuration from/to a configuration frame : extact string from the frame and call DoConfigure
	/*!
	 By default, configuration by frames are most probably re-configuration. 
	 As such they should only modifiy the data members i.e. the Reset() method should not be called before streaming the content of the conf frame
	 
	 */		
	virtual Bool_t Configure(ConfigurationFrame *,  const char *option = "in", Bool_t allow_init = false);
	//! configuration from/to a file : open the file and call DoConfigure
	virtual Bool_t Configure(const char *name, const char *option = "in", Bool_t allow_init = true);
}; 

} // namespace ADF

#endif








