/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file FactoryItem.h header file for FactoryItem.cpp */

#ifndef ADF_FactoryItem
#define ADF_FactoryItem

#ifndef ADF_ADFConfig
#include "ADFConfig.h"
#endif
#ifndef ADF_Version
#include "Version.h"
#endif

#include <string>
#include <iostream>

namespace ADF
{ 

//! Base class that described an item in a Factory
/*!
	An item in a factory should be uniquely identified by :
 
 - the name of the factory (unique, Registered in the MainFactory)
 - an item name/label/type.
 - a version of that item. In this case two numbers (UShort_t) Major / Minor
 
 Additional informations concerning the factory are stored there
 - the version of the factory. A factory inherits from DotConf so that it could be initialised from files/ConfFrame. As such the version number is just a short
 (-1 is for unversionned things). The factory itself (New/DoNew) must deal with version number. For instance, asking for an item to a factory older should in principle return 0x0.
 - the address (pointer) of the instance that has created / should create a object. 0x0 means the global factory while non null corresponds to clones of the global factory 
*/
class FactoryItem
{
private:
	std::string fWhichFactory;
	Short_t		fWhichFactoryVersion;
	std::string fWhichItem;	
	Version     fWhichVersion;
	
private:
	//! internal use for clone / allocation to know which instance of the factory has created this
	/*!
		0x0 means no factory so in case of first allocation, it asks the main factory
	 */
	void *		fPtrFactory;
	

public:
	FactoryItem();	
	FactoryItem(const Char_t* factname,const Char_t *whichitem, Version itemversion);
	FactoryItem(const Char_t* factname, Short_t factversion, const Char_t *whichitem, Version itemversion);
	FactoryItem(const FactoryItem &);	
	virtual ~FactoryItem() 
		{;}	
			
	const std::string &GetFactoryName() const 
		{ return fWhichFactory; }
	const std::string &GetItemName() const 
		{ return fWhichItem; }
	
	//! Ask the factory version
	Short_t GetFactoryVersion() const 
		{ return fWhichFactoryVersion; }
	//! Change the factory version 
	void SetFactoryVersion(Short_t v)
		{ fWhichFactoryVersion = v; }	
	
	Version GetVersion() const 
		{ return fWhichVersion; }
	//! change 
	void SetVersion(const Version v)
		{ fWhichVersion = v; }
	
	void *GetFactoryPtr() const 
		{ return fPtrFactory; }
	//! change 
	void SetFactoryPtr(void *ptr)
		{ fPtrFactory = ptr; }
			
	//! to change an item
	void SetItem(const Char_t *whichfactory, const Char_t *whichitem, Version itemversion, Short_t factversion = -1, void *ptr = 0x0);
	void SetItem(const Char_t *whichfactory, Short_t factversion, const Char_t *whichitem, Version itemversion, void *ptr = 0x0);
	
	void SetFactoryName(const char *fname)	
		{ fWhichFactory = fname; }
		
	void SetItemName(const char *iname)	
		{ fWhichItem = iname; }
			
	//! Compare two items
	/*!
		They have to have the same name, type and version number (see also IsAVersion)
	*/
	Bool_t operator == (const FactoryItem &) const;
	
	//! Compare two items
	/*!
	 Check if this is a version of the item given as argument. For that, they have to have the same factory name and the same type whatever the item version number.
	 This item should have a factory version number lower or equal to the passed one.
	*/	
	Bool_t IsAVersion(const FactoryItem &) const;
		
	void Print() const;
};

//! prototype function to control the evolution of a factory item.
typedef Bool_t (*PF_FactoryItemChange)(const FactoryItem &, const FactoryItem &);

//! change from item_from to item_to is allowed if they differ only by the item version number
inline Bool_t ChangeOfVersion(const FactoryItem &item_from, const FactoryItem &item_to)
{ 
	return item_from.IsAVersion(item_to); 
}
	
//! No changes allowed	
inline Bool_t NoChange(const FactoryItem &item_from, const FactoryItem &item_to)
	{ return item_from == item_to; }
	
//! Any changes (factory,type,version) allowed		
inline Bool_t AllChange(const FactoryItem & /*item_from*/, const FactoryItem & /*item_to*/)
	{ return true; }
		
} // namespace ADF

// operator ta read/write in a std stream
namespace std {
	std::ostream & operator << (std::ostream &  os, const ADF::FactoryItem &item) ;
	std::istream & operator >> (std::istream &  is, ADF::FactoryItem &item);
}

#endif





