/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file Frame.h header file for Frame.cpp */

#ifndef ADF_Frame
#define ADF_Frame
#define ADF_FRAME_H ADF_Frame

#ifndef ADF_Key
#include "Key.h"
#endif

#ifndef ADF_FrameFactory
#include "FrameFactory.h"
#endif

#ifndef ADF_PtrStack
#include "PtrStack.h"
#endif

#include <string>
#include <cstdio>

namespace ADF
{ 

class ConfigurationFrame;

//! Base class for a Frame
/*!
	A Frame is an association of a key and a buffer to fill in/out data.
	The creation of a new Frame is performed through FrameFactory.
	
	A block of memory can be linked to a Frame via the Link method. Once successfully linked, 
	any methods 
	provided by a particular Frame are available to deal with the content of the Frame.
	
	A Frame could be composed of sub-frames. 
	In this case IsComposite returns true (false otherwise = default).
	If a Frame is a composite frame, the Scan method operates on the data part to extract 
	a list of keys 
	for each sub-frame. This list of keys is kept in a stack. 
	The GetNbSubFrame gives the number of sub-frames found.
	
	The LinkSubFrame and CopySubFrame methods could be then used to extract a particular sub-frame.
	
	A "non-composite" frame contains data needed by an algorithm. 
	Thus high-level structures (kind of sub-events) are
	associated with a particular frame (see for instance PSAFrame). 
	These structures are all filled from the 
	Frame by calling the Read method. 
	As well, once you have fill them with the appropriate results from
	your algorithm, their content is written in the Frame to be sent back on the data flow.
*/
class Frame
{
public:
 	friend class FrameFactory;
private:
	//! To get the describtion of this frame as an element of a Factory
	FactoryItem fSignature; 
	//! true if a Frame is a Zombie    
	Bool_t fZombie;  
	//! Function used to decide if a key change is acceptable or not
// Migration SharedFP	PF_FactoryItemChange fKeyChange;
	//! Function used to decide if a Frame change is acceptable or not	
// Migration SharedFP	PF_FactoryItemChange fFrameChange;
	
private:
	Short_t fCheckLevel;
		        
protected:
	//! To avoid direct creation of a Frame
	Frame();  
	Frame(Key *);  
		
	//! to change the signature of that frame
	void SetSignature(FactoryItem item)
		{ fSignature = item ; }
		
protected:
	//! configure this from an input stream
	virtual Bool_t DoConfigure(std::istream &) 
		{ return false ; }
	//! write the configuration in an output stream	
	virtual Bool_t DoConfigure(std::ostream &) 
		{ return false ; }
	    
public:
	virtual ~Frame();
	
	//! The number of tests (and the speed !) once writing/reading in Frame increase with this level
	/*!
		Default value is 0, it means the lowest number of test but garantee at some point 
		the integrity of the data.
		
		More tests are checked by increasing the value 
		
		A negative value means potential corruption forseen ... 
		but available to really speed up the reading/writing !
	*/		
	virtual void SetCheckLevel(Short_t checklevel)
		{ fCheckLevel = checklevel; }
		
	virtual Short_t GetCheckLevel() const
		{ return fCheckLevel; }	
			
	//! Signature of that frame
	const FactoryItem &GetSignature() const
		{ return fSignature; }
	
	//! give access to the underlying buffer of the data part
	virtual const BufferIO *GetRealBuffer() const = 0;      
	
	//! To get the Key associated to this frame   
	virtual Key *GetKey() = 0; 
	
	//! To get the Key associated to this frame       
	virtual const Key *GetKey() const = 0; 
		
	//! To get the current status of this frame       
	virtual BaseBuffer::EStatus GetFrameStatus() const = 0;  
	
	//! To clear to kGood the current status of this frame           
	virtual void ClearFrameStatus() = 0; 
	
	//! total length for that frame 
	virtual UInt_t GetLength() const = 0; 
	
	//! Link a buffer to this Frame
	/*!
		It links both the key and the data part. Returns false if it cannot be correctly linked.
		
		The Link method is used to avoid copy of buffers. It should be used mainly if you only need to
		read the content of the Frame.
	*/        
	virtual Bool_t Link(const Char_t *, UInt_t, const char = 'b') = 0; 

	//! copy a buffer to this Frame
	/*!
	*/        
 	virtual Bool_t Copy(const Char_t *, UInt_t, const char = 'b') = 0; 
	
	//! Link a buffer to this Frame
	/*!
		it links both the key and the data part. Returns false if it cannot be correctly linked.
		The Link method is used to avoid copy of buffers. It should be used mainly if you only need to
		read the content of the Frame.
	*/        
	virtual Bool_t Link(const BufferIO &/*buf*/, UInt_t /*pos*/, const char /*opt*/ = 'b') = 0; 
			
	//! UnLink a buffer to this Frame	
	virtual void Unlink() = 0; 
	
	//! true if the underlying buffers have been linked.
	virtual Bool_t IsLinked() const = 0;    
	
	//! Export the content of the frame in a buffer
	/*!
		Export means a copy of the full content of the frame into a BufferIO
		return false if not enought place in the destination buffer.
	*/  		 
	virtual Bool_t Export(BufferIO &buf) const = 0;
	
	//! Export the content of the frame (data part only) in a buffer 
	/*!
		Export means a copy of the full content of the frame into a BufferIO
		return false if not enought place in the destination buffer.
	*/  		 
	virtual Bool_t ExportData(BufferIO &buf) const = 0; 
	
	//! copy a frame into this one. Theyy have to have the same prototype
	virtual Bool_t Copy(const Frame *) = 0;
	//! Link a frame into this one. They have to have the same prototype
	virtual Bool_t Link(const Frame *) = 0;
	
	//! It reads the content into dedicated structures from the Frame (data part)
	/*!
		return the number of bytes read (data part)
	*/    
	virtual UInt_t Read() 
		{ return 0u;}
	
	//! It writes to the Frame the content of the dedicated structures
	/*!
		return the number of bytes written (data part)
	*/    
	virtual UInt_t Write() 
		{ return 0u;}
			
	//! tells if this frame is a composite frame i.e. if it is composed of sub-frames
	/*!
	*/   
	virtual Bool_t IsComposite() const 
		{ return false; }
	
	//! Scan this Frame. If it is a composite frame, it looks for the keys of sub-frames
	/*!
		returns the sum of all the data parts of the sub-frames
	*/   
	virtual UInt_t Scan() 
		{ return 0u; }      
	
	//! Returns one of the sub-key in case this is a composite frame
	virtual const Key *GetSubKey(UInt_t) const = 0;
	
	//! Returns the number of sub-frames composing this frame. Scan have to be called first
	/*!
	*/   
	virtual UInt_t GetNbSubFrame() const 
		{ return 0u; }     
		
	//! to check if a subframe could be added to the composite frame   
	/*!
	
	*/
	virtual Bool_t IsASubFrame(const Frame &) 
		{ return false; }
			
	//! Add a subframe to that frame (only if composite)     
	/*!
	
	*/
	virtual Bool_t AddSubFrame(const Frame &) 
		{ return false; }
	
	//! Attach the sub-frame corresponding to the given subkey # to the Frame given in the second argument
	/*!
		The data part of the sub-frame is linked to the frame given as an argument.
		Its key is modified just to set the correct length using GetDataLength. The Key part is not linked
		because one may need to get a sub-frame 
	*/   
	virtual Bool_t LinkSubFrame(UInt_t , Frame *) 
		{ return false; }       
		
	//! copy the sub-frame #i to the Frame given in the second argument
	/*!
	*/   
	//virtual Bool_t CopySubFrame(Int_t, Frame *) { return false; }   
	
	//! transfert subframe whichsub from from_frame to this 

	/*!
	This is used, once you have added particular subframe with the AddSubFrame method of
	Frame, to transfert a subframe from the input frame to the output frame without taking
	care of the convertion of the key. The sub frame is simply added at the end of the destination
	frame. 
	If you need to access the added sub frame, apply the Scan method first.  
	*/
	virtual Bool_t TransfertSubFrame(const Frame & /*from_frame*/, UInt_t /*whichsub*/)
		{ return false; }       	
	
	//! Reset the current frame
	/*!
		remove all keys from the list of sub-keys, set the full buffer to 0 and offset to 0
		Set length in key to 0.
		
		If you call the method often, setting the full buffer to 0 may take time. If you don't need it, 
		better to re-implement this method in your frame and to use FastReset() 
	*/   
	virtual void Reset() = 0; 
	
	//! FastReset the current frame, means the data part keep the previous values 
	virtual void FastReset() = 0; 
		
	//! Dump the content of the frame in a file (for debugging)
	/*!
		if true, the file is open in writing mode otherwise in append mode
	*/
	virtual void Dump(const Char_t *filename, Bool_t anew = false) const = 0; 
	virtual Bool_t Dump(FILE *) const = 0;
		
	//! load a Frame from a C file (for debugging)
	virtual Bool_t Load(FILE *) = 0;  
		
	//! Fill the frame with a given pattern (used for ADF test)  
	/*!
		default is a sequence (repetition times) based on the definition of the frame : \n	
		|factoryname:factoryitem[Major,Minor]|
	*/
	virtual void Stallion(UInt_t repetition = 4u) = 0;
		
	//! configuration from/to a file
	virtual Bool_t Configure(const char *name, const char *option);
	//! to reconfigure the interface for that frame
	virtual Bool_t Configure(ConfigurationFrame *,  const char *option); 	
};

class ConcreteFrame : public Frame
{           
protected:
	BufferIO *fBuffer;	// the buffer that streams in/out the content of the Frame	
	Key *fKey;		// the key to map the frame
        
protected:    
	virtual BufferIO *RealBuffer() 
		{ return fBuffer; }   

private:
	//! 
	// ConcreteFrame(); 
    
protected:    
	//!     
	ConcreteFrame(Key *);    
    
	virtual Bool_t LinkKey(const Char_t *buf, UInt_t size)   
		{ return fKey->Link(buf,size) ; } 
		
	virtual Bool_t CopyKey(const Char_t *buf, UInt_t size)   
		{ return fKey->Copy(buf,size) ; }
    	
	virtual Bool_t LinkFrame(const Char_t *buf, UInt_t size)   
		{ return fBuffer->Link(buf,size); } 
		
	virtual Bool_t CopyFrame(const Char_t *buf, UInt_t size)   
		{ return fBuffer->Copy(buf,size); }
		      
public:      
	virtual ~ConcreteFrame();
	
	virtual const BufferIO *GetRealBuffer() const 
		{ return fBuffer; } 
	
	//! To get the Key associated to this frame   
	virtual Key *GetKey() 
		{ return fKey; }  
	
	//! To get the Key associated to this frame       
	virtual const Key *GetKey() const 
		{ return fKey; }  
		
	//! To get the current status of this frame       
	virtual BaseBuffer::EStatus GetFrameStatus() const 
		{ return fBuffer->GetStatus(); } 
	
	//! To clear to kGood the current status of this frame           
	virtual void ClearFrameStatus() 
		{ fBuffer->SetStatus(BaseBuffer::kGood); }
	
	//! total length for that frame 
	virtual UInt_t GetLength()  const 
		{ return fKey->GetFrameLength(); }    
	
	//! Link a buffer to this Frame
	/*!
		It links both the key and the data part. Returns false if it cannot be correctly linked.
		
		The Link method is used to avoid copy of buffers. It should be used mainly if you only need to
		read the content of the Frame.
	*/        
// 	virtual Bool_t Link(Char_t *buf, UInt_t size) 
// 	{ 
// 		return LinkKey(buf,size) && LinkFrame(buf+fKey->GetKeyLength(),size-fKey->GetKeyLength()); 
// 	}
	virtual Bool_t Link(const Char_t *buf, UInt_t size, const char opt = 'b')
	{
		if ( opt == 'k' )
			return LinkKey(buf,size);
		if ( opt == 'f' )
			return LinkFrame(buf,size);	
			
		return (LinkKey(buf,size) && LinkFrame(buf+fKey->GetKeyLength(),size-fKey->GetKeyLength())); 
	} 
// 	virtual Bool_t Copy(Char_t *buf, UInt_t size)
// 	{
// 		return false;
// 	}
	virtual Bool_t Copy(const Char_t *buf, UInt_t size, const char opt = 'b') 	
	{
		if ( opt == 'k' )
			return CopyKey(buf,size);
		if ( opt == 'f' )
			return CopyFrame(buf,size);	
			
		return (CopyKey(buf,size) && CopyFrame(buf+fKey->GetKeyLength(),size-fKey->GetKeyLength())); 
	}
		
	//! copy a buffer to this Frame
	/*!
	*/        
//   	virtual Bool_t Copy(const Char_t *buf, UInt_t size)
// 	{
// 		return CopyKey(buf,size) && CopyFrame(buf+fKey->GetKeyLength(),size-fKey->GetKeyLength());
// 	}
	
	//! Link a buffer to this Frame
	/*!
		it links both the key and the data part. Returns false if it cannot be correctly linked.
		The Link method is used to avoid copy of buffers. It should be used mainly if you only need to
		read the content of the Frame.
	*/        
	virtual Bool_t Link(const BufferIO &buf, UInt_t pos, const Char_t opt = 'b') 
	{ 
		if ( opt == 'k' )
			return fKey->Link(buf,pos,fKey->GetKeyLength());
		if ( opt == 'f' )
			return fBuffer->Link(buf,pos+fKey->GetKeyLength(),fKey->GetDataLength());	
			
		return fKey->Link(buf,pos,fKey->GetKeyLength()) 
			&& fBuffer->Link(buf,pos+fKey->GetKeyLength(),fKey->GetDataLength()); 
	} 
			
	//! Copy a buffer to this Frame
	/*!
		it links both the key and the data part. Returns false if it cannot be correctly linked.
		The Link method is used to avoid copy of buffers. It should be used mainly if you only need to
		read the content of the Frame.
	*/        
// 	virtual Bool_t Copy(BufferIO &buf, UInt_t pos) 
// 		{ return fKey->Copy(buf,pos,fKey->GetKeyLength()) 
// 			&& fBuffer->Copy(buf,pos+fKey->GetKeyLength(),fKey->GetDataLength()); }
		
	//! UnLink a buffer to this Frame	
	virtual void Unlink() 
		{ fKey->Unlink(); fBuffer->Unlink(); }
	
	virtual Bool_t IsLinked() const
		{ return fBuffer->IsLinked() && fKey->IsLinked(); }       
	
	//! copy the content of a Frame into this one
	virtual Bool_t Copy(const Frame *);
	//! Link a frame into this one. They have to have the same prototype
	virtual Bool_t Link(const Frame *);
	
	//! Export the content of the frame in a buffer
	/*!
		Export means a copy of the fill content of the frame into a BufferIO
		return false if not enought place in the destination buffer.
	*/  		 
	virtual Bool_t Export(BufferIO &buf) const 
	{ 
		Bool_t ok = fKey->Export(buf);
		UInt_t length =  fKey->GetDataLength();

//		ok = ok	&& buf.Import((*fBuffer),fKey->GetDataLength()) == fKey->GetDataLength();
		ok = ok	&& (buf.Import((*fBuffer),length) == length);
		return ok;
	}
	
	//! Export the content of the frame (data part only) in a buffer 
	/*!
		Export means a copy of the fill content of the frame into a BufferIO
		return false if not enought place in the destination buffer.
	*/  		 
	virtual Bool_t ExportData(BufferIO &buf) const
		{ return BufferIO::Copy((*fBuffer),buf,fKey->GetDataLength()); }   
	
				
	//! tells if this frame is a composite frame i.e. if it is composed of sub-frames
	/*!
	*/   
	virtual Bool_t IsComposite() const 
		{ return false; }
	
	//! Scan this Frame. If it is a composite frame, it looks for the keys of sub-frames
	/*!
		returns the sum of all the data parts of the sub-frames
	*/   
	virtual UInt_t Scan() 
		{ return 0u; }      
	
	//! Returns one of the sub-key in case this is a composite frame
	virtual const Key *GetSubKey(UInt_t) const 
		{ return fKey; }
	
	//! Returns the number of sub-frames composing this frame. Scan have to be called first
	/*!
	*/   
	virtual UInt_t GetNbSubFrame() const 
		{ return 0u; }       
	
	//! Add a subframe to that frame (only if composite)     
	/*!
	
	*/
	virtual Bool_t AddSubFrame(const Frame &) 
		{ return false; }
	
	//! Attach the sub-frame corresponding to the given subkey to the Frame given in the second argument
	/*!
		The data part of the sub-frame is linked to the frame given as an argument.
		Its key is modified just to set the correct length using GetDataLength. 
		The Key part is not linked
		because one may need to get a sub-frame 
	*/   
	virtual Bool_t LinkSubFrame(UInt_t , Frame *) 
		{ return false; }        
		
	//! copy the sub-frame #i to the Frame given in the second argument
	/*!
	*/   
	//virtual Bool_t CopySubFrame(Int_t, Frame *) { return false; }   
	
	//! Reset the current frame
	/*!
		remove all keys from the list of sub-keys, set the full buffer to 0 and offset to 0
		Set length in key to 0.
		
		If you call the method often, setting the full buffer to 0 may take time. 
		If you don't need it, 
		better to re-implement this method in your frame and to use FastReset() 
	*/   
	virtual void Reset() 
		{ fBuffer->Reset(); fBuffer->SetOffset(); fKey->SetDataLength(0u); }  
	
	//! FastReset the current frame, means the data part keep the previous values 
	virtual void FastReset() 
		{ fBuffer->SetOffset(); fKey->SetDataLength(0u); } 
		
	//! Dump the content of the frame in a file (for debugging)
	/*!
		A file is open the first time this method is called. Once open, any new call will write 
		the frame in this file so that all the frame will be in it. If you want to change close the 
		file and open a new one, set anew to true.
		
		The file is open in append mode, so remember to delete it if you don't want to keep 
		previous records.
	*/
	virtual void Dump(const Char_t *filename, Bool_t anew = false) const ;  
	virtual Bool_t Dump(FILE *) const;
	
	//! load a Frame from a C file (for debugging) - return the number of bytes read
	virtual Bool_t Load(FILE *);  
		
	//! Fill the frame with a given pattern (used for debugging)  
	/*!
		default is a sequence based on the definition of the frame : \n	
		factoryname:factoryitem
	*/
	virtual void Stallion(UInt_t repetition = 4u);
}; 
// ConcreteFrame inline members //////////////////////////////////////////////////////

//! A RawFrame gives direct access to the underlying buffer
/*!
	\warning It is the responsability of the user to Read/Write correctly from/to the buffer
*/
class RawFrame : public ConcreteFrame
{
public:      
	RawFrame(Key *key): ConcreteFrame(key) 
		{;}	
	virtual ~RawFrame()
		{;}	
		
	//! It just returns the current size of the buffer
	/*!
	*/  
	virtual UInt_t Read() 
		{ return GetKey()->GetDataLength(); }
	
	//! It just set the size of the frame from the current position in the buffer
	/*!
		Starting from an offset equal to zero, when the buffer is filled 
		just call this method.
	*/    
	virtual UInt_t Write() 
		{ GetKey()->SetDataLength(GetRealBuffer()->Offset()); return GetKey()->GetDataLength(); }	
	
	//! This method gives access in reading/writing mode to the underlying data buffer
	virtual BufferIO &RawBuffer()
		{ return (*RealBuffer()); } 
};

//! A Shared Frame Pointer
/*!
	The Frame is owned by the pointer. In case of delete, the frame is deleted.
	To avoid this:
 
	\code
	SharedFP *sfp;
	...
	Frame *f = sfp->GetFrame(0x0);
	delete sfp;
	// nwow it is the charge of the user to delete f
	\endcode
*/	
class SharedFP
{	
private:
	Frame *fFrame;
private:
	//! true if valid, can be used in case a pointer is valid but the frame is not set properly
	Bool_t fIsValid;
private:
	//! Function used to decide if a key change is acceptable or not
	PF_FactoryItemChange fKeyChange;
	//! Function used to decide if a Frame change is acceptable or not	
	PF_FactoryItemChange fFrameChange;
	
public:	
	explicit SharedFP( Frame *f = 0x0 );
	virtual ~SharedFP() 
		{ if (fFrame) delete fFrame ; fFrame = 0x0; }

	//! true if it is a valid pointer
	virtual Bool_t IsValid() const
		{ return fIsValid; }
	//! true if it is a valid pointer
	virtual void SetValid(Bool_t validity) 
		{ fIsValid = validity; }
	
	//! change the frame, returns the current one 
	virtual Frame *SetFrame(Frame *);
	
	virtual Frame * GetFrame() const
		{ return fFrame; }
	
	//! true if for that Frame the key modification proposed is acceptable 
	/*!
	 */	
	Bool_t IsKeyChangeable(FactoryItem item_from, FactoryItem item_to) const
		{ return fKeyChange(item_from,item_to); }
	
	virtual void SetKeyChangeFunction(PF_FactoryItemChange pf)
		{ fKeyChange = pf; }	
	
	const PF_FactoryItemChange &GetKeyChangeFunction() const
		{ return fKeyChange; }
	
	//! true if for that Frame the frame modification proposed is acceptable
	/*!
	 */	
	Bool_t IsFrameChangeable(FactoryItem item_from, FactoryItem item_to) const
		{ return fFrameChange(item_from,item_to); }
	
	virtual void SetFrameChangeFunction(PF_FactoryItemChange pf)
		{ fFrameChange = pf; }
	
	const PF_FactoryItemChange &GetFrameChangeFunction() const
		{ return fFrameChange; }	
};

} // namespace ADF
#endif




