/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file FrameBlock.h header file for FrameBlock.cpp */

#ifndef ADF_FrameBlock
#define ADF_FrameBlock

#ifndef ADF_ConfAgent
#include "ConfAgent.h"
#endif
#ifndef ADF_Frame
#include "Frame.h"
#endif

namespace ADF
{ 
	
	class Key;
	
	//! A Block of Frames
	/*!	
	 A block of Frames is a source composed of sequential Frames in reading or writing mode. 
	 Methods to add/get frames are to be provided by the concrete implementation.
	 */   
	class FrameBlock 
	{
	private:
		//! This block is in reading or writing mode
		ConfAgent::EMode fMode;
	private:	
		//! True is this is at the end of the block
		/*!
		 In reading mode, no more Frame coming 
		 In writing mode, cannot write any mode ...
		 */
		Bool_t fEoB; 
	private:	
		//! block name, should be unique !
		// std::string fName;
	private:
		//! current owner of the frame block : allow checking that the ressource has been free
		void *fOwner;	
		
	protected:
		//! return true if the output is expandable
		/*!
		 */
		virtual Bool_t Expand(UInt_t) = 0;
		
	public:
		FrameBlock(ConfAgent::EMode mode = ConfAgent::kRead): 
		fMode(mode), 
		fEoB(false),
		/* fName("AFrameBlock"), */
		fOwner(0x0)
		{;}  
		virtual ~FrameBlock() 
		{;}
		
		//! Return the operating mode 		
		ConfAgent::EMode GetModeIO() const 
		{ return fMode; }
		virtual void SetModeIO(ConfAgent::EMode mode) 
		{ fMode = mode; }
		Bool_t IsMode(ConfAgent::EMode mode) const 
		{ return fMode == mode; }
		
		Bool_t IsEoB() const
		{ return fEoB; }
		
		void SetEoB(Bool_t b)
		{ fEoB = b; }
		
		//! Name of the block ... used to distinguish several blocks in a topology
		/*
		 virtual void SetName(std::string name)
		 {
		 fName = name;
		 }
		 virtual std::string GetName() const 
		 {
		 return fName;
		 }
		 */
		//! change the owner of this block. In case of multi-threading, be careful to lock properly this part
		virtual void *SetOwner(void * owner = 0x0)
		{
			void *current = fOwner; fOwner = owner; return current;
		}
		virtual Bool_t IsOwner(void * owner = 0x0) const
		{
			return fOwner == owner;
		}	
		
		//! To get the address of the underlying buffer
		virtual BufferIO *GetBufferIO(UInt_t which = 0u) = 0;
		// virtual const BufferIO *GetBufferIO(UInt_t which = 0u) = 0;
		
		//!  true if there is a new frame available in the block
		/*!
		 */	
		virtual Bool_t NextFrame(Key *) = 0; 
		
		//! the new frame is wrapped in a BufferIO object for transportation between the different classes
		/*!
		 which explanation: in case of several entries.
		 */
		virtual BufferIO &GetFrame(UInt_t which = 0u) = 0; 
		
		//! Reserve the given size for writing 
		/*!
		 in case you would like to add several frames and you would like to
		 be sure it is possible.
		 
		 if it is not possible, return false. 
		 
		 If switch_eob is set to true, it set end of block if Reserve is not possible
		 */	
		virtual Bool_t Reserve(UInt_t, Bool_t /*switch_eob*/ = false, UInt_t /*which*/ = 0u) 
		{ return false; }
		
		//! size of the current block
		virtual Long64_t GetSize(UInt_t = 0u) const
		{ return 0; }	
		//! max size of the block
		virtual Long64_t GetMaxSize(UInt_t = 0u) const
		{ return 0; }			
		//! Current position
		virtual Long64_t GetOffset(UInt_t = 0u) const
		{ return 0; }
		
		//! Add a Frame to the block.  
		/*!
		 It returns false is cannot add the Frame to the Block
		 */
		virtual Bool_t AddFrame(const BufferIO &, UInt_t = 0u) 
		{ return false; }
		
		virtual Bool_t AddFrame(const Frame &, UInt_t = 0u)
		{ return false; }	
		
		//! Rewind to be ready to be used again
		/*!
		 if this in in writing mode and rewinded in reading mode, 
		 then the current position is set as the size of the block.
		 
		 Thus the same block could be used, filled first then 
		 once the filling is performed, just Rewind in reading mode to start reading
		 */
		virtual void  Rewind(ConfAgent::EMode mode = ConfAgent::kRead) = 0;
		
		//! keep the current position	
		virtual void Snapshot() 
		{;}
		//! back to the last position	
		virtual void BackToSnapshot() 
		{;}	
		
		virtual UInt_t NumberOfChannels() const = 0; 
		
		//! Attach a external block to this
		/*!
		 Should be possible only for certain type of blocks.
		 
		 if r is set as an option (default) then rewind is called :
		 - if the mode is kRead then the attached buffer should be full of data
		 and with Rewind called, the FrameBlock is ready to be read \n
		 - if the mode is kWrite then whatever there is in the attached buffer
		 it will be overwritten because Rewind is called. It is a way to redirect 
		 the output to a given piece of memory.
		 
		 if r is not set is means the EoB has been reached. An explicit call of Rewind should be done.
		 
		 To detach the external block, call SetBlock()
		 */	
		virtual Bool_t SetBlock(Char_t * /*buf */ = 0x0, UInt_t /*size*/ = 0u, char /*opt_rewind*/ = 'r', char /*opt_link_or_copy*/ = 'l', UInt_t /*which*/ = 0u) = 0; 	
		
		//! It is the charge of the user to be sure the source and destination are compatible 
		//virtual Bool_t MoveBlock( FrameBlock &, char /*opt_rewind*/ = 'r', char /*opt_link_or_copy*/ = 'l', UInt_t /*which*/ = 0u)
		//	{ return false; } 	
		
	};
	
	//! It implements an in-memory block of Frames
	/*!	
	 By default, it is not Expandable. See ExpandableBlock for such a block.
	 */   
	class OneBlock : public FrameBlock 
	{		
	protected:
		UInt_t fLastSnapshot;
		
		//! The buffer to look for Frames
		/*!
		 Just used for global manipulations so no need for an endian 
		 adapted buffer. 
		 Readings are performed though the Key which has the rigth endian (hopefully)
		 */
		BufferIO fBuffer;
		//! to wrap the current Frame 
		BufferIO fFrame;
		
	protected:
		//! This kind of Block is not expandable.
		/*!
		 */
		virtual Bool_t Expand(UInt_t) 
		{ return false ; }
		
	public:
		OneBlock(UInt_t size = aMByte, ConfAgent::EMode mode = ConfAgent::kRead): FrameBlock(mode), fLastSnapshot(0u), fBuffer(size), fFrame()
		{;}  
		virtual ~OneBlock() 
		{;}
		
		virtual Long64_t GetSize(UInt_t = 0u) const
		{ 
			if (GetModeIO()==ConfAgent::kRead) {
				return fBuffer.Size();
			}
			else return fBuffer.Offset(); 
		}
		virtual Long64_t GetMaxSize(UInt_t = 0u) const
		{ 
			return fBuffer.Size(); 
		}
		
		//! Current position
		virtual Long64_t GetOffset(UInt_t = 0u) const
		{ 
			return fBuffer.Offset(); 
		}
		
		virtual BufferIO *GetBufferIO(UInt_t /*which*/ = 0u)
		{ return &fBuffer; }
		
		//! return true if there is a new frame available (in reading mode)
		/*!
		 */	
		virtual Bool_t NextFrame(Key *); 
		
		//! Return a buffer that is linked with the part of the buffer that corresponds to the next frame
		/*!
		 */
		virtual BufferIO &GetFrame(UInt_t = 0u) 
		{ return fFrame; } 
		
		//! check if enough place to write something (writing mode)
		virtual Bool_t Reserve(UInt_t, Bool_t /*switch_eob*/ = false, UInt_t /*which*/ = 0u);
		
		//! keep the position	
		virtual void Snapshot() 
		{ fLastSnapshot = fBuffer.Offset(); }
		//! back to the last position	
		virtual void BackToSnapshot() 
		{ fBuffer.SetOffset(fLastSnapshot); }	
		
		//!  
		/*!
		 */
		virtual Bool_t AddFrame(const BufferIO &, UInt_t i = 0u);
		virtual Bool_t AddFrame(const Frame &, UInt_t i = 0u);
		
		//!  
		/*!
		 */
		virtual void  Rewind(ConfAgent::EMode mode = ConfAgent::kRead);	
		
		virtual UInt_t NumberOfChannels() const 
		{ return 1u; }
		
		//! Attach a external block to this
		/*!
		 */	
		virtual Bool_t SetBlock(Char_t * /*buf*/ = 0x0, UInt_t /*size*/ = 0u, char /*opt_rewind*/ = 'r', char /*opt_link_or_copy*/ = 'l', UInt_t /*which*/ = 0u); 	
		
		//! It is the charge of the user to be sure the source and destination are compatible i.e avoid using with FileBlock in argument
		// virtual Bool_t MoveBlock( FrameBlock &, char /*opt_rewind*/ = 'r', char /*opt_link_or_copy*/ = 'l', UInt_t /*which*/ = 0u); 		
		// 
		/*
		 virtual void View()
		 {
		 UInt_t off; 
		 
	    fBuffer.Offset();
		 fBuffer.SetOffset();	
		 
		 UInt_t length, message;
		 fBuffer >> length >> message;
		 std::cout << std::dec	
	    << length 
	    << " 0x" 
	    << std::hex
	    << message << std::dec << std::endl;
		 
		 fBuffer.SetOffset(off);	
		 }
		 */
	};
	
	//! It implements a black hole for Frames
	/*!	
	 Cannot get any Frame and all Frames could be added 
	 */   
	class NullBlock : public OneBlock 
	{
	protected:
		virtual Bool_t Expand(UInt_t) 
		{ return true ; }
		
	public:
		NullBlock(ConfAgent::EMode mode = ConfAgent::kRead): OneBlock(aByte,mode)
		{ ; }  
		virtual ~NullBlock() {;}
		
		virtual Bool_t NextFrame(Key *) 
		{ return false; }
		virtual BufferIO &GetFrame(UInt_t = 0u)
		{ fFrame.SetEffectiveSize(0u); return fFrame; } 
		
		virtual Bool_t Reserve(UInt_t, Bool_t /*switch_eob*/ = false, UInt_t /*which*/ = 0u)
		{ return true; }
		
		virtual Bool_t AddFrame(const BufferIO &, UInt_t = 0u)
		{ return true; }
		virtual Bool_t AddFrame(const Frame &, UInt_t = 0u)
		{ return true; }	
	};
	
	//! It implements an expandable block of Frame
	/*!	
	 */ 
	class ExpandableBlock : public OneBlock 
	{
	protected:
		//! implements Expand method
		virtual Bool_t Expand(UInt_t);
		
	public:
		ExpandableBlock(UInt_t size = aMByte, ConfAgent::EMode mode = ConfAgent::kRead): OneBlock(size,mode)
		{;}  
		virtual ~ExpandableBlock() 
		{;}
	};
	
	//! Frames are extracted/added directly, one by one from/to a file
	/*!	
	 */ 
	class FileBlock : public OneBlock
	{
	private:
		Long64_t fCurrentSize;	
		Long64_t fMaxSize;
		
		Long64_t fFileSnapshot;
		
	private:
		FILE *fFile;
		
	protected:
		void Reset();
		
	public:
		FileBlock(UInt_t size = aMByte, ConfAgent::EMode mode = ConfAgent::kRead);   
		virtual ~FileBlock() 
		{;}
		
		virtual Long64_t GetSize(UInt_t = 0u) const
		{
			if (GetModeIO()==ConfAgent::kRead) {
				return fMaxSize;
			}
			else return fCurrentSize; 
			
		}
		virtual Long64_t GetMaxSize(UInt_t = 0u) const
		{ 
			return fMaxSize; 
		}
		
		//! Current position
		virtual Long64_t GetOffset(UInt_t = 0u) const
		{ 
			return fCurrentSize; 
		}	
		
		virtual void SetFile(FILE *file, Long64_t max_size = kMaxUInt_t)
		{ 
			Reset(); 
			if (GetModeIO()==ConfAgent::kRead) { // reading mode, take the value from the file 
#if  !defined(_MSC_VER)
				// keep current position 
				Long64_t c = ::ftell(file);
				// now get the end of file
				::fseek(file, 0L, SEEK_END);
				fMaxSize = ::ftell(file);
				// back to current position			
				::fseek(file, c, SEEK_SET);
#else
				// WINDOWS, where the use of 64 bits is not automatic as in LINUX 
				// keep current position
				Long64_t c = _ftelli64(file);
				// now get the end of file
				_fseeki64(file, __int64(0), SEEK_END);
				fMaxSize = _ftelli64(file);
				// back to current position			
				_fseeki64(file, c, SEEK_SET);
#endif
			}
			else fMaxSize = max_size; 
			fFile = file; 
		}
		
		virtual Bool_t Reserve(UInt_t, Bool_t /*switch_eob*/ = false, UInt_t /*which*/ = 0u);
		
		//! Read next frame from file
		/*!
		 In this case, GetBufferIO contains only one Frame and the content is then identical to what is obtained from GetFrame
		 */
		virtual Bool_t NextFrame(Key *);
		
		//! keep the position	
		virtual void Snapshot() 
		{ 
			fFileSnapshot = fCurrentSize; 
		}
		//! back to the last position	
		virtual void BackToSnapshot() 
		{ 
			if ( fFile ) 
#if !defined(_MSC_VER)
				::fseek(fFile,fFileSnapshot,SEEK_SET); 
#else
			_fseeki64(fFile,fFileSnapshot,SEEK_SET); 
#endif
			fCurrentSize = fFileSnapshot; 
		}	
		
		virtual void  Rewind(ConfAgent::EMode = ConfAgent::kRead) 
		{ 
			fCurrentSize = 0u; 
			if ( fFile ) 
				::fseek(fFile,0u,SEEK_SET); 
		}
		
		virtual Bool_t SetBlock(Char_t * /*buf*/= 0x0, UInt_t/*size*/= 0u, char /*opt_rewind*/ = 'r', char /*opt_link_or_copy*/ = 'l', UInt_t/*which*/ = 0u)	
		{ return false; }
		
		virtual Bool_t AddFrame(const BufferIO &, UInt_t i = 0u);
		virtual Bool_t AddFrame(const Frame &, UInt_t i = 0u);
	};
	
} // namespace ADF
#endif


