/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file FrameFactory.h header file for FrameFactory.cpp */

#ifndef ADF_FrameFactory
#define ADF_FrameFactory
#define ADF_FRAMEFACTORY_H ADF_FrameFactory

#ifndef ADF_KeyFactory
#include "KeyFactory.h"
#endif
#ifndef ADF_PtrStack
#include "PtrStack.h"
#endif
#ifndef ADF_DotConf
#include "DotConf.h"
#endif
#include <string>


namespace ADF
{ 

class Frame;
class SharedFP;
class MainFrameFactory;

//! Base class for a FrameFactory
/*!
*/
class FrameFactory : public DotConf
{
	friend class MainFrameFactory;

private:
	// used only by MainFactory
	FrameFactory();
	
private:
	//! Unique name for this FrameFactory
	std::string fName;		
	
protected: 
	//! for classes inheriting of this
	FrameFactory(std::string name); 

	//! Do really the job of allocating the frame   
	virtual SharedFP *DoNew(Key *, const FactoryItem &) 
		{ return 0x0; }
	
protected: 
	//! realised clone of this factory
	virtual FrameFactory *Clone(const std::string & = "full") = 0;
	
	// change the name of the factory in case default constructor has been used
	void SetName(std::string name) 
	{
		fName = name;
	}
	
public:  
	//! 
	virtual ~FrameFactory()
		{;}
	
	//! name of the factory 
	/*!
	 It should be a unique name (class name related) since the name is used to look for a particular factory.
	 */      
	virtual const std::string &GetName() const 
	{
		return fName;
	}
	
	//! Ask this factory for a new Frame
	/*!
		Key method for all New* like methods. It should never be virtual. Call NewImp which is Implementation dependent
	*/  
	virtual SharedFP *NewSharedFrame(const FactoryItem &, const FactoryItem &);  
	//
	virtual Frame *New(const FactoryItem &key_item, const FactoryItem &frame_item);
};
// FrameFactory inline members //////////////////////////////////////////////////////
	
//! Base class for a MainFactory i.e. a factory composed of factories
/*!
 */
class MainFrameFactory : public FrameFactory  
{	
private:
	static MainFrameFactory gTheMainFactory;
	
private:
	//! list of Factories owned by this main factory
	std::vector <FrameFactory *> fFactories;
	//! list of Factories cloned by this main factory
	std::vector <FrameFactory *> fClonedFactories;

private:
	virtual FrameFactory *Clone(const std::string & = "full") 
		{ return 0x0 ; } // the main factory cannot be cloned
	
	//! Ask for a particular factory from is name  
	/*!
	 */     
	FrameFactory *GetClonedFactory(std::string which_factory, void *) const;
	
public: 
	//!     
	MainFrameFactory(std::string factory_name);  
	virtual ~MainFrameFactory();  
	
	//! direct access to the list of factories 
	const std::vector <FrameFactory *> &ListOfFactories() const
		{ return fFactories; }
	
	//! Ask the main for a new Frame
	/*!
	 Using the item it first search for the factory then send the job to it.
	 */  
	virtual SharedFP *NewSharedFrame(const FactoryItem &, const FactoryItem &);  
	//! just a frame
	virtual Frame *New(const FactoryItem &key_item, const FactoryItem &frame_item);
	
	//! To Clone a SharedFrame / Frame
	/*!
	 */   
	virtual SharedFP *Clone(const SharedFP *);  
	//!
	virtual Frame *Clone(const Frame *frame);
	
	//! Clone a factory
	/*!
	 With option full, a complete clone of the factory is returned. Without, only a new, empty instance.
	 This can be used if the configure is called later on for insyance in the DFAgent.
	 */
	FrameFactory *Clone(std::string which_factory, const std::string &opt = "full");
	//! Clone all factories. return the number of successfully cloned factories and the list of pointer added to the argument 
	Int_t Clone(std::list<FrameFactory *> &, const std::string &opt = "full");	
	
	//! Add one factory to this collection of keyfactories: name has to be unique so return false if a factory with the same name exists
	Bool_t Register(FrameFactory *);  
	//! Add one factory pre-defined factory by its name
	Bool_t Register(std::string factory_name);  
	
	//! Ask for a particular factory from is name  
	/*!
	 */     
	FrameFactory *GetFactory(std::string which_factory) const;
	
	//! check a factory has been registered    
	Bool_t IsKnown(std::string which_factory) const ;   
	
	//! the main (global) keyfactory
	static MainFrameFactory &theMainFactory();
};
	
	
//! Base class for any kind of Frame factory
/*!
*/
template <typename Factory_T>
class AFrameFactory : public FrameFactory
{
private:
	static Factory_T *gTheFactory;
    
protected:
	AFrameFactory (const char *factory_name) : FrameFactory(factory_name) 
	{ 
		if ( gTheFactory == 0x0 )
			MainFrameFactory::theMainFactory().Register(this);  // only the global one is registered. Others are cloned
	}
	virtual ~AFrameFactory () 
	{ 
		if ( fLog.IsDebug(5) ) {

		}
		/* std::cout << "destroying singleton." << std::endl;*/ 
	}
	
	//!
	virtual SharedFP *DoNew(Key *, const FactoryItem &) = 0; 

	//! Clone
	virtual FrameFactory *Clone(const std::string &opt = "full")
	{
		fLog.SetProcessMethod("ADF::AFrameFactory<>::Clone(const std::string &)"); 

		Factory_T * newfact = new Factory_T ();
		if ( newfact == 0x0 ) {
			fLog << error << "Cannot allocate a new cloned factory of " << GetName() << dolog; 
			return 0x0;
		}
		//
		newfact->SetName(GetName());
		//
		if ( opt.find("full") == std::string::npos  ) { // only constructor = empty instance, otherwise full clone
			
			fLog << info << "A new empty clone has been created for the factory " << GetName() << " " << nline; 
			
			if ( fLog.IsDebug(2) ) {
				fLog.SetLevel(LogMessage::kInfo,1);
				DoConfigure( fLog() );
				fLog.SetLevel(LogMessage::kInfo,0);
			}
			fLog << dolog;
			
			return newfact;
		}
		// use DotConf facilities to clone
		Bool_t ok = true;
		//
		std::ostringstream o; 
		ok = ok &&
			this->DoConfigure(o);
		//
		// std::string tmp;
		// tmp = o.str();
		std::istringstream i(o.str());
		ok = ok &&
			newfact->DoConfigure(i,true); // 	
		
		if ( !ok ) { 
			fLog << error << "the clone configuration has failed, delete it !  " << nline; 
			
			delete newfact; 
			newfact = 0x0; 
		}
		else {
			fLog << info << "the clone configuration has succeded !  " << nline; 

			if ( fLog.IsDebug(2) ) {
				fLog.SetLevel(LogMessage::kInfo,1);
				DoConfigure( fLog() );
				fLog.SetLevel(LogMessage::kInfo,0);
			}			
		}

		fLog << dolog;
		
		return newfact;
	}
	
public:
	//! to create the main (singleton) factory of that type
	static Factory_T *theFactory ()
	{
		if (gTheFactory == 0x0) {
#ifdef DEBUGSTATIC				
			std::clog << "[[static]]  Creation in Framefactory ]\n[...\n";
#endif			
			gTheFactory = new Factory_T;
#ifdef DEBUGSTATIC			
			if ( gTheFactory ) 
				std::clog << gTheFactory->GetName() << " done\n...]\n";
			else 
				{ std::clog << " *** ERROR *** \n"; std::clog << "...]\n"; }
#endif
		}
		else { /* std::cout << "singleton already created!" << std::endl; */ }
	
		return (static_cast<Factory_T*> (gTheFactory));
	}
	//! Call Configure for all the registered factories
	virtual Bool_t DoConfigure(std::ostream &out) 
		{ return DotConf::DoConfigure(out); }
	//! Call Configure for all the registered factories
	virtual Bool_t DoConfigure(std::istream &in, Bool_t allow_init) 
		{ return DotConf::DoConfigure(in, allow_init); } 
};

template <typename Factory_T>
Factory_T *AFrameFactory<Factory_T>::gTheFactory = 0x0;

} // namespace ADF
#endif





