/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file FrameIO.h header file for FrameIO.cpp */

#ifndef ADF_FrameIO
#define ADF_FrameIO
#define ADF_FRAMEIO_H ADF_FrameIO

#ifndef ADF_FrameBlock
#include "FrameBlock.h"
#endif
#ifndef ADF_ConfAgent
#include "ConfAgent.h"
#endif
#ifndef ADF_PtrStack
#include "PtrStack.h"
#endif
#ifndef ADF_Counter
#include "Counter.h"
#endif

#include <iostream>
#include <vector>

namespace ADF
{ 

class Frame;
class DFTrigger;

//! Base class to get/write Frames
/*!
	This is the base class to get/write Frames in a data flow. It has a unique type (see GetType()).
	
	It is a kind of state machine with several possible status. 
		- kUndefined
		- kRunning
		- kStopped
		- kSuspended
		- kIdle
		- kFinished
		
	It could be used just to get frames (kRead), to send frames (kWrite) or both at the same time (kReadWrite).
	
	A FrameIO processes a data flow with many different kind of frames going through. To get the particular
	frame you are interested in, you should Register using a FrameTrigger. 
	
	\code
	
	\endcode
*/
class BaseFrameIO
{   
public:
	//! current state of the FrameIO
	enum EStatus 
		{ kUndefined = 0, kRunning = 1, kStopped = 2, kSuspended = 3, kIdle = 4, kFinished = 5 };

private:
	//! Name of this Frame IO manager
	std::string fName;	
	//! current status 
	EStatus fStatus;
    
private:
	//! Agent that keeps the definition of the data flow as seen by this 
	ConfAgent *fConfAgent; 
    
protected:
	//! to send messages to the log server   
	LogMessage Log;  
	   
public:    
	BaseFrameIO(const char *);  
	virtual ~BaseFrameIO();
	
	//! string corresponding to the current status
	virtual const char *GetStringStatus(EStatus) const;
	
	//! Name of this FrameIO
	virtual const std::string &GetName() const 
		{ return fName; } 
		
	virtual void SetName(const char *name)
		{ fName = name; } 
					
	virtual const Char_t *GetType() const 
		{ return "BaseFrameIO"; }   
	
	void  SetModel(ConfAgent::EModel model) ;        
	
	void  SetStatus(EStatus stat) 
		{ fStatus = stat; }
		
	//! Return the agent which keeps the current configuration for that FrameIO 
	/*!
	*/    
	ConfAgent *GetConfAgent() const 
		{ if ( fConfAgent == 0x0 ) ConfAgent::theGlobalAgent(); return fConfAgent; } 
		
	//! 
	/*!
	*/
	ConfAgent::EModel GetModel() const ; 
	//! 
	/*!
	*/
	EStatus GetStatus() const 
		{ return fStatus ; }     
	
	//! Print some informations (statistics)
	/*!
	*/ 	
	virtual void Print(std::ostream &out = std::cout) const;    
};


//! One stream as an input and one in output
/*!
	Used for Producer (in this case input is a FileBlock), Filter and Consumer (in this case output is a FileBlock)
*/
class FrameIO : public BaseFrameIO
{    
private:    
	//! Default Input and output in which frames are read/written (NullBlock)
	FrameBlock *fDefaultInFB;     
	FrameBlock *fDefaultOuFB;   
		
	//! Current Input and output in which frames are read/written
	FrameBlock *fInFB;     
	FrameBlock *fOuFB;  
	
	//! Primary key to read the data flow
	Key *fKeyIn;
	//! trigger on a global reconfiguration frame  
	DFTrigger *fConfIO;
	//! stack of all registered (valid) triggers.
	PtrStack<DFTrigger> fRegisteredTriggers; 
	//! stack of running (on the input) triggers means suitable for the current dataflow
	PtrStack<DFTrigger> fRunningTriggers; 
	//! stack of write only triggers i.e. for triggers that just add Frames on the Data flow or check things
	PtrStack<DFTrigger> fWriteOnlyTriggers; 		
	//! stack of triggers that have became zombie
	PtrStack<DFTrigger> fZombieTriggers;
			
	//! counters
	//! the number of time a global configuration frame has been found
	Counter fGlobalCounter;	
	//! untriggered frame read/written		
	Counter fNoTrigCounterIN;
	Counter fNoTrigCounterOUT;
	//! details per trigger	
	std::vector<Counter> fTrigCountersIN;
	std::vector<Counter> fTrigCountersOUT;
	
	//! number of frames read, write, and global trigger rate  
	UInt_t fRead, fWritten, fTriggerRate; 
	
	//! Option : Rewind or not in case a new block is attached. Defaut is true
	Bool_t fDoRewind;

protected:
	Bool_t InitFromAgent();
     
protected:
	//! to know if there is a next frame of the dataflow 
	/*!
		fInFB->Snapshot() is called in case something the output cannot be filled
	*/
	virtual Bool_t NextFrame();

	Bool_t DoGlobalReconfiguration();
// 	void DoFrameReconfiguration();	  	
	//     void DoNotATrigger();
	
	//! check if a trigger is suitable for the dataflow
	virtual Bool_t Check(DFTrigger *, PtrStack<DFTrigger> &, PtrStack<DFTrigger> &);	
              
public:   
	FrameIO(const char *);  
	virtual ~FrameIO() ;
	
	//! if set true, rewind is done once a block is attached. Return the previous value
	Bool_t DoRewind(Bool_t rewind = true) 
	{
		Bool_t old = 
			fDoRewind;
		fDoRewind = rewind;
		
		return old;
	}
	
	//! to check whether or not a global reconfiguration has occured
	virtual const DFTrigger *GetGlobalConfigurationTrigger()
		{ return fConfIO; }
	
	//! write the current configuration in the output data
	/*!
		if a path is given, it is also saved in a file (to control things)
	*/
	virtual Bool_t RecordGlobalConfiguration(const std::string = "NO");
	
	const FrameBlock *GetCurrentBlockIN() const
		{ return fInFB; }
	const FrameBlock *GetCurrentBlockOUT() const
		{ return fOuFB; }
	UInt_t GetFramesRead() const
		{ return fRead; }
	UInt_t GetFramesWritten() const
		{ return fWritten; }
	
	//! What kind of data flow is it ?
	virtual const Char_t *GetType() const 
		{ return "FrameIO"; }
	
	//! To register a trigger 
	/*!
		It returns true if this trigger is correctly registered
	*/
	virtual Bool_t Register(DFTrigger *);
	
	//! To retrieve a trigger by its unique name	
	virtual DFTrigger *GetTrigger(const Char_t *);
	
	//! To know the number of registered triggers   
// 	virtual UInt_t NumberOfTriggers() const 
// 		{ return fRegisteredTriggers.GetSize() ; }
		
	//! To Notify that at least one trigger has fired 
	/*!
	 It takes the current Frame on the data flow and checks whether or
	 not it fullfills one of the Registered Trigger. It returns true if al least one has fired.
	 
	 External (added by users with Register) and internal Triggers exist (for instance the one to take
	 into account a global reconfiguration of the data flow). 
	 By default only external triggers returns true. To allow this method to Notify 
	 an internal trigger has been fired, call it with true as argument (default is false).
	 
	 Using this, a user can be informed of a global reconfiguration, that comes from instance
	 for a new run, with :
	 
	 \code 
	 while ( GetFrameIO().Notify(true) ) {
		...
		if ( GetFrameIO().GetGlobalConfigurationTrigger()->IsFired() ) {
			// a global reconfiguration Frame has been found ... 
			// check for instance the run number and do something appropriate in case of changes

			continue;
		}
	 }
	 \endcode
	*/    
	virtual Bool_t Notify(Bool_t with_inner_trigger = false);
	
	//! To record a frame trigger
	/*!
		In writting mode, once a particular frame has been created, it is add to the data flow 
		with this Record method.
	*/        
	virtual Bool_t Record(DFTrigger *);
	virtual Bool_t Record();    
		
	//! Print some informations (statistics)
	/*!
	*/ 	
	virtual void Print(std::ostream &out = std::cout) const;   
	
	//! Attach a block to this. Rewind called if DoRewind set to true (Default)
	virtual void Attach(FrameBlock *in, FrameBlock *out);
	virtual void Detach(FrameBlock *in, FrameBlock *out);     
}; 

} // namespace ADF
#endif





