/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file GObjects.h header file for GObject.cpp */

#ifndef ADF_GObject
#define ADF_GObject

#include "DataHandling.h"
#include "BaseBuffer.h"

#include <vector>
	
namespace ADF
{ 

//! Base class for a Global Object
/*!
	Data Frame are composed of structures that are related to the algorithm which produces them.
	For intance a PSAFrame is composed of a list of Hits, whatever is the real data members stored 
	in the Frame (ex: with or without errors).
	
	Data cannot be always organized into specific structures, but could be just a list of data 
	identified by a unique name and a given basic type.
	This object is used to help the user to read/modify such data item.
	
            - S  : a 16 bit signed integer (Short_t)
            - US : a 16 bit unsigned integer (UShort_t)
            - I  : a 32 bit signed integer (Int_t)
            - UI : a 32 bit unsigned integer (UInt_t)
            - F  : a 32 bit floating point (Float_t)
            - D  : a 64 bit floating point (Double_t)
            - L  : a 64 bit signed integer (Long64_t)
            - UL : a 64 bit unsigned integer (ULong64_t)
            - B  : a boolean (Bool_t)
	    
	A GObject is then a list of NamedItem that represents the current definition of that Object.
	To add a item to the list:
	\code
	GObject ob1;
	
	// add items to ob1
	ob1.AddItem( new NamedItem<UShort_t>("Crys",0u) );
	ob1.AddItem( new NamedItem<Float_t>("E",0.0) );
	ob1.AddItem( new NamedItem<Int_t>("T0",0) );
	\endcode
	Items are owned by ob1 and are thus deleted by it ... so the user should not care about that.
	As a consequence, the following is forbidden:
	\code
	GObject ob1; NamedItem<UShort_t> cry("Crys",0u);
	
	// add items to ob1
	ob1.AddItem( &crys );
	\endcode
	
	NamedItem are known by their name and type ... as a consequence, you cannot 
	add twice an item with the same name and type.
	
	In your program, you can establish a link between a NamedItem and a local 
	variable (of the same type !) so that changing the value of your local 
	variable will changed the item's one. As well you can get back the value of the itemin you program
	To do that:
	\code	
	GObject ob1;
	
	// add items to ob1
	ob1.AddItem( new NamedItem<UShort_t>("Crys",0u) );
	
	// link it with a variable in youy code:
	UShort_t Crys_link; 
	ob1.LinkItem("Crys",&Crys_link);
	
	Crys_link = 1u; 
	ob1.SetItems(); // now the value Crys in ob1 is equal to 1u
	
	// ... another code changes the Crys value in ob1 to 2u
	
	ob1.GetItems(); // now your local variable Crys_link is equal to 2u
	\endcode	
	You can only link once a NamedItem ... otherwise it would be a mess !
	
	You can however consult (reading mode) a NamedItem using proxy
	This is done like this :
	\code	
	GObject ob1;
	
	// add items to ob1
	ob1.AddItem( new NamedItem<UShort_t>("Crys",0u) );
	const NamedItem<UShort_t> *Crys_prox = ob1.Get<UShort_t>("Crys");
	// warning, it returns 0x0 is not a good name with the right type
	
	cout << Crys_prox->Get() << endl;
	\endcode
		
	see ADFObject.C for concrete exemples
	\author O.Stezowski
*/
class GObject
{ 
public:	
	//! a full item is composed of a pointer to an item, an pointer to a proxy and a pointer to a datalink
	typedef std::pair<ANamedItem *, std::pair<ANamedItem *,DataLink *> > FullItem;
		
private:
	//! List of data items composing this Object - used for 'fast' streaming
	std::vector <ANamedItem *> fDataItems;
	//! List of registered datalink, used for 'fast' synchronization of internal/external values
	std::vector <DataLink *> fDataLinks;
private:
	//! safely remove this link from the list of active links (fDataLinks) and delete it !
	void UnLink(DataLink *);
private:
	//! to keep the full information of items (idle), links and proxy
	std::vector< FullItem > fArrayOfItems;
	std::vector< FullItem > fArrayOfIdleItems;
	
private:	
	//! to check if the given item is defined in this object
	FullItem *IsAnItem(const Char_t *, const Char_t *);
	FullItem *IsAnItem(const ANamedItem *);
	
	//! to check if the given item is idle in this object	
	FullItem *IsAIdleItem(const Char_t *, const Char_t *);

public:   
	//! Constructor
	GObject();
	
	//! Destructor
	/*!
		All of them are deleted by this destructor (item and idleitems)
		together with the links and proxy
	*/    
	virtual ~GObject();
		
	//! Add a named item to this. The item belongs to GObject and will be thus deleted
	/*!
		If a item with the same name and type has already been added, 
		the passed item is not added and the method returns false.
		
		To add an item:
	\code
	GObject ob1;
	// add items "Crys" of type unsigned short to ob1
	ob1.AddItem( new NamedItem<UShort_t>("Crys",0u) );
	\endcode
		the following is forbidden !	
	\code
	GObject ob1; NamedItem<UShort_t> cry("Crys",0u);
	// add items "Crys" of type unsigned short to ob1
	ob1.AddItem( &crys );
	\endcode
	*/
	virtual Bool_t AddItem(ANamedItem *item);
	
	//! to get a given item (reading mode only !!) 
	template <typename Data_t> const NamedItem<Data_t> *Get(const char *name)
	{ 
		static Data_t t; ANamedItem *proxy = 0x0; FullItem *fitem = 0x0;

		// ckeck if the asked item is an item for that object or not 
		fitem = IsAnItem(name,NamedItemTypeOf(t));  
		if ( fitem ) {
			if ( fitem->second.first == 0x0 )    // not yet created
				proxy = fitem->second.first = fitem->first->MakeAProxy();
			else 
				proxy = fitem->second.first; // already there
		}
		else { // add it on the list of passed items
			// check is already in the proxy list
			fitem = IsAIdleItem(name,NamedItemTypeOf(t));	
			if ( fitem ) // item in the passed item collection
				if ( fitem->second.first )
					proxy = fitem->second.first;
				else
					proxy = fitem->second.first = fitem->first->MakeAProxy();
			 else { // create an item, its proxy and add it to the list of IdleItems
			 
				FullItem p;
				p.first		= new NamedItem<Data_t>(name);
				p.second.first	= p.first->MakeAProxy(); 
				p.second.second	= 0x0; 
				
				fArrayOfIdleItems.push_back( p );
				
				proxy = p.second.first;
			}	
		}
		if ( proxy ) 
			return dynamic_cast< const NamedItem<Data_t> * >(proxy); 
		else
			return 0x0;	
	}
	//! General method to Link one item with an external address
	template <typename Data_t> Bool_t Link(const char *name, Data_t *external_address)
	{ 
		Bool_t ok = false; FullItem *fitem = 0x0;
		
		// ckeck if the asked item is an item for that object or not 
		fitem = IsAnItem(name,NamedItemTypeOf(*external_address));  
		if ( fitem ) {
			if ( fitem->second.second == 0x0 ) {   // not yet linked so can do it
			
				fitem->second.second = fitem->first->MakeADataLink(external_address);
				fDataLinks.push_back(fitem->second.second);
				
				ok = true;
			} /* else {;} NO, BECAUSE ONLY ONE LINK POSSIBLE ! */	
		}
		else { // add it on the list of passed items
			// check is already in the passed list
			fitem = IsAIdleItem(name,NamedItemTypeOf(*external_address));	
			if ( fitem ) { //  already there
				if ( fitem->second.second == 0x0 ) {   // not yet linked so can do it
			
					fitem->second.second = fitem->first->MakeADataLink(external_address);
					fDataLinks.push_back(fitem->second.second);

				} /* else {;} NO, BECAUSE ONLY ONE LINK POSSIBLE ! */
			}
			else { // create an item, its proxy and link add it to the list of IdleItems

				FullItem p;
				p.first		= new NamedItem<Data_t>(name);
				p.second.first	= 0x0; 
				p.second.second	= p.first->MakeADataLink(external_address); 
				
				fArrayOfIdleItems.push_back( p );
			}	
		}		
		return ok;
	}
	
	//! Only for basic type ... otherwise better to use the most general Link method
	Bool_t LinkItem(const Char_t *, const Char_t *, void *);
// 	Bool_t LinkItem(const Char_t *name, Bool_t *address);
	Bool_t LinkItem(const Char_t *name, Short_t *address);
	Bool_t LinkItem(const Char_t *name, UShort_t *address);
	Bool_t LinkItem(const Char_t *name, Int_t *address);
	Bool_t LinkItem(const Char_t *name, UInt_t *address);
	Bool_t LinkItem(const Char_t *name, Float_t *address);
	Bool_t LinkItem(const Char_t *name, Double_t *address);
	Bool_t LinkItem(const Char_t *name, Long64_t *address);      
	Bool_t LinkItem(const Char_t *name, ULong64_t *address);  

	//! to check if all the items are linked with external variables	
	virtual Bool_t IsFullyLinked() const
		{ return fDataLinks.size() == fDataItems.size(); }    
			
	//! write the value of all items into the buffer	
	virtual void SetItems(BufferIO &) const;	
	//! read the value of all item from the buffer	
	virtual void GetItems(const BufferIO &);
	
	//! Copy external variables into locals
	virtual void SetItems();
	//! Copy local variables into externals
	virtual void GetItems() const ;      
	
	//! Set local to default 
	virtual void  Reset();
	
	//! transfert proxy and link for one GObject to another one
	/*!
		\warning The destination object should be a newly created objects
		because in this function, the destination object is supposed to be with no proxy 
	*/
	void TransferLinkAndProxy(GObject &to);
	
	//! print some informations about this object	
	virtual void Print(std::ostream &out = std::cout) const ;
};

//! 
/*!
*/
class Anonymous
{
	static const UShort_t kMaxSize = 256u;
public: 
	UShort_t fRealSize;
	Char_t   fData[kMaxSize];
public:
	Anonymous(Char_t c = 0) { fRealSize = 0u; ::memset(fData,c,kMaxSize); }
	virtual ~Anonymous() 
		{;}
		
	template<typename Data_T> Data_T Get(UShort_t offset = 0u) const
	{	
		if ( (offset+sizeof(Data_T)) < kMaxSize ) 
			return (*(Data_T*)(fData+offset));
		else 
			return 0;
	}
	template<typename Data_T> void Set(const Data_T data, UShort_t offset = 0u)
	{ 
		if ( (offset+sizeof(Data_T)) < kMaxSize ) 
			*((Data_T*)(fData+offset)) = data;
	}		
	
	friend std::ostream & operator << (std::ostream &, const Anonymous & );
	friend BufferIO & operator << (BufferIO &, const Anonymous & );
 	friend const BufferIO & operator >> (const BufferIO &, Anonymous &);

	Anonymous operator = (const Anonymous & a);
	
//	virtual void Print(std::ostream &out = std::cout) const {;}

};
	
std::ostream & operator << (std::ostream &, const Anonymous & );
BufferIO & operator << (BufferIO &, const Anonymous & );
const BufferIO & operator >> (const BufferIO &, Anonymous &);

const char *NamedItemTypeOf(Anonymous/*s*/);

//! Base class for a DataInterface
/*!
	A DataInterface is composed of Objects that have a sense for the given algorithms
	
	A DataInterface is at least composed of a list of NamedItem through GObject
*/
class DataInterface 
{
private: 
	GObject fGObject;
protected:
	void ResetGlobal()
		 { fGObject.Reset(); }
public:
	DataInterface(): 
		fGObject() {;}
	virtual ~DataInterface() 
		{;}

	virtual GObject *Global()
		{ return &fGObject; }
};

} // namespace ADF
#endif





