/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file Key.h header file for Key.cpp */

#ifndef ADF_Key
#define ADF_Key
#define ADF_KEY_H ADF_Key

#ifndef ADF_BufferIO
#include "BufferIO.h"
#endif

#ifndef ADF_FactoryItem
#include "FactoryItem.h"
#endif

#include <typeinfo>

namespace ADF
{ 

class KeyFactory;
class ConcreteFrame;

//! Base class for a Key
/*!
	A key is a (short) buffer that helps to map the data flow. One of the most simple Key is, 
	for instance, just an integer giving the size of the frame in which are stored data.
	
	A Key should not be allocated directly but through a KeyFactory

	To add new keys:
		For the consistancy of the system, your key have to inherits of this base class and
		you MUST use the constructor Key(BaseBuffer::EEndian, UInt_t s, const Char_t *); 
		(see for instance the class AgataKey)
	 
*/
class Key
{   
public:
 	friend class KeyFactory;  
 	friend class ConcreteFrame;  
	   
private:
	//! to keep needed informations to clone and compare keys
	/*
	class KeySignature : public FactoryItem
	{
	public:
		KeySignature(std::string);
		
		std::string fType;	// type of this key ... corresponding to the interface
		UInt_t	fUniqueID;	// KeyID as delivered by the KeyFactory, helps to compare Keys
	};
	KeySignature fSignature; */
	FactoryItem fSignature;
	
protected:  
	Version GetVersion() const 
		{ return fSignature.GetVersion() ; }
	    
protected:      
	//! Used by specific methods
	/*!
		The buffer in which is written the real content of the data part of the frame 
		should stay unreachable by users. 
		However, it might be useful to reach it directly in some cases (see TransfertSubFrame)
	*/         
	virtual BufferIO *RealBuffer() 
		{ return fBuffer; }     
    
private:    
	//! to prevent people trying to put signed value in the buffer for Length
	void SetDataLength(Int_t)  
		{;}
	void SetDataLength(Short_t)
		{;}
            
protected:
	//! the buffer that fills in/out the Key
	BufferIO *fBuffer;
	   
private:
	//! to avoid direct allocation
	Key(); 
    
protected:    
	//! Constructor to be used by any Key that inherits from this class
	/*!
		It fixes the endian type, the length of the key.
	*/
	Key(BaseBuffer::EEndian, UInt_t s);   
      
public:      
	virtual ~Key();
	
	const FactoryItem &GetSignature() const
		{ return fSignature; }
			
	virtual const BufferIO *GetRealBuffer() const 
		{ return fBuffer; }
		
	virtual BaseBuffer::EStatus GetKeyStatus() const 
		{ return fBuffer->GetStatus();}
	virtual void ClearKeyStatus()              
		{ fBuffer->SetStatus(BaseBuffer::kGood);}
	
	virtual BaseBuffer::EEndian GetEndian() const 
		{ return fBuffer->GetEndian(); }
	
	virtual void SetDataLength(UShort_t) 
		{ fBuffer->SetStatus(BaseBuffer::kCorrupt) ;} 
	virtual void SetDataLength(UInt_t)   
		{ fBuffer->SetStatus(BaseBuffer::kCorrupt) ;}   
	
	virtual UInt_t   GetDataLength() const 
		{ return 0u; }	
	virtual UInt_t  GetFrameLength() const 
		{ return GetKeyLength() + GetDataLength(); }	  
	
	virtual const Char_t *GetAddress() const 
		{ return fBuffer->GetAddress(); }     
		
	//! Link an external buffer to this key
	/*!
		The second argument is here just to be sure the attached buffer does not have a size 
		smaller that the key because it might be a problem.
		
		It returns false in case the buffer cannot be linked, true otherwise
	*/        
	virtual Bool_t Link(const Char_t *, UInt_t);	
		
	//! Link an external buffer to this key
	/*!
		The second argument is here just to be sure the attached buffer does not have a size 
		smaller that the key because it might be a problem.
		
		It returns false in case the buffer cannot be linked, true otherwise
	*/        
	virtual Bool_t Link(const BufferIO &buf, UInt_t pos, UInt_t size);
	
	//! Copy an external buffer to this key	
	/*!
		It is similar to Link in the behavior
	*/
	virtual Bool_t Copy(const Char_t *, UInt_t);
			
	//! Unlink  the buffer attached to this key.
	virtual void Unlink() 
		{ fBuffer->Unlink(); }
	
	virtual Bool_t IsLinked() const
		{ return fBuffer->IsLinked(); }   
	
	//! copy the content of an external key to this key
	/*!
		Partial copy are not allowed. So the two keys have to have the same signature (and the same length).
		If you do need such a behavior, overwrite this method in your key.
		It returns false in case the buffer cannot be copied, true otherwise
	*/        
	virtual Bool_t Copy(const Key *);     
	
	//! to copy the content of the key to an external buffer
	/*!
		The key is exported at the current position in the buffer. Try to expand if needed
	*/
	virtual Bool_t Export(BufferIO &buf) const 
	{ 
		Bool_t ok = 
			(buf.Import((*fBuffer),fBuffer->Size()) == fBuffer->Size());
		return ok;
	}  
	
	//! Unique number corresponding to a type of Key
	/*!
		This unique id is attributed by the KeyFactory. All key that have the same interface
		shares a common id
	*/       
//	UInt_t GetUniqueID() const 
//		{ return fSignature.fUniqueID; }
		
	//! Key name as a char string i.e. the name of the class
	/*!
		MANDATORY: used by the KeyFactory to assigned a unique number to a kind of key
	*/    
//	virtual const std::string &GetKeyType() const 
//		{ return fSignature.fType ; }
		
	//! Length of the key in bytes
	/*!
	*/   
	virtual UInt_t GetKeyLength() const 
		{ return fBuffer->Size(); }
	
	//! Compares two keys.     
	/*!
		By default two keys are equal if they have the same signature
	*/      
	virtual Bool_t IsAKey(const Key *akey) const
	{
		return fSignature == akey->GetSignature(); 
	}
	
	//! convert a key to another key ... not an easy task ... use it carefully
	/*!
		By default, converting a key means just copy the data length
	*/
	virtual Bool_t Convert(const Key *akey) 
	{ 
		// to test if the length could be stream from one key to another one (ex : length could be too high converting int to short)
		fBuffer->SetStatus(); 
		// copy data /frame length
		SetDataLength(akey->GetDataLength()); 
		
		return fBuffer->GetStatus() == BaseBuffer::kGood; 
	}
		
	//! Print the content of the key        
	virtual void Print(std::ostream &out = std::cout) const;
};
// Key inline members //////////////////////////////////////////////////////

} // namespace ADF
#endif





