/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file KeyFactory.h header file for KeyFactory.cpp */

#ifndef ADF_KeyFactory
#define ADF_KeyFactory
#define ADF_KEYFACTORY_H ADF_KeyFactory

#include "BaseBuffer.h"
#include "FactoryItem.h"
#include "DotConf.h"

#include <iostream>
#include <string>
#include <vector>

namespace ADF
{ 

class Key;
class MainKeyFactory;
		
//! Base Key factory
/*!
*/	
class KeyFactory : public DotConf  
{		
	friend class MainKeyFactory;

private:
	//! Name of the factory ... should be UNIQUE !
	std::string fName;
	//! Endian type (needed for Key initialisations)
	BaseBuffer::EEndian fEndian;
	
private:
	KeyFactory()
		{;} 
	
protected:
	//! Do really the job of allocating the key
	virtual Key *DoNew(const FactoryItem &) 
		{ return 0x0; }

protected:
	//! realised clone of this factory
	virtual KeyFactory *Clone(const std::string &) = 0;
	
	// change the name of the factory in case default constructor has been used
	void SetName(std::string name) 
	{
		fName = name;
	}
	
public: 
	//! Constructor
	KeyFactory(std::string factory_name, BaseBuffer::EEndian e = BaseBuffer::kLittle) : DotConf("ADF::KeyFactory"), fName(factory_name), fEndian(e)
		{;}  
	//! destructor
	virtual ~KeyFactory()
		{;}
	
	//! build a key using item. 
	/*!
	 it should not be overwritten since it takes care to set the signature of the key i.e. maily it keeps the address
	 of the factory that creates the key. this is used by the key clone procedure !
	 The real job of allocating the key is done by DoNew which could be overwritten in class that inherits from this.
	 This DoNew method focuses only on creating the key.
	*/
	virtual Key *New(const FactoryItem &);
	
	//! name of the factory 
	/*!
		It should be a unique name (class name related) since the name is used to look for a particular factory.
	 */      
	virtual const std::string &GetName() const 
	{
		return fName;
	}
	//! Endian type for the keys delivered by this factory  
	virtual BaseBuffer::EEndian GetEndian() const
	{
		return fEndian;
	}	
	//! Set Endian type for the keys delivered by this factory  
	virtual void SetEndian(BaseBuffer::EEndian e)
	{
		fEndian = e;
	}	
	
// some static to access the global main key factory
public:	
	//! to add a new default KeyFactory to the main factory  
	/*!
	 Once registered, the factory will be deleted here. The user MUST NOT do it himself
	 */      
	//static Bool_t Register(const Char_t *);  
	
	//! to get a pointer to the main factory   
	/*!
	 It should be called at least once at the beginning to build the main factory. 
	 By default, the Default and the Agata factory are also built and added to the main factory.
	 */      
	//static MainKeyFactory &theMainFactory();
	
	//! Ask the main factory for a Key
	/*!
	 It looks for a particular factory (by its name) and sent the request to it
	 i.e calls its New(const Char_t *which_key, Version which_key_version, Short_t in_factory_version) method
	 */       
	// static Key *New(const FactoryItem &item); 
	//static Key *New(const Char_t *which_factory, const Char_t *which_key, Version version);
	
	//! to get a particular factory from the main factory    
	/*!
	 */     
	//static KeyFactory *GetFactory(std::string which_factory);
	
	//! check if a factory has been registered  to the main factory  
	//static Bool_t IsKnown(std::string which_factory) 
	//	{ return KeyFactory::GetFactory(which_factory) != 0x0 ; }    
};

//! Base class for a MainFactory i.e. a factory composed of factories
/*!
*/
class MainKeyFactory : public KeyFactory  
{	
private:
	static MainKeyFactory gTheMainFactory;
	
private:
	//! list of Factories owned by this main factory
	std::vector <KeyFactory *> fFactories;
	//! list of Factories cloned by this main factory
	std::vector <KeyFactory *> fClonedFactories;
	
private:	
	MainKeyFactory(std::string factory_name, BaseBuffer::EEndian e = BaseBuffer::kLittle);  

private:
	virtual KeyFactory *Clone(const std::string &) 
		{ return 0x0 ; } // the main factory cannot be cloned

	//! Ask for a particular factory from is name  
	/*!
	 */     
	KeyFactory *GetClonedFactory(std::string which_factory, void *) const;
	
public: 
	//!     
	virtual ~MainKeyFactory();  

	//! build a key using item. 
	/*!
		Using the item it first search for the factory then send the job to it.
	 */
	virtual Key *New(const FactoryItem &);
	
	//! Clone an existing key. Use the key signature to get the factory used to create the passed key, create a new one and copy the content
	Key *Clone(const Key *key);
	
	//! Clone a factory. 
	KeyFactory *Clone(std::string, const std::string &opt = "full");
	//! Clone all factories. return the number of successfully cloned factories and the list of pointer added to the argument 
	Int_t Clone(std::list<KeyFactory *> &, const std::string &opt = "full");	
	
	//! direct access to the list of factories 
	const std::vector <KeyFactory *> &ListOfFactories() const
		{ return fFactories; }
	
	//! Add one factory to this collection of keyfactories: name has to be unique so return false if a factory with the same name exists
	Bool_t Register(KeyFactory *);  
	//! Add one factory pre-defined factory by its name
	Bool_t Register(std::string factory_name);  
	
	//! Ask for a particular factory from is name  
	/*!
	 */     
	KeyFactory *GetFactory(std::string which_factory) const;
	
	//! check a factory has been registered    
	Bool_t IsKnown(std::string which_factory) const ;   
	
	//! the main (global) keyfactory
	static MainKeyFactory &theMainFactory();
	
	//! list all factories 
	void ls(std::string o) const;
};
	
//! Base class for any KeyFactory implementation
/*!
	It takes care of creating a global factory and registers it to the global main key factory
*/
template <typename Factory_T>
class AKeyFactory : public KeyFactory
{	
private:
	static Factory_T *gTheFactory;
	
protected:
	AKeyFactory (const char *factory_name, BaseBuffer::EEndian e = BaseBuffer::kLittle) : KeyFactory(factory_name,e) 
	{ 
		if ( gTheFactory == 0x0 )
			MainKeyFactory::theMainFactory().Register(this); // only the global one is registered. Others are cloned 
	}
	virtual ~AKeyFactory () { /* std::cout << "destroying singleton." << std::endl; */  }
	
	//! Have to be overwritten
	virtual Key *DoNew(const FactoryItem &) = 0;
	
	//! Clone
	virtual KeyFactory *Clone(const std::string &opt = "full")
	{
		fLog.SetProcessMethod("ADF::AKeyFactory<>::Clone(const std::string &)"); 

		Factory_T * newfact = new Factory_T ();
		if ( newfact == 0x0 ) {
			fLog << error << "Cannot allocate a new cloned factory of " << GetName() << dolog; 			
			return 0x0;
		}
		//
		newfact->SetName(GetName()); newfact->SetEndian(GetEndian());
		//
		if ( opt.find("full") == std::string::npos  ) { // only constructor = empty instance, otherwise full clone

			fLog << info << "A new empty clone has been created for the factory " << GetName() << " " << nline; 
			
			if ( fLog.IsDebug(2) ) {
				fLog.SetLevel(LogMessage::kInfo,1);
				DoConfigure( fLog() );
				fLog.SetLevel(LogMessage::kInfo,0);
			}
			fLog << dolog;
			
			return newfact;
		}
		// use DotConf facilities to clone
		Bool_t ok = true;
		//
		std::ostringstream o; 
		ok = ok &&
				this->DoConfigure(o);
		//
		std::istringstream i(o.str());
		ok = ok &&
				newfact->DoConfigure(i,true); // 	
		
		if ( !ok ) { 
			fLog << error << "the clone configuration has failed, delete it !  " << nline; 
			
			delete newfact; 
			newfact = 0x0; 
		}
		else {
			fLog << info << "the clone configuration has succeded !  " << nline; 
			
			if ( fLog.IsDebug(2) ) {
				fLog.SetLevel(LogMessage::kInfo,1);
				DoConfigure( fLog() );
				fLog.SetLevel(LogMessage::kInfo,0);
			}			
		}
		
		fLog << dolog;
		return newfact;
	}
	
public:
	//! to create the main global factory for that type
	static Factory_T *theFactory ()
	{
		if (0x0 == gTheFactory) {
#ifdef DEBUGSTATIC					
			std::clog << "[[static]]  Creation in Keyfactory ]\n[...\n";
#endif
			gTheFactory = new Factory_T;
#ifdef DEBUGSTATIC					
			if ( gTheFactory ) 
				std::clog << gTheFactory->GetName() << " done\n...]\n";
			else 
				{ std::clog << " *** ERROR *** \n"; std::clog << "...]\n" ;}	
#endif			
		}
		else { /* std::cout << "singleton already created!" << std::endl; */ }
	
		return (static_cast<Factory_T*> (gTheFactory));
	}
	
	//! just call DotConf method, may be be overwitten
	virtual Bool_t DoConfigure(std::ostream &out)
		{ return DotConf::DoConfigure(out); }
	//! just call DotConf method
	virtual Bool_t DoConfigure(std::istream &in, Bool_t allow_init) 
		{ return DotConf::DoConfigure(in, allow_init); }
};

template <typename Factory_T>
Factory_T *AKeyFactory<Factory_T>::gTheFactory = 0x0;

} // namespace ADF
#endif





