
/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file MetaFrame.h header file for MetaFrame.cpp */

#ifndef ADF_MetaFrame
#define ADF_MetaFrame

#include "AgataDataFrame.h"
#include "GObject.h"
#include "Hits.h"

#include <cmath>

namespace ADF
{ 

//! A vertex
class VertexInterface : public DataInterface
{
protected:
	//! Vertex definition for internal use
	template<typename Precision_T> struct Vertex {
	
		Vertex() { ::memset(this,0,sizeof(Vertex<Precision_T>)); }
		
		Precision_T P[3];
		
		Precision_T D[3];
		Precision_T Beta;
		
		Precision_T DP[3];
		
		Precision_T DD[3];
		Precision_T DBeta;
		
		void Reset() { ::memset(this,0,sizeof(Vertex<Precision_T>)); } 
	};
	
public:
	VertexInterface() : DataInterface() 
		{;} 	
	virtual ~VertexInterface() 
		{;} 

	//! Set the position of the source (last argument is used in case the position depends on time)
	virtual void SetPosition(Double_t  , Double_t  , Double_t   ,Double_t  = 0.0) = 0;
	//! get the position of the source (last argument is used in case the position depends on time)	
	virtual void GetPosition(Double_t &, Double_t &, Double_t & ,Double_t  = 0.0) const = 0;
	//! set the direction of the source (last argument is used in case the position depends on time)	
	virtual void SetDirection(Double_t  , Double_t  , Double_t  ,Double_t  = 0.0) = 0;
	//! get the direction of the source (last argument is used in case the position depends on time)	
	virtual void GetDirection(Double_t &, Double_t &, Double_t &,Double_t  = 0.0) const = 0;
	
	//! set recoil velocity
	virtual void SetBeta(Double_t, Double_t  = 0.0) = 0;
	//! get recoil velocity	
	virtual Double_t GetBeta(Double_t = 0.0) const = 0;
	virtual Double_t GetGamma(Double_t = 0.0) const
		{ return 1.0 / ::sqrt ( 1 - GetBeta()*GetBeta() ); };
		
	virtual void DopplerShift(Hit *) = 0;
	virtual void DopplerCorrection(Hit *) = 0;
	
	virtual Double_t DopplerShift(const Hit *) const = 0;
	virtual Double_t DopplerCorrection(const Hit *) const = 0;
		
	//! apply doppler shift
	/*!
		-e : energy
		-beta  : v/c
		-gamma : 1/(1-beta*beta)1/2
		-costheta : NOT theta !
	*/
	static Double_t DopplerShift(Double_t e, Double_t beta, Double_t gamma, Double_t costheta);
	//! apply doppler shift correction
	/*!
		-e : energy
		-beta  : v/c
		-gamma : 1/(1-beta*beta)1/2
		-costheta : NOT theta !
	*/	
	static Double_t DopplerCorrection(Double_t e, Double_t beta, Double_t gamma, Double_t costheta);	
};

//! Base class for a Vertex i.e. a position, recoil direction and velocity
class AVertexFrame : public AgataDataFrame< VertexInterface >
{  
protected:
	AVertexFrame(Key *akey):
		AgataDataFrame< VertexInterface >(akey) {;} 
public:
	virtual ~AVertexFrame() 
		{;} 
};
	
//! Base class Frame that helps controls of the data flow
class ASyncFrame : public AgataDataFrame< DataInterface >
{  
protected:
	ASyncFrame(Key *akey):
		AgataDataFrame< DataInterface >(akey) {;} 
public:
	virtual ~ASyncFrame() 
		{;} 
};

} // namespace ADF
#endif






