/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file NarvalInterface.h header file for NarvalInterface.cpp */

#ifndef ADF_NarvalInterface
#define ADF_NarvalInterface
#define ADF_NARVALINTERFACE_H ADF_NarvalInterface

#include "FrameIO.h"

namespace ADF
{ 
	//! It defines the general interface needed to be a narval actor.
	/*!
	 To run in Narval, an algorithm has to have a given interface i.e. has to define some methods.
	 
	 A narval algorithm has also a version number, a process id (unique number) and a system to send log messages to a LogCollector.
	 
	 Because your own implementation inherits from one the particular actor 
	 (NarvalProducer, NarvalConsumer or NarvalFilter)
	 some protected methods helps you to get informations about the current data flow. 
	 In particular, what are the current
	 version number of the primary key (GetCurrentKeyVersion()) that structures the data flow 
	 and the current frame versions with GetCurrentFrameVersion().
	 
	 This informations should be used to know if your algorithm is compliant to process the current data flow.
	 */
	class NarvalInterface 
	{	
	private:
		//! needed by Narval, unique number per PC ! MUST BE FIRST DATA MEMBER
		//! it corresponds to instance #i 
		UInt_t fProcessID;
		
	protected:	
		virtual ConfAgent *GetConfAgent() const ;
		
	private:
		//! Path given in the process_config method.
		std::string fConfPath;
		
	protected:
		//! Algo version (used to check determine suitable frame version ?)
		Version fVersion; 
		
		//! to send messages to the log server
		LogMessage Log;	
		
	private:	
		//! some useful globals
		struct G {	
			//! configuration path of the algorithm
			static std::string ConfPath;
			//! to have a unique id per processes	
			static UInt_t ProcessID;
		};
		
	public:
		NarvalInterface() ;
		virtual ~NarvalInterface() ; 
		
		//! To get the ID number for that algorithm
		UInt_t GetPID() const 
		{ return fProcessID; }
		
		//! to get from narval a parameter () ... IS IT mandatory to have there a static method ????
		static Int_t GetGlobalParameter(const char *name, UShort_t &val);
		static Int_t GetGlobalParameter(const char *name, Int_t &val);
		static Int_t GetGlobalParameter(const char *name, std::string &val);
		
		static Int_t SetGlobalParameter(const char *name, const UShort_t &);
		static Int_t SetGlobalParameter(const char *name, const std::string &);	
		
		//! global conf path keep since passed in static method and needed in init method
		static std::string GetGlobalConfPath() 
		{ return G::ConfPath; }
		
		//! To get the algo path associated with the current actor	
		const std::string &GetConfPath() 
		{ return fConfPath; }
		
		//! for this actor, it gives the max number of entry lines it can handle (max included).
		virtual Int_t GetMaxInput() const
		{ return 0; }
		//! for this actor, it gives the max number of output lines it can handle (max included).
		virtual Int_t GetMaxOutput() const
		{ return 0; }
		//! for this actor, it gives the current number of entry lines attached.	
		virtual Int_t GetNbInput() const
		{
			return 0;
		}
		//! for this actor, it gives the current number of output lines attached.	
		virtual Int_t GetNbOutput() const 
		{
			return 0;
		}	
		//! Attach a block as input for a given slot number 
		/*!
		 slot < 0 means attach it on the top of the stack in case the actor can handle many input. As a consequence, GetNbInput is increased by 1.
		 The returned value is the slot position
		 if slot < 0 and >= GetNbInput, the block cannot be attached and it returns -1
		 */ 	
		virtual Int_t AttachInput(FrameBlock * /*block*/, Int_t /*slot*/ = -1)
		{
			return -1;
		}
		//! Attach a block as output for a given slot number 
		/*!
		 slot < 0 means attach it on the top of the stack in case the actor can handle many output. As a consequence, GetNbOutput is increased by 1.
		 The returned value is the slot position
		 if slot < 0 and >= GetNbOutput, the block cannot be attached and it returns -1 
		 */
		virtual Int_t AttachOutput(FrameBlock * /*block*/, Int_t /*slot*/ = -1)
		{
			return -1;
		}
		//! generic ProcessBlocks method for any kind of actor. Intput/Output should be attached using the AttachInput/AttachOutput methods
		/*!
		 It allows moreover to interrupt and restart without waiting for the end of one block.
		 It should :
		 - Check the actor is the owner of the block through the SetOwner method
		 - Process the block(s) in case
		 - Free the block once fully consumed/produced using again the SetOwner ( 0x0 ) method
		 */
		virtual UInt_t ProcessBlocks()
		{
			return 1u;
		}
		
		//! Have to be overwritten and called in your implementation
		/*!
		 With this methods, global parameters/variables are set. The first argument is a path to a 
		 directory in which you may have a particular file (or set of files) required to help you algorithm
		 running. 
		 
		 It could be for instance cross section tables for tracking, base signals for PSA etc ...
		 
		 This method is static because it is called once per PC (with the possibility to have 
		 one thread per core/CPU in multiple core/CPU systems). It would not be efficient to load 
		 the same informations several times, especially if this informations is huge.
		 
		 \warning 
		 As a consequence, this method could not be virtual pure to oblige the user to overwrite it. It is
		 then the charge of the user to define it properly
		 
		 NarvalInterface::process_config() has to be called in your implementation 
		 do_adf_conf = 0 (default) means adf is not, in any cases, reconfigured here
		 do_adf_conf = 1 means adf is reconfigured only if ADF_CONF_PATH is not defined \n
		 (look for ADF.conf in path)
		 do_adf_conf = 2 means adf is reconfigured even if ADF_CONF_PATH is defined \n	
		 (allows a specific actor to have a specific configuration)	 
		 */
		static void process_config(const Char_t * /*path*/, UInt_t * /*error_code*/, Short_t do_adf_conf = 0);	
		
		//! To init the internal values (real constructor)
		/*!
		 Just after the call of the constructor, the object should be in a 'NULL state'. 
		 The real initialisation should be done
		 here since Narval must be able to reset the algorithm (with process_reset()) and 
		 thus to re-init it with this 
		 method.
		 */
		virtual void process_initialise (UInt_t *error_code) = 0;
		
		//! To reset the internal values (real destructor)
		/*!
		 Reset the algorithm. Everything should be set to 0 (or NULL), ready to be initialised again by
		 calling process_initialise()
		 */ 
		virtual void process_reset (UInt_t *error_code) = 0;
		
		//! This method is called every time the daq stops
		/*!
		 */	
		virtual void process_stop (UInt_t *error_code)  
		{ *error_code = 0u;
			std::cout << "NarvalInterface::process_stop called" << std::endl;
		}	
		//! This method is called every time the daq starts
		/*!
		 */	
		virtual void process_start (UInt_t *error_code) 
		{ *error_code = 0u;}
		//! This method is called every time the system pauses data acquisition 
		/*!
		 */		
		virtual void process_pause (UInt_t *error_code)
		{ *error_code = 0u;}
		//! this method is called every time the system resumes data acquisition 
		/*!
		 */	
		virtual void process_resume (UInt_t *error_code)
		{ *error_code = 0u;}
	};
	
	//! It defines the interface needed to be a consumer.
	/*!
	 */
	class NarvalConsumer : public NarvalInterface
	{
	private:
		//! external block attached to this
		FrameBlock *fAttachedInputBlock;
	private:
		//! used in process_block to link the external buffer
		OneBlock   fOneBlockIn;
		
	protected:
		//! Frame IO 
		FrameIO fFrameIO;
		
	protected:
		virtual ConfAgent *GetConfAgent() const 
		{ return fFrameIO.GetConfAgent(); } 
		
	public:
		NarvalConsumer();
		virtual ~NarvalConsumer() ; 
		
		FrameIO &GetFrameIO()
		{ return fFrameIO; }
		
		//! for this actor, it gives the max number of entry lines it can handle : one for a consumer
		virtual Int_t GetMaxInput() const
		{ return 1; }
		
		//! Attach a block as input for a given slot number 
		/*!
		 slot < 0 means attach it on the top of the stack in case the actor can handle many input. As a consequence, GetNbInput is increased by 1.
		 The returned value is the slot position
		 if slot < 0 and >= GetNbInput, the block cannot be attached and it returns -1
		 */ 	
		virtual Int_t AttachInput(FrameBlock *block, Int_t slot = -1)
		{
			if ( slot < 1 ) {
				fAttachedInputBlock = block;
				return 0;
			}
			return -1;
		}	
		//! for this actor, it gives the current number of entry lines attached.	
		virtual Int_t GetNbInput() const
		{
			if ( fAttachedInputBlock ) {
				return 1;
			}
			return 0;
		}
		//! process blocks attached by AttachInput/AttachOutput. No need to overwrite it if ProcessBlock(FrameBlock &). Otherwise both are to be implemented.
		virtual UInt_t ProcessBlocks();
		
		//! virtual method to be implemented
		virtual UInt_t ProcessBlock (FrameBlock &) = 0;
		
		//! Narval interface
		/*!
		 */
		static void process_config (const Char_t*, UInt_t *);	
		
		//! Narval interface, Ask the algorithm to process the input data block 
		/*!
		 */
		virtual void process_block (void *input_buffer,
											 UInt_t size_of_input_buffer,
											 UInt_t *error_code);
	};
	
	
	//! Base class for an actor with N input and M output : main purpose is to structure the dataflow
	/*!
	 \warning The max number of input/output should be defined in the constructor and not in the init method.
	 */
	class NarvalShunter  : public NarvalInterface
	{
	protected:
		virtual Bool_t SetNbInput(Int_t /*size*/)
		{ 
			return false; 
		}
		virtual Bool_t SetNbOutput(Int_t /* size */)  
		{ 
			return false; 
		}
		
	public:
		NarvalShunter();
		virtual ~NarvalShunter() ; 
		
		static void process_config (const Char_t *algo_path, UInt_t *error_code) 
		{ 
			NarvalInterface::process_config(algo_path,error_code); 
		}	
		
		virtual Int_t GetNbInput() const
		{
			return 0;
		}
		virtual Int_t GetNbOutput() const 
		{
			return 0;
		}
		
		//! generic ProcessBlocks method for any kind of actor. Intput/Output should be attached using the AttachInput/AttachOutput methods
		/*!
		 It allows moreover to interrupt and restart without waiting for the end of one block.
		 It should :
		 - Check the actor is the owner of the block through the SetOwner method
		 - Process the block(s) in case
		 - Free the block once fully consumed/produced using again the SetOwner ( 0x0 ) method
		 */
		virtual UInt_t ProcessBlocks() = 0 ;
	};
	
	//! It defines the interface needed to be a narval actor (Filter).
	/*!
	 */
	class NarvalFilter  : public NarvalInterface
	{
	private:
		FrameBlock *fAttachedInputBlock;
		FrameBlock *fAttachedOutputBlock;
	private:
		//! used in process_block to link the external buffer
		OneBlock   fOneBlockIn;
		OneBlock   fOneBlockOut; 
		
	protected:
		//! Frame IO 
		FrameIO fFrameIO;
		
	protected:	
		virtual ConfAgent *GetConfAgent() const 
		{ return fFrameIO.GetConfAgent(); } 
		
	public:
		NarvalFilter();
		virtual ~NarvalFilter() ; 
		
		FrameIO &GetFrameIO()
		{ return fFrameIO; }
		
		//! for this actor, it gives the max number of entry lines it can handle ==> Only one for a filter
		virtual Int_t GetMaxInput() const
		{ return 1; }
		
		//! for this actor, it gives the max number of output lines it can handle ==> Only one for a filter
		virtual Int_t GetMaxOutput() const
		{ return 1; }
		
		//! Attach a block as input for a given slot number 
		/*!
		 slot < 0 means attach it on the top of the stack in case the actor can handle many input. As a consequence, GetNbInput is increased by 1.
		 The returned value is the slot position
		 if slot < 0 and >= GetNbInput, the block cannot be attached and it returns -1
		 */ 		
		virtual Int_t AttachInput(FrameBlock *block, Int_t slot = -1)
		{
			if ( slot < 1 ) {
				fAttachedInputBlock = block;
				return 0;
			}
			return -1;
		}	
		//! for this actor, it gives the current number of entry lines attached.	
		virtual Int_t GetNbInput() const
		{
			if ( fAttachedInputBlock ) {
				return 1;
			}
			return 0;
		}
		//! Attach a block as output for a given slot number 
		/*!
		 slot < 0 means attach it on the top of the stack in case the actor can handle many output. As a consequence, GetNbOutput is increased by 1.
		 The returned value is the slot position
		 if slot < 0 and >= GetNbOutput, the block cannot be attached and it returns -1 
		 */
		virtual Int_t AttachOutput(FrameBlock *block, Int_t slot = -1)
		{
			if ( slot < 1 ) {
				fAttachedOutputBlock = block;
				return 0;
			}
			return -1;
		}	
		//! for this actor, it gives the current number of entry lines attached.	
		virtual Int_t GetNbOutput() const
		{
			if ( fAttachedOutputBlock ) {
				return 1;
			}
			return 0;
		}	
		//! process blocks attached by AttachInput/AttachOutput. No need to overwrite it if ProcessBlock(FrameBlock &). Otherwise both are to be implemented.
		virtual UInt_t ProcessBlocks();
		
		//! virtual method to be implemented
		virtual UInt_t ProcessBlock (FrameBlock &, FrameBlock &) = 0;
		
		//! Have to be overwritten and called in your implementation
		/*!
		 It set the global variables to connect 
		 */
		static void process_config (const Char_t*, UInt_t *);
		
		virtual void process_reset (UInt_t *error_code);	
		
		//! Ask the algorithm to process the data block 
		/*!
		 
		 */	
		virtual void process_block (void *input_buffer,
											 UInt_t size_of_input_buffer,
											 void *output_buffer,
											 UInt_t size_of_output_buffer,
											 UInt_t *used_size_of_output_buffer,
											 UInt_t *error_code);
	};
	
	//! It defines the interface needed to be a narval actor (producer).
	/*!
	 */
	class NarvalProducer  : public NarvalInterface
	{
	private:
		FrameBlock *fAttachedOutputBlock;
	private:
		//! used in process_block to link the external buffer
		OneBlock   fOneBlockOut; 
	protected:	
		//! Frame IO ... should be private ...
		FrameIO fFrameIO;
		
	protected:		
		virtual ConfAgent *GetConfAgent() const 
		{ return fFrameIO.GetConfAgent(); } 
	public:
		NarvalProducer() ;
		virtual ~NarvalProducer() ; 
		
		FrameIO &GetFrameIO()
		{ return fFrameIO; }
		
		//! for this actor, it gives the max number of output lines it can handle ==> Only one for a producer
		virtual Int_t GetMaxOutput() const
		{ return 1; }
		
		//! Attach a block as output for a given slot number 
		/*!
		 slot < 0 means attach it on the top of the stack in case the actor can handle many output. As a consequence, GetNbOutput is increased by 1.
		 The returned value is the slot position
		 if slot < 0 and > GetNbOutput, the block cannot be attached and it returns -1 
		 */
		virtual Int_t AttachOutput(FrameBlock *block, Int_t slot = -1)
		{
			if ( slot < 1 ) {
				fAttachedOutputBlock = block;
				return 0;
			}
			return -1;
		}	
		//! for this actor, it gives the current number of entry lines attached.	
		virtual Int_t GetNbOutput() const
		{
			if ( fAttachedOutputBlock ) {
				return 1;
			}
			return 0;
		}
		//! process blocks attached by AttachInput/AttachOutput. No need to overwrite it if ProcessBlock(FrameBlock &). Otherwise both are to be implemented.
		virtual UInt_t ProcessBlocks();
		
		//! process one block, virtual method to be implemented
		virtual UInt_t  ProcessBlock (FrameBlock &) = 0;	
		
		//! Have to be overwritten and called in your implementation
		/*!
		 */
		static void process_config (const Char_t*, UInt_t *);	
		
		//! Ask the algorithm to process the data block 
		/*!
		 Not need to overwrite if ProcessBlock is defined.
		 */	
		virtual void process_block ( void *output_buffer,
											 UInt_t size_of_output_buffer,
											 UInt_t *used_size_of_output_buffer,
											 UInt_t *error_code);
	};
	
} // namespace ADF

/* 
 // c interface for Narval. Only Register and process_config are not there. 
 // They should be in the BaseClass for a given type of algi (see TrackingFilter)
 
 // Filter specific
 void process_block (ADF::NarvalFilter *algo_data,
 void *input_buffer,
 UInt_t size_of_input_buffer,
 void *output_buffer,
 UInt_t size_of_output_buffer,
 UInt_t *used_size_of_output_buffer,
 UInt_t *error_code)
 {
 algo_data->process_block (input_buffer, size_of_input_buffer,
 output_buffer, size_of_output_buffer,
 used_size_of_output_buffer, error_code);
 }
 // Consumer specific
 void process_block (ADF::NarvalConsumer *algo_data,void *input_buffer, UInt_t size_of_input_buffer, UInt_t *error_code)
 {
 algo_data->process_block (input_buffer, size_of_input_buffer,error_code);
 }
 // Producer specific
 void process_block (ADF::NarvalProducer *algo_data, void *output_buffer,
 UInt_t size_of_output_buffer,
 UInt_t *used_size_of_output_buffer,
 UInt_t *error_code)
 {
 algo_data->process_block 
 (output_buffer,size_of_output_buffer,used_size_of_output_buffer,error_code);
 }
 // any kind of algo
 void process_initialise (ADF::NarvalInterface *algo_data, UInt_t *error_code)
 {
 algo_data->process_initialise (error_code);
 }
 void process_reset (ADF::NarvalInterface * *algo_data,UInt_t *error_code)
 {
 algo_data->process_reset (error_code);
 }
 void process_start (ADF::NarvalInterface * *algo_data, UInt_t *error_code)
 {
 algo_data->process_start (error_code);
 }
 void process_stop (ADF::NarvalInterface * *algo_data,UInt_t *error_code)
 {
 algo_data->process_stop (error_code);
 }
 void process_pause (ADF::NarvalInterface * *algo_data,UInt_t *error_code)
 {
 algo_data->process_pause (error_code);
 }
 void process_resume (ADF::NarvalInterface * *algo_data, UInt_t *error_code)
 {
 algo_data->process_resume (error_code);
 }
 
 */
#endif


