/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file Signals.h header file for Signals.cpp */

#ifndef ADF_Signals
#define ADF_Signals

#include "ADFConfig.h"
#include "BufferIO.h"

namespace ADF
{ 

//! Base class for a Signal
/*!
	A signal is composed of an array of basic values : short, integer, or real
	
	\warning To avoid continual checkings and to not have to deal with error messages,
	if an out-of-range value is asked, the first position is returned. 
*/
class Signal
{
protected:	
	mutable UInt_t fLast;	//! temporary     
    
protected:    
	Signal(): fLast(0u) 
		{;}
    
public:
	virtual ~Signal() 
		{;}
		
	//! to convert a Short to UShort taking into account type's ranges
	static UShort_t UShort(Short_t val) 
	{ 
		UShort_t rval; 
		if ( val < 0 )	
			rval = 0u;
		else
			rval = UShort_t(val);  
		return rval; 
	}
	//! to convert a UShort to UShort taking into account type's ranges
	static UShort_t UShort(UShort_t val) 
	{ 
		return val; 
	}	
	//! to convert an Int_t to UShort taking into account type's ranges
	static UShort_t UShort(Int_t val) 
	{ 
		UShort_t rval; 
		if ( val > Int_t(kMaxUShort_t) )	
			rval =  kMaxUShort_t;
		else {
			if ( val < 0 )
				rval = 0u;
			else
				rval = UShort_t(val);
		}
		return rval; 
	}
	//! to convert a UInt_t to UShort taking into account type's ranges
	static UShort_t UShort(UInt_t val) 
	{ 
		UShort_t rval; 
		if ( val > kMaxUShort_t )	
			rval = kMaxUShort_t;
		else
			rval = UShort_t(val);
		return rval; 
	}
	//! to convert a Float_t to UShort taking into account type's ranges
	static UShort_t UShort(Float_t val) 
	{ 
		UShort_t rval; 
		if ( val > kMaxUShort_t )	
			rval = kMaxUShort_t;
		else {
			if ( val < 0 )
				rval = 0u;
			else
				rval = UShort_t(val);
		}
		return rval; 
	}
	//! to convert a Double_t to UShort taking into account type's ranges
	static UShort_t UShort(Double_t val) 
	{ 
		UShort_t rval; 
		if ( val > kMaxUShort_t )	
			rval = kMaxUShort_t;
		else {
			if ( val < 0 )
				rval = 0u;
			else
				rval = UShort_t(val);
		}
		return rval; 
	}
	//! to convert a Short to Short taking into account type's ranges
	static Short_t Short(Short_t val) 
	{  
		return val; 
	}
	//! to convert a UShort to Short taking into account type's ranges
	static Short_t Short(UShort_t val) 
	{ 
		Short_t rval; 	
		if ( val > kMaxShort_t )	
			rval = kMaxShort_t;
		else
			rval = Short_t(val); 	
		return rval; 
	}	
	//! to convert an Int_t to Short taking into account type's ranges
	static Short_t Short(Int_t val) 
	{ 
		Short_t rval; 
		if ( val > kMaxShort_t )	
			rval = kMaxShort_t;
		else {
			if ( val < kMinShort_t )
				rval = kMinShort_t;
			else
				rval = Short_t(val);
		}
		return rval; 
	}
	//! to convert a UInt_t to Short taking into account type's ranges
	static Short_t Short(UInt_t val) 
	{ 
		Short_t rval; 
		if ( val > UInt_t(kMaxShort_t) )	
			rval = kMaxShort_t;
		else
			rval = Short_t(val); 
		return rval; 
	}
	//! to convert a Float_t to Short taking into account type's ranges
	static Short_t Short(Float_t val) 
	{ 
		Short_t rval; 
		if ( val > kMaxShort_t )	
			rval = kMaxShort_t;
		else {
			if ( val < kMinShort_t )
				rval = kMinShort_t;
			else
				rval = Short_t(val);
		}
		return rval; 
	}
	//! to convert a Double_t to Short taking into account type's ranges
	static Short_t Short(Double_t val) 
	{ 
		Short_t rval; 
		if ( val > kMaxShort_t )	
			rval = kMaxShort_t;
		else {
			if ( val < kMinShort_t )
				rval = kMinShort_t;
			else
				rval = Short_t(val);
		}
		return rval; 
	}	
	//! to convert a Short to UInt_t taking into account type's ranges
	static UInt_t UInt(Short_t val) 
	{ 
		UInt_t rval; 
		if ( val < 0 )	
			rval = 0u;
		else
			rval = UInt_t(val);  
		return rval; 
	}
	//! to convert a UShort to UInt_t taking into account type's ranges
	static UInt_t UInt(UShort_t val) 
	{ 
		return UInt_t(val);
	}	
	//! to convert an Int_t to UInt_t taking into account type's ranges
	static UInt_t UInt(Int_t val) 
	{ 
		UInt_t rval; 
		if ( val < 0 )	
			rval = 0u;
		else
			rval = UInt_t(val); 
		return rval; 
	}
	//! to convert a UInt_t to UInt_t taking into account type's ranges
	static UInt_t UInt(UInt_t val) 
	{ 
		return val; 
	}
	//! to convert a Float_t to UInt_t taking into account type's ranges
	static UInt_t UInt(Float_t val) 
	{ 
		UInt_t rval; 
		if ( val > kMaxUInt_t )	
			rval = kMaxUInt_t;
		else {
			if ( val < 0.0 )
				rval = 0u;
			else
				rval = UInt_t(val);
		}
		return rval; 
	}
	//! to convert a Double_t to UInt_t taking into account type's ranges
	static UInt_t UInt(Double_t val) 
	{ 
		UInt_t rval; 
		if ( val > kMaxUInt_t )	
			rval = kMaxUInt_t;
		else {
			if ( val < 0.0 )
				rval = 0u;
			else
				rval = UInt_t(val);
		}
		return rval; 
	}
	//! to convert a Short to Int_t taking into account type's ranges
	static Int_t Int(Short_t val) 
	{ 
		return Int_t(val); 
	}
	//! to convert a UShort to Int_t taking into account type's ranges
	static Int_t Int(UShort_t val) 
	{ 
		return Int_t(val); 
	}	
	//! to convert an Int_t to UShortInt_t taking into account type's ranges
	static Int_t Int(Int_t val) 
	{ 
		return val; 
	}
	//! to convert a UInt_t to Int_t taking into account type's ranges
	static Int_t Int(UInt_t val) 
	{ 
		Int_t rval; 
		if ( val > UInt_t(kMaxInt_t) )	
			rval = kMaxInt_t;
		else {
			rval = Int_t(val);
		}
		return rval; 
	}
	//! to convert a Float_t to Int_t taking into account type's ranges
	static Int_t Int(Float_t val) 
	{ 
		Int_t rval; 
		if ( val > kMaxInt_t )	
			rval = kMaxInt_t;
		else {
			if ( val < kMinInt_t )
				rval = kMinInt_t;
			else
				rval = Int_t(val);
		}
		return rval; 
	}
	//! to convert a Double_t to Int_t taking into account type's ranges
	static Int_t Int(Double_t val) 
	{ 
		Int_t rval; 
		if ( val > kMaxInt_t )	
			rval = kMaxInt_t;
		else {
			if ( val < kMinInt_t )
				rval = kMinInt_t;
			else
				rval = Int_t(val);
		}
		return rval; 
	}
	//! to convert a Short to Float_t taking into account type's ranges
	static Float_t Float(Short_t val) 
	{  
		return Float_t(val); 
	}
	//! to convert a UShort to Float_t taking into account type's ranges
	static Float_t Float(UShort_t val) 
	{ 
		return Float_t(val); 
	}	
	//! to convert an Int_t to Float_t taking into account type's ranges
	static Float_t Float(Int_t val) 
	{ 
		return Float_t(val); 
	}
	//! to convert a UInt_t to Float_t taking into account type's ranges
	static Float_t Float(UInt_t val) 
	{ 
		return Float_t(val); 
	}
	//! to convert a Float_t to Float_t taking into account type's ranges
	static Float_t Float(Float_t val) 
	{ 
		return val; 
	}
	//! to convert a Double_t to Float_t taking into account type's ranges
	static Float_t Float(Double_t val) 
	{ 
		return Float_t(val); 
	}	
	//! to convert a Short to Double_t taking into account type's ranges
	static Double_t Double(Short_t val) 
	{  
		return Double_t(val); 
	}
	//! to convert a UShort to Double_t taking into account type's ranges
	static Double_t Double(UShort_t val) 
	{ 
		return Double_t(val); 
	}	
	//! to convert an Int_t to Double_t taking into account type's ranges
	static Double_t Double(Int_t val) 
	{ 
		return Double_t(val); 
	}
	//! to convert a UInt_t to Double_t taking into account type's ranges
	static Double_t Double(UInt_t val) 
	{ 
		return Double_t(val); 
	}
	//! to convert a Float_t to Double_t taking into account type's ranges
	static Double_t Double(Float_t val) 
	{ 
		return Double_t(val); 
	}
	//! to convert a Double_t to Double_t taking into account type's ranges
	static Double_t Double(Double_t val) 
	{ 
		return val; 
	}	
		
	//!     
	virtual UInt_t GetLength() const = 0; 
	virtual Bool_t Resize(UInt_t, Char_t) = 0; 
		
	//! Return the real last position used 
	/*!
		When asking for a given value in the array, one may ask for out of range values. To avoid
		throwing exceptions (even if it could be better), the first value is always return in case 
		of out of range. To check if you get the value you want:
		\code	
			Signal s(10u);
			
			s.Set(1,1u);             // s.GetLast() = 1
			Double_t d = s.Get(d,9); // s.GetLast() = 9		
			s.Set(-2,11u) = -2.      // s.GetLast() = 9 !! not 11 
		\endcode		
	*/    
	virtual UInt_t GetLast() const 
		{ return fLast; }    
	
	//! In memory size for the signal  
	virtual UInt_t SizeOf(const Char_t opt = '_') = 0;  
		
	//! method to allocate a new signal
	/*!
		the first argument is the size of the signal	
		the second argument is the type of signal:
		
		- S  : a 16 bit signed integer (Short_t)
		- US : a 16 bit unsigned integer (UShort_t)
		- I  : a 32 bit signed integer (Int_t)
		- UI : a 32 bit unsigned integer (UInt_t)
		- F  : a 32 bit floating point (Float_t)
		- D  : a 64 bit floating point (Double_t)
	*/             
	static Signal *New(UInt_t length, const char *type = "US");
	
	//! To attach an external array (to avoid copies)
	/*!
		if external_sig == 0x0 returns to the inner allocated array
		and, \n
		- with length == 0 returns to with the full length of the allocated inner array
		- with length != 0, the given length if it does not exceed the allocate size ... 
		otherwise the allocated size. 
		
		\warning if used to bypass the streaming methods, the swapping is not garantee  
	*/
	virtual Bool_t SetAddress(Char_t* /*external_sig*/ = 0x0, UInt_t /*length*/ = 0u) 
		{ return false;}
	
	//! To have a direct access to the underlying array
	/*!
		The argument is there just to avoid sending an incompatible type.
		By default it returns 0x0. 
		Only if the type of the argument is the same that the underlying array it returns the true address
	*/
	virtual UShort_t *Address(const UShort_t *)
		{ return 0x0; }
	virtual  Short_t *Address(const  Short_t *)
		{ return 0x0; }	
	virtual   UInt_t *Address(const   UInt_t *)
		{ return 0x0; }	
	virtual    Int_t *Address(const    Int_t *)
		{ return 0x0; }	
	virtual  Float_t *Address(const  Float_t *)
		{ return 0x0; }	
	virtual Double_t *Address(const Double_t *)	
		{ return 0x0; }
		
	//! Method to get the value for a particular bin of the signal
	/*!
		Because the returned value is not used by the compiler to resolve overloading,
		one has to trick it with such a method. Usage: \n
		\code
			Signal s(10u);
			
			UShort_t value_bin_5;
			value_bin_5 = s.Get(value_bin_5,5);	
		\endcode
		
		see also GetLast()
	*/        
	virtual UShort_t Get(UShort_t, UInt_t) const = 0;
	virtual  Short_t Get( Short_t, UInt_t) const = 0;
	virtual   UInt_t Get(  UInt_t, UInt_t) const = 0;
	virtual    Int_t Get(   Int_t, UInt_t) const = 0;
	virtual  Float_t Get( Float_t, UInt_t) const = 0;
	virtual Double_t Get(Double_t, UInt_t) const = 0;
	
	//! Method to extract a subset of the signal with conversion
	/*!
		It allows partial copy and returns the number of values extracted so that 
		things could be checked 
	*/
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const = 0;
	virtual UInt_t Get( Short_t *sig, UInt_t n, UInt_t istart = 0u) const = 0;
	virtual UInt_t Get(  UInt_t *sig, UInt_t n, UInt_t istart = 0u) const = 0;
	virtual UInt_t Get(   Int_t *sig, UInt_t n, UInt_t istart = 0u) const = 0;
	virtual UInt_t Get( Float_t *sig, UInt_t n, UInt_t istart = 0u) const = 0;
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const = 0;
	
	//! Method to set the value for a particular bin of the signal
	/*!
	*/        
	virtual void Set(UShort_t, UInt_t) = 0;
	virtual void Set( Short_t, UInt_t) = 0;
	virtual void Set(  UInt_t, UInt_t) = 0;
	virtual void Set(   Int_t, UInt_t) = 0;
	virtual void Set( Float_t, UInt_t) = 0;
	virtual void Set(Double_t, UInt_t) = 0;
	
	//! Method to initialize a subset or the full signal from an array
	/*!
		It allows partial copy and returns the number of values extracted so that 
		things could be checked 
	*/	
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u) = 0;
	virtual UInt_t Set(const  Short_t *sig, UInt_t n, UInt_t istart = 0u) = 0;
	virtual UInt_t Set(const   UInt_t *sig, UInt_t n, UInt_t istart = 0u) = 0;
	virtual UInt_t Set(const    Int_t *sig, UInt_t n, UInt_t istart = 0u) = 0;
	virtual UInt_t Set(const  Float_t *sig, UInt_t n, UInt_t istart = 0u) = 0;
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u) = 0;
	
	//!     
 	virtual Double_t operator[](UInt_t i) const 
 		{ Double_t r_val = 0.0; r_val = Get(r_val,i); return r_val; } 
		
	//! Get the signal from the given array
	/*!
		GetLast returns the real size copied
	*/
	// virtual void Get(Double_t *, UInt_t) const = 0;   
	
	//! Set the current signal to the given array
	/*!
		GetLast returns the real size copied   
	*/
	// virtual void Set(const Double_t *, UInt_t) = 0;
		
	//! Reset this signal (set all values to O)
	/*!
	*/    
	virtual void Reset(UInt_t l = 0u) = 0;         
	
	//! Write the signal into a Buffer  
	/*!
		if opt = 'l', the length is written first and after all the values of the signal.
		Otherwise, the length is not written (to be used in case of array of fixed length)
	*/         
	virtual UInt_t Export(BufferIO *buf, const Char_t opt = '_') const = 0; 
	
	//! Read the signal into a Buffer   
	/*!
		if opt = 'l', the length is written first and after all the values of the signal.
		Otherwise, the length is not written (to be used in case of array of fixed length)    
	*/      
	virtual UInt_t Import(const BufferIO *buf, const Char_t opt = '_') = 0;         
};


template <typename T>
class SignalImp : public Signal
{
	friend class Signal;
	
private:
	UInt_t fInnerLength;	// Real length of the inner signal
	T *fInnerSignal;	// Signal
	
protected:
	UInt_t fLength;		//! length of the signal
	UInt_t fRealLength;	//! Real length of the signal
	
	T *fSignal;		//[fLength] signal
	
protected:
	SignalImp(UInt_t length):  
		Signal(), 
		fInnerLength(0u), 
		fInnerSignal(0x0),
		fLength(0u), 
		fRealLength(0u), 
		fSignal(0x0) 
		{ Reserve(length); }
	
protected:
	//! to change the size of the current array 
	/*!
		Expand is called if required.
		It takes care if the current signal is the inner one or an allocated one
	*/
	virtual Bool_t Reserve(UInt_t length, char c = '0')
	{
		UInt_t l; 
		( length == 0u ) ? l = 1u : l = length;
		
		// check first if fSignal exists and if the RealSize is enough 
		if ( fSignal ) {
			if ( fRealLength >= l ) {
				fLength = l; 
				if ( c == '0' )
					BaseBuffer::Zero(fSignal,fLength);
				return true;
			}
		}
		
		// need to expand ... can be done only if the current signal is the inner one
		if ( fSignal != fInnerSignal )
			return false;
			
		T *sig = 0x0; sig = BaseBuffer::New(sig,l);
		if ( sig ) {
			if ( fInnerSignal ) 
				{ BaseBuffer::Delete(fInnerSignal); }
				
			fInnerSignal = sig;
			fSignal = sig;
			fLength = fRealLength = fInnerLength = l;
			
			return true;
		}
		else return false;	
	}
	
public:
	virtual ~SignalImp() 
	{ 
		SetAddress(); // unlink eventual external address
		if ( fSignal ) 
			BaseBuffer::Delete(fSignal); 
	}
	
	virtual UInt_t GetLength() const 
		{ return fLength; }
		
	virtual Bool_t Resize(UInt_t length, char c = '0')
		{ return Reserve(length,c); }
		
	virtual UInt_t SizeOf(const Char_t opt = '_')
	{
		UInt_t size = GetLength()*sizeof(T); 
		if ( opt == 'l') 
			size += sizeof(UInt_t); 
		return size;	
	}
	
	virtual void Reset(UInt_t l = 0u)  
  	{ 
		if ( l == 0u ) 
			BaseBuffer::Zero(fSignal,fLength); 
		else Reserve(l); 
	}	
	
	virtual Bool_t SetAddress(Char_t *external_sig = 0x0, UInt_t length = 0u)
	{
		// back to inner signal
		if ( external_sig == 0x0 || external_sig == (Char_t *)fInnerSignal ) {
		
			fSignal = fInnerSignal; fRealLength = fInnerLength;
			
			if ( length == 0u ) 
				fLength = fRealLength;
			else 
				( length > fInnerLength ) ? fLength = fInnerLength : fLength = length;
			
			return true;
		}
		
		// try to cast first
		T *c_external_sig = 
			(T*)(external_sig);
		if ( c_external_sig == 0x0 )
			return false;
		
		// casting ok ... replace fSignal par external
		fSignal = c_external_sig;
		fRealLength = fLength = length;
		
		return true;
	}
		
	//! Write the signal into a Buffer  
	/*!
		if opt = 'l', the length is written first and after all the values of the signal.
		Otherwise, the length is not written (to be used in case of array of fixed length)
	*/         
	virtual UInt_t Export(BufferIO *buf, const Char_t opt = '_') const
	{
		UInt_t off_init = buf->Offset(), needed = fLength*sizeof(T); 
		
		// write the length of the signal (if needed) and the signal
		if (opt == 'l') 
			{ (*buf) << fLength; needed += sizeof(UInt_t); }
		buf->Import(fSignal,fLength);
		
		// back to original offset if something goes wrong
		if ( buf->Offset() - off_init != needed ) 
			buf->SetOffset(off_init);
			
		return buf->Offset() - off_init;
	}
	//! Read the signal into a Buffer   
	/*!
		if opt = 'l', the length is written first and after all the values of the signal.
		Otherwise, the length is not written (to be used in case of array of fixed length)    
	*/ 
	virtual UInt_t Import(const BufferIO *buf, const Char_t opt = '_')
	{
		Bool_t ok_length = true;
		UInt_t off_init = buf->Offset(), needed = fLength*sizeof(T), reallength = fLength; 
		
		// read the length of the signal (if needed) and the signal
		if (opt == 'l') { 
	
			(*buf) >> reallength; 
			needed = reallength*sizeof(T) + sizeof(UInt_t); 
			
			// Expand if required, otherwise reset because the current length could be 
			// smaller that the allocated one.
			ok_length = Reserve(reallength,'n');
		}
		if ( ok_length )
			buf->Export(fSignal,reallength);
		
		// back to original offset if something goes wrong
		if ( buf->Offset() - off_init != needed ) 
			buf->SetOffset(off_init);
			
		return buf->Offset() - off_init;   
	} 
};

//! 
/*!
*/
class SignalUS : public SignalImp<UShort_t>
{
public:
	SignalUS(UInt_t length): 
		SignalImp<UShort_t>::SignalImp(length) {;}
	virtual ~SignalUS()
		{;}
		
	virtual UShort_t *Address(const UShort_t *)
		{ return fSignal; }
	virtual  Short_t *Address(const  Short_t *)
		{ return 0x0; }	
	virtual   UInt_t *Address(const   UInt_t *)
		{ return 0x0; }	
	virtual    Int_t *Address(const    Int_t *)
		{ return 0x0; }	
	virtual  Float_t *Address(const  Float_t *)
		{ return 0x0; }	
	virtual Double_t *Address(const Double_t *)	
		{ return 0x0; }
				
	virtual UShort_t Get(UShort_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast]; 
	}
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... 
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = shifted_signal[i];
		}
		return tocopy; 
	}
	virtual Short_t Get(Short_t /*val*/, UInt_t i) const
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Short(fSignal[fLast]); 
	}
	virtual UInt_t Get(Short_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... 
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Short(shifted_signal[i]);
		}
		return tocopy; 
	}
 	virtual UInt_t Get(UInt_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::UInt(fSignal[fLast]);
 	}
	virtual UInt_t Get(UInt_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... 
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UInt(shifted_signal[i]);
		}
		return tocopy; 
	}
	virtual Int_t Get(Int_t /*val*/, UInt_t i) const
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Int(fSignal[fLast]);
 	}
	virtual UInt_t Get(Int_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... 
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Int(shifted_signal[i]);
		}
		return tocopy; 
	}
 	virtual Float_t Get(Float_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Float(fSignal[fLast]);
 	}
	virtual UInt_t Get(Float_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... 
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Float(shifted_signal[i]);
		}
		return tocopy; 
	}	
 	virtual Double_t Get(Double_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Double(fSignal[fLast]);
 	}
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... 
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Double(shifted_signal[i]);
		}
		return tocopy; 
	}
	virtual void Set(UShort_t val, UInt_t i)
	{ 
		( i < fLength ) ? fLast = i : fLast = fLength - 1u; 
		fSignal[fLast] = Signal::UShort(val);
	}
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = sig[i];
		}
		return tocopy; 	
	}	
	virtual void Set(Short_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UShort(val);
	}
	virtual UInt_t Set(const Short_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UShort(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(UInt_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UShort(val);
 	}
	virtual UInt_t Set(const UInt_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UShort(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Int_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UShort(val);
 	}
	virtual UInt_t Set(const Int_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UShort(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Float_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UShort(val);
 	}
	virtual UInt_t Set(const Float_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UShort(sig[i]);
		}
		return tocopy; 	
	}	
 	virtual void Set(Double_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UShort(val);
 	}
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		UShort_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UShort(sig[i]);
		}
		return tocopy; 	
	}
};

class SignalS : public SignalImp<Short_t>
{
public:
	SignalS(UInt_t length): 
		SignalImp<Short_t>::SignalImp(length) {;}
	virtual ~SignalS()
		{;}

	virtual UShort_t *Address(const UShort_t *)
		{ return 0x0; }
	virtual  Short_t *Address(const  Short_t *)
		{ return fSignal; }
	virtual   UInt_t *Address(const   UInt_t *)
		{ return 0x0; }	
	virtual    Int_t *Address(const    Int_t *)
		{ return 0x0; }	
	virtual  Float_t *Address(const  Float_t *)
		{ return 0x0; }	
	virtual Double_t *Address(const Double_t *)	
		{ return 0x0; }
				
	virtual UShort_t Get(UShort_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::UShort(fSignal[fLast]); 
	}
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {	
			sig[i] = Signal::UShort(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Short_t Get(Short_t /*val*/, UInt_t i) const
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast]; 
	}
	virtual UInt_t Get(Short_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {	
			sig[i] = shifted_signal[i];	
		}
		return tocopy; 
	}
	virtual UInt_t Get(UInt_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::UInt(fSignal[fLast]); 
	}
	virtual UInt_t Get(UInt_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {	
			sig[i] = Signal::UInt(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Int_t Get(Int_t /*val*/, UInt_t i) const
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Int(fSignal[fLast]);
 	}
	virtual UInt_t Get(Int_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {	
			sig[i] = Signal::Int(shifted_signal[i]);	
		}
		return tocopy; 
	}	
 	virtual Float_t Get(Float_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Float(fSignal[fLast]);
 	}	
	virtual UInt_t Get(Float_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {	
			sig[i] = Signal::Float(shifted_signal[i]);	
		}
		return tocopy; 
	}
 	virtual Double_t Get(Double_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Double(fSignal[fLast]);
 	}
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {	
			sig[i] = Signal::Double(shifted_signal[i]);	
		}
		return tocopy; 
	}	
	virtual void Set(UShort_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Short(val);
	}
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Short(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Short_t val, UInt_t i)
	{ 
		( i < fLength ) ? fLast = i : fLast = fLength - 1u; 
		fSignal[fLast] = val;
	}
	virtual UInt_t Set(const Short_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Short(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(UInt_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Short(val);
 	}
	virtual UInt_t Set(const UInt_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Short(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Int_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Short(val);
 	}
	virtual UInt_t Set(const Int_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Short(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Float_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Short(val);
 	}
	virtual UInt_t Set(const Float_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Short(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Double_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Short(val);
 	}
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... memcpy to be tested in this particular case to see if it speed up things	
		Short_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Short(sig[i]);
		}
		return tocopy; 	
	}		
};

class SignalUI : public SignalImp<UInt_t>
{
public:
	SignalUI(UInt_t length): 
		SignalImp<UInt_t>::SignalImp(length) {;}
	virtual ~SignalUI()
		{;}

	virtual UShort_t *Address(const UShort_t *)
		{ return 0x0; }
	virtual  Short_t *Address(const  Short_t *)
		{ return 0x0; }	
	virtual   UInt_t *Address(const   UInt_t *)
		{ return fSignal; }	
	virtual    Int_t *Address(const    Int_t *)
		{ return 0x0; }	
	virtual  Float_t *Address(const  Float_t *)
		{ return 0x0; }	
	virtual Double_t *Address(const Double_t *)	
		{ return 0x0; }
						
	virtual UShort_t Get(UShort_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::UShort(fSignal[fLast]);
	}
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UShort(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Short_t Get(Short_t /*val*/, UInt_t i) const
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Short(fSignal[fLast]);
	}
	virtual UInt_t Get(Short_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Short(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual UInt_t Get(UInt_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast]; 
	}
	virtual UInt_t Get(UInt_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = shifted_signal[i];	
		}
		return tocopy; 
	}
	virtual Int_t Get(Int_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Int(fSignal[fLast]);
	}
	virtual UInt_t Get(Int_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Int(shifted_signal[i]);	
		}
		return tocopy; 
	}
 	virtual Float_t Get(Float_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Float(fSignal[fLast]);
 	}
	virtual UInt_t Get(Float_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Float(shifted_signal[i]);	
		}
		return tocopy; 
	}
 	virtual Double_t Get(Double_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Double(fSignal[fLast]);
 	}
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ...
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Double(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual void Set(UShort_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UInt(val);
	}
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UInt(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Short_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UInt(val);
	}
	virtual UInt_t Set(const Short_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UInt(sig[i]);
		}
		return tocopy; 	
	}	
	virtual void Set(UInt_t val, UInt_t i)
	{ 
		( i < fLength ) ? fLast = i : fLast = fLength - 1u; 
		fSignal[fLast] = val;
 	}
	virtual UInt_t Set(const UInt_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UInt(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Int_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UInt(val); 
 	}
	virtual UInt_t Set(const Int_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UInt(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Float_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
 		fSignal[fLast] = Signal::UInt(val);
 	}
	virtual UInt_t Set(const Float_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UInt(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Double_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::UInt(val);
 	}
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		UInt_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::UInt(sig[i]);
		}
		return tocopy; 	
	}	
};

class SignalI : public SignalImp<Int_t>
{
public:
	SignalI(UInt_t length): 
		SignalImp<Int_t>::SignalImp(length) {;}
	virtual ~SignalI()
		{;}

	virtual UShort_t *Address(const UShort_t *)
		{ return 0x0; }
	virtual  Short_t *Address(const  Short_t *)
		{ return 0x0; }	
	virtual   UInt_t *Address(const   UInt_t *)
		{ return 0x0; }	
	virtual  Int_t *Address(const  Int_t *)
		{ return fSignal; }
	virtual  Float_t *Address(const  Float_t *)
		{ return 0x0; }	
	virtual Double_t *Address(const Double_t *)	
		{ return 0x0; }
						
	virtual UShort_t Get(UShort_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;	
		return Signal::UShort(fSignal[fLast]); 
	}
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UShort(shifted_signal[i]);	
		}

		return tocopy; 
	}
	virtual Short_t Get(Short_t /*val*/, UInt_t i) const
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::Short(fSignal[fLast]);
	}
	virtual UInt_t Get(Short_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Short(shifted_signal[i]);	
		}

		return tocopy; 
	}
	virtual UInt_t Get(UInt_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::UInt(fSignal[fLast]); 		
	}
	virtual UInt_t Get(UInt_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UInt(shifted_signal[i]);	
		}

		return tocopy; 
	}
	virtual Int_t Get(Int_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast]; 
	}
	virtual UInt_t Get(Int_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = shifted_signal[i];	
		}

		return tocopy; 
	}
 	virtual Float_t Get(Float_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Float(fSignal[fLast]);
 	}
	virtual UInt_t Get(Float_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Float(shifted_signal[i]);	
		}

		return tocopy; 
	}	
 	virtual Double_t Get(Double_t /*val*/, UInt_t i) const 
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Signal::Double(fSignal[fLast]);
 	}
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Double(shifted_signal[i]);	
		}

		return tocopy; 
	}		
	virtual void Set(UShort_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Int(val);
	}
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Int(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Short_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Int(val); 
	}
	virtual UInt_t Set(const Short_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Int(sig[i]);
		}
		return tocopy; 	
	}	
	virtual void Set(UInt_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Int(val); 
 	}
	virtual UInt_t Set(const UInt_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Int(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Int_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = val; 
 	}
	virtual UInt_t Set(const Int_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Int(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Float_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Int(val); 
 	}
	virtual UInt_t Set(const Float_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Int(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Double_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Signal::Int(val); 
	}
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Int_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Int(sig[i]);
		}
		return tocopy; 	
	}
};

class SignalF : public SignalImp<Float_t>
{
public:
	SignalF(UInt_t length): 
		SignalImp<Float_t>::SignalImp(length) {;}
	virtual ~SignalF()
		{;}
		
	virtual UShort_t *Address(const UShort_t *)
		{ return 0x0; }
	virtual  Short_t *Address(const  Short_t *)
		{ return 0x0; }	
	virtual   UInt_t *Address(const   UInt_t *)
		{ return 0x0; }	
	virtual    Int_t *Address(const    Int_t *)
		{ return 0x0; }	
	virtual  Float_t *Address(const  Float_t *)
		{ return fSignal; }
	virtual Double_t *Address(const Double_t *)	
		{ return 0x0; }

	virtual UShort_t Get(UShort_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::UShort(fSignal[fLast]);
	}
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UShort(shifted_signal[i]);		
		}
		return tocopy; 
	}
	virtual Short_t Get(Short_t /*val*/, UInt_t i) const
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::Short(fSignal[fLast]);
	}
	virtual UInt_t Get(Short_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Short(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual UInt_t Get(UInt_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::UInt(fSignal[fLast]);
	}
	virtual UInt_t Get(UInt_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UInt(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Int_t Get(Int_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::Int(fSignal[fLast]);
	}
	virtual UInt_t Get(Int_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Int(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Float_t Get(Float_t /*val*/, UInt_t i) const 
	{
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast];	
	}	
	virtual UInt_t Get(Float_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = shifted_signal[i];	
		}
		return tocopy; 
	}
	virtual Double_t Get(Double_t /*val*/, UInt_t i) const 
	{
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast];	
	}
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Double(shifted_signal[i]);	
		}
		return tocopy; 
	}	
	virtual void Set(UShort_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Float_t(val);
	}
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Float(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Short_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Float_t(val); 
	}
	virtual UInt_t Set(const Short_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Float(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(UInt_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Float_t(val); 
 	}
	virtual UInt_t Set(const UInt_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Float(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Int_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Float_t(val); 
 	}
	virtual UInt_t Set(const Int_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Float(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Float_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = val; 
 	}
	virtual UInt_t Set(const Float_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Float(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Double_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Float_t(val); 
	}
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Float_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Float(sig[i]);
		}
		return tocopy; 	
	}
};

class SignalD : public SignalImp<Double_t>
{
public:
	SignalD(UInt_t length): 
		SignalImp<Double_t>::SignalImp(length) {;}
	virtual ~SignalD()
		{;}
		
	virtual UShort_t *Address(const UShort_t *)
		{ return 0x0; }
	virtual  Short_t *Address(const  Short_t *)
		{ return 0x0; }	
	virtual   UInt_t *Address(const   UInt_t *)
		{ return 0x0; }	
	virtual    Int_t *Address(const    Int_t *)
		{ return 0x0; }	
	virtual  Float_t *Address(const  Float_t *)
		{ return 0x0; }	
	virtual  Double_t *Address(const Double_t *)
		{ return fSignal; }
			
	virtual UShort_t Get(UShort_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::UShort(fSignal[fLast]);
	}
	virtual UInt_t Get(UShort_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UShort(shifted_signal[i]);		
		}
		return tocopy; 
	}
	virtual Short_t Get(Short_t /*val*/, UInt_t i) const
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::Short(fSignal[fLast]);
	}
	virtual UInt_t Get(Short_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Short(shifted_signal[i]);		
		}
		return tocopy; 
	}
	virtual UInt_t Get(UInt_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::UInt(fSignal[fLast]); 
	}
	virtual UInt_t Get(UInt_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::UInt(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Int_t Get(Int_t /*val*/, UInt_t i) const 
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u;
		return Signal::Int(fSignal[fLast]);
	}
	virtual UInt_t Get(Int_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Int(shifted_signal[i]);	
		}
		return tocopy; 
	}
	virtual Float_t Get(Float_t /*val*/, UInt_t i) const 
	{
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return Float_t(fSignal[fLast]);	
	}
	virtual UInt_t Get(Float_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = Signal::Float(shifted_signal[i]);	
		}
		return tocopy; 
	}	
	virtual Double_t Get(Double_t /*val*/, UInt_t i) const 
	{
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		return fSignal[fLast];	
	}
	virtual UInt_t Get(Double_t *sig, UInt_t n, UInt_t istart = 0u) const
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
		
		// copy ... memcpy to be tested in this particular case to see if it speed up things
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			sig[i] = shifted_signal[i];	
		}
		return tocopy; 
	}		
	virtual void Set(UShort_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Double_t(val);
	}
	virtual UInt_t Set(const UShort_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Double(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Short_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Double_t(val); 
	}
	virtual UInt_t Set(const Short_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Double(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(UInt_t val, UInt_t i)
	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Double_t(val); 
 	}
	virtual UInt_t Set(const UInt_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Double(sig[i]);
		}
		return tocopy; 	
	}
	virtual void Set(Int_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Double_t(val); 
 	}
	virtual UInt_t Set(const Int_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Double(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Float_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = Double_t(val); 
 	}
	virtual UInt_t Set(const Float_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = Signal::Double(sig[i]);
		}
		return tocopy; 	
	}
 	virtual void Set(Double_t val, UInt_t i)
   	{ 
		( i < GetLength() ) ? fLast = i : fLast = GetLength() - 1u; 
		fSignal[fLast] = val; 
	}
	virtual UInt_t Set(const Double_t *sig, UInt_t n, UInt_t istart = 0u)
	{
		UInt_t tocopy;
		// starting point in this signal should be in the range [0,GetLength()[
		if( istart + 1u > GetLength() ) 
			return 0u;
			
		// compute last value and number of values to be copied
		fLast  = istart + n; 
		tocopy = n;
		if ( fLast > GetLength() ) 
			{ fLast = GetLength(); tocopy = (fLast-istart); }
			
		// copy ... 	
		Double_t *shifted_signal = 
			fSignal + istart;
		for(UInt_t i = 0u; i < tocopy; i++) {
			shifted_signal[i] = sig[i];
		}
		return tocopy; 	
	}
};	

} // namespace ADF
#endif





