/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file Trigger.h header file for Trigger.cpp */

#ifndef ADF_Trigger
#define ADF_Trigger
#define ADF_TRIGGER_H ADF_Trigger

#ifndef ADF_FrameIO
#include "FrameIO.h"
#endif
#ifndef ADF_Frame
#include "Frame.h"
#endif
#include "GenericFrame.h"

#include <vector>

namespace ADF
{ 

class Key;

//! General trigger
/*!
	Keys are extracted from the data flow to map it. It is the right time then to check
	if the current frame is needed by a particular algorithm.
	
	Thus, an algorithm Register through a Trigger to a Data Flow. If the conditions 
	are filled, the trigger is set Fired and the data flow notifies the algorithm.
*/
class Trigger
{   
private:
	//! to associate an unique ID for a trigger
	static UInt_t gUniqueID; 
    
private:    
	//! the unique ID
	UInt_t fUniqueID;

protected:
	//! to send messages to the log server   
	LogMessage fLog;  
	
	//! true if this is an active trigger
	Bool_t fIsActive;
	//! true if this has just been fired 
	Bool_t fIsFired; 
	//! true if this is a zombie (EX: objects not properly allocated)     
	Bool_t fIsZombie; 
	//! true if this is an universal trigger (means suitable for any dataflow)     
	Bool_t fIsUniversal; 
	
	//! the trigger name
	std::string fName;
	
private:
	Trigger() : 
		fUniqueID(0u), 
		fLog("Trigger"), 
		fIsActive(true), 
		fIsFired(false), 
		fIsZombie(false), 
		fIsUniversal(false),
		fName("Base") {;}
         
public:   
	Trigger(const Char_t *name):
		fUniqueID(0u), 
		fLog("Trigger"), 
		fIsActive(true), 
		fIsFired(false), 
		fIsZombie(false),
		fIsUniversal(false), 
		fName(name) 
		 { fUniqueID = gUniqueID++; }  
	virtual ~Trigger() 
		{;}
	
	UInt_t GetUniqueID() 
		{ return fUniqueID; }
	
	const std::string &GetName() const 
		{ return fName; }
	
	//! Fired (default) or not this trigger
	void Fired(Bool_t b = true) 
		{ fIsFired = b; }
	
	//! Check if this trigger is in fired state   
	Bool_t IsFired() const 
		{ return fIsFired ; }   
		
	//! Check if this trigger is in fired state   
	Bool_t IsZombie() const 
		{ return fIsZombie; }  
		
	//!  this trigger
	void Zombie(Bool_t b = true) 
		{ fIsZombie = b; }
			
	//! Active this trigger
	void Active(Bool_t b = true) 
		{ fIsActive = b; }
	
	//! Check if this trigger is in fired state   
	Bool_t IsActive() const 
		{ return fIsActive ; } 

	//! Set this trigger as universal or not
	void Universal(Bool_t b = true) 
		{ fIsUniversal = b; }
	
	//! Check if this trigger is in fired state   
	Bool_t IsUniversal() const 
		{ return fIsUniversal; } 
			
	//! Check if the currently read key fired this trigger. If yes, IsFired is true        
	virtual Bool_t Check() 
		{ return fIsFired = false;  }    
	
	//! Check if this trigger is in fired state      
	virtual void Reset() 
		{ Fired(false); }   
	  
}; 

//! Base class for a trigger on a data flow
/*!
	Keys are extracted from the data flow to map it. It is the right time then to check
	if the current frame is needed by a particular algorithm.
	
	Thus, an algorithm Register through a Trigger to a Data Flow. If the conditions 
	are fullfilled, the trigger is set Fired and the data flow notifies the algorithm.
*/
class DFTrigger : public Trigger
{ 
public:
	explicit DFTrigger(const Char_t *trig_name) : Trigger(trig_name) 
		{;} 
	virtual ~DFTrigger()
		{;} 
		
	//! number of input frames that define this trigger 
	virtual UInt_t GetNbInputFrame() const = 0;
		
	//! true if this is a trigger for embedded frames
	virtual Bool_t HasEmbedded() const = 0;  
	
	//! data flow option for that particular input
	virtual std::string GetDFOption(UInt_t /*which*/ = 0u) const 
	{ 
		std::string r("");
		return r;
	}

	//! Add a frame to the Embedded list of required frames
	/*!
		the frame is added with a flag to tell if the given frame is consumable.
		
		\todo more explanations.
	*/
	virtual SharedFP *Add(const FactoryItem &key_item, const FactoryItem &frame_item, 
				bool iscons = true, const Char_t *option = "") = 0;
	
	//! Add a configuration frame to the input frame #which
	virtual SharedFP *AddAutoConf(const FactoryItem &key_item, const FactoryItem &frame_item, UInt_t which) = 0;
	
	//! to get back the frames that define this trigger         
	virtual Frame *GetInputFrame(UInt_t which = 0u) = 0;
	//! to get back the Shared pointer that define this trigger         
	virtual SharedFP *GetInputSharedFP(UInt_t which = 0u) = 0;	
	
	//! to get back the frames that define this trigger         
	virtual const Frame *GetConstInputFrame(UInt_t which = 0u) const = 0; 
	//! to know if the frame is consumable or not
	virtual Bool_t IsConsumable(UInt_t which = 0u) const = 0;  
	
	//! fine tuning of the trigger current conditions         
	virtual Bool_t IsIndividualFired(UInt_t which = 0u) const = 0;  

	
	//! Once a trigger has fired, the result of the algorithm is set through this      
	virtual SharedFP *SetOutputFrame(Frame *frame) = 0;
	//! Once a trigger has fired, the result of the algorithm is set through this 
	virtual SharedFP *SetOutputFrame(const FactoryItem &key_item, const FactoryItem &frame_item) = 0;  
	
	//! to get back the output frame   
	virtual Frame *GetOutputFrame() = 0; 
	
	//! Return the true output frame 
	/*!
		From the current state of the trigger, it builds the true output Frame. 
		In case the input frame is a composite frame it adds the output one if this output is suitable for the composite frame, otherwise add it after in the data flow st
	 
		in case the output frame should be embedded in the input one. 
	*/
	virtual Frame *GetEmbeddedOutputFrame(ConfAgent::EModel model = ConfAgent::kStrict) = 0;   
	
	//! Modify the Frame definitions respecting the new data flow definition and the acceptable Frame changes	
	virtual Bool_t Reconfigure( ConfAgent * ) = 0;
};

//! Base class for a generic frame trigger
/*!
	Processing the data flow, an algorithm expects a set of frames. It could be a single input frame, for instance 
	the PSA algorithm expect a crystal frame (data:crystal), or composite one, for instance tracking is expecting an event 
	(event:data composite frame) in which there is a event:data:psa frame.
	A Frametrigger fully defines the required condition. To add a condition, the Add method is used:
	The first given frame is the frame expected on the data flow while the next ones are expected to be inside the first one given :
	
	\code
	MainFrame (#0)
		SubFrame (#1)
		SubFrame (#2)
	\endcode
 
	Of course at least the first one is required while the next one are optionnal. With the Add method, through the option the number of expected
	Frame of each kind could be set as well as if it is a coincidence or an anti-coincidence.
 
	To be notified of such an occurence, the trigger has to register a trigger to a FrameIO object. 
	Once the expected frame(s) has been given (through the Add method), and the trigger registered to a FrameIO, the following tests are done:
	
	The Check() method is used to determine if the current Key on the data flow 
	corresponds to the kind of frames the algorithm is interested in. It set the trigger to true if it is the case, false otherwise.
	
	
	
*/
class FrameTrigger : public DFTrigger
{   
private:	
	//! stack of frames. The first one is the main frame, the others one are for sub-frames of the main one
	PtrStack<SharedFP> fInputFrames; 
	//! stack of keys. used to determine if the current frame fullfilled the trigger conditions.
	std::vector<const Key *> fInputKeyConst;		
	//! a bool is associated to each frame to know if a subframe is consumable 
	std::vector<Bool_t> fIsConsumable;	
	//! value to be check to know if the current condition fullfilled the trigger condition
	std::vector<Short_t> fFiredOn;		
	//! a bool is associated to each frame to know if a subframe has fired or not 
	std::vector<Bool_t> fIndividualFired;  
	//! to have a reference between a subframe and an input frame for that trigger      
	std::vector<UInt_t> fScanPattern;    
private:	
	//! stack of utilities, used to encode/decode composite frame
	PtrStack<SharedFP> fUtilities; 
private:
	//! configuration frames associated to frames that define the trigger
	PtrStack<SharedFP> fInputConfFrames;
	//! to have a correspondance between a frame and the associated configuration frame
	std::vector<UShort_t> fMap;  
	//! active configuration frames (suitable for the data flow)
	std::vector < std::pair<ConfigurationFrame *, UShort_t> > fActiveInputConfFrames;	
	
private:	
	//! true if trigger is the owner of the output frame
	/*!
		if the output has been added using SetOutputFrame(Frame *), 
		fIsOwnerOfOutput is set to false and the frame is not deleted in the destructoo
		Otherwise, fIsOwnerOfOutput is true (default) and the frame is deleted.
	*/
	Bool_t fIsOwnerOfOutput;
	//! Frame pointer on the single output frame 	
	SharedFP *fOutputFrame;  
	
private:	
	//! to build the complete outpout frame in case of composite frame
	Frame *fEmbeddedOutputFrame;
	
protected:
	//! Check if the currently read key fired this trigger. If yes, IsFired is true 
	virtual Bool_t CheckConf();   
	
protected:
	virtual void FrameTransfert (Frame *from, Frame *to);
	
public:   
	explicit FrameTrigger(const Char_t *trig_name); 
	virtual ~FrameTrigger();
	
	//! fine tuning of the trigger current conditions         
	virtual Bool_t IsIndividualFired(UInt_t which = 0u) const
	{ 
		if ( which < fIndividualFired.size() ) 
			return fIndividualFired[which]; 
		return false; 
	}   
	//! get option for the input frame ( consumed: {-} , anti : {!} mandatory {} or not {|}          
	virtual std::string GetDFOption(UInt_t which = 0u) const
	{ 
		std::string result = "{";
		if ( which < GetNbInputFrame()  ) {
			switch ( fFiredOn[which] ) {
				case -1:
					result += "|";
					break;
				case  0:
					result += "!";
					break;
				default:
					break;
			}
		}
		if ( IsConsumable(which) ) {
			result += "-";
		}
		result += "}";
		
		return result; 
	}   	
	virtual UInt_t GetNbInputFrame() const
		{ return fInputFrames.GetSize(); }
	
	//! true if this is a trigger for embedded frames
	virtual Bool_t HasEmbedded() const
		{ return fInputFrames.GetSize() > 1u ; } 
	
	//! Add a frame to the list of required frames
	/*!
		the frame is added with a flag to tell if the given frame is consumable or not i.e. if it is still on the output dataflow in case the trigger defines also an output frame.
			isconc true if 

		A FrameTrigger has a tree (stack) structure as followed :
	 \code
	 MainFrame (#0)
		SubFrame (#1)
		SubFrame (#2)
		...
	 \endcode
		The mainframe is mandatory, other ones (subframes inside the main one) are optional. 
	 
		Options are :
		- none (default) means one and exactly one such a kind of frame is expected to trig 
		- ! : anti coincidence means the current trigger does not contain the frame
		- | : means the frame is there or not. In such case, check if the frame is there with IsIndividualFired
	 
	*/
	virtual SharedFP *Add(const FactoryItem &key_item, const FactoryItem &frame_item, 
				bool iscons = true, const Char_t *option = "");
	
	//! Add utilities, used to decode/encode composite frames	
	virtual SharedFP *AddUtility(const FactoryItem &key_item, const FactoryItem &frame_item);
	
	//! Add a configuration frame to the input frame #which
	virtual SharedFP *AddAutoConf(const FactoryItem &key_item, const FactoryItem &frame_item, UInt_t which);
		
	//! to get back the frames that define this trigger         
	virtual Frame *GetInputFrame(UInt_t which = 0u)
	{ 
		if ( which < fInputFrames.GetSize() ) 
			return fInputFrames.At(which)->GetFrame(); 
		return 0x0; 
	}   
	//! to get back the Shared pointer that define this trigger         
	virtual SharedFP *GetInputSharedFP(UInt_t which = 0u)	
		{ return fInputFrames.At(which); } 	 
	//! to get back the frames that define this trigger         
	virtual const Frame *GetConstInputFrame(UInt_t which = 0u) const
	{ 
		if ( which < fInputFrames.GetSize() ) 
			return fInputFrames.At(which)->GetFrame(); 
		return 0x0;
	}   	
	//! to know if the frame is consumable or not
	virtual Bool_t IsConsumable(UInt_t which = 0u) const;  
	//! to change the consumable flag. 
	virtual void SetConsumable(UInt_t which, Bool_t value = true)
	{
		if ( which < fIsConsumable.size() ) 
			fIsConsumable[which] = value;  		
	}
	
	//! to extract a sub frame from one event frame of the current trigger  
	/*!
		scan the composite frame only if do_scan is set to true (in principle the main frame is scanned at checking time)
	 so in case which_iput is not 0, do :
	 \code
		// trigger defintion 
		FrameTrigger *trigger = AgataFrameTrigger::Build( ... );
		// utility to get sub frames 
		SharedFP *sub = trigger->AddUtility( ... );
	 
		// in case the trigger has fired 
		if  ( trigger->IsIndividualFired(which_input) ) {
			trigger->GetInputSharedFP(which_input)->Scan();
			for (UInt_t i = 0; i < trigger->GetInputSharedFP(which_input)->GetNbSubFrame(); i++) {
				Frame *data_frame = trigger->GetSubFrameFrom(which_input,i,sub);
				if ( data_frame ) {
					// get the data and put them in your local data
				}
			}
		}
	 \endcode
	 */	
	virtual Frame *GetSubFrameFrom(UInt_t which_input, UInt_t which_sub, SharedFP *utility, Bool_t do_scan = false);
	//! Same as GetSubFrameFrom but to get directly a data part
	template <typename Data_T>
	Data_T *GetSubDataFrom(UInt_t which_input, UInt_t which_sub, SharedFP *utility, Bool_t do_scan = false)
	{
		return GetDataPointer<Data_T>( GetSubFrameFrom(which_input,which_sub,utility,do_scan) );
	}
	
	//! Once a trigger has fired, the result of the algorithm is set through this      
	virtual SharedFP *SetOutputFrame(Frame *frame);
	//! Once a trigger has fired, the result of the algorithm is set through this 
	virtual SharedFP *SetOutputFrame(const FactoryItem &key_item, const FactoryItem &frame_item);  
	
	//! to get back the output frame   
	virtual SharedFP *GetOutputSharedFP() 
	{ 
		return fOutputFrame;
	} 
	//! to get back the output frame   
	virtual Frame *GetOutputFrame() 
	{ 
		if (fOutputFrame == 0x0) 
			return 0x0; 
		return fOutputFrame->GetFrame();
	} 
	
	//! In case the output frame should be embedded
	/*!
		From the current state of the trigger, it builds the true output Frame
		in case the output frame should be embedded in the input one. 
	*/
	virtual Frame *GetEmbeddedOutputFrame(ConfAgent::EModel model = ConfAgent::kStrict);   
	
	//! Check the content of the frame if the primary frame is a composite one 
	/*!
		The frame has to be filled properly
	*/
	virtual Bool_t Check();   
	
	//! Modify the Frame definitions respecting the new data flow definition and the acceptable Frame changes	
	virtual Bool_t Reconfigure( ConfAgent * );
};

//! To be used for Agata Data Flow
/*!
	
*/
class AgataFrameTrigger : public FrameTrigger
{   
public:
	explicit AgataFrameTrigger(const Char_t *trig_name): 
		FrameTrigger(trig_name) {;}
	virtual ~AgataFrameTrigger() 
		{;}	
	
	//! Build a new AgataFrame Trigger
	/*!
		The name is mandatory. If in or out are empty, it means write / read only trigger. if agent is 0x0, the global one is selected.
	 Options ("","!" or "|" see Add(const FactoryItem &key_item, const FactoryItem &frame_item, bool iscons = true, const Char_t *option = "")) 
	 for input frames are passed by attached to the type of agata frame. Ex:
	 \code
		AgataFrameTrigger::Build("mytrigger1","event:data event:data:psa !data:ranc1");
	 \endcode
	 the trigger triggs for an event:data frame always containing an event:data:psa frame AND without an ancillary frame 
	 \code
		AgataFrameTrigger::Build("mytrigger2","event:data event:data:psa |data:ranc1");
	 \endcode
	 the trigger triggs for an event:data frame always containing an event:data:psa frame, with OR without an ancillary frame 
	 \code
		AgataFrameTrigger::Build("mytrigger3","event:data |event:data:psa |data:ranc1");
	 \endcode
	 the trigger triggs for an event:data frame always containing or not an event:data:psa frame or an ancillary frame
	 
	 Whether the frame is really there or not could be tested using IsIndividualFired() or from the associated SharedFP (IsValid method) which is set to valid is the frame is there, not valid otherwise
	 */
	static AgataFrameTrigger *Build(const Char_t *name, const Char_t *in, const Char_t *out = "", ConfAgent *agent = 0x0);

	virtual SharedFP *Add(const FactoryItem &key_item, const FactoryItem &frame_item, 
				bool iscons = true, const Char_t *option = "")
			{ return FrameTrigger::Add(key_item,frame_item,iscons,option); }
			
	SharedFP *Add(const char *, const Version &, const Version &, 
			   bool iscons = true, const Char_t *option = "");
	
	//! Add a Frame using its definition in the ConfAgent
	SharedFP *Add(const char *, const char *, 
			   const ConfAgent *agent = 0x0, bool iscons = true, const Char_t *option = "");
			   
	//! Add utilities, used to decode/encode composite frames	
	virtual SharedFP *AddUtility(const FactoryItem &key_item, const FactoryItem &frame_item, const ConfAgent *agent);
	virtual SharedFP *AddUtility(const FactoryItem &key_item, const FactoryItem &frame_item)
	{
		return FrameTrigger::AddUtility(key_item, frame_item);
	}
	// for pure agata frame
	virtual SharedFP *AddUtility(const char *, const ConfAgent *agent);

	virtual SharedFP *SetOutputFrame(Frame *frame)
		{ return FrameTrigger::SetOutputFrame(frame); }
		
	//! Once a trigger has fired, the result of the algorithm is set through this 
	virtual SharedFP *SetOutputFrame(const FactoryItem &key_item, const FactoryItem &frame_item)
		{ return FrameTrigger::SetOutputFrame(key_item,frame_item); }
		
	SharedFP *SetOutputFrame(const char *, const Version &, const Version &);
	//! Set Output Frame using its definition in the ConfAgent	
	SharedFP *SetOutputFrame(const char *, const char *, const ConfAgent *agent = 0x0);
};
	
} // namespace ADF
#endif





