/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   aLong_t with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file LogMessage.h header file for LogMessage.cpp 

*/

#ifndef _LogMessage
#define _LogMessage
#define _LOGMESSAGE_H _LogMessage

#include <iostream>
#include <sstream>

class LogCollector;

//! Base class for a Log message
/*!
	A log message helps to send messages to a LogCollector. Depending of the current collector, the message
	could be treated differently (print out on the standard output, send it to a server etc ...).
	
	A Process name, a process method and a process id (unsigned int) are associated to a LogMessage and can 
	be changed with GetProcessName(), GetProcessMethod() and SetPID().
	As well an address could be set to allow the separation between identical processes.
	By default it is the pointer on the allocated LogMessage but it could be configured by the user with SetProcessPtr
	
	A log message has a level and a verbosity. Currently defined:
		- info : an information concerning the running conditions
		- warning : warning, something may be not good
		- error : potential crash of the system
		- out : could be used to write output results of intermediate results
	
	The default value for the verbosity is 0. Use SetLevel to change the values.
 
	Any basic type could be written in a log message. Here are some illustrations
	\code
	void afunction ( ) {
	  
		// open a log message
		LogMessage log; log.GetProcessName() = ""; 
 
		// give the name of the function, method that defines a message. It starts a new ticket to be sent to the collector
		log.SetProcessMethod("afunction");
		
		log << info << "Hello world" << nline; // add hello world to the message
		
		unsigned int i = 0;
		log << info << " the value of i is " << i << nline; // add a value set in this function
		
		Float_t ar[5];
		
		if ( i > 4 )
			log << error << " trying to access undefined slot in ar " << i << nline; // change status to error 
 
		log << dolog; // it sends the message to the collector
	}
	\endcode
 
	
*/
class LogMessage
{   
	friend LogMessage &info(LogMessage &);
	friend LogMessage &warning(LogMessage &);    
	friend LogMessage &error(LogMessage &);
	friend LogMessage &debug(LogMessage &);
	friend LogMessage &dolog(LogMessage &); 
	friend LogMessage &clear(LogMessage &); 
	friend LogMessage &endl(LogMessage &); 
	friend LogMessage &nline(LogMessage &); 
			
	friend LogMessage &sep(LogMessage &); 

public:
	//! message level
	enum ELevel { kError, kWarning, kInfo, kOut } ; 
	
private:
	//! to fix the global level for debugguing
	static short gDebug; 
	
private:	
	//! maximum depth 
	static short gkMaxDepth;
	//! maximum depth before automatic call of clear
	static short gkAutoClear;
	
private:
	//! this message comes from the process fProcessName (classname)
	std::string fProcessName;
	//! this message comes from the method fProcessMethod
	std::string fProcessMethod;
	//! the current ticket has started with that method
	std::string fProcessMethodTicket;
	
	//! A process id could be specified 	
	unsigned int fProcessID;
	//! A process could be specified by its pointer
	void *fProcessPtr;

private:
	//! to allow using debug manipulator ie debug for a single line 
	short fDebug;
	
private:
	//! unique ticket number
	unsigned int fTicketNumber;
	//! current level info	
	ELevel fLevel;
	//! lowest level reached during the last ticket  	
	ELevel fLowestLevel;
	//! current verbosity, for the current level
	unsigned short fVerbosity;
	
private:
	//! current indentation 
	unsigned short fIndentation;
	//! current real indentation 
	unsigned short fRealIndentation;
	//! true a good endl (nline) is required
	bool  fIsANewLine;

private:
	//! header for the current ticket
	std::ostringstream fHeader;
	//! core of the log message	
	std::ostringstream fCoreMsg;
	//! footer for the current ticket
	std::ostringstream fFooter;
	
protected:
	void NewTicket();
	
protected:
	//! to format the output string ... increase indent level
	virtual void Indent();
	//! indent an ostream	
	virtual void Indent(std::ostream &);
	//! really clear the message
	virtual void DoClear();	
        
public:   
	LogMessage(const char *pname, unsigned int pid = 0u);
	virtual ~LogMessage();
	
	//! to modify the global parameters 
	static void SetLogParameters(short global_debug_level = 0, short max_indent = 5, short max_before_clear = 7);
	static void SetDebug(short global_debug_level = 0);
	
	//! to test if the required debug level is lower that the global one
	bool IsDebug(short debug_to_test) const
		{ return debug_to_test <= gDebug; }
	
	//! Allow to pass the underlying stream to funtions/methods   
	/*!
		For instance, within a TestDump class, the Print method is defined and required an ostream.
		Default is to print out on the standard output. However, the content of the message
		could be redirected into the log with:
	\code
		...
		LogMessage mylog;
		...
		...
		// we would like to write in the log system, the current statics
		mylog << info << " Statics because I would like to see it " ;
		dumpfile.Print( mylog() ); 
		mylog << dolog;
		...
	\endcode
	*/ 
	virtual std::ostringstream &operator() () 
		{ return fCoreMsg; }
	
	//! To get the Process ID number
	virtual unsigned int GetPID() const 
		{ return fProcessID; }
	//! To set the Process ID number
	virtual void SetPID(unsigned int pid) 
		{ fProcessID = pid; }  
	//! To set the Process address
	/*!
		By default the address is the one of the currently allocated LogMessage.
	 One may change this by giving for instance the address of an object that inherits 
	 for many classes and content many objects so that all the log messages will be referenced
	 by the same address which is the one of the object that groups all the sub-systems.
	*/
	virtual void SetProcessPtr(void *address) 
		{ fProcessPtr = address; }  
	
	//! To get the Process name
	virtual std::string &GetProcessName() 
		{ return fProcessName; }
	//! To get the current method
	virtual std::string &GetProcessMethod() 
		{ return fProcessMethod; }
	
	//! To set the current method
	/*!
		Using this method one can use indentation. After being cleared the message starts with not indentation.
	 Once SetProcessMethod is called, a ticket is open start with the line which means
	 ticket #5, state |info,verbose=0| open for ProcessName TestLOG_3 into ProcessMethod ToLog_3, corresponding to address 0x7fff5fbd9f28
	 \code
	 [+[#5] |i,0| TestLOG_3::ToLog_3 : 0x7fff5fbd9f28 ]
	 \endcode 
	 
	 Each time a SetProcessMethod is called, the indentation is increased ... while a call of dolog decreases it.
	 Once the log ticket is closed (same number of calls of SetProcessMethod / dolog), the message is sent 
	 to the central log.
	 Using this, the ouput reflects the successive calls of methods inside methods. See for instance:
	 
	 \code
	 [+[#5] |i,0| TestLOG_3::ToLog_3 : 0x7fff5fbd9f28 ]
	  [ -> ::ToLog_2 ]
	   [ -> ::ToLog_1 ]
	    [ -> ::ToLog_0 ]
	     0 set to 0 
	   |E,0|
	   1 cannot be set to 1 
	  |w,1|
	  2 cannot be set to 2 
	   |i,0|
	  3 cannot be set to 3 
	 [-[#5] TestLOG_3::ToLog_3 : 0x7fff5fbd9f28 ]
	 \endcode
	 
	 which is part of the output of the demos macro TestLOG.C
	*/
	virtual void SetProcessMethod(const char *);
	
	//! To get the current level of the message
	virtual void SetLevel(ELevel lev, unsigned short verbosity = 0u);
	//! To get the current level of the message
	virtual ELevel GetLevel() const 
		{ return fLevel; }   
	//! To get the current level of the message
	virtual ELevel GetLowestLevel() const 
		{ return fLowestLevel; }  
	//! To get the current level of the message
	virtual void SetVerboseLevel(unsigned short verbosity)
		{ fVerbosity = verbosity; } 
	//! To get the current level of the message
	virtual unsigned short GetVerboseLevel() const 
		{ return fVerbosity; }  
	
	//! To send the message to the log collector
	virtual void Send();      
	//! To get the current message    
	virtual std::string Message() const 
		{ return fCoreMsg.str() ; }   
	//! To clear the current message    
	virtual void ClearMessage();
	
	//! To get the header of the current message    
	virtual std::string Header() const 
		{ return fHeader.str() ; }  
	//! To get the Footer of the current message    
	virtual std::string Footer() const 
		{ return fFooter.str() ; }  
	
	//! send any request to the underlying ostringstream           
	template<class T> LogMessage &operator<<(T t)  
	{ 
		if ( fDebug > gDebug )
				return (*this);
			
		// it requires a new ticket to be open
		if ( fIndentation == 0 )
			NewTicket();
		
		if ( fIsANewLine ) {
			Indent(fCoreMsg); fIsANewLine = false;
		}
		fCoreMsg << t; 
		return (*this); 
	}
	
	//! To allow manipulators    
	virtual LogMessage &operator<< ( LogMessage &(*pf)(LogMessage&) ) 
		{ (*pf)(*this); return (*this); }     
}; 

//! manipulator to modify the LogMessage
LogMessage &info(LogMessage &);
LogMessage &warning(LogMessage &);
LogMessage &error(LogMessage &);
LogMessage &debug(LogMessage &);
LogMessage &dolog(LogMessage &);

//! others
LogMessage &clear(LogMessage &);    
LogMessage &nline(LogMessage &);
LogMessage &sep(LogMessage &); 

std::ostream &hline (std::ostream &);

// deprecated
LogMessage &endl(LogMessage &);

#endif





