/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** @file NOMADBuilder.h header file for NOMADBuilder.cpp */

#ifndef _NOMADBuilder
#define _NOMADBuilder

#include "NarvalInterface.h"
#include "NOMADMap.h"

#include <deque>
#include <stack>
#include <utility>

class TRandom;

//! NOMADBuilder 
/*!
 Process the data stream provided by the NOMAD DAQ and transform it. 
 For that it requires a configuration file.
 
 What is done is 
 - to transform a couple board,cell (electronic point of view) into groupe,item (useful data point of view)
 - for each couple groupe/item, it applies a shift in TS to align them in case of delays
 - it calibrates the energy, giving the true energy
 
 \author O.Stezowski
*/
class NOMADBuilder : public ADF::NarvalFilter
{  
public:
	// group of Ge, AC, La
	static UShort_t kGEThreshold;
	static UShort_t kACThreshold; 
	static UShort_t kLAThreshold;
    
    /*
 	static std::pair< Int_t, std::pair<UShort_t, UShort_t> > kGERange;
	static std::pair< Int_t, std::pair<UShort_t, UShort_t> > kACRange;
	static std::pair< Int_t, std::pair<UShort_t, UShort_t> > kLaRange; */
	
	// AC pattern to know for an item what other item should be added in clovers: all composed of 4 crystals
	static Bool_t kACPattern[4*4];
	
private:
	//! use for calibration
	TRandom *fRandom;

protected:
	//! input buffer as delivered by NOMADMap that contains the NOMAD stream composed of 2 consecutives words of 32 bits.
	ADF::BufferIO *fCurrentIN;
	
protected:
	//! Mapping + Calibration
	NOMADMap fMap;
	//! current input event
	NOMADEvent *fCEvent;
protected:
	//! to select different algorithm
	Short_t fRunningMode;
	//! to select different output mode
	Short_t fOutputMode;	
	
protected:
	//! Width of the coinc window
	Int_t fFineSlidingWindowWidth;
	Int_t fTwiceCoincWindowWidth;
	std::pair < ULong64_t, ULong64_t > fFineSlidingWindow;
	// fNbOfFineWindowWidthInRoughSlidingWidth is used to compute fRoughSlidingWindowWidth from fFineSlidingWindowWidth 
	// fRoughSlidingWindowWidth = fNbOfFineWindowWidthInRoughSlidingWidth * fFineSlidingWindowWidth
	//! With of the sliding window (always in unit of fFineSlidingWindow)
	Int_t fNbOfFineWindowWidthInRoughSlidingWidth;
	Int_t fRoughSlidingWindowWidth;
	std::pair < ULong64_t, ULong64_t > fRoughSlidingWindow;	
	//! Add to the read TS so that in case of chained files there is no come back in time. done once back to time strongly !
	ULong64_t fLocalTSOffset;
	//! gaps that has to be added between two files to avoid creating correlations.
	ULong64_t fChainTSGAP;
	
protected:
	UInt_t fNbEventsInSlicing;
	UInt_t fNbEventsInContainer;
	//! lowest time in event container
	ULong64_t fPendingEventsLowTS;
	
protected:
	/*
	//! Number of detectot that are alive. one at least all alive detector have one event pending, could SetOutput
	Int_t fNumberOfAlive;
	//! 
	Int_t fNumberOfPending;
	//!
	Int_t fMaxPendingSize;
	 */
	
	//! array of all allocated events.
	std::vector < NOMADEvent > fListOfAllocatedEvents;
	//! stack use to get/put free events
	std::stack < NOMADEvent * > fListOfFreeEvents;
	
	/* 
	//! Contains pending events waiting for a sliding window
	std::vector < std::deque<NOMADEvent *> > fPendingEvents;
	
	//! for each detector, contains the numbers of events waiting in the stack
	std::vector < Int_t > fPendingUID;
	
	//! Current sliding window
	std::stack< std::deque< NOMADEvent *> * > fAllocatedListOfSliging;
	std::deque< std::deque< NOMADEvent *> * > fListOfSliding;
	 */
	
	//! a time gate in which in each slot is kept a pointer to an array of events
	/*!
		The current time window definition is given by [fRoughSlidingWindow.first, fRoughSlidingWindow.second[  
	 */
	std::vector < std::vector <NOMADEvent *> * > fRoughArrayOfSliding;
	//! Stack of array of events. 
	std::stack  < std::vector <NOMADEvent *> * > fStackOfCoinc;
	//! pending events i.e. evnt in the future compared to 
	std::deque  < NOMADEvent * > fQueueOfPending;
	
protected:	
	//! in twice a coinc window, it helps deciding whether a event is in coincidence with the others ones
	/*
		The parameter is the time for coincidence. if set to 4, for instance, it means in the following  
		|x|-|-|-|-|x|
	 
		that the two xx are separated by more that 4 thus are not in coincidence.
	 
		The procedure appiled is then to build a builder histo which consist to add around the event x
		a pattern which could be for instance a traditional rectangle. Such a pattern is however not very 
		precise as it built events without taking into account the density of counts.
		For instance : \n
		|a|-|-|-|b|c| gives two events a and bc \n
	 
		Current what is done is to add to an histogram a weight which is equal to coinc window and with values that decrease 
		moving away. Ex : \n
		|-|-|-|-|x|-|-|-|-| --> |0|1|2|3|4|3|2|1|0| \n
	 
		The procedure is then the following. Once the histo is built. The time line is scaned 
		if a event is found, the weight Wx is read and the coinc is closed at the next time a weight less that Wx is found.
		Such procedure is equivalent to build events based one local gravity centers. \n
		For instance, with this procedure \n
		|a|b|-|-|-|c|d|e|-| gives two events (ab) and (cde) since the gravity centers of the two paquets are separated by more that 4
	 
		Note : the algorithm is equivalent to contruct a triangular shape around a time position.
		One may thing of a gaussian shape ... 
	 */
	std::vector < std::vector<Float_t> > fCoincPattern;
	std::vector <Float_t> fCoincBuilderHisto; 
	std::vector < std::vector <NOMADEvent *>  > fCoincBuilderEvent;
	
protected:	
	//! list of events, produced by Process waiting to be stream out by SetOutput
	std::deque <Int_t> fNumberOfCoinc;
	std::deque < NOMADEvent * > fListOfCoinc;
	//! to help building an events
	std::vector <NOMADEvent *> fUnknownEvent;
	std::vector <NOMADEvent *> fGEEvent;
	std::vector <NOMADEvent *> fACEvent;
	std::vector <NOMADEvent *> fLAEvent;
    
	std::vector <NOMADEvent *> fEvent;
	std::vector <NOMADEvent *> fOutEvent;
	
protected:
	ULong64_t fTotNumberOfEvents;
	ULong64_t fBadEvents;
	
protected:
	ADF::SharedFP *fOut;
	
protected:
	//! init collection taking into account the config file
	void InitOnceConfig();
	
protected:
	//! read the data from the trigger and put them in local variable
	virtual UInt_t SetInput();
	//! process this with the current conditions
	virtual UInt_t Process();	
	//! compute output of the process 
	virtual UInt_t SetOutput(ADF::FrameBlock &);
	
protected:
	//! push an event
	void Push(NOMADEvent *);
	//! 
	//Bool_t Front(UShort_t, NOMADEvent &);
	//!
	//Bool_t Pop(UShort_t, NOMADEvent &);
	
public:
 	NOMADBuilder();
 	virtual ~NOMADBuilder(); 
			
	//! to init globals (static) from a directory
	static void process_config (const Char_t *, UInt_t *) ;
	
	//! Constructor implementation 	
	virtual	void process_initialise (UInt_t *error_code);
	
	//! Destructor implementation  
	virtual	void process_reset (UInt_t *error_code);
		
	//! To process just a block of data
	/*!
	*/	
	virtual UInt_t ProcessBlock(ADF::FrameBlock &, ADF::FrameBlock &);	
	
};

#endif
