/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file NOMADFile.cpp compiled in libADF.so */

#include "NOMADFile.h"

#include <iostream>
#include <iomanip>
#include <fstream>

using namespace ADF; 

NOMADFile::NOMADFile(): 
	NarvalProducer(),
	fCurrentFile(0x0),
	fCurrentFileNumber(0u),
	fFiles2Read(),
	fTotEventsExpected(0UL),
	fTotBytesRead(0UL),
	fEventsExpected(0UL),
	fEventsRead(0UL),
	fProgressBar(0,1000)
{
	Log.GetProcessName() = "NOMADFile"; 
}

NOMADFile::~NOMADFile() 
{  
	// in principle not needed ... just in case reset it has not been called by narval
	UInt_t error = 0u; process_reset(&error) ; 
} 

Bool_t NOMADFile::NewFile()
{	
	Bool_t ok = false;
	
	if ( fCurrentFile ) {
		::fclose(fCurrentFile);
		fCurrentFile = 0x0;
		Log << info << "the input file " << fFiles2Read[fCurrentFileNumber-1] << " has just been closed " << nline;
	}
	
	while ( fCurrentFileNumber < fFiles2Read.size() ) {
		
		fCurrentFile = ::fopen(fFiles2Read[fCurrentFileNumber].data(),"rb");
		if ( fCurrentFile == 0x0 ) {
			Log << warning << "File " << fFiles2Read[fCurrentFileNumber] << " is not a valid file " << nline ;
			fCurrentFileNumber++; 
			continue;
		}
		Log << info << "A new input file has just been open " << fFiles2Read[fCurrentFileNumber] << nline;
		// decode the header part. For the moment just jump

		UInt_t number_of_adc, number_of_channel_per_adc, tmp;
		
		::fread ( &number_of_adc, sizeof(UInt_t), 1, fCurrentFile);
		::fread ( &number_of_channel_per_adc, sizeof(UInt_t), 1, fCurrentFile);
		::fread ( &tmp, sizeof(UInt_t), 1, fCurrentFile);
		::fread ( &fEventsExpected, sizeof(UInt_t), 1, fCurrentFile);
		
		Log << " Number of ADC " << number_of_adc << nline;
		Log << " Number of Channel per ADC " << number_of_channel_per_adc << nline;
		Log << " Number of events " << fEventsExpected << nline;
		
		fTotEventsExpected += fEventsExpected; fProgressBar.SetRange(0,fEventsExpected);

		//
		GetFrameIO().SetStatus(BaseFrameIO::kIdle);
		ok = true;
		fCurrentFileNumber++;
		break;
	}
	if ( fCurrentFile == 0x0 ) {
		GetFrameIO().SetStatus(BaseFrameIO::kFinished);
		Log << info << "All input file have been read " << nline;
	}
	//
	Log << dolog;
	
	fProgressBar.Start();
	//
	return ok;
}

void NOMADFile::process_config (const Char_t *directory_path, UInt_t *error_code)
{	
	// first init narval and ADF stuff (if required)
//	NarvalInterface::process_config(directory_path, error_code);
	
	if ( (*error_code) == 0u ) {
		// now init your stuff
	}
}

UInt_t NOMADFile::ProcessBlock(ADF::FrameBlock &out) 
{
	Log.ClearMessage(); Log.SetProcessMethod("process_block");
	
	// make sure length is 0 before starting anything
	out.Rewind(ConfAgent::kWrite);
	//
	if ( GetFrameIO().GetStatus() == BaseFrameIO::kFinished ) { // set in NewFile once list is over
		out.SetEoB(true);
		return 0u;
	}
	
	//  files produced bu nomad are in unit of 64 bits
	UInt_t size = 
		8 *(out.GetMaxSize() / 8), real_size;
	//
	BufferIO *buf = out.GetBufferIO(); buf->SetOffset(); real_size = buf->Import(fCurrentFile,size);
	
	if ( real_size == 0u ) { // Pb @ reading
		if ( ::feof(fCurrentFile) ) {
			NewFile();
		}
		else return 1u;
	}
	else fTotBytesRead+=real_size; fProgressBar += real_size/8; // printf("%d bytes read \n ", real_size);
	//
	out.SetEoB(true);
	
	Log << dolog; return 0u;
}

void NOMADFile::process_initialise (UInt_t *error_code)
{
	std::string tmp, option, filename;
	//
	*error_code = 0u; Log.ClearMessage(); Log.SetProcessMethod("process_initialise");
	
	// read an input file to get the base filename and the path  
	std::string conffile = GetConfPath() + "NOMADFile.conf";
	
	std::ifstream filein(conffile.data());
	if ( filein.is_open() == true ) { 
		getline(filein,tmp);
		tmp += " ";
		while ( filein.good() && !filein.eof() ) { // read input stream line by line
			if ( tmp[0] == '#' ) {
				getline(filein,tmp);
				tmp += " ";
				continue; 
			} // this line is a comment 
			
			std::istringstream decode(tmp);
			
			if( tmp[0] == 'l' ) { // the .conf file contains the list of file to be read
				decode >> option >> filename;
				if( decode.good() )
					fFiles2Read.push_back(filename);				
			} 
			getline(filein,tmp);
			tmp += " ";
		}
	}
	filein.close();
	
	Log << " " << fFiles2Read.size() << " file(s) is(are) to be read by NOMADFile Producer " << nline;

	// open the first file, try up to the first good one. if nothing retrun error
	if ( !NewFile() )
		*error_code = 1u;
	
	Log << dolog;	 
}


void NOMADFile::process_reset (UInt_t *error_code)
{
	*error_code = 0u;
	Log.ClearMessage(); Log.SetProcessMethod("process_reset");
	
	if ( fCurrentFile ) { 
		::fclose(fCurrentFile); fCurrentFile = 0x0;
	}	
	
	Log 
		<< fCurrentFileNumber << " file read (each has an header of 16 bytes) "  
		<< " Total number of events expected " << fTotEventsExpected << " within " << fTotBytesRead << " bytes read" 
		<< dolog;
	
}
/*
 void NOMADFile::process_unload (UInt_terror_code)
 {
 std::cout << "process_unload called with GetPID()" << GetPID() << std::endl;
 *error_code = 0;
 } */





