/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file NOMADFilter.cpp compiled in libGWILL.so */

#include "NOMADFilter.h"
#include "BufferIO.h"
#include "DefaultKeyFactory.h"

#include "TH3.h"
#include "TRandom2.h"

#include <fstream>

using namespace ADF; 

namespace  {

	
	// temp
	TH1F *ADC;
	TH3F *IDS;	
	
//	TH1F *IDS;	
	
}

NOMADFilter::NOMADFilter() : 
	NarvalFilter(),
	fRandom(new TRandom2(0)),
	fCurrentIN(0x0),
	fMap(),
	fCCrate(0),
	fCBoard(0),
	fCCell(0),
	fCRawTS(0),
	fCGroup(0),
	fCItem(0),
	fCCalibValue(0),
	fCCalibTS(0),
	fOut(0x0),
	fTotNumberOfEvents(0UL),
	fBadEvents(0UL)
{	
	Log.GetProcessName() = "NOMADFilter"; 
	
	IDS = new TH3F("NOMADMAP","NOMADMAP",
						NOMADMap::MAX_NUMBER_OF_CRATE,0,NOMADMap::MAX_NUMBER_OF_CRATE,
						NOMADMap::MAX_NUMBER_OF_BOARD,0,NOMADMap::MAX_NUMBER_OF_BOARD,
						NOMADMap::MAX_NUMBER_OF_CELLS,0,NOMADMap::MAX_NUMBER_OF_CELLS);
}

NOMADFilter::~NOMADFilter() 
{  
	// in principle not needed ... just in case reset has not been called by narval
	UInt_t error = 0u; process_reset(&error) ; 
} 

UInt_t NOMADFilter::SetInput()
{
	Int_t w0, w1; ULong64_t ts_fast, ts_slow, ts_slow_tmp; 
	// stream composed of 2 words of 32 bits 	
	// decode each word
	(*fCurrentIN) >> w0 >> w1;
	//
	fCBoard = (w0 & 0xF0000000) >> 28;
	fCCell  = (w0 & 0x0E000000) >> 25;
	fCCrate = (w1 & 0xC0000000) >> 30;
	
	// Be sure it does not exceed map 
	if ( !fMap.IsValid(fCCrate,fCBoard,fCCell) ) {
		Log << error << " Non expected [crate,board,cell] read " << Form("[%d,%d,%d]",fCCrate,fCBoard,fCCell) << dolog;
		fBadEvents++;
		return 1u;
	}	
	//
	fCRawValue = w0 & 0x00007FFF;
	if ( !fMap.IsValid(fCCrate,fCBoard,fCCell) ) {
		Log << error << " ADC is 0 for [crate,board,cell] " << Form("[%d,%d,%d]",fCCrate,fCBoard,fCCell) << dolog;
		fBadEvents++;
		return 1u;
	}	
	//
	ts_slow_tmp = w0 & 0x01FF8000; ts_slow = ts_slow_tmp << 15; ts_fast = w1 & 0x3FFFFFFF;
	fCRawTS = ts_slow | ts_fast;
	
//	std::cout << "C  " << fMap.toUID[fCCrate][fCBoard][fCCell] << " " << fCCrate << " " << fCBoard << " " << fCCell << " "; 
//	std::cout << fCRawValue << " " << fCRawTS << std::endl; 

	
	fTotNumberOfEvents++;
		
	return 0u;
}


void NOMADFilter::process_config (const Char_t *directory_path, UInt_t *error_code)
{
	// In case your algorithm needs to initiate globals, it should be done here
//	NarvalInterface::process_config(directory_path,error_code);
  
	if ( (*error_code) == 0u ) { // now init your stuff
		MainKeyFactory::theMainFactory().Register("Default");		
		MainFrameFactory::theMainFactory().Register("Default");
	}
}

UInt_t NOMADFilter::Process()
{
	// TS correction, apply calib and fill output
	UShort_t uid = fMap.toUID[fCCrate][fCBoard][fCCell]; 
	
	fCGroup = fMap.allDigitizer[ uid ].Group;
	 fCItem = fMap.allDigitizer[ uid ].Item;
	
	TParameter<Int_t> *shift = 
		(TParameter<Int_t> *)fMap.tsShift.At(uid);
	fCCalibTS = fCRawTS - shift->GetVal();
	
	TF1 *calib = 
		(TF1 *)fMap.adcCalib.At(uid);
	fCCalibValue = calib->Eval(fCRawValue+fRandom->Uniform());
	
	return 0u;
}

UInt_t NOMADFilter::SetOutput()
{	
	static const Int_t data_size = 8u;

	// frame to stream out the current
	RawFrame *f = 
		(RawFrame *)fOut->GetFrame(); 
	f->Reset();
	
	// check if enough place to add to the output block
	if ( !f->RawBuffer().Reserve(data_size) )
		return 1u;
	
	// 
	(f->RawBuffer()) << fMap.toUID[fCCrate][fCBoard][fCCell] << fCCalibValue;
	UIDKey *key = (UIDKey *)f->GetKey();
	key->SetUID(fCCalibTS);
	f->Write();
			
	// ADC->Fill(fCRawValue);
	// IDS->Fill(fBoard*MAX_NUMBER_OF_CELLS+fCell);
	IDS->Fill(fCCrate,fCBoard,fCCell);
	
	// printf("--> %d %d %d %lu \n",fBoard,fCell,fRawValue,fRawTS);
	
	return 0u;	
}

UInt_t NOMADFilter::ProcessBlock(ADF::FrameBlock &in, ADF::FrameBlock &out)
{
	Log.ClearMessage(); Log.SetProcessMethod("process_block"); UInt_t result = 0u;

	in.Rewind(ConfAgent::kRead); out.Rewind(ConfAgent::kWrite);
	
	fCurrentIN = in.GetBufferIO();
//	fCurrentIN->SetOffset();
	//
	while ( 1 ) {
		if ( fCurrentIN->FreeSize() < 4u ) {
			in.SetEoB(true);
			break;
		}
		// test here to stop if no enough place for next out
		if ( !out.Reserve(16u,true) ) {
			Log << warning << " Out is full while input not yet completely treated " << nline;
			break;
		}
		UInt_t inp = SetInput();
		switch ( inp ) {
			case 0u:
				if ( Process() == 0u ) {
					if ( SetOutput() == 0u ) {
						// it requires to have at least a minimum of bytes for at least one out
						out.AddFrame((*fOut->GetFrame()));			
					}
				}					
				break;
			case 1u:
				break; // just bad event, do nothing but try next event
				
			default:
				result = 1u;
				break;
		}
	} 

//	printf("NOMADFilter::ProcessBlock %d %d %d %d \n", in.GetOffset(),out.GetOffset(), in.GetSize(),out.GetSize());
	
	out.SetEoB(true);
	
	// send messages to the central log
	Log << dolog;
	return 0u;
}	      


void NOMADFilter::process_initialise (UInt_t *error_code)
{
	*error_code = 0u; Log.ClearMessage(); Log.SetProcessMethod("process_initialise"); 
	//
	
	// read an input file to get the base filename and the path  
	std::string conffile = 
		GetConfPath() + "NOMADFilter.conf";
	
	std::ifstream filein(conffile.data());
	if ( filein.is_open() == true ) { 
		
		std::string keyword, tmp; TString nomad_map;
	
		getline(filein,tmp);
		tmp += " ";
		while ( filein.good() && !filein.eof() ) { // read input stream line by line
			//
			if ( tmp[0] == '#' ) {
				getline(filein,tmp);
				tmp += " ";
				continue; 
			} // this line is a comment 
			
			std::istringstream decode(tmp);
			decode.clear();
			decode >> keyword;
			if ( keyword == "NOMADMap" ) {
				decode >> nomad_map;
				
				nomad_map.Prepend(GetConfPath().data());
				
				if ( !fMap.Load(nomad_map.Data()) ) {
					Log << warning << " cannot load configuration file " << nomad_map.Data() << dolog;
					*error_code = 1u;
					return;
				}
			}
			
			getline(filein,tmp);
			tmp += " ";
		}			
		filein.close();
	}
		
	// Output frame produced which is a raw data : 
	// [ FS(ushort:2bytes) | TS(ulong64:8bytes)] [ uid(ushort:2bytes) | e(Float:4bytes) ]
	fOut = MainFrameFactory::theMainFactory().
			NewSharedFrame(FactoryItem("Default","UID",Version(1,1)),FactoryItem("Default","data",Version(0,0)));
	
	Log << dolog;		
}


void NOMADFilter::process_reset (UInt_t * /*error_code*/)
{
	Log.SetProcessMethod("process_reset"); 
	
	// some log
	Log << " Number of events treated "  << fTotNumberOfEvents << dolog;
	
	if (fOut) {
		delete fOut; fOut = 0x0;
	}

}




