/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file NOMADGene.cpp compiled in libADF.so */

#include "NOMADGene.h"

#include <iostream>
#include <iomanip>
#include <fstream>

#include "TRandom2.h"

using namespace ADF; 

UShort_t current_id = 0; ULong64_t NUMBEROFSIMULATEDEVENTS = 1000000UL;

NOMADGene::NOMADGene(): 
	NarvalProducer(),
	fRandom(new TRandom2(0)),
	fMap(),
	fNumberOfEventsToCreate(NUMBEROFSIMULATEDEVENTS),
	fNumberOfEventsSimulated(0UL),	
	fDTRate(1000),
	fDTCoinc(5),	
	fFold(3),
	fCurrentTS(0UL),
	fProgressBar(0,NUMBEROFSIMULATEDEVENTS)
{
	Log.GetProcessName() = "NOMADGene"; 
}

NOMADGene::~NOMADGene() 
{  
	// in principle not needed ... just in case reset it has not been called by narval
	UInt_t error = 0u; process_reset(&error) ; 
} 

void NOMADGene::process_config (const Char_t *directory_path, UInt_t *error_code)
{	
	// first init narval and ADF stuff (if required)
//	NarvalInterface::process_config(directory_path, error_code);
	
	if ( (*error_code) == 0u ) {
		// now init your stuff
	}
}

UInt_t NOMADGene::ProcessBlock(ADF::FrameBlock &out) 
{
	ULong64_t ts_slow, ts_fast, ts_tmp;

	Log.ClearMessage(); Log.SetProcessMethod("process_block");
	
	// make sure length is 0 before starting anything
	out.Rewind(ConfAgent::kWrite);
	//
	if ( fNumberOfEventsSimulated > fNumberOfEventsToCreate ) { // 	
		printf("NB SIMULATED %d \n ",(Int_t)fNumberOfEventsSimulated);
		out.SetEoB(true);
		return 0u;
	}
	
	//  files produced bu nomad are in unit of 64 bits
	UInt_t size = 
		8 *(out.GetMaxSize() / 8), simulated_size = 0u, free_size ;
	free_size = size;
	//
	BufferIO *buf = out.GetBufferIO(); buf->SetOffset(); 

	while ( simulated_size < size ) {

		UInt_t w0, w1, fCBoard, fCCell, fCCrate, fCRawValue;

		if ( fFold*8 > free_size ) {
			break;
		}

		for (UShort_t i = 0; i < fFold; i++) {

			ts_tmp = fCurrentTS+i*i*fDTCoinc; 
			ts_slow = (ts_tmp & 0xFFC0000000) >> 15; ts_fast = ts_tmp & 0x3FFFFFFF; // highest part to be moved to be in w0  
			
			w0 = (ts_slow);
			w1 = (ts_fast);
			
			fCBoard = fMap.allDigitizer[ current_id ].Board ;
			fCCell  = fMap.allDigitizer[ current_id ].Cell ;
			fCCrate = fMap.allDigitizer[ current_id ].Crate ;

			fCBoard = (fCBoard << 28) & 0xF0000000;
			fCCell  = (fCCell  << 25) & 0x0E000000;
			fCCrate = (fCCrate << 30) & 0xC0000000;

			fCRawValue = ( (i+1)*500 ) & 0x00007FFF;
			
			w0 |= fCBoard;
			w0 |= fCCell;
			
			
		//	w0 |= ((fCurrentTS+i*fDTCoinc) & 0x01FF8000);
			w0 |= fCRawValue ;

			w1 |= fCCrate ;
		//	w1 |= ((fCurrentTS+i*fDTCoinc) & 0x3FFFFFFF) ;

			(*buf) << w0 << w1;

			simulated_size += 8; free_size -= 8;
			current_id++;
			if ( current_id == 46u )
				current_id = 0u;

			fNumberOfEventsSimulated++;
		}
		fCurrentTS += fDTRate;
	}

	fProgressBar += fNumberOfEventsSimulated; 

	//
	out.SetEoB(true);
	
	Log << dolog; return 0u;
}

void NOMADGene::process_initialise (UInt_t *error_code)
{
	std::string tmp, option, filename;
	//
	*error_code = 0u; Log.ClearMessage(); Log.SetProcessMethod("process_initialise");
	
	// read an input file to get the base filename and the path  
	std::string conffile = 
		GetConfPath() + "NOMADGene.conf";
	
	std::ifstream filein(conffile.data());
	if ( filein.is_open() == true ) { 
		
		std::string keyword, tmp; TString nomad_map;
	
		getline(filein,tmp);
		tmp += " ";
		while ( filein.good() && !filein.eof() ) { // read input stream line by line
			//
			if ( tmp[0] == '#' ) {
				getline(filein,tmp);
				tmp += " ";
				continue; 
			} // this line is a comment 
			
			std::istringstream decode(tmp);
			decode.clear();
			decode >> keyword;
			if ( keyword == "NOMADMap" ) {
				decode >> nomad_map;
				
				nomad_map.Prepend(GetConfPath().data());
				
				if ( !fMap.Load(nomad_map.Data()) ) {
					Log << warning << " cannot load configuration file " << nomad_map.Data() << dolog;
					*error_code = 1u;
					return;
				}
			}
			
			getline(filein,tmp);
			tmp += " ";
		}			
		filein.close();
	}

	Log << dolog;	 
}


void NOMADGene::process_reset (UInt_t *error_code)
{
	*error_code = 0u;
	Log.ClearMessage(); Log.SetProcessMethod("process_reset");
	
}
/*
 void NOMADGene::process_unload (UInt_terror_code)
 {
 std::cout << "process_unload called with GetPID()" << GetPID() << std::endl;
 *error_code = 0;
 } */





