/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file NOMADMap.cpp compiled in libGWILL.so */

#include "NOMADMap.h"
#include <fstream>
#include <sstream>
#include <iomanip>

NOMADMap::NOMADMap() :
	toUID(),
	maxUID(0),
	allDigitizer(MAX_NUMBER_OF_CRATE*MAX_NUMBER_OF_BOARD*MAX_NUMBER_OF_CELLS),
	tsShift(MAX_NUMBER_OF_CRATE*MAX_NUMBER_OF_BOARD*MAX_NUMBER_OF_CELLS),
	adcCalib(MAX_NUMBER_OF_CRATE*MAX_NUMBER_OF_BOARD*MAX_NUMBER_OF_CELLS)
{
	// be careful, default is Crate, Board etc = -1 
	for (Int_t i = 0; i < MAX_NUMBER_OF_CRATE*MAX_NUMBER_OF_BOARD*MAX_NUMBER_OF_CELLS; i++) {
		Digitizer adigitizer;
		allDigitizer.push_back(adigitizer);
	}
	// make sure enough space
	allDigitizer.reserve(MAX_NUMBER_OF_CRATE*MAX_NUMBER_OF_BOARD*MAX_NUMBER_OF_CELLS);
	
	// empty slot are to be checked using tsShift which also contains the unique name for the corresponding adc
	tsShift.SetOwner(true); adcCalib.SetOwner(true);
}


Bool_t NOMADMap::Load(const Char_t *fname)
{
	std::ifstream in(fname); std::string tmp; Int_t time_offset, min_ts_offset = kMaxInt ;
	
	if ( in.is_open() == false ) {
		return false;
	}
	
	getline(in,tmp);
	tmp += " ";
	while ( in.good() && !in.eof() ) { // read input stream line by line
		
		if ( tmp[0] == '#' ) {
			getline(in,tmp); 
			tmp += " ";
			continue; 
		} // this line is a comment 
		
		
		{
			UShort_t crate, board, cell, group, item, number_of_parameters; UShort_t uid; std::string uname, calib_exp; 
			//
			std::istringstream decode(tmp);
			decode.clear();
			decode >> crate >> board >> cell >> group >> item >> uname >> uid >> time_offset >> calib_exp >> number_of_parameters;
			if (!decode.good() ) {
				std::cout << " *** WARNING *** decoding line | " << tmp << " |" << std::endl;
				getline(in,tmp);
				tmp += " ";
				continue;
			}	
			// check map :
			if ( crate >= MAX_NUMBER_OF_CRATE || board >= MAX_NUMBER_OF_BOARD || cell >= MAX_NUMBER_OF_CELLS ) {
				std::cout << " *** ERROR *** crate, board, cell to high line " << tmp << std::endl;
				return false;
			}
		
			// map
			toUID[crate][board][cell] = uid;
			if ( maxUID < uid )
				maxUID = uid+1;
			
			// family/group/item/uname
			allDigitizer[uid].Crate = crate;
			allDigitizer[uid].Board = board;
			allDigitizer[uid].Cell  =  cell;
            
            if ( uname.find("EXO") != std::string::npos || uname.find("ILL") != std::string::npos || uname.find("GASP") != std::string::npos ) {
                allDigitizer[uid].Family = 0;
            }
            else {
                if ( uname.find("AC") != std::string::npos ) {
                    allDigitizer[uid].Family = 1;
                }
                else {
                    if ( uname.find("LaBr3") != std::string::npos ) {
                        allDigitizer[uid].Family = 2;
                    }
                }
            }
            allDigitizer[uid].Group = group;
			allDigitizer[uid].Item  =  item;
			
			// TS : keep both unique name and TS shift for alignment
			TParameter<Int_t> *para = new TParameter<Int_t>(uname.data(),time_offset);
			tsShift.AddAt(para,uid);
			if ( time_offset < min_ts_offset ) {
				min_ts_offset = time_offset;
			}
			// ADC
			TF1 *calib = new TF1(uname.data(),calib_exp.data(),0,4*4096);
			adcCalib.AddAt(calib,uid);
			
			for (Short_t i = 0; i < number_of_parameters; i++) {
				Double_t par;
				decode >> par;
				calib->SetParameter(i,par);
			}
			//
			if (!decode.good()) {
				std::cout << " *** ERROR *** Cannot get calib parameter for line " << tmp << std::endl;
				return false;
			}
		}
		// add null character at the end since getline
		// does not store \n and it may give troubles when decoding
		getline(in,tmp);
		tmp += " ";
	}	
	in.close();
	
	// align all TS shift so that the latest one is on 0.
	/*
	for (Int_t i = 0; i < tsShift.GetSize(); i++) {
		TParameter<Int_t> *para = (TParameter<Int_t> *)tsShift.At(i);
		if ( para ) {
			Int_t value = para->GetVal() + abs(min_ts_offset);
			para->SetVal(value);
		}
	}
	 */
	
// 	tsShift.ls();
	
	return true;
}

Bool_t NOMADMap::Save(const Char_t *fname, Option_t *opt) const
{
	std::ofstream out(fname); TString tmp(opt);
	
	if ( out.is_open() == false ) {
		std::cout << " *** WARNING *** decoding line | " << tmp << " |" << std::endl;
		return false;
	}
	// some comments
	out << "# a line starting with # is a comment " << std::endl;
	out << "# each line corresponds to one timestamp unique crate/board/cell to which is associated : " << std::endl;
	out << "# - a group/item pair that can be the same that board/cell or more oriented to users's denomination that electronics " << std::endl;
	out << "# - a unique name " << std::endl;
	out << "# - a unique id (int) " << std::endl;
	out << "# - a shift in time to correct for delays between the different timestamp items " << std::endl;
	out << "# - an expression for the calibration of the ADC in TF1 root format " << std::endl;
	out << "# - the number of parameters followed by the parameters " << std::endl;
	out << "# " << std::endl;	
	out << "# Format : crate board cell group item uname uid shift_ts [0]+[1]*x 2 0.0 1.0 " << std::endl;	
	out << "# " << std::endl;	
	
	if ( maxUID == 0 ) {
		// default
	}
	else {
		for (UInt_t i = 0; i < tsShift.GetSize(); i++) {
			if ( tsShift.At(i) == 0x0 ) {
				continue;
			}
			if ( tmp.Contains("init") ) {
				out << std::setw(1) << allDigitizer[i].Crate << "  " 
					<< std::setw(2) << allDigitizer[i].Board << " "  
					<< std::setw(1) << allDigitizer[i].Cell  << " " 
					<< std::setw(3) << allDigitizer[i].Group << " "  
					<< std::setw(3) << allDigitizer[i].Item  << " " 
					<< "Det_" << std::setw(3) << i <<  " " << i << " +0 [0]+[1]*x 2 0.0 1.0" 
					<< std::endl;					
			}
			else {
				out << std::setw(1) << allDigitizer[i].Crate << "  " 
					<< std::setfill('0') << std::setw(2) << allDigitizer[i].Board << " "  
					<< std::setw(1) << allDigitizer[i].Cell  << " " 
					<< std::setw(4) << allDigitizer[i].Group << " "  
					<< std::setw(4) << allDigitizer[i].Item  << " " 
					<< tsShift.At(i)->GetName() << " " << std::setw(4) << i << " " ;
				if (((TParameter<Int_t> *)tsShift.At(i))->GetVal()<0) {
					out << "-" << std::setw(2) << TMath::Abs(((TParameter<Int_t> *)tsShift.At(i))->GetVal()) << " ";
				}
				else out << "+" << std::setw(2) << ((TParameter<Int_t> *)tsShift.At(i))->GetVal() << " ";
				
				out << adcCalib.At(i)->GetTitle() << std::setw(1) <<  " " << ((TF1 *)adcCalib.At(i))->GetNpar() << "\t";			
				for (Int_t j = 0; j < ((TF1 *)adcCalib.At(i))->GetNpar(); j++) {
					out << ((TF1 *)adcCalib.At(i))->GetParameter(j) << "\t";
				}
				out << std::endl;
			}

		}
	}	
	out.close();
	
	return true;
}


