/***************************************************************************
 *   Copyright (C) 2004-2006 by Olivier Stezowski & Christian Finck        *
 *   stezow(AT)ipnl.in2p3.fr, cfinck(AT)ires.in2p3.fr                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file ALoadedActor.h header file for ALoadedActor.cpp */

#ifndef Gw_ALoadedActor
#define Gw_ALoadedActor

// std headers
#include <string>

// ADF headers
#include "NarvalInterface.h"
#include "DataHandling.h"
#include "AnActorConnection.h"

//! the function used to configure the underlying actor. Added to allow configuration by calling explicitely one of the base class instead of the class itself. 
typedef void (*ConfigFunction)(const Char_t *, UInt_t *);

using namespace ADF;
namespace Gw {

class ALoadedActor;

//! base class in charge of configurating, allocating the true actor loaded in a LoadedActor
class AConfigurator  
{
protected:
	//! last error 
	UInt_t fError;
protected:
	//! Ref to the loader using this in order to have access to the list of global (containing ConfDir)
	ALoadedActor *fRefToLoadedActor;
protected:
	//! real method to be used to configure the actor
	ConfigFunction fConfigFunction;	
	
public:
	AConfigurator(ALoadedActor *, ConfigFunction);
	virtual ~AConfigurator() 
		{;}
	
	void SetConfigMethod(ConfigFunction f)
	{
		fConfigFunction = f;
	}
	//! Reset error 
	virtual void SetError( UInt_t err = 0u )
	{
		fError = err;
	}	
	virtual UInt_t GetError( ) const
	{
		return fError;
	}		
	//! creates an instance of the actor
	virtual NarvalInterface *Register()
	{
		return 0x0;
	}
	//! configure environment for the actor. 
	virtual void Configure();
	
};
	
//! An actor looded in an emulator
/*!
 This class wraps an actor (NarvalInterface). It has then a name (the same one that in Narval)
 and keeps two lists of loaded actors corresponding to the actors producing and consuming the data for this actor
 */
class ALoadedActor
{
public:
	//! type of the loaded actor. Required since some interfaces are different for different type of actors. All however are based on the NarvalInterface
	enum EType { kUnknown, kProducer, kFilter, kConsumer, kShunter };
public:
	//! current status
	/*!
		- undefined : not set properly
		- configure : has just been configured
		- loaded : has just been registered
		- init : has just been initialised
		- dorunning : read to call the embedded actor
		- running : the actor is running
		- HasRun : has consumed/produces some data ... otherwise stay in do running state
		- stop : stop has been called, resume produce do running
		- idle : waiting 
		- finished (in principle only for producer) : no data any more (for producer)
	*/
	enum EStatus { kUndefined, kConfigured, kLoaded, kInitialised, kDoRunning, kRunning, kHasRun, kStopped, kIdle, kFinished, kDead };
	
protected:	
	LogMessage fLog;	
protected:	
	//! Actor's name
	std::string fName;
protected:
	//! Globals for this actor ... not owned by this
	std::vector < ANamedItem * > fGlobals;	
	//! From the global it keeps the generation here for efficiency. Init @ SetGlobal
	NamedItem<Short_t> *fGeneration;
protected:
	//! used to configure the actor (config + load)
	/*!
		This is required since config and register are delayed after the full topology is defined
		and because the two methods don't have a unique interface (NarvalInterface).
	*/
	AConfigurator *fConfigurator;
protected:	
	//! link to actors providing data ... not owned by this 
	std::vector < Gw::AnActorConnection * > fInputs;
	//! link to actors for sending data ... not owned by this 
	std::vector < Gw::AnActorConnection * > fOutputs;
protected:	
	//! depth in the toplogy in which this actor is active.
	Int_t fDepth;
protected:
	NarvalInterface *fActor;
protected:
	//! to know what kind of actor it is (done @ registration time, see RegisterActor) :
	EType fActorType;
protected:
	//! to know what kind of actor it is (done @ registration time, see RegisterActor) :
	EStatus fActorStatus;
protected:
	//! to know what kind of actor it is (done @ registration time, see RegisterActor) :
	UInt_t fLastError;
	
protected:
	//! to deal with generation for which the interface depends of the kind of actor
	/*!
	*/
	void ProcessActorGeneration_2_1_0 (Short_t generation);
	
public:
	ALoadedActor(const Char_t *name);
	//! delete the embedded actor. Connections are owned by the emulator.
	virtual ~ALoadedActor()
	{
		if ( fActor ) delete fActor; // explicit call of process_reset ?
	}
	//! get loaded actor, should have the narval interface
	virtual NarvalInterface *GetActor() const
	{ 
		return fActor; 
	}
	//! get loaded actor, should have the narval interface
	virtual EType GetActorType() const
	{ 
		return fActorType; 
	}	
	//! get dpeth of this in a topology: -1 if not connected
	virtual EStatus GetStatus() const
	{ 
		return fActorStatus; 
	}	
	virtual void SetStatus(EStatus status)
	{ 
		fActorStatus = status; 
	}	
	//! get dpeth of this in a topology: -1 if not connected
	virtual Int_t GetDepth() const
	{ 
		return fDepth; 
	}	
	virtual void SetDepth(Int_t depth)
	{ 
		fDepth = depth; 
	}	
	//! to get the last error code, whatever the action
	virtual UInt_t GetLastError() const
	{ 
		return fLastError; 
	}	
	//! to set the last error code
	virtual void SetLastError(UInt_t error)
	{ 
		fLastError = error; 
	}	
	//! change all globals simultaneously 
	virtual void SetGlobals( const std::vector < ANamedItem *> & );
	//! change one global 
	virtual void SetGlobal( const ANamedItem & );
	//! to get back one global
	virtual Bool_t GetGlobal( ANamedItem & ) const;
	
	//! adopt it so it is delete by this. It returns the previous one so that the user can delete it once the change is done
	virtual AConfigurator *SetConfigurator(AConfigurator *);
		
	//! compute depth (the highest one) of this in the data flow. 
	void ComputeDepth(Int_t &depth);
		
	//! get loaded actor's name
	/*!
	 */
	const std::string &GetName() const
	{ 
		return fName; 
	}	
	//! get number of input connected actors 
	virtual Int_t GetNbInput() const
	{ 
		return fInputs.size(); 
	}
	//! get number of ouput connected actors 
	virtual Int_t GetNbOutput() const
	{ 
		return fOutputs.size(); 
	}	
	//! returns a given input of this
	virtual ALoadedActor *GetInput(Int_t i) const
	{ 
		return fInputs[i]->GetProducer(); 
	}
	//! returns a given ouput of this
	virtual ALoadedActor *GetOutput(Int_t i) const
	{ 
		return fOutputs[i]->GetConsumer(); 
	}	
	//! try to see if the two actors are linked: 0 no, -1 yes as input, +1 yes as output
	Short_t IsLinked(ALoadedActor *, AnActorConnection *&) const;
	
	//! Add one input line to the actor. It does not check if possible ... it has to be done by the user before calling this using GetMaxInput/GetMaxOutput
	/*!
	 */	
	virtual void AddOutput(AnActorConnection *);
	//! Add one input line to the actor. It does not check if possible ... it has to be done by the user before calling this using GetMaxInput/GetMaxOutput
	/*!
	 */		
	virtual void AddInput(AnActorConnection *);
	//! configure this actor i.e call static methods. The way it is done depends on the Globals set by using SetEnvironment (most of the time through the emulator)
	virtual void ConfigureActor();
	//! Register this actor. The way it is done depends on the Globals set by using SetEnvironment (most of the time through the emulator)
	virtual void RegisterActor();
	//! Init this actor. The way it is done depends on the Globals set by using SetEnvironment (most of the time through the emulator)
	virtual void InitActor();
	//! do the job  
	/*!
	 */	
	virtual void ProcessActor ();
};
	
template<typename Actor_T> class Configurator : public AConfigurator
{
public:
	Configurator(ALoadedActor *lactor, ConfigFunction f) : AConfigurator(lactor,f) {;}
	virtual ~Configurator() {;}
	
	//! creates an instance of the actor
	virtual NarvalInterface *Register()
	{
		return new Actor_T;
	}	
};

} // end namespace

#endif