/***************************************************************************
 *   Copyright (C) 2004-2006 by Olivier Stezowski & Christian Finck        *
 *   stezow(AT)ipnl.in2p3.fr, cfinck(AT)ires.in2p3.fr                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file BaseEmulator.h header file for BaseEmulator.cpp */

#ifndef Gw_BaseEmulator
#define Gw_BaseEmulator

#include <vector>
#include <sstream>

#include "ALoadedActor.h"
#include "AnActorConnection.h"
#include "NarvalInterface.h"
#include "DataHandling.h"

#include "GwLogMessage.h"

namespace Gw {

//! Base class for an emulator
/*!
	An emulator regroups actors that are connected together i.e. they exchange buffers of data, whatever is inside the buffer.
	It is organised such that it starts with producers and, in principle, ends with consumers.
 
	The topology could be complex. It is however organized in level in the data processing. Here is a simple example of such a topology with different levels.
 
	one line producer -> filter -> consumer
 
	The data go then through a chain of actors : producer, filter, consumer and dispatcher. 
	Dispatchers are special actors since they handle multiple input and/or ouput while the other ones handle at maximum one input and/or output.
 
	The organisation in layers of the topology constrains then the maximum depth of the topology. Here is a more complex example :
 
	several lines producer -> filter -> consumer -> EB with producer @ different levels
 
	The global behavior of the emulator can be modified using SetGlobal(). By default the following options are added to the emulator :
	- Name: "ConfDir" , type: STR, default "./"
	- Name: "ReadDir" , type: STR, default "./"
	- Name: "SaveDir" , type: STR, default "./"
 
	
 */
class BaseEmulator
{
private:
	LogMessage fLog;
private:
	//! Name of this emulator
	std::string fName;
protected:
	//! to configure the emulator. The name of each item should be unique !
	std::vector < ANamedItem * > fGlobals;
protected:		
	//! list of all actors to start building the map and to delete them @ the end
	std::vector < ALoadedActor * > fAllActors;
protected:		
	//! list of all connections. Owned by the emulator thus delete by this 
	std::vector < AnActorConnection * > fAllConnections;
protected:			
	//! Actors are memorized depending on their depth in the data processing chain
	std::vector < std::vector< ALoadedActor * > > fTopology; 

protected:
	//! to get back an item from its name
	template <class T_item> 
	NamedItem<T_item> *GetItem(const Char_t *name, std::vector <ANamedItem *> &v)
	{
		NamedItem<T_item> *item = 0x0;
		
		for (size_t i = 0; i < v.size(); i++) {
			if ( v[i]->GetName().compare(name) == 0 ) {
				item = dynamic_cast< NamedItem<T_item> * > (v[i]);
				break;
			}
		}
		return item;
	}
	
protected:
	//! link middle (new actor in the emulator) with left and right (could be null) ... the type of connecion depends on the emulator's implementation
	virtual void LinkActors(ALoadedActor *left, ALoadedActor *middle, ALoadedActor *right) = 0;
	
	//! Actors are memorized depending on their depth in the data processing chain
	/*!
	 to start a new chain, add a consumer using Add("-","AUniqueActorName","-",...)
	 */
	template <class T_loader, class T_actor> 
	T_loader *AddLoader(const Char_t *name_left, const Char_t *name_new_actor, const Char_t *name_right, const Char_t *conf_dir, ConfigFunction f_config)
	{
		T_loader *new_node = 0x0; ALoadedActor *input_node = 0x0, *output_node = 0x0; 
		
		// check whether the name has already been used ... an actor is unique by name !
		if ( GetLoaderByName(name_new_actor) ) {
			std::cout << " " << name_new_actor << " already in the topology, name has to be unique " << std::endl;
			return new_node;
		}
		// try and creat a loaded actor
		new_node = new T_loader( name_new_actor );
		if ( new_node == 0x0 ) {
				return new_node;
		}
		else {
			// add it to the list of loaded actors in the emulator
			fAllActors.push_back(new_node);
			// the loaded actor inherits from all the current globals
			new_node->SetGlobals( fGlobals ); 
			// and change the one to be changed here
			new_node->SetGlobal( NamedItem<std::string>("ConfDir",std::string(conf_dir)) );
			// Set configurator
			AConfigurator *configurator = 
					new Configurator<T_actor>(new_node,f_config);
			new_node->SetConfigurator(configurator);
		}

		// look for left and right 
		input_node  = GetLoaderByName(name_left); output_node = GetLoaderByName(name_right);
		// and link them
		LinkActors(input_node, new_node, output_node); 
		
		return new_node;
		/*
		// check if it can be added i.e. allocated and connected with the network, input and/or output accept
		new_node = new T_loader( name_new_actor, new T_actor() );
		if ( new_node ) {
			if ( new_node->GetActor() == 0x0 ) 
				return new_actor;
		}
		else return new_actor;
		
		// now connect the new actor to the network
		new_actor = dynamic_cast<T_actor *>(new_node->GetActor());
		
		std::cout << "[+] Trying to add " << name_new_actor << " to the topology " << std::endl;
		
		// to be done : check with the max # of input output
		Bool_t is_link_possible = true;
		if ( input_node ) {
			if ( (input_node->GetNbOutput()+1) > input_node->GetActor()->GetMaxOutput() ) { // cannot add a link with input
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot have " 
				<< input_node->GetName() 
				<< " as input since all its output channels are occupied : "
				<< (input_node->GetNbOutput()+1) 
				<< " asked, " 
				<< input_node->GetActor()->GetMaxOutput() << " available " << std::endl;
			}
			if ( new_node->GetActor()->GetMaxInput() < 1 ) { // this is a producer ! cannot have a input
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot be linked as output of " 
				<< input_node->GetName() 
				<< " , it is probably a producer !" 
				<< std::endl;
			}			
		}
		else {
			// std::cout << " Cannot find actor with name " << name_left << " in the topology, ==> try to add new actor " << name_new_actor << " without input " << std::endl;
		}
		if ( output_node ) {
			if ( (output_node->GetNbInput()+1) > output_node->GetActor()->GetMaxInput() ) { // cannot add a link with input
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot have " 
				<< output_node->GetName() 
				<< " as output since all its input channels are occupied : "
				<< (output_node->GetNbInput()+1) 
				<< " asked, " 
				<< output_node->GetActor()->GetMaxInput() << " available " << std::endl;				
			}
			if ( new_node->GetActor()->GetMaxOutput() < 1 ) { // cannot add a link with new node
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot be linked as input of " 
				<< output_node->GetName() 
				<< " , it is probably a consumer !" 
				<< std::endl;	
			}					
		}
		else {
			// std::cout << " Cannot find actor with name " << name_right << " in the topology, ==> try to add new actor " << name_new_actor << " without output " << std::endl;
		}
		if ( !is_link_possible ) {
			std::cout << "[-] Trying to add " << name_new_actor << " to the topology : IMPOSSIBLE !! " << std::endl;
			
			delete new_node;
			return 0x0;
		}
		
		// the new actor is filter, consumer or dispatcher
		if ( input_node ) { 
			
			fAllActors.push_back(new_node);
			
			// check nb input/output
			std::cout << new_node->GetName() << " has for input node " << input_node->GetName() << std::endl;
			//
			LinkActors(input_node, new_node, buffer_size_left); 
			if ( output_node ) {
				//
				std::cout << new_node->GetName() <<  " has for output node " << output_node->GetName() << std::endl;
				//
				LinkActors(new_node, output_node, buffer_size_right);	
			}
		}
		else { //  the new actor is a producer ! ... 
			
			fAllActors.push_back(new_node);
			if ( output_node ) {
				std::cout << "Consumer " << new_node->GetName() <<  " has for output node " << output_node->GetName() << std::endl;
				LinkActors(new_node, output_node, buffer_size_right);	
			}			
			else std::cout << "Consumer " << new_node->GetName() << " added to the emulator " << std::endl;
		}
		std::cout << "[-] Trying to add " << name_new_actor << " to the topology " << std::endl;
		// return new actor so that it can be modified in case particular methods should be called 
		return new_actor;
		 */
	}		
	/*
	template <class T_actor,class T_loader> 
	T_actor *Add(const Char_t *name_left, const Char_t *name_new_actor, const Char_t *name_right = "-", 
					 const Char_t *conf_path = "./", UInt_t buffer_size_left = 100u, UInt_t buffer_size_right = 100u)
	{
		ALoadedActor *new_node = 0x0, *input_node = 0x0, *output_node = 0x0; T_actor * new_actor = 0x0; 
		
		// check whether the name has already been used ... an actor is unique by name !
		if ( GetLoaderByName(name_new_actor) ) {
			std::cout << " " << name_new_actor << " already in the topology, name has to be unique " << std::endl;
			return new_actor;
		}
		// look for left and right 
		//
		input_node  = GetLoaderByName(name_left);			
		//
		output_node = GetLoaderByName(name_right);
		
		// check if it can be added i.e. allocated and connected with the network, input and/or output accept
		new_node = new T_loader( name_new_actor, new T_actor() );
		if ( new_node ) {
			if ( new_node->GetActor() == 0x0 ) 
				return new_actor;
		}
		else return new_actor;
		
		// now connect the new actor to the network
		new_actor = dynamic_cast<T_actor *>(new_node->GetActor());

		std::cout << "[+] Trying to add " << name_new_actor << " to the topology " << std::endl;

		// to be done : check with the max # of input output
		Bool_t is_link_possible = true;
		if ( input_node ) {
			if ( (input_node->GetNbOutput()+1) > input_node->GetActor()->GetMaxOutput() ) { // cannot add a link with input
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot have " 
				<< input_node->GetName() 
				<< " as input since all its output channels are occupied : "
				<< (input_node->GetNbOutput()+1) 
				<< " asked, " 
				<< input_node->GetActor()->GetMaxOutput() << " available " << std::endl;
			}
			if ( new_node->GetActor()->GetMaxInput() < 1 ) { // this is a producer ! cannot have a input
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot be linked as output of " 
				<< input_node->GetName() 
				<< " , it is probably a producer !" 
				<< std::endl;
			}			
		}
		else {
			// std::cout << " Cannot find actor with name " << name_left << " in the topology, ==> try to add new actor " << name_new_actor << " without input " << std::endl;
		}
		if ( output_node ) {
			if ( (output_node->GetNbInput()+1) > output_node->GetActor()->GetMaxInput() ) { // cannot add a link with input
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot have " 
				<< output_node->GetName() 
				<< " as output since all its input channels are occupied : "
				<< (output_node->GetNbInput()+1) 
				<< " asked, " 
				<< output_node->GetActor()->GetMaxInput() << " available " << std::endl;				
			}
			if ( new_node->GetActor()->GetMaxOutput() < 1 ) { // cannot add a link with new node
				is_link_possible = false;
				std::cout << new_node->GetName() 
				<< " cannot be linked as input of " 
				<< output_node->GetName() 
				<< " , it is probably a consumer !" 
				<< std::endl;	
			}					
		}
		else {
			// std::cout << " Cannot find actor with name " << name_right << " in the topology, ==> try to add new actor " << name_new_actor << " without output " << std::endl;
		}
		if ( !is_link_possible ) {
			std::cout << "[-] Trying to add " << name_new_actor << " to the topology : IMPOSSIBLE !! " << std::endl;

			delete new_node;
			return 0x0;
		}
		
		// the new actor is filter, consumer or dispatcher
		if ( input_node ) { 
			
			fAllActors.push_back(new_node);

			// check nb input/output
			std::cout << new_node->GetName() << " has for input node " << input_node->GetName() << std::endl;
			//
			LinkActors(input_node, new_node, buffer_size_left); 
			if ( output_node ) {
				//
				std::cout << new_node->GetName() <<  " has for output node " << output_node->GetName() << std::endl;
				//
				LinkActors(new_node, output_node, buffer_size_right);	
			}
		}
		else { //  the new actor is a producer ! ... 
			
			fAllActors.push_back(new_node);
			if ( output_node ) {
				std::cout << "Consumer " << new_node->GetName() <<  " has for output node " << output_node->GetName() << std::endl;
				LinkActors(new_node, output_node, buffer_size_right);	
			}			
			else std::cout << "Consumer " << new_node->GetName() << " added to the emulator " << std::endl;
		}
		std::cout << "[-] Trying to add " << name_new_actor << " to the topology " << std::endl;
		// return new actor so that it can be modified in case particular methods should be called 
		return new_actor;
	}
	 */

public:
	//! Built the map, starting from producer is calculates the depth of each actor in the data flow
	void BuildTopology();
	
public:
 	BaseEmulator(std::string name);
 	virtual ~BaseEmulator();
	
	//! This method allows to modified some options of the emulator
	/*!
		Be carefull that depending on the item it might affect the emulator only in its initialisation phase.
		
		- If an item with the same name and type already exists in the list, it just modified the value. 
		Note that the name of an item should be unique in the list.
		- if the name doest not exist, it adds it to the list
	 */	
	virtual void SetGlobal( const ANamedItem & );
	// virtual void SetGlobalParameter( const ANamedItem & );
	
	//! Name of the emulator
	std::string &GetName()
	{
		return fName;
	}
	//! search for a node in the emulator
	virtual ALoadedActor *GetLoaderByName(const Char_t *) const;
	//! search for an actor in the emulator
	virtual NarvalInterface *GetActorByName(const Char_t *) const;
	//! search for an actor in the emulator
	template <class T_actor> 
	T_actor *GetActorByName(const Char_t *name) const
	{
		NarvalInterface *actor = GetActorByName(name); T_actor *t_actor= 0x0;
		if( actor ) {
			t_actor = dynamic_cast<T_actor *> (actor);
		}
		return t_actor;
	}
	
	//! Init the emulator. To be called before calling Run
	/*!
		To be flexible, it relies on an init sequence.
	 - [config load init] means for each actors { config, register, init } i.e. the three actions are called for the first actor, then the second one ... and so on
	 - the order of the sequence could be change i.e. load, config, and init ... carefull it is not the way of doing !
	 - another way to is to call first conf for all actors, then register for all actors ... here is the code to do it
	 \code
	 emulator.Init( NamedItem<std::string>("InitSequence","config") );
	 emulator.Init( NamedItem<std::string>("InitSequence","load") );
	 ...
	 \endcode 
	*/
	virtual void Init( const NamedItem<std::string> &init_sequence = NamedItem<std::string>("InitSequence","config load init") );
	
	//! Run max loop
	virtual void Run(Int_t max_loop = -1 /*, Bool_t is_interrupt_on  ... or some kind of option*/); 
		
	//! print topology
	virtual void PrintTopology();
	
	//! Draw ... default is print
	virtual void Draw()
		{ PrintTopology(); }
};

} // end namespace

#endif


