/***************************************************************************
 *   Copyright (C) 2010 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file BasicReplica.h header file for BasicReplica.cpp */

#ifndef Gw_BasicReplica
#define Gw_BasicReplica

#include "NarvalInterface.h"
#include "FrameBlock.h"
#include "ADFConfig.h"

using namespace ADF;

namespace Gw {
		
//! It has a single input line and copies it to all the output lines attached
/*!
 */
class BasicReplica : public ADF::NarvalShunter 
{
private:
	ADF::FrameBlock *fAttachedInputBlock;
	std::vector < ADF::FrameBlock * > fAttachedOutputBlocks;
private:
	ULong64_t fBytesAttachedInputBlock;
	std::vector < ULong64_t > fBytesAttachedOutputBlock;
	
public:
	BasicReplica() : fAttachedInputBlock(0x0), fAttachedOutputBlocks(), fBytesAttachedInputBlock(0UL), fBytesAttachedOutputBlock()
		{;}
	virtual ~BasicReplica() { UInt_t error = 0u; process_reset(&error); } // nothing to be done, Blocks are owned by the emulator
	
	//! Only one input
	virtual Int_t GetMaxInput() const
	{ 
		return 1; 
	}	
	//! for this actor, it gives the max number of output lines it can handle (max included).
	virtual Int_t GetMaxOutput() const
	{ 
		return kMaxInt_t; 
	}
	//! for this actor, it gives the current number of output lines attached.	
	virtual Int_t GetNbOutput() const 
	{
		return Int_t(fAttachedOutputBlocks.size());
	}		
	//! Attach a block as input for a given slot number 
	/*!
	 slot < 0 means attach it on the top of the stack in case the actor can handle many input. As a consequence, GetNbInput is increased by 1.
	 The returned value is the slot position
	 if slot < 0 and >= GetNbInput, the block cannot be attached and it returns -1
	 */ 	
	virtual Int_t AttachInput(ADF::FrameBlock * block, Int_t slot = -1)
	{
		if ( slot < 1 ) {
			fAttachedInputBlock = block;
			return 0;
		}
		return -1;
	}
	//! Attach a block as output for a given slot number 
	/*!
	 slot < 0 means attach it on the top of the stack in case the actor can handle many output. As a consequence, GetNbOutput is increased by 1.
	 The returned value is the slot position
	 if slot < 0 and >= GetNbOutput, the block cannot be attached and it returns -1 
	 */
	virtual Int_t AttachOutput(ADF::FrameBlock *block, Int_t slot = -1)
	{
		Int_t rslot = -1; ULong64_t init = 0UL;
		
		if ( slot < 0 ) { // push_back
			fAttachedOutputBlocks.push_back(block); fBytesAttachedOutputBlock.push_back(init);
			//
			rslot = GetNbOutput() - 1;
		}
		else {
			if ( slot < GetNbOutput() )
				fAttachedOutputBlocks[slot] = block;
		}
			
		return rslot;
	}	
	
	//! To init the internal values (real constructor)
	/*!
	 Just after the call of the constructor, the object should be in a 'NULL state'. 
	 The real initialisation should be done
	 here since Narval must be able to reset the algorithm (with process_reset()) and 
	 thus to re-init it with this 
	 method.
	 */
	virtual void process_initialise (UInt_t *error_code) {;}
	
	//! To reset the internal values (real destructor)
	/*!
	 Reset the algorithm. Everything should be set to 0 (or NULL), ready to be initialised again by
	 calling process_initialise()
	 */ 
	virtual void process_reset (UInt_t *error_code) ;
	
	virtual UInt_t ProcessBlocks() ; 

};

} // namespace

#endif