/***************************************************************************
 *   Copyright (C) 2004-2006 by Olivier Stezowski & Christian Finck        *
 *   stezow(AT)ipnl.in2p3.fr, cfinck(AT)ires.in2p3.fr                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file DEmulator.h header file for DEmulator.cpp */

#ifndef Gw_DEmulator
#define Gw_DEmulator

#include <vector>
#include <string>
#include <sstream>

#include "BaseEmulator.h"

namespace Gw 
{

//! Default connection between actors  : to be used in non-multi threaded programs
/*!
*/
class DActorConnection : public AnActorConnection 
{
protected:
	//! left actor ==> Producer
	ALoadedActor *fProducer;
	//! right actor ==> Consumer
	ALoadedActor *fConsumer;
	//! Block use to exchange data between the producer and the consumer
	FrameBlock *fBlock;
	
public:
 	DActorConnection(ALoadedActor *producer, ALoadedActor *consumer, Int_t size, const Char_t *type = "") : AnActorConnection(), fProducer(producer), fConsumer(consumer), fBlock(0x0)
	{ 
		std::string opt = type;
		//
		if ( opt == "expand" ) {
			fBlock = new ExpandableBlock(size); 
		}
		if ( fBlock == 0x0 ) {
			fBlock = new OneBlock(size); 
		}
		// set name using producer, consumer name and size of the buffer 
		
		GetName();
	}
 	virtual ~DActorConnection()
	{ 
			if ( fBlock ) 
				delete fBlock ; 
	}	
	//! to get back the producer 
	virtual ALoadedActor *GetProducer() 
	{ 
		return fProducer ; 
	}
	//! to get back the consumer 
	virtual ALoadedActor *GetConsumer()
	{ 
		return fConsumer ; 
	}
	
	//! test if some data are ready to be sent to the consumer (right part of the connection) 
	virtual Bool_t IsReadyforConsumer() ; 
	//! Prepare the connection so that data can be consumed by the right actor: return false if someting wrong
	virtual Bool_t SetReadyforConsumer(); 
	
	//! test if some data can be collected be the producer (left part of the connection) 
	virtual Bool_t IsReadyforProducer() ; 
	
	//! Prepare the connection so that data can be produced by the left actor	: return false if someting wrong
	virtual Bool_t SetReadyforProducer(); 

	//! name of the connection, based on actors's names and size of the block
	/*!
		be careful in case this is called in a loop, the name is computed each time
	*/
	virtual std::string GetName() const
	{
		std::stringstream 
			name_link;
		name_link << "[" << fProducer->GetName() << "]-" << fBlock->GetMaxSize() << "-[" << fConsumer->GetName() << "]";
		
		return name_link.str();
	}
	
	virtual FrameBlock *GetBlock(const Char_t /* opt */)
	{		
		return fBlock;
	}

};

//! Default loader  : to be used in non-multi threaded programs
/*!
*/
class DLActor : public ALoadedActor
{
public:
 	DLActor(const Char_t *name) : ALoadedActor(name) 
 		{;}
 	virtual ~DLActor()
 		{;}
};

//! An general emulator, could run or not in multi-threading environement 
/*!
 */
class DEmulator : public BaseEmulator
{	
protected:
	//! link left and right	... depends on the emulator's implementation
	virtual void LinkActors(ALoadedActor *left, ALoadedActor *middle, ALoadedActor *right);
	
public:
 	DEmulator(std::string name = "DEmulator") : BaseEmulator(name)
		{;}
 	virtual ~DEmulator()
		{;}

	template <class T_actor> 
	DLActor *Add(const Char_t *name_left, const Char_t *name_new_actor, const Char_t *name_right = "-", const Char_t *conf_dir = "./")
	{
		return BaseEmulator::AddLoader<DLActor,T_actor>(name_left,name_new_actor,name_right,conf_dir,T_actor::process_config);
	}
	template <class T_actor, class T_Configurator> 
	DLActor *Add(const Char_t *name_left, const Char_t *name_new_actor, const Char_t *name_right = "-", const Char_t *conf_dir = "./")
	{
		return BaseEmulator::AddLoader<DLActor,T_actor>(name_left,name_new_actor,name_right,conf_dir,T_Configurator::process_config);
	}	
	//! Init the emulator
	// virtual Bool_t Init(); 
	//! Run max loop
	// virtual Bool_t Run(UInt_t max_loop); 
};

} // end namespace


#endif

