

#ifndef Gw_FrameDispatcher
#define Gw_FrameDispatcher

// Gw includes
#ifndef Gw_Watchers
#include "Watchers.h"
#endif

// ADF includes
#ifndef ADF_Trigger
#include "Trigger.h"
#endif
#ifndef ADF_NarvalInterface
#include "NarvalInterface.h"
#endif

// ROOT includes
#include "TH1.h"
#include "TStopwatch.h"
#include "TROOT.h"

// standard includes
#include <vector>

namespace Gw {

//! Process a buffer, trigg and dispachs the work to sub-watchers.
/*!
	As a Narval consumer, it receives a buffer to be processed through the ProcessBlock methods.
	The block is processed and look for triggers.
	
	A default trigger HAS to be set using the SetTrigger method. It is owned by FrameDispatcher
	so will be deleted once this is deleted !!
	
	Only once a (valid) default trigger is set, Watchers could be added to the FrameDispatcher using
	the Add method. Thus once the trigger fired in the ProcessBlock method, the associated
	watcher is executed.
	
	As well once can associate a watcher to a particular trigger with the method
	Add (Watcher *, DFTrigger *). 
	The passed trigger is owned by this so it will be deleted.
	 
	On can associate the same trigger to different watchers, the FrameDispacher
	takes care about deleteing correctly without problems (in principle ..)
	
	\code
		AgataFrameTrigger *trig;
		Watcher *w1, *w2; 
		
		// allocate trig, w1 and w2 
		// ...
		trig = new AgataFrameTrigger( .. )
		w1 = new Watcher ( .. );
		w2 = new Watcher ( .. );	
		//
		{
		FrameDispatcher dispatcher;
 
		dispatcher.SetTrigger( trig );
		
		dispatcher.Add ( w1, trig );
		dispatcher.Add ( w2, trig );	
		
		} // => OK, dispatcher takes care to delete trig just once !
	\endcode
	
	As triggers are owned by the FrameDispatcher, a given pointer could not be 
	shared between two FrameDispatcher. However one can create two of them with
	the same conditions.
	
	\author Olivier Stezowski	
*/
class FrameDispatcher : public Watcher, public ADF::NarvalConsumer 
{	
private:
	//! to 
	static Short_t gDoTimer;
	
private:
	//! List of triggers owned by this 
	ADF::PtrStack<ADF::DFTrigger> fListOfTriggers;
	
protected:
	//! Histograms that displays some statistics
	TStopwatch fTimeControl;
	
private:
	static const UInt_t kMaxCycle = 1000u;
	UInt_t fWhichBin;	
	
private:	
	//! Trigger rate : number of time one of watcher has been called
	TH1D *fTriggerRate;
	//! 
	TH1F *fBufferSize;	
	//! one histogram per watcher for Real and cpu time used to processed
	TList *fRealTime;
	TList *fCPUTime;
	
	//! to keep the means values
	std::vector< std::pair<Double_t, Double_t> > fCurrentValues;
	
protected:
	//! To be overwritten by real implementation if a canvas is produced.
	virtual void DoCanvas(TCanvas *c, Option_t *);
	
	//! 
	// void BeginOfBlock();
	// void EndOfBlock();
	
public:
	FrameDispatcher();
	FrameDispatcher(const char *, const char *, TDirectory *mother_watcher_dir = 0x0, TDirectory *mother_tag_watcher_dir = 0x0);
	virtual ~FrameDispatcher();
	
	//! To set on/off stopwatch that gives informations about cpu/user time to execute. Switch on only if asked (system call, time consuming)
	/*!
		- 0 (default) means no timer active
		- 1 means only dispacher
		- 2 means dispacher + all individual watchers attached to this dispacher
	 */
	static Short_t DoTimer(Short_t dotimer)
	{
		Short_t old = gDoTimer;
		switch (dotimer) {
			case 0:
			case 1:
				gDoTimer = dotimer;
				break;				
			default:
				gDoTimer = 2;
				break;
		}
		return old;
	}
		
	//! do nothing ... not a complete consumer
	virtual void process_initialise (UInt_t */*error_code*/)
		{;}
		
	//! do nothing ... not a complete consumer
	virtual void process_reset (UInt_t */*error_code*/)	
		{;}	
		
	//! look for a Frame using the Trigger and dispatch it to sustasks		
	virtual UInt_t ProcessBlock (ADF::FrameBlock &);
	
	//! This is the default trigger for any added watcher 
	/*!
		Using the Add(Watcher *) method, a new watcher is added to the FrameDispatcher.
		
		
		The trigger is adopted, deleted with this.
	*/
	virtual Bool_t SetTrigger(ADF::DFTrigger *);

	//! allocate a new watcher so that it creates corectly its directory under the dispatcher directory 
	template<typename Watcher_T> Watcher_T * Add(const Char_t *name, const Char_t *title, ADF::DFTrigger *t = 0x0)
	{
		//set the watcher. Directories are set in Add since no need to have them at construction to load spectra
		Watcher_T *watcher = new Watcher_T(name,title); 
		
		if ( t ) 
			Add( watcher , t );	
		else 
			Add( watcher );
		
		return watcher;
	}
	
	//! change directories and Load objects if required
	virtual void SetDirectory(TDirectory *mother_dir_of_watcher, TDirectory *mother_dir_for_tag, Bool_t load_objects = true);
	
	//! add only Watchers to the list of tasks to avoid problems.
	virtual void Add(TTask *);
	
	//! To set a particular watcher to a particular trigger 
	/*!
		The trigger is adopted, deleted with this.
	*/	
	virtual void Add(Watcher *, ADF::DFTrigger *);	
	
	//! call Zero for all the watchers
	virtual void Zero(Option_t * hname = "pool", Option_t *binning="");  //*MENU*
		
	ClassDef(FrameDispatcher,0) // From a block, dispatchs specific Triggers to other Watchers
};

} // end namespace

#endif


