/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file TTreeWatcher.h header file for TTreeWatcher.cpp */

#ifndef Gw_TTreeWatchers
#define Gw_TTreeWatchers

#include "Watchers.h"
#include "GCondition.h"

#include "TTree.h"
#include "TH1.h"
#include "TH2.h"
#include "TLeaf.h"

namespace Gw 
{

//! Base class for a Watcher that fill a TTree or some branches of a TTree
/*
 
 DoSpectra (default = 2) : 
 - 0 user spectra are not filled. Spectra under GCondition are always filled
 - 1 user spectra are filled only if the global conditions for this watcher is passed
 - 2 user spectra are always filled 
 
 DoTree (default = 2) :
 - 0 Tree is not filled 
 - 1 Tree is filled only if condition 
 - 2 Tree is filled and TEntryList is build
	
*/
class TTreeBuilder
{
protected:
    //! some global strings to fix the Output directory where the Tree is to be written and its default name
    static TString gDefaultPathOfOut;
    static TString gDefaultTTreeName;
  
private:
	// the tree to be filled by the watcher
	TTree *fMyTree; 
	// true if the tree belongs to this 
	Bool_t fIsOwner;
	
protected:
	// 0 no, 1 fill spectra only if condition is filled, 2 fill specta in any cases
	Short_t fDoSpectra;
	// 0 no, 1 fill the tree only is all conditions passed , 2 fill the tree and the TEntryList 
	Short_t fDoTree;
	
protected:
	// number of container of conditions (GCondition) on this watcher
	Short_t fNbConditions;
	
protected:
	// Should be implemented to set properly the branches in new or update mode
	virtual void SetBranches() = 0;
	//! Fill the tree if it is the owner of the tree
	virtual void FillTree();
	
public:
	TTreeBuilder(const char *name, const char *title, TTree *tree = 0x0, Short_t do_spectra = 2, Short_t do_tree = 2);
	virtual ~TTreeBuilder();
    
	//! to know spectra mode for this watcher 
	Short_t GetDoSpectra() const
	{
		return fDoSpectra;
	}	
	//! to know spectra mode for this watcher 
	Short_t GetDoTree() const
	{
		return fDoTree;
	}	
	//! to know spectra mode for this watcher 
	void SetDoSpectra(Short_t do_spectra) 
	{
		fDoSpectra = do_spectra;
	}	
	//! to know spectra mode for this watcher 
	void SetDoTree(Short_t do_tree)
	{
		fDoTree = do_tree;
	}	
	
	//! return the number of conditions registered in case this tree/branch has some 
	virtual Int_t GetNbConditions() const
	{
		return fNbConditions;
	}
	
	//! to get the current Tree
	TTree *GetTree()
		{ return fMyTree; }
};	

//! Watcher to build a tree with branches processed in several watchers
/*
 
*/
class TTreeMaster : public Watcher, public TTreeBuilder
{
private:
	// list of Watchers filling a branch
	TList fListOfWatchers;
    
protected:
    //! to get errors while filling the different branches
    TH1 *fErrors;

protected:
	// Should be implemented to set properly the branches in new or update mode
	virtual void SetBranches()
		{;}
	
public:
	TTreeMaster(const char *name, const char *title, Short_t do_spectra = 2, Short_t do_tree = 2);
	virtual ~TTreeMaster();

    //! to change the default value of path for output. It returns the old one
    static void SetDefaultPathAndName(const char *new_path = "./Out",const char *treename = "TreeMaster");
	
    /*
    virtual void SetDirectory(TDirectory *mother_dir_of_watcher, TDirectory *mother_dir_for_tag, Bool_t load_objects = true)
    {
        Watcher::SetDirectory(mother_dir_of_watcher,mother_dir_for_tag,load_objects);
    }
     */
    
	//! Save all histograms belonging to this watcher. Should be overloaded in case of spectra not registered in pool
	/*!
	 if option contains !tag, tagged spectra are not saved.
	 if option contains *, all spectra are saved, otherwise assume wildcard
	 
	 Ex: 
	 - * !tag
	 - *_E
	 - *_E !tag 
	 */
	virtual UInt_t Snapshot(Option_t * = "*"); //*MENU*
	
	//! Allocate a watcher in charge of filling one part of the tree ==> Should inherits from TreeBuilder  
	template<typename Watcher_T> Watcher_T * Add(const Char_t *name, const Char_t *title, Short_t do_spectra, Short_t do_tree)
	{
		Watcher_T *watcher = new Watcher_T(name,title,GetTree(),do_spectra,do_tree);		
		Add( watcher );
		
		return watcher;		
	}
	//! Allocate a watcher in charge of filling one part of the tree ==> Should inherits from TreeBuilder  	
	template<typename Watcher_T> Watcher_T * Add(const Char_t *name, const Char_t *title)
	{
		Watcher_T *watcher = new Watcher_T(name,title,GetTree());
		Add( watcher );
		
		return watcher;
	}
	
	//! Allocate a standard watcher i.e. not a tree builder 
	template<typename Watcher_T> Watcher_T * AddWatcher(const Char_t *name, const Char_t *title)
	{
		Watcher_T *watcher = new Watcher_T(name,title);
		Add( watcher );
		
		return watcher;
	}

	//! Fill the tree if it is the owner of the tree
	virtual void FillTree()
	{ 
		if ( GetTree() )
			GetTree()->Fill(); 
	}
	
    //! add only Watchers to the list of tasks to avoid problems.
	virtual void Add(TTask *);
	
	//! To set a particular watcher to a particular trigger
	/*!
     The trigger is adopted, deleted with this.
     */
	virtual void Add(Watcher *, ADF::DFTrigger *);
    
	//! watch the current frame
	virtual void Exec(Option_t *option="");
    
	ClassDef(TTreeMaster,0)  // Watcher that owns the tree and call Fill 
};
	
//! To handle a simple branch 
class SimpleGCondHandler1D : public GCondHandler
{
private:
	TLeaf *fData;
	
private:
	TH1 *fHisto;
	
public:
	SimpleGCondHandler1D(GCondition *cond, TLeaf *data, TH1 *h) : GCondHandler(cond), fData(data), fHisto(h)
		{;}
	virtual ~SimpleGCondHandler1D() 
		{;}			
	
	virtual Bool_t Check();
	virtual void Fill()
	{
		for (Int_t i = 0; i < fData->GetLen(); i++) {
			fHisto->Fill(fData->GetValue(i));
		}
	}	
};

class SimpleGCondHandler2D : public GCondHandler
{
private:
	TLeaf *fData1;
	TLeaf *fData2;
	
private:
	TH1 *fHisto;
	
public:
	SimpleGCondHandler2D(GCondition *cond, TLeaf *data1, TLeaf *data2, TH1 *h) : GCondHandler(cond), fData1(data1), fData2(data2), fHisto(h)
	{;}
	virtual ~SimpleGCondHandler2D()	
	{;}			
	
	virtual Bool_t Check()
	{
		Bool_t is_passed = false;
		//
		for (Int_t i = 0; i < fData1->GetLen(); i++) {
			for (Int_t j = 0; j < fData2->GetLen(); j++) {
				if ( fCond->IsPassed(fData1->GetValue(i),fData2->GetValue(j)) )
					is_passed = true;
			}
		}
		return is_passed;
	}
	virtual void Fill()
	{
		for (Int_t i = 0; i < fData1->GetLen(); i++) {
			for (Int_t j = 0; j < fData2->GetLen(); j++) {
				fHisto->Fill(fData1->GetValue(i),fData2->GetValue(j));
			}
		}
	}	
};

//! A watcher filling TTree, Spectra under conditions 
/*!

 
*/
class TTreeWithConditions : public Watcher, public TTreeBuilder 
{
protected:
	//! Clone of the tree part filled by this watcher with only one entry. Used to test Formulas
	TTree *fClone;
protected:
	//! conditions and spectra are automaticaly handled by the base class
	std::list < GCondHandler * > fCondHandler; 
protected:
	//! list of conditions defined by the user. spectra are built/filled by the user
	std::list < GCondition * > fUserGCond;		
protected:
	//! list of formula defined by the user to apply conditions 
	std::vector < TTreeFormula * > fUserTFCond;
protected:
	//! list of all conditions. Used for external and to delete them   		
	std::vector < GCondition * > fAllGCond;
    
protected:
	//! Should be used to reset a various data filled by this watcher
	virtual void Clear(Option_t* = "") = 0;
protected:
	//! Add the specific branches filled by this watcher to the given tree 
	virtual void AddBranches(TTree *tree) = 0;
	//! Should be used fill the current event
	virtual void FillBranches() = 0;
	
protected:
	//! add branches to the tree and the ciruclar tree
	virtual void SetBranches() 
	{ 
		if ( fDoTree > 0 )
			AddBranches(GetTree()); 
		AddBranches(fClone); 
	}
protected:
	//! define user spectra ... to be implementted by user only if it defines its ones spectra without conditions
	virtual void BuildSpectraByUser() 
		{;}	
	//! Fill user spectra ... to be implementted by user only if it defines its ones spectra without conditions
	virtual void FillSpectraByUser() 
		{;}

protected:
	//! Build a 1D condition using the given leaf. It allocates also the spectrum
	virtual std::pair< GCond1D *, TH1 * > MakeGCond1D(
							const Char_t *name_cond, const Char_t *title_cond, const Char_t *name_leaf, 
							const Char_t *name_hist, const Char_t *title_hist , Int_t bin, Double_t min, Double_t max, 
							Option_t *opt = "F",
							const Char_t *sub = "");
	//! Build a 2D condition using the given leaves. It allocates also the spectrum 
	/*
	 The leaf name should have the form X:Y otherwise it fails
	 */
	virtual std::pair< GCond2D *, TH2 * > MakeGCond2D(
							const Char_t *name_cond, const Char_t *title_cond, const Char_t *name_leaf,  
							const Char_t *name_hist, const Char_t *title_hist, Int_t binx, Double_t xmin, Double_t xmax, Int_t biny, Double_t ymin, Double_t ymax, 
							Option_t *opt = "F",
							const Char_t *sub = "");		
	
	//! Add a new condition. delete it done by the base class. It is the duty of the user to deal with the user and fill the needed spectra
	virtual GCond1D * MakeGCond1D(const Char_t *name_cond, const Char_t *title_cond);
	virtual GCond2D * MakeGCond2D(const Char_t *name_cond, const Char_t *title_cond);
	
public:
	virtual Int_t CheckCondByHandler();
	virtual Int_t CheckTreeFormula();
	
	//		virtual void FillSpectraByHandler();
	
public:
	TTreeWithConditions(const char *name, const char *title, TTree *tree = 0x0, Short_t do_spectra = 2, Short_t do_tree = 2);
	virtual ~TTreeWithConditions();
		
	//! Save all histograms belonging to this watcher + conditions unless !cond is added
	/*!
	 */
	virtual UInt_t Snapshot(Option_t * = "*"); //*MENU*
	
	const std::vector < GCondition * > & GetListOfGConditions() const
	{
		return fAllGCond;
	}
	/*
	 const std::vector < TTreeFormula * > & GetListOfFormula() const
	 {
	 return fUserTFCond;
	 }	
	 */
	// 
	virtual void SetDirectory(TDirectory *mother_dir_of_watcher, TDirectory *mother_dir_tag = 0x0, Bool_t load_objects = true);
	
	//! Add a general condition (see TTree::Draw()). Return the total number of expression in this watcher once added if success 
	virtual Int_t AddTreeFormula(const Char_t *name, const Char_t *expression); //*MENU*	
	
	//! to be overwitten in case you add your own conditions but TTreeWithConditions::CheckCond() should be called in the user method
	virtual Int_t CheckCond()
	{
		return CheckCondByHandler();
	}
	
	//! to be overwitten in case you add your own conditions/spectra
	virtual void FillSpectra()
	{
		// spectra declared with conditions are automatically filled 
		FillSpectraByUser();
	}
	
	//! In principle works for any watchers that inherits from it 
	virtual void Exec(Option_t *option="");
	
	ClassDef(TTreeWithConditions,0)  // Watcher that owns the tree and call Fill 
};

	
} // namespace

#endif

