/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file Watcher.h header file for Watcher.cpp */

#ifndef Gw_Watchers
#define Gw_Watchers

// ADF includes
#include "NarvalInterface.h"
#include "Trigger.h"
#include "GwLogMessage.h"

// ROOT includes
#include "TTask.h"
#include "TStopwatch.h"
#include "TObjArray.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TH1.h"
#include "TTimer.h"
#include "TROOT.h"
#include "TFolder.h"

namespace Gw {

//! Base class for a Watcher
/*!
	A watcher is working on a particular type of Frame or several Frame.
    For that it is associated to a Trigger that content what needs the Watcher to do a job 
    (see the Exec method that should be redefined in any particular Watcher).
	It could be simply just to display the content of a Frame or to perform more complex analysis.
	
	Be careful however that several watchers are likely to be chained so avoid heavy processing
	of the data in one of them if it does not fit your analysis.
	For that, use a FrameDispacher, it gives the time requested for any Watcher to fullfil the task. 
	
	It inherits of TTask so that it could be displayed in a ROOT browser, 
	and on the menu it could be set active/inactive.
	
	\author Olivier Stezowski
*/
class Watcher : public TTask 
{	
public: 
	
private:
	//! List of all registered watchers
	static TList *fgListOfWatchers;
	//! add/Remove watcher to the list of all 
	/*!
	 if opt == "-", remove the wather from the list, otherwise add with potentially an option
	 */
	static void Register(Watcher *, Option_t *opt = "+");
	
public:
	//! to retrieve the last registered watcher of one type.
	/*!
	 if opt == baseclass, return the last watcher that inherits from basseclass
	 if opt == class, return the last wathcer that is of type classname
	 */	
	static Watcher *GetLastRegistered(const Char_t *classname, Option_t *opt = "baseclass");
public:
	//! move a watcher on top of the list
	static void SetFirst(Watcher *);
	
protected:
	mutable LogMessage fLog; //! to send log messages to the log system
	
private:
	//! global flag used for the initialisation of the data member fDeletePool
	static Bool_t fgDeletePool;
	
private:
	//! histograms of the pool are deleted or not depending of the flag 
	Bool_t fDeletePool;
	
	//! pool of object ... objects in there are deleted by default!
	TObjArray fPoolOfObjects;
	//! list of tagged histograms
	TObjArray fTagged;
	//! Reference for any tagged histogram	
	TObjArray fReference;	
	
protected:
	//! Objects are in folders. They can be saved/load in/from directory. This is the mother in which this watcher directory is set 
	TDirectory *fDirectory; //! 
	//! Tagged Objects are in folders. They can be saved/load in/from directory. This is the mother in which this watcher directory is set 
	TDirectory *fTagDirectory; //!
	//! Top folder in which ae stored all spectra
	TFolder *fTopFolder;
	
private:
	//! Status to know if an error occured during Exec
	Short_t fLastExecStatus;
	
protected:
	//! trigger associated to this watcher
	ADF::DFTrigger *fTrigger;
	
protected:
	//! to specify what kind of watcher 
	/*!
		default is SpectraBuilder. Other possibilities are VertexBuilder, TreeBuilder
		The string could contain more that one kind
	 */
	// TString fKind;
  	
protected:
	//! if an object is found in the dir, its content is copied into new_obj
	/*!
		It takes care of checking if an object with the same name already exists
		in dir. if yes, it copied its content into new_obj 
	*/
	void LoadObject(TObject *new_obj, TDirectory *root_dir);
	
	//! facility for other watchers: keep a list of all histograms (objects) for global operations
	/*!
		- check if a histogram with the same name exist in the directory associated with this Watcher. if yes and it has the same range, load the content
		- set the TDirectory to the current one
	*/
	TObject *AddToPool(TObject *ob);
	
	//! It creates a new embedded canvas
	/*!
		In case root is run with the new version of the root browser, it adds a new Tab.
	*/
	TCanvas *NewCanvas(Option_t *);
	
	//! utilities to add/move/remove objects from one dir to another one
	void DirToDir(TDirectory *, TDirectory *, TObjArray &, Bool_t load_objects = true);
	void RemoveFromDir(TObject *, TDirectory *);
	void AddToDir(TObject *, TDirectory *);
	//! in order to save/load, the structure of the top folders should exists in the root dir
	Bool_t MakeDir(TFolder *f, TDirectory *mother_dir, Bool_t do_top = true);  
	//! in a folder look for the path to an elmement that is not a folder
	/*
		In case of success, path has the form a/b/c/
		Otherwise empty string. The path is not reset, the user gas to give an empty one 
	 */
	Bool_t GetPathOf(TFolder *f, TObject *searched, TString &path);
	
protected:
	//! for a new histogram, it returns the folder it belongs to. It creates sub and all intermediates if required
//	TDirectory * GetSubFolder(const Char_t *sub = "");
	TFolder *GetSubFolder(TFolder *topfolder, const Char_t *sub = "");
	
	//! Build an histogram, set it to sub directories if required and add it to the pool
	template <typename Histo_T> Histo_T *
		MakeTH1(const Char_t *hname, const Char_t *htitle, Int_t binx, Double_t xmin, Double_t xmax, const Char_t *sub = "")
	{
		Bool_t def = TH1::AddDirectoryStatus();
		TH1::AddDirectory(false); 	
		
		Histo_T *h = 0x0; TString subfolder = sub; 
		//
		h = new Histo_T(hname,htitle,binx,xmin,xmax);
		if ( h ) {
			TFolder *folder = GetSubFolder(fTopFolder,sub);
			folder->Add((TObject *)h); 
			AddToPool(h);
			// does not belong to directory
			h->SetDirectory(0);
		}
		TH1::AddDirectory(def); 	
		
		return h;
	}
	template <typename Histo_T> Histo_T *	
		MakeTH2(const Char_t *hname, const Char_t *htitle, 
				  Int_t binx, Double_t xmin, Double_t xmax, 
				  Int_t biny, Double_t ymin, Double_t ymax, 
				  const Char_t *sub = "")
	{
		Bool_t def = TH1::AddDirectoryStatus();
		TH1::AddDirectory(false); 	
		
		Histo_T *h = 0x0; TString subfolder = sub; 
		//
		h = new Histo_T(hname,htitle,binx,xmin,xmax,biny,ymin,ymax);
		if ( h ) {
			TFolder *folder = GetSubFolder(fTopFolder,sub);
			folder->Add(h); 
			AddToPool(h);
			// does not belong to directory
			h->SetDirectory(0);
		}
		TH1::AddDirectory(def); 	

		return h;
	}	
	template <typename Histo_T> Histo_T *
		MakeTH3(const Char_t *hname, const Char_t *htitle, 
				  Int_t binx, Double_t xmin, Double_t xmax, 
				  Int_t biny, Double_t ymin, Double_t ymax, 
				  Int_t binz, Double_t zmin, Double_t zmax, 				  
				  const Char_t *sub = "")
	{
		Bool_t def = TH1::AddDirectoryStatus();
		TH1::AddDirectory(false); 	
		
		Histo_T *h = 0x0; TString subfolder = sub; 
		//
		h = new Histo_T(hname,htitle,binx,xmin,xmax,biny,ymin,ymax,binz,zmin,zmax);
		if ( h ) {
			TFolder *folder = GetSubFolder(fTopFolder,sub);
			folder->Add(h); 
			AddToPool(h);
			// does not belong to directory
			h->SetDirectory(0);
		}
		TH1::AddDirectory(def); 	

		return h;
	}	
	
protected:
	//! Add this histogram to the list of tagged histograms
	void TagOn(TObject *);
	
	//! To display tagged histograms and their reference
	void DrawTag(TCanvas *, Option_t *);	
	
protected:	
	//! To be overwritten by real implementation if a canvas is produced.
	virtual void DoCanvas(TCanvas *, Option_t *)
		{ ; }
	
	//! Extract from the trigger and given frame pointer (usefull for other watchers)
	/*!
		To get from the trigger the sharedFP with a given type and put th results in the 
		last parameter
	*/
	Bool_t GetFromTrigger(ADF::DFTrigger *, const char *, ADF::SharedFP *&);		
	
public:
	Watcher();	
	Watcher(const char *name, const char *title, TDirectory *mother_dir_of_watcher = 0x0, TDirectory *mother_dir_tag = 0x0);
	virtual ~Watcher();
	
	//! overloaded for efficiency reasons. allocation of TIter is see in Shark (macos) as time consuming
	virtual void CleanTasks();
	
	//! change kind of Watcher
	/*
	void SetKind(const char *kind)
	{
		fKind = kind;
	}
	
	Bool_t IsKind(const char *kind) const
	{
		return fKind.Contains(kind);
	}
	 */
	
	//! Save all histograms belonging to this watcher. Should be overloaded in case of spectra not registered in pool
	/*!
		if option contains !tag, tagged spectra are not saved.
		if option contains *, all spectra are saved, otherwise assume wildcard
	 
	 Ex: 
		- * !tag
		- *_E
		- *_E !tag 
	 */
	virtual UInt_t Snapshot(Option_t * = "*"); //*MENU*
	
	//! TMP : migration to TreeWatcher with Cuts : should be removed  
	void SetLastError(Short_t s = 0)
	{ fLastExecStatus = s ; }
	Short_t GetLastError() const
	{ return fLastExecStatus ; }		
	
	//! reset last status. 0 means no error, 0 < means error, > 0 means ok with conditions
	void SetLastExecStatus(Short_t s = 0)
	{ 
		fLastExecStatus = s ; 
	}
	
	//! to get the last status 
	Short_t GetLastExecStatus() const
	{ 
		return fLastExecStatus ; 
	}		
		
	//! to each watcher a TDirectory is associated 
	//virtual TDirectory *GetDirectory(Option_t * /*opt*/ = "")
	//	{ return fDirectory; }
	virtual TDirectory *GetDirectory(Option_t * opt = "")
	{ 
		TString o = opt; 
		if ( o.Contains("tag") ) 
			return fTagDirectory; 
		return fDirectory; 
	}
	
	virtual TObjArray &GetPoolOfObject()
		{ return fPoolOfObjects; }
	virtual TObjArray &GetReference()
		{ return fReference; }
	
	//! to change the directory in which watcher's objects are stored
	/*!
	- if mother_dir_of_watcher == 0x0 all objects are removed from the directory they belong to
	- try to creat a subdir of mother_dir_of_watcher (or check if it already exists) with the name of the watcher
	 in case of success all objects in the pool belongs to the new directory otherwise objects don't belong to any directory
	*/
	virtual void SetDirectory(TDirectory *mother_dir_of_watcher, TDirectory *mother_dir_tag, Bool_t load_objects = true);
		
	//! force fHasExecuted (recursively) to true
	/*!
		In case in a tree a condition is not good to execute 
		the subtasks, the HasExecute flag is set to true so that the subtasks are not executed.
	*/
	void Pass();
	
	//! to remove them from the TTask menu
	virtual void Continue()	
		{;}	
	virtual void Abort() 
		{;}	
	virtual void SetBreakin(Int_t /*breakin*/ = 1) 
		{;}
	virtual void SetBreakout(Int_t /*breakout*/ = 1)
		{;}

	//! add only Watchers to the list of tasks to avoid problems.
	/*!
		It set the trigger to this one if not already set
	*/
	virtual void Add(TTask *);

	//! To set the Frames (through a trigger) associated to this watcher
	/*!
		This method should be overloaded by the real implementation of the watcher.
		The trigger does not belongs to that watcher so it will not be deleted
	*/
	virtual Bool_t SetTrigger(ADF::DFTrigger *);
		
	//! To know the trigger in which the frame to be watched is embedded
	virtual ADF::DFTrigger *GetTrigger() const
		{ return fTrigger; }
  
	//! set static flag to delete objects in pool
	static void SetGlobalDeletePool(Bool_t d) 
		{ fgDeletePool = d; }
  
	//! set flag to delete objects in pool
	virtual void SetDeletePool(Bool_t d) 
		{ fDeletePool = d; }
  
	//! get flag to delete objects in pool
	virtual Bool_t GetDeletePool() const 
		{ return fDeletePool; }
	
	//! watch the current frame ... to be overwritten by the watcher
 	// virtual void Exec(Option_t *option="") 
	
	//! Reset the histograms 
	/*!
	 The first argument is the histogram or the collection of histograms (pool/tag)
	 to be Reset to 0. 
	 Widcards could be given to applied to several histograms at the same time.
	 
	 The second argument, if given is to modified the definition of the histogram i.e. the binning, min and max. 
	 with the followwing format: \n
	 binx minx miny [biny miny maxy]
	 
	 In case a collection is used, it applies the same binning to all the histograms
	*/
 	virtual void Zero(Option_t *hname="pool", Option_t *binning="");  //*MENU*
	
	//! Display some results
	void ShowCanvas(Option_t *option="") ; //*MENU*
	
	//! Compare the current spectra with the tagged ones using Kolmogorov test
	virtual Bool_t TestTag(Option_t *opt);   //*MENU*
	
	//! Tag the current watcher	
	virtual void Tag(Option_t *opt = ""); //*MENU*
	
	ClassDef(Watcher,0)  // base class for a watcher
};

class WatcherWithTag : public Watcher
{	
private:
	//! list of tagged histograms
	//TObjArray fTagged;
	//! Reference for any tagged histogram	
	//TObjArray fReference;	
	
protected:
	//TObjArray &GetReference()
	//	{ return fReference; }
	
protected:
	//! Tag Histograms belong to a ROOT TDirectory
//	TDirectory *fTagDirectory; //!
	
protected:
	//! Add this histogram to the list of tagged histograms
	//void TagOn(TObject *);
	
	//! To display tagged histograms and their reference
	//void DrawTag(TCanvas *, Option_t *);
	
public:
	WatcherWithTag(); 	
	WatcherWithTag(const char *name, const char *title, TDirectory *mother_dir_of_watcher, TDirectory *mother_dir_tag);
	virtual ~WatcherWithTag() 
		{;}
	
	//! to change the directory 
	/*!
	 Histograms that belongs to pool belongs now to the main_dir
	 */
	// virtual void SetDirectory(TDirectory *mother_dir_of_watcher, TDirectory *mother_dir_tag = 0x0, Bool_t load_objects = true);

	//! to each watcher a TDirectory is associated 
	//virtual TDirectory *GetDirectory(Option_t * opt = "")
	//	{ TString o = opt; if ( o.Contains("tag") ) return fTagDirectory; return Watcher::GetDirectory(); }
	
	//! Reset the histograms stored in the pool of histograms
 	//virtual void Zero(Option_t *hname="pool", Option_t *binning="");  //*MENU*
	
	//virtual void Add(TTask *);

	//! Compare the current spectra with the tagged ones using Kolmogorov test
	//virtual Bool_t TestTag(Option_t *opt);   //*MENU*
	
	//! Tag the current watcher	
	//virtual void Tag(Option_t *opt = ""); //*MENU*
	
	ClassDef(WatcherWithTag,0)  // To handle watchers with tagged histograms
};

class LoopControl : public WatcherWithTag
{
private:
	//! Histograms of the time needed to execute one loop	
	TH1 *fLoopTime;
private:	
	//! to get the time between two loops
	TStopwatch	fTimeControl;
	//! min time for one loop (in milliseconds)
	Double_t	fDelta;	
private:
	TTask *fMother;
private:
	TDirectory *fMotherDirWatcher;
	TDirectory *fMotherDirTag;
	
private:
	//! time to auto-restart Save 
	TTimer  *fTimer;
	//! Last arguments of Save. Use by AutoSave 
	TString fLastSaveArg1;
	TString fLastSaveArg2;
	TString fLastSaveArg3;
	  Int_t fLastSaveArg4;
	
protected:	
	//! To be overwritten by real implementation if a canvas is produced.
	virtual void DoCanvas(TCanvas *, Option_t *);
	
public:
	LoopControl(TTask *, TDirectory *, TDirectory *);
	virtual ~LoopControl();
	
	//! call Zero for all the watchers
	virtual void Zero(Option_t * hname = "pool", Option_t *binning="");  //*MENU*
	
	//! Save the spectra in files 
	/*!
	 - main_file is the name of the root file to save spectra
	 -  tag_file is the name of the root file to save tagged spectra
	 - option has the following format :	
		- ; means save spactra in the current TDirectory
		- ( means open a new file
		- ) means close the file
		- ch means works with a chain i.e. sequence of names with the same pattern.
		- option to Zero spectra is given using {opt1:opt2}. 
	 - autotime : call automatically this command again after autotime seconds. Off if set to 0.
	 
	 Here are some examples.
	 - if you have open your watcher is a root directory and would like just to save there the current histograms : "" "0" ";" 0
	 - to save in the current root dir, open a new one and move all hitograms there : "my_new.root" "0" ";(" 0
	 - to dump histograms in a root file (open, save and close the file) : "my_dump" "0" "(;)" 0
	 - to dump histograms in a chain of root files (open, save and close the file) : "my_dump" "0" "ch(;)" 0 \n
	    it looks for existing my_dumpXXXX.root files in the current dir XXXX = [0, ...[. It takes the first non existing one.
	 - to open a new files that becomes the mother file of all watchers, and save histo there : "my_dump" "0" "(;" 0
	 - to set spectra of the pool to zero :  "my_dump" "0" "(; {pool}" 0
	 - to set all spectra to zero and change the binning : "my_dump" "0" "(; {pool:100 0 100}" 0
	*/
//	void Save(const Char_t *main_file = "", const Char_t *tag_file = "0", Option_t *option = "", Int_t autotime = 0); //*MENU* *ARGS={main_file=>fLastSaveArg1,tag_file=>fLastSaveArg2,option=>fLastSaveArg3,autotime=>fLastSaveArg4}
	void Save(const Char_t *main, const Char_t *tag, const Char_t *option, Int_t autotime); // *MENU* *ARGS={main=>fLastSaveArg1,tag=>fLastSaveArg2,option=>fLastSaveArg3,autotime=>fLastSaveArg4}
	
	//! Auto save in case of timer set. To stop timer, use SetAutoTime(0)
	void AutoSave();
	//! change time to autosave
	void SetAutoTime(Int_t autotime = 0); //*MENU* 
	//!
	const Char_t *GetLastSaveArg1() 
		{ return fLastSaveArg1.Data(); }
	const Char_t *GetLastSaveArg2() 
		{ return fLastSaveArg2.Data(); }
	const Char_t *GetLastSaveArg3() 
		{ return fLastSaveArg3.Data(); }
	Int_t GetLastSaveArg4() 
		{ return fLastSaveArg4; } 
	
	//! Display some results
	//void ShowCanvas(Option_t *option = "") ;//*MENU*
	
	virtual void Exec(Option_t *option); 
	
	ClassDef(LoopControl,0)  // to control the execution of LoopOnTasks 	
};

//! To start a loop on tasks
/*!
	By default in root TTask are executed once when ExecuteTask is called.
	This task starts an infinite loop if the option l is given. 
	The infinit loop can be stopped by calling Abort
	
	To know if the loop is running, just pass in the browser on it and in the 
	tooltip you will see [ON] or [OFF] depending on the status.
	
	\author Olivier Stezowski
*/
class LoopOnTasks : public TTask 
{
private:
	//! true is the loop tasks is currently looping on task
	Bool_t	fIsLoop;
	//! timer used to really start the loop on task. It avoids having the start window stacked
	TTimer  *fTimer;
	
private:	
	virtual Bool_t HandleTimer(TTimer* timer);
	
private:
	LoopControl *fControl;
	
public:
	LoopOnTasks(TDirectory *dir_watchers, TDirectory *dir_tag) : 
		TTask("LoopOnTask","0"), 
		fIsLoop(false), 
		fTimer(new TTimer()), 
		fControl(new LoopControl(this,dir_watchers,dir_tag)) { Add(fControl); }
	virtual ~LoopOnTasks()
	{ 
		/*  delete fControl */
		; 
		// no need to do it, done by task destructor
	}	

	//! to remove them from the menu
	virtual void Continue()	
		{;}	
	virtual void SetBreakin(Int_t /*breakin*/ = 1) 
		{;}
	virtual void SetBreakout(Int_t /*breakout*/ = 1)
		{;}

	//! To set this active/inactive	
  	virtual void SetActive(Bool_t active=kTRUE); // *TOGGLE*
	
	//! Abort looping if started	
	virtual void Abort(); //*MENU*
	
	virtual void ExecuteTask(Option_t* option = "l"); // *MENU*
	
	//! the main loop			
	virtual void Exec(Option_t *option); 
	
	ClassDef(LoopOnTasks,0)  // to loop on task once or infinitely
};
	
} // end namespace

#endif


