/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file BaseNucleus.h header file for a BaseNucleus */

#ifndef Gw_BaseNucleus
#define Gw_BaseNucleus
#define GW_BASENUCLEUS_H Gw_BaseNucleus
	
#ifndef ROOT_TObject
#include <TObject.h>
#endif

#ifndef ROOT_TMap
#include <TMap.h>
#endif

using namespace std;

namespace Gw {

/** Base class for a nucleus which means its composition

	This class will be used widely to define a nucleus as composed on nucleons. Because it will be extended 
	to strange nucleus, 

	@author Olivier Stezowski
*/
class BaseNucleus : public TObject
{
private:
    static TMap *gTable;

protected:
    Int_t fA;	// total number of nucleons
    Int_t fN;	// number of neutrons
    Int_t fZ;	// number of protons
 
public:
   //! Set any components to 0 
   BaseNucleus() ;
   explicit BaseNucleus(Int_t, Int_t, Int_t);
   virtual ~BaseNucleus();
    
   Int_t GetA() const;
   Int_t GetZ() const;
   Int_t GetN() const;

   void SetA(Int_t A);
   void SetZ(Int_t Z);
   void SetN(Int_t N);

   //! Load the correspondence table between the symbol of the nucleus and Z from a XML file 
   /*!	
	The user has the possibility to define its own XML file following the "TableNucleus.dtd" template (see UserGuide ndb section) 
	@param xmlfile (default) "TableNucleus.xml" (see UserGuide ndb section)
   */
   static bool LoadTable(const char *xmlfile = "");

   //! 
   /*!
	 
  */ 
   bool SetAZN(Int_t A, Int_t Z);

   //! Set A, Z, N from the symbol of the nucleus (see the correspondence table) 
   /*!
	@param s Symbol of the nucleus
	@see LoadTable 
  */
   bool SetAZN(const char* s);
   /*
	It returns false (without changing anything) if it cannot be a nucleus (i.e. A < Z).
	If N != A - Z, it keeps A-Z and send a warning.
   */
   bool SetAZN(Int_t A, Int_t Z, Int_t N);

   //! Set A, Z, N from the symbol of the nucleus (see the correspondence table) and A
   /*!
	N is fixed by N = A-Z (A must be >= Z)
	@param s Symbol of the nucleus
	@param A Number of nucleons
	@see LoadTable
  */
   bool SetAZN(const char *s, Int_t A);

   //! return the symbol corresponding to Z if the nucleus is found in the correspondence table 
   /*!
	if no nucleus is found, return "" 
	@param Z Number of protons
	@see LoadTable
   */
   static const char *WhichSymbol(Int_t Z);
   static const char *WhichSymbol(Int_t Z, Int_t A);

   //! return the number of protons (Z) corresponding to a symbol if the nucleus is found in the correspondence table, else return -1
   /*!
	@param s Symbol of the nucleus
	@see LoadTable
   */ 
   static Int_t WhichZ(const char *s);

   //! return the full name corresponding to a symbol if it is found in the correspondence table, else return ""
   /*!
	@param s Symbol of the nucleus
	@see LoadTable
   */ 
   static const char* WhichFullName(const char *s);

   //! return the full name corresponding to the number of proton Z if the nucleus is found in the correspondence table, else return -1
   /*!
	@param Z Number of protons
	@see LoadTable
   */ 
   static const char* WhichFullName(Int_t Z);
   
   //! Root dictionary related
   ClassDef(BaseNucleus,1) // Base class for a nucleus (just composition)
};
// inline members
inline Int_t BaseNucleus::GetA() const { return fA; }
inline Int_t BaseNucleus::GetZ() const { return fZ; }
inline Int_t BaseNucleus::GetN() const { return fN; }
inline void BaseNucleus::SetA(Int_t A) {fA = A;}
inline void BaseNucleus::SetZ(Int_t Z) {fZ = Z;}
inline void BaseNucleus::SetN(Int_t N) {fN = N;}

}
#endif
