#ifndef GW_GAMMATRACKER_H
#define GW_GAMMATRACKER_H

#ifndef ROOT_Rtypes
#include "Rtypes.h"
#endif

#ifndef ROOT_TNamed
#include "TNamed.h"
#endif


namespace Gw {

/** Base class to build tracker families

	Any tracking algorithm must inherit from GammaTracker to be used in the GammaWare. \n
	Once a new tracker is loaded (see TrackerLoader.C), it must be allocated by using the static method GetTracker. \n
	Because GammaTracker defines an general interface, it is kept as simple as possible. Inputs and outputs are just buffers of doubles.
	Inputs are energies and positions of the different impacts in the detector (with their associated errors). Outputs are the
	reconstructed energies, the emission angles (theta1, phi1) ands the direction of the second impact (theta2, phi2) for
	polarisation measurements. \n
	Different tracking may work with different unit system. It can be checked/modified through the DistFactor methods.
	Once an event has been initialised (see SetEvent methods), the tracking is called with DoTracking(). The output buffers contain then the tracked gamma-rays. 


	@author Olivier Stezowski
 */
class GammaTracker : public TNamed
{	
private:
	UInt_t MaxImpacts;	//! Max # of impacts 
		
	Double_t DistFactor; 	//! Units for lengths 
	Double_t EnerFactor;	//! Units for energies

protected:
	UInt_t    InputN;	//! # of impacts 
	Double_t *InputE;	//! energies 	
	Double_t *InputDE;	//! errors on energies 

	Double_t *InputX;	//! X positions 
	Double_t *InputY;	//! Y positions 
	Double_t *InputZ;	//! Z positions 
	Double_t *InputDX;	//! DX positions 
	Double_t *InputDY;	//! DY positions 
	Double_t *InputDZ;	//! DZ positions 
	
	UInt_t    OutputN;	//! # of reconstructed gamma-rays 
	Double_t *OutputE;	//! energies of the reconstructed gamma-rays 
	Double_t *OutputTheta1;	//! incoming direction (theta) of the reconstructed gamma-rays 
	Double_t *OutputPhi1;	//! incoming direction   (phi) of the reconstructed gamma-rays 	
	Double_t *OutputTheta2;	//! direction (theta) of the first scattered gamma-ray for polarisation 
	Double_t *OutputPhi2;	//! direction  (phi)  of the first scattered gamma-ray for polarisation	

//	Bool_t fDoPacking;	// true if  packing must be done (data comes from Geant)
//	Bool_t fDoSmearing;	// true if smearing must be done (data comes from Geant)
	
public:
	GammaTracker(const char *name = "GammaTracker", const char *title = "Base to track gamma-rays") ;
	virtual ~GammaTracker();
	
	//! To load an interface to a particular tracker 
	static bool Load(const char *classname, const char *pathtopackage, const char *libraryname); 
	
	//! To know if a tracker is already loaded 
	/*!
		It returns true if a tracker is already loaded
	*/
	static bool IsTracker(const char *classname) { return GetTracker(classname) != NULL; }
	
	//! Interface to get a new tracker
	/*
	Trackers are external programs that should be linked. According to DAQ specifications, any tracking 
	algorithm must be compiled in a single dynamic library 
	*/
	static GammaTracker *GetTracker(const char *classname);
	
	//! Internal unit system for lengths and energies
	/*!
	To know the unit system needed by the tracker. 
	*/
	Double_t GetDistFactor() const { return DistFactor; } 
	Double_t GetEnerFactor() const { return EnerFactor; }
	
	virtual void SetDistFactor(Double_t f) { DistFactor = f; } 
	virtual void SetEnerFactor(Double_t f) { EnerFactor = f; }
	
	Double_t GetMaxImpacts() const { return MaxImpacts; } 
	
//	virtual Int_t Expand(Int_t);

	virtual UInt_t    GetInputN()  const { return InputN; }
	virtual const Double_t *GetInputE()  const { return InputE; }
	virtual const Double_t *GetInputDE() const { return InputDE;}
	virtual const Double_t *GetInputX()  const { return InputX; }
	virtual const Double_t *GetInputY()  const { return InputY; }
	virtual const Double_t *GetInputZ()  const { return InputZ; }
	virtual const Double_t *GetInputDX() const { return InputDX; }
	virtual const Double_t *GetInputDY() const { return InputDY; }
	virtual const Double_t *GetInputDZ() const { return InputDZ; }
	
	//!
	/*
	Energies and Positions are copied, multiplied by a factor to take into account the tracking may work with units 
	different from the ones given by the data flow.
	*/
	virtual UInt_t SetEvent(UInt_t n, const Double_t *e, const Double_t *x, const Double_t *y, const Double_t *z, Double_t efactor = 1.0, Double_t pfactor = 1.0);
	
	void ShowCurrentEvent();

	virtual UInt_t    GetOutputN() const { return OutputN; }
	virtual const Double_t *GetOutputE() const { return OutputE;  }
	virtual const Double_t *GetOutputTheta1() const { return OutputTheta1;  }
	virtual const Double_t *GetOutputPhi1()   const { return OutputPhi1;  }
	virtual const Double_t *GetOutputTheta2() const { return OutputTheta2;  }
	virtual const Double_t *GetOutputPhi2()   const { return OutputPhi2;  }

	virtual void SetPacking()  {;}
	virtual void SetSmearing() {;}

	//! to track the gamma-rays
	/*!
		Once the inputs have been initialised, the tracking is called. It returns the number of reconstructed
		gamma-rays. 
	*/
	virtual Int_t DoTracking();

	//! Root dictionary related
	ClassDef(GammaTracker,0); // Base class for tracking algorithms
};

//! to set the size of the input 
/*
Must be used in conjonction with SetInput(UInt_t i, Double_t ...)
*/	
// virtual UInt_t SetInput(UInt_t ninput) { InputN=(ninput>MaxImpacts)?MaxImpacts:ninput; return InputN; }
	
/*
virtual void SetInput(UInt_t i, Double_t e, Double_t x, Double_t y, Double_t z, Double_t de = 0.0, Double_t dx = 0.0, Double_t dy = 0.0, Double_t dz = 0.0) 
{ if ( i > InputN && i < MaxImpacts ) {InputN=i+1; InputE[i]=e; InputDE[i]=de; InputX[i]=x; InputDX[i]=dx; InputY[i]=y; InputDY[i]=dy; InputZ[i]=z; InputDZ[i]=dz;} }
*/
	
}
#endif



