/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file RandObj.h header file for a RandObj */  

#ifndef GW_RANDOBJ_H
#define GW_RANDOBJ_H

#ifndef ROOT_TObjArray
#include <TObjArray.h>
#endif

#ifndef GW_RANDOM_H
#include "Random.h"
#endif

#include <vector>

class TH1;

namespace Gw {

/**
A class to select randomly an object in a TObjArray of objects

Each object added to RandObj is added with a weight. The distribution and the cumulative distribution are computed 
each time the Add/Remove member are called so that the user doesn't need to keep the weight information. \n
Only pointers to object are stored so <em>they are not deleted in the destructor</em>.
The cumulative distribution is also available (STD vector) to check if it has been correctly computed. \n

The underlying random generator it Rand. You can change it at the beginning even at the
running time !. See Rand for more informations.  \n

<em>This class has only pointer to objects</em>
 
@author Olivier Stezowski
*/
class RandObj : public TObject
{
private:
    Random *fRand;			// unique instance to Rand
    
    TObjArray fObj;			// used to store a array of pointers to objects 
       
    std::vector<Float_t> fCumu;		// used to store the cumulative function
    std::vector<Float_t> fWeight;	// used to store the weight distribution

    /*! determine the cumulative distribution */
    void ComputeCumu();

public:
    RandObj() ;
    virtual ~RandObj() ;

    //! it returns a pointer to a randomly selected object from the collection 
    virtual TObject * Rand();
    
    //! it returns a pointer to a randomly selected object from the collection 
    /*!
      An additional number is returned as an argument and corresponds to the slot number in the TObjArray collection.
      This might be useful when other collections are related to this one by the slot number  
    */
    virtual TObject * Rand(Int_t &);

    //! it adds an object with a weight to this collection.
    /*!    
    What happens if:
    	- the object already exists in the collection:
    	  -# only its weight is modified (sum with the old one)
	  -# if the new weight is positive, the weight of the object is changed. if it becomes
	     negative, the corresponding object is removed from the collection.

    	- the object doesn't exist in the collection:
	  -# It is added to the collection even if the weight is negative. In this case, it is assumed as a mistake and
	     the stored weight is the absolute value.

    Because this class used std::vector, you should avoid adding/removing objects too often
    like in a loop. \n
    */
    virtual void Add(TObject *, Float_t);
    
    //! clear everything (the objects in the collection are not deleted) 
    virtual void Clear(Option_t *opt = "");
    
    virtual Int_t GetSize() const { return fObj.GetEntries(); }
    virtual TObject * At(Int_t slot) const { return fObj.At(slot); }
    
    //! Intensity with which the object at slot is associated 
    virtual Float_t WeightAt(Int_t slot) const { if ( 0 <= slot && slot < fObj.GetEntries() ) return fWeight[slot] ; return 0.0; }
    
    //! Allocate and fill the histogram randomly
    /*!
    	The histogram is allocated with one bin (slot) per object stored in the collection. It might stay NULL if the collection is 
	empty or if ntimes is less than 0. It is the charged of the user to delete the histogram.
    */
    virtual void FillRandom(TH1 *h, Int_t ntimes = 5000);
     
    virtual void ls(Option_t *) const;
    
    //! rootcint dictionary
    ClassDef(RandObj,0); // a class to randomly select object in a TObjArray collection
};
// inline members

}

#endif
