/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file Random.h header file for Random */

#ifndef GW_RANDOM_H
#define GW_RANDOM_H

#ifndef ROOT_TRandom
#include <TRandom.h>
#endif

class TList;

namespace Gw {

/** To change the internal GammaWare random generator. 

Any classes of the GammaWare that relies on random numbers use the Random object. It is a unique 
object (singleton) and was introduced because gRandomom (the ROOT global
pointer to a TRandomom object) is not protected and might be NULL. To get the single instance, see Random::Instance() \n

Random is in fact a collection of TRandomom objects. It proposes the Current() method that returns a 
global TRandomom object that always exists. \n
Several different TRandomom objects are available through this class and the current one can be changed
by using the SetCurrent method. Each random generator has a unique name and six are defined by default: \n
 - basic  \n
 	corresponds to a TRandomom object with a seed = 65539  Periodicity = 10**8 
 - basic_clock\n
 	corresponds to a TRandomom object with a seed based on the clock Periodicity = 10**8 
 - medium    \n  
 	corresponds to a TRandomom2 object with a seed = 65539 Periodicity > 10**14 
 - medium_clock \n
 	corresponds to a TRandomom2 object with a seed based on the clock Periodicity > 10**14  
 - large  \n
  	corresponds to a TRandomom3 object with a seed = 65539 Periodicity = 2**19937-1
 - large_clock\n
  	corresponds to a TRandomom3 object with a seed based on the clock Periodicity = 2**19937-1
	  
See ROOT documentation for more explanations about the different TRandomom implementations.

It is also possible to add new random generator (see Random::Add()). It also 
synchronize gRandomom with the current generator. gRandomom is restored once this is deleted.

@author Olivier Stezowski
*/
class Random
{
private:
    static Random  *fgRandom;

    TList   *fList;	// list of TRandomom objects
    TRandom *fCurrent;	// current TRandomom 
    TRandom *fROOT;	// current ROOT gRandomom 
            
    Random();
//    Random(const Random &r) { ; } 
//    Random& operator= (const Random &r) { return (*this); }   
    
public:
    virtual ~Random();
    
    //! to access the unique instance
    /*!
    	Random *therand = Random::Instance(); // give a pointer to the unique instance
	
	After that, you can use the Random pointer to get the current random generator 
	and to generate numbers.
		Ex: therand->Current()->Rndm(); // to get uniform random numbers between 0 and 1
    */      
    static Random *Instance(); 
    
    //! Add a new TRandomom object to the list 
    /*!
    	if the name already exists in the collection, this method has no effects. Otherwise it adds the new
	TRandomom object to the list. 
	
	/warning Random owns its TRandomom objects, so, once added, you MUST NOT delete the TRandomom
	object you have created. 
    */ 
    void Add(const char *, TRandom *);
    
    //! to set the current TRandomom object
    /*!
    	If the given name is not known, the current generator is not changed. Default is basic:
	To change the initial conditions of your GEM simulations, just do, for instance:
	
    	Random *therand = Random::Instance(); // give a pointer to the unique instance	\n
	Random::SetCurrent("medium");         // now medium is the current \n
	
	{ // start your gammaware simulation \n
		BaseGEM mynucleus; .... \n
	}	
    */   
   void SetCurrent(const char *name = "basic");
    
    //! to get the current TRandomom object 
    /*!
    	
    */      
    TRandom *Current() const { return fCurrent; }

    //! rootcint dictionary
    ClassDef(Random,0); // Collection of TRandomom objects
};

}
// inline members
#endif
