#ifndef GW_XTERMPROGRESSBAR_H
#define GW_XTERMPROGRESSBAR_H

#ifndef ROOT_Rtypes
#include "Rtypes.h"
#endif

namespace Gw {

/** Class to display a progress bar on the command line

	This class implements a progress bar that is shown on the command line (so that it could be called in a programs
that has not graphical interface). You must provide a starting value and an ending value through the constructor. 
Once it is done call the Start member before the loop . It will display an empty bar: \n 
[&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;] \n
Inside the loop call the operator ++. It increments the counter by one and if needed update the progress bar \n
[--------&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;] \n
IsFinished returns true if the process must be stopped. Here it a example how to use an XtermProgressBar in a loop: \n
\code 
	XtermProgressBar progress(0,MaxEvents);
	while( ) { // loop on events
		progress++;
		...
		...
		...
		...
		if ( progress.IsFinished() ) break;
	}
\endcode
	See FisrtTracking.C for a copmlete example

	\author Olivier Stezowski
*/
class  XtermProgressBar 
{
private:	
	UInt_t fMaxSize;
	UInt_t fCurrentSize;

	Double_t fMin;
	Double_t fMax;
	
	Double_t fCurrent;

public:
	//! Defines a ProgressBar that counts events from min to max	
	XtermProgressBar(Double_t min, Double_t max);
	virtual ~XtermProgressBar();
	
	//! 
	virtual void SetRange(Double_t min, Double_t max)
	{
		fMin = min; fMax = max;
	}
	
	//! Display current status on the standard output
	virtual void UpdateCurrentStatus();
	
	//! Re-initialised the counters
	virtual void Start() ;
	
	//! to get the current event number
	virtual Double_t GetCurrentPosition() const 
	{ 
		return fCurrent; 
	}
	
	//! to get the current event number
	virtual bool IsFinished() const { return (fCurrent > fMax); }	
	
	//! Add one event and change the display if needed
	virtual XtermProgressBar operator++() 
	{ fCurrent++; if ( fMaxSize*(fCurrent-fMin)/(fMax-fMin) > fCurrentSize+1 ) { fCurrentSize++; UpdateCurrentStatus();} return *this; }
	
	//! Add one event and change the display if needed
	virtual XtermProgressBar operator++(int)
	{ fCurrent++; if ( fMaxSize*(fCurrent-fMin)/(fMax-fMin) > fCurrentSize+1 ) { fCurrentSize++; UpdateCurrentStatus();} return *this; }	
	
	//! Add one event and change the display if needed
	virtual XtermProgressBar operator+=( int i )
	{ fCurrent+=i; if ( fMaxSize*(fCurrent-fMin)/(fMax-fMin) > fCurrentSize+1 ) { fCurrentSize++; UpdateCurrentStatus();} return *this; }
};

}

#endif
