#include<BaseCalib.h>
#include<TCanvas.h>
#include<TDatime.h>
#include<TString.h>
#include<TSpectrum.h>
#include<TFile.h>
#include<Peak.h>
#include<TLine.h>
#include<TList.h>
#include<TSystem.h>
#include <TMath.h>


/*#define NBCANAUX 16384
#define RANGE 10    //Range to fit a peak --> Fit between Max-RANGE and Max+RANGE in keV

#define SCALEENERGY 6000   //scale for the calibrated histogram
#define KEVCANAL 0.366   //We want a calibrated histogram with KEVCANAL keV/canal

#define MINCOUNTPEAK 100 //Minimal counts in the maximum of the peak to do the calibration*/

const Int_t	BaseCalib::NBCANAUX     = 16384;  
const Float_t	BaseCalib::SCALEENERGY  =  6000;  
const Float_t	BaseCalib::KEVCANAL     = 0.366;  
const Float_t	BaseCalib::RANGE        =    10;  
const Int_t	BaseCalib::MINCOUNTPEAK	=   100;  


Double_t SkewedGaussian(Double_t *x, Double_t *par){  //SkewedGaussian function (See Radford WebSite for details)
	return(par[0]*TMath::Exp(-0.5*TMath::Power(((x[0]-par[1])/par[2]),2)) +
	       par[3]*TMath::Exp((x[0]-par[1])/par[4])*TMath::Erfc(((x[0]-par[1])/(TMath::Sqrt(2.0)*par[2])) +
	       (par[2]/(TMath::Sqrt(2.0)*par[4]))) + par[5]*TMath::Erfc((x[0]-par[1])/(TMath::Sqrt(2.0)*par[2])));
}

Double_t Gaussian(Double_t *x, Double_t *par){
	return (par[0]*TMath::Exp(-0.5*TMath::Power(((x[0]-par[1])/par[2]),2)));
}


BaseCalib::BaseCalib() {  //Defaut Constructor
	fOut = kFALSE;
	fLog = kFALSE;
	fErr = kFALSE;
	
	fNbSpectra=0;
	fNbGood = 0;
	fNbWrong = 0;

	fCheck = kFALSE;

	fGaussian = new TF1("Gauss",Gaussian,0,SCALEENERGY,3);
	fGaussian->SetParNames("Constant","Mean","Sigma");
	fGaussian->SetLineColor(2);
	fSkewedGaussian = new TF1("Skewed",SkewedGaussian,0,SCALEENERGY,6);
	fSkewedGaussian->SetParNames("ConstantGaus","Mean","Sigma","ConstantSkewed","Beta","ConstantStep");
	fSkewedGaussian->SetParameter(5,0.0); //no constantStep for the moment
	fSkewedGaussian->SetLineColor(3);

	fCollHistoCalib = new TObjArray(1);

	gROOT->GetListOfFunctions()->Remove(fGaussian);
	gROOT->GetListOfFunctions()->Remove(fSkewedGaussian);
}

BaseCalib::~BaseCalib(){  //Destructor
	if (fGaussian != NULL) delete fGaussian;
	if (fSkewedGaussian != NULL) delete fSkewedGaussian;

	//Close the opened files
	if (fOut) {CloseOutFile();}
	if (fErr) {CloseErrFile();}
	if (fLog) {CloseLogFile();}

	if (fCollHistoCalib != NULL) delete fCollHistoCalib;

}


TString BaseCalib::FileNextName(const Char_t *basename, const Char_t *ext){
  Char_t strdate[50]; int i; TDatime date;
  Char_t str[50];
  Char_t filename[50];
  TString string;
  TFile *infile;

  sprintf(strdate,"%d_%d_%d",date.GetDay(),date.GetMonth(),date.GetYear());
  for (i = 1; i <= 999; i++) {

    sprintf(str, "%s_%s_%d.%s", basename, strdate, i, ext);

    // test if it exists
    infile = new TFile(str,"NEW");
    if (infile->IsOpen()==0) {
	delete infile;
	continue;//Le fichier existe deja
    }

    sprintf(filename,"%s",str);
    string.Insert(0,filename);
    infile->Close();
    delete infile;
    break;
  }
  return (string);
}

void BaseCalib::OpenLogFile(){
	TString string(FileNextName("Calib","log"));
	fLogFile.open(string.Data(),ios_base::out);
	if (!fLogFile.good()) {
		cerr << "Cannot open file " <<string.Data()<<endl;
		fLog = kFALSE;
	}
	else {
		fLogFile << "-------------- Fichier "<<string.Data()<<" ---------------" <<endl<<endl;
		fLog = kTRUE;
	}
}



void BaseCalib::CloseLogFile(){
	fLogFile.close();
	fLog = kFALSE;
}

void BaseCalib::OpenErrFile(){
	TString string(FileNextName("Calib","err"));
	fErrFile.open(string.Data(),ios_base::out);
	if (!fErrFile.good()) {
		cerr << "Cannot open file " <<string.Data()<<endl;
		fErr = kFALSE;
	}
	else {
		fErrFile << "-------------- Fichier "<<string.Data()<<" ---------------" <<endl<<endl;
		fErr = kTRUE;
	}
}

void BaseCalib::CloseErrFile(){
	fErrFile.close();
	fErr = kFALSE;
}

void BaseCalib::OpenOutFile(){
	TString string(FileNextName("Calib","out"));
	fOutFile.open(string.Data(),ios_base::out);
	if (!fOutFile.good()) {
		cerr << "Cannot open file " <<string.Data()<<endl;
		fOut = kFALSE;
	}
	else {
		fOutFile << "-------------- Fichier "<<string.Data()<<" ---------------" <<endl<<endl;
		fOut = kTRUE;
	}
}

void BaseCalib::CloseOutFile(){
	fOutFile.close();
	fOut = kFALSE;
}


Bool_t BaseCalib::GetOut() const{
	return(fOut);
}

Bool_t BaseCalib::GetLog() const{
	return(fLog);
}

Bool_t BaseCalib::GetErr() const{
	return(fErr);
}

Bool_t BaseCalib::GetCheck() const{
	return(fCheck);
}

void BaseCalib::SetCheck(const Bool_t b){
	fCheck = b;
}


void BaseCalib::Setup(){
	char rep;
	cout << "-------- SETUP --------"<<endl;
	cout << "Fichier .out ? y/n : ";
	cin >> rep;
	if (rep == 'y') {OpenOutFile();} else {fOut = kFALSE;}
	cout << "Fichier .log ? y/n : ";
	cin >> rep;
	if (rep == 'y') {OpenLogFile();} else {fLog = kFALSE;}
	cout << "Fichier .err ? y/n : ";
	cin >> rep;
	if (rep == 'y') {OpenErrFile();} else {fErr = kFALSE;}
}


TF1* BaseCalib::GetGaussian() {return(fGaussian);}

TF1* BaseCalib::GetSkewedGaussian() {return(fSkewedGaussian);}


void BaseCalib::Fit1Peak(TH1D *histo, const Char_t *option){
	//To Fit a TH1 histo with a Gaussian or a SkewedGaussian (See Radford website for more details)
	//The fit is performed in the histogram range
	//So you have to make histo->SetAxisRange(xmin, xmax) if necessary to select a peak before calling this function
	//In all cases, a Gaussian Fit is realised first
	//Then, if the option = "Skewed" is selected, a SkewedGaussian function is performed to fit
	//If option contains "Bg", you perform a background substraction before fitting
	//A copy of histo is made in the beginning of the method, and all operations are performed on this copy (See TH1::Copy)
	//At the end, the fit function named "Gauss" or "Skewed" is added to the list of functions of the histogram histo

	//Parameters option accepted :
	// 	"Gauss" : defaut option --> Gaussian Fit without background
	//	"" is set to "Gauss"
	//	"GaussBg" --> Gaussian Fit with background
	//	"Skewed" --> Gaussian Fit first, then SkewedGaussian Fit without background substraction
	//	"SkewedBg" --> Gaussian Fit first, then SkewedGaussian Fit with background substraction
	//	"JustBg" --> Do the fit with a gaussian with Background substraction (fitted function is not stored)
	//	"Just" --> Do the fit with a gaussian without Background substraction (fitted function is not stored)


	//RAZ Fit functions
	fGaussian->SetParameters(0.0,0.0,0.0);
	fSkewedGaussian->SetParameters(0.0,0.0,0.0,0.0,0.0);

	if (histo == NULL) {
		if (fErr) fErrFile << "No histogram --> abort the fit ... "<<endl;
		cout <<"No histogram --> abort the fit ... "<<endl;
		return;
	}

	TString string(option);
	TH1D histocopy;
	/*if (histocopy == NULL) {
		if (fErr) fErrFile << "Dynamical allocation failed in Fit1Peak method ..."<<endl;
		return;
	}*/

	//(*histocopy) = histo;
	histo->Copy(histocopy);
	histocopy.SetNameTitle("histocopy","histocopy");

	//Check the validity of the parameter "option"
	if (!string.CompareTo("")) string.Replace(0,0,"Gauss");
	if ((string.CompareTo("Gauss")) && (string.CompareTo("GaussBg"))
		&& (string.CompareTo("Skewed")) && (string.CompareTo("SkewedBg"))
		&& (string.CompareTo("JustBg")) && (string.CompareTo("Just"))){
			cout << "Option '"<<string.Data()<<"' is not define to fit --> Abort ... "<<endl;
			return;
	}


	if (string.Contains("Bg")){
		DoBg(&histocopy);
	}

	//gaussian parameter initialisation for the fit
	// Constant --> Maximum in the histogram range
	// Mean --> bin correspondant to the maximum in the histogram range
	// sigma --> 1
	fGaussian->SetParameters(histocopy.GetMaximum(),histocopy.GetMean(),1);

	//Fit with a gaussian
	if (string.Contains("Gauss")) histocopy.Fit(fGaussian,"QO+");   //store the fit function in the histogram ListOfFunctions
	if (string.Contains("Skewed")) histocopy.Fit(fGaussian,"NQ");	//Don't store the fit function in the histogram ListOfFunctions
	if (string.Contains("Just")) histocopy.Fit(fGaussian);

	//store results of gaussian fit in the Log File
	if (string.Contains("Gauss")) {
		if (fLog) {
			fLogFile << "	******* GAUSSIAN FIT RESULTS *******"<<endl;
			fLogFile << "	* Constant = "<<fGaussian->GetParameter(0)<<" +- "<<fGaussian->GetParError(0)<<endl;
			fLogFile << "	* Mean = "<<fGaussian->GetParameter(1)<<" +- "<<fGaussian->GetParError(1)<<endl;
			fLogFile << "	* Sigma = "<<fGaussian->GetParameter(2)<<" +- "<<fGaussian->GetParError(2)<<endl;
			fLogFile << "	* Chi2/NDF = "<<fGaussian->GetChisquare()<<" / "<<fGaussian->GetNDF()<<" = "<<fGaussian->GetChisquare()/fGaussian->GetNDF()<<endl;
			fLogFile << "	************************************"<<endl;
		}
	}

	if (string.Contains("Skewed")){
	//SkewedGaussian parameter initialisation with the Gaussian function parameters
	//ConstantGauss --> Constant of the fitted Gaussian function
	//Mean --> Mean of the fitted Gaussian function
	//Sigma --> Sigma of the fitted Gaussian function
	//ConstantSkewed --> Constant of the fitted Gaussian function divided by 10 (See Radford Website)
	//Beta --> Set to 5 (See Radford Website)
	//ConstantStep --> Constant of the fitted Gaussian function divided by 100 (See Radford Website)
		fSkewedGaussian->SetParameter(0,fGaussian->GetParameter(0));
		fSkewedGaussian->SetParameter(1,fGaussian->GetParameter(1));
		fSkewedGaussian->SetParameter(2,fGaussian->GetParameter(2));
		fSkewedGaussian->SetParameter(3,fGaussian->GetParameter(0)/10);
		fSkewedGaussian->SetParameter(4,5);
		fSkewedGaussian->SetParameter(5,fGaussian->GetParameter(0)/100);

	//Fit peak with the SkewedGaussian function
		histocopy.Fit(fSkewedGaussian,"Q+");

	//Store results of SkewedGaussian Fit in the Log File
		if (fLog) {
			fLogFile << "	******* SKEWED GAUSSIAN FIT RESULTS *******"<<endl;
			fLogFile << "	* ConstantGaus = "<<fSkewedGaussian->GetParameter(0)<<" +- "<<fSkewedGaussian->GetParError(0)<<endl;
			fLogFile << "	* Mean = "<<fSkewedGaussian->GetParameter(1)<<" +- "<<fSkewedGaussian->GetParError(1)<<endl;
			fLogFile << "	* Sigma = "<<fSkewedGaussian->GetParameter(2)<<" +- "<<fSkewedGaussian->GetParError(2)<<endl;
			fLogFile << "	* ConstantSkewed = "<<fSkewedGaussian->GetParameter(3)<<" +- "<<fSkewedGaussian->GetParError(3)<<endl;
			fLogFile << "	* Beta = "<<fSkewedGaussian->GetParameter(4)<<" +- "<<fSkewedGaussian->GetParError(4)<<endl;
			fLogFile << "	* ConstantStep = "<<fSkewedGaussian->GetParameter(5)<<" +- "<<fSkewedGaussian->GetParError(5)<<endl;
			fLogFile << "	* Chi2/NDF = "<<fSkewedGaussian->GetChisquare()<<" / "<<fSkewedGaussian->GetNDF()<<" = "<<fSkewedGaussian->GetChisquare()/fSkewedGaussian->GetNDF()<<endl;
			fLogFile << "	*******************************************"<<endl;
		}

	}

	//cout << "***** Click on TCanvas to continue ..."<<endl;

	if (gDebug > 0) {
		histocopy.DrawCopy();
		gPad->Update();
		//gPad->WaitPrimitive();
	}


	TF1 *f1 = new TF1();
	if(string.Contains("Gauss")) {

		//BE CAREFUL !!! The Gaussian function is added to the histogram ListOfFunctions
		fGaussian->Copy(*f1);
		if (f1 != NULL) histo->GetListOfFunctions()->Add(f1);
	}

	if (string.Contains("Skewed")) {

		//BE CAREFUL !!! The SkewedGaussian function is added to the histogram ListOfFunctions
		((TF1 *)histocopy.FindObject("Skewed"))->Copy(*f1);
		if (f1 != NULL) histo->GetListOfFunctions()->Add(f1);
	}

	// Delete all local pointers
	//if (histocopy != NULL) delete histocopy;
}


void BaseCalib::DoBg(TH1D *histo){
	//Do the background substraction with the Background1 method of TSpectrum
	//Parameter "number of iterations" is fixed to 20
	//BE CAREFUL !!! the TH1 histo is modified, the final histogram is background substracted
	TSpectrum spectrum;
	Float_t *spectre = new Float_t[histo->GetNbinsX()];
	if (spectre == NULL) {
		if (fErr) fErrFile << "Dynamical allocation failed in DoBg method ..."<<endl;
		return;
	}

	for(Int_t i = 1 ; i < histo->GetNbinsX() ; i++){
		spectre[i] = histo->GetBinContent(i);
	}

   //Option_t* opt = "";
	spectrum.Background(histo,20, "");
	for(Int_t i = 1 ; i < histo->GetNbinsX() ; i++){
		histo->SetBinContent(i,histo->GetBinContent(i)-spectre[i]);
	}


	// Delete all local pointers
	if (spectre != NULL) delete[] spectre;
}


void BaseCalib::FitGraph(TGraphErrors *graph, TF1 *formula){
	//Check formula
	if (formula == NULL) {
		if (fErr) fErrFile <<"pointer formula null in FitGraph method ..."<<endl;
		cout<<"pointer formula null in FitGraph method ..."<<endl;
		return;
	}

	//Check graph
	if (graph == NULL) {
		if (fErr) fErrFile <<"pointer graph null in FitGraph method ..."<<endl;
		cout<<"pointer graph null in FitGraph method ..."<<endl;
		return;
	}

	//Check the number of points in the graph
	if (graph->GetN() == 0){
		if (fErr) fErrFile <<"0 point in the graph - Cannot Fit ..."<<endl;
		cout<<"0 point in the graph - Cannot fit ..."<<endl;
		return;
	}

	//Do a copy of formula
	TF1 *copyformula = new TF1();
	formula->Copy(*copyformula);

	// Now fit the Graph with copyformula in the range of the function (option "R")
	if (gDebug > 0 ) {graph->Fit(copyformula,"R");} else {graph->Fit(copyformula,"ROQ");}
	if (fLog) {
		fLogFile <<"	----------- GRAPH FIT RESULT -----------"<<endl;
		for (Int_t i=0 ; i<copyformula->GetNpar(); i++){
			fLogFile << "		- param : "<<copyformula->GetParName(i)<<" = "<<copyformula->GetParameter(i)<<endl;
		}
		fLogFile <<"		- Chi2/NDF = "<<copyformula->GetChisquare()<<" / "<<copyformula->GetNDF()<<" = "<<copyformula->GetChisquare()/copyformula->GetNDF()<<endl;
		fLogFile <<"	----------------------------------------"<<endl;
	}

	//Now if fCheck = kFALSE, copy copyformula in formula
	if (!fCheck) {
		for (Int_t i=0 ; i<formula->GetNpar() ; i++){
			formula->SetParameter(i,copyformula->GetParameter(i));
		}
	}

	//Delete all local pointers
	if (copyformula != NULL) delete copyformula;
}


void BaseCalib::Resolution(TObjArray *tabpeaks, Float_t threshold) {
	Float_t mean=0.0; //mean resolution

	//Check tabpeaks
	if (tabpeaks == NULL) {
		if (fErr) fErrFile << "pointer null TObjArray in Resolution method"<<endl;
		cout << "pointer null TObjArray in Resolution method"<<endl;
		return;
	}
	
	//Check entries of tabpeaks
	if (tabpeaks->GetEntries() == 0) {
		if (fErr) fErrFile << "TObjArray Entries == 0 in Resolution method"<<endl;
		cout << "TObjArray Entries == 0 in Resolution method"<<endl;
		return;
	}
	
	//Calcul of mean resolution
	for (Int_t i=0 ; i<tabpeaks->GetEntries() ; i++){ //loop on entries of the TObjArray
		mean += ((Peak *)tabpeaks->At(i))->GetRes(); 
		if (fLog){ 
			fLogFile << "Resolution of peak "<<i<<" = "<<((Peak *)tabpeaks->At(i))->GetRes()<<endl;
			fLogFile << endl << "Mean Resolution = "<<mean;
		}	
		
	} //i
	
	//Now Check mean resolution
	if (mean == 0.0) {
		if (fErr) fErrFile << "Mean resolution == 0 in Resolution method"<<endl;
		cout << "Mean Resolution == 0 in Resolution method"<<endl;
		return;
	}
	
	//Compare the resolution of each peak
	Float_t relres = 0.0 ; //relative resolution of a peak 
	for (Int_t i=0 ; i<tabpeaks->GetEntries() ; i++){ //loop on entries of the TObjArray
		relres = ( ((Peak *)tabpeaks->At(i))->GetRes() - mean )	/ mean ;
		if (relres > threshold) {
			if (fLog) fLogFile << "Peak "<<i<< " --> Bad Resolution"<<endl;
		}
	} //i
}


void BaseCalib::AlignMax(TObjArray *collhisto,TObjArray *collformula, Int_t binvalue, Float_t xmin, Float_t xmax){

	//Perform the alignment of the maximum of a TH1D on value
	//The search for maximum if performed in the range [xmin,xmax] of the histogram  
	//If xmin=0, xmax=0 (by default), the search for maximum if performed in the total range of the histogram
	
	//on prend le bin max et on aligne sur value
	//
	
	TH1D *htemp;
	TF1 *ftemp;
	Int_t maxbin, shift;


	//Check collhisto
	if (collhisto == NULL) {
		if (fErr) fErrFile <<"pointer collhisto null in AlignMax method ..."<<endl;
		cout<<"pointer collhisto null in AlignMax method ..."<<endl;
		return;
	}
	//Check collhisto2
	if (collformula == NULL) {
		if (fErr) fErrFile <<"pointer collformula null in  AlignMax method ..."<<endl;
		cout<<"pointer collformula null in AlignMax method ..."<<endl;
		return;
	}
	
	//Check size of the 2 collections
	if (collhisto->GetSize() > collformula->GetSize()) {
		if (fErr) fErrFile <<"Size of collhisto > size of collformula in AlignMax method ..."<<endl;
		cout<<"Size of collhisto > size of collformula in AlignMax method ..."<<endl;
		return;
	}
	
	
	for (Int_t i=0 ; i<collhisto->GetEntries() ; i++){ //loop on the TH1D
		if ((collhisto->At(i) == NULL) || (!collhisto->At(i)->InheritsFrom("TH1D"))){
			cout << "Warning --> slot "<<i<<" is not a TH1D --> Pass this slot ..."<<endl;
			continue;
		}
		
		if (collformula->At(i) == NULL){
			cout << "Warning --> slot "<<i<<" is null for collformula --> Pass this slot ..."<<endl;
			//TF1 *f = new TF1("","");
			continue;
		}
		
		htemp = (TH1D *)collhisto->At(i);
		htemp->GetXaxis()->SetRangeUser(xmin,xmax);
		ftemp = (TF1 *)collformula->At(i);
		maxbin = htemp->GetMaximumBin();
		shift = binvalue - maxbin; 
		
		ftemp->SetParameters(shift,1);   //just a shift --> no dilatation
		
	}
	
	collformula->Write("FormulaT",TObject::kSingleKey);

}

void BaseCalib::Display(TObjArray *collhisto, Int_t nx, Int_t ny, Float_t xmin, Float_t xmax){
	Int_t n = 0;

	//Check collhisto
	if (collhisto == NULL) {
		if (fErr) fErrFile <<"pointer collhisto null in Display method ..."<<endl;
		cout<<"pointer collhisto null in Display method ..."<<endl;
		return;
	}
	
	
	//Create a new Canvas 
	TCanvas *c1 = new TCanvas("Display","Display");
	c1->Divide(nx,ny,0.0,0.0);
	//c1->cd(1);  // go to the first pad
	
	for (Int_t i=0 ; i<collhisto->GetEntries() ; i++){
		if (collhisto->GetEntries() <= nx*ny) {
			c1->cd(i+1);
			((TH1D *)collhisto->At(i))->GetXaxis()->SetRangeUser(xmin,xmax);
			((TH1D *)collhisto->At(i))->Draw();
		} else {
			n = i%(nx*ny);
			if ((i != 0) && (n==0)){
				printf("Click on the pad to continue ...\n");
				gPad->Update();
				gPad->WaitPrimitive();
				for (Int_t j=0 ; j<nx*ny ; j++){
					c1->cd(j+1);
					gPad->Clear();
				}
				
			}
			c1->cd(n+1);
			((TH1D *)collhisto->At(i))->GetXaxis()->SetRangeUser(xmin,xmax);
			((TH1D *)collhisto->At(i))->Draw();
		}
		
	} 

}


void BaseCalib::DisplayAlign(TObjArray *collhisto, Float_t value, Float_t xmin, Float_t xmax){
	//Display a collection of histograms to check alignment on a value

	TSystem sys;
	
	
	//Check collhisto
	if (collhisto == NULL) {
		if (fErr) fErrFile <<"pointer collhisto null in Display method ..."<<endl;
		cout<<"pointer collhisto null in Display method ..."<<endl;
		return;
	}
	
	//Create a new TLine
	TLine *line = new TLine();
	line->SetLineColor(2); line->SetLineStyle(2);
	
	
	//Create a new Canvas 
	TCanvas *c1 = new TCanvas("DisplayAlign","DisplayAlign");
	
	for (Int_t i=0 ; i<collhisto->GetEntries() ; i++){
		
		//Check xmin, xmax
		((TH1D *)collhisto->At(i))->GetXaxis()->SetRangeUser(xmin,xmax);
		((TH1D *)collhisto->At(i))->Draw();
		line->DrawLine(value,0.0,value,((TH1D *)collhisto->At(i))->GetMaximum());
		printf("Double Click on the pad to continue ...\n");
		gPad->Update();
		gPad->WaitPrimitive();	
		gSystem->Sleep(300);
	} 
	
	if (line!=NULL) delete line;
	if (c1!=NULL) delete c1;
}

void BaseCalib::CheckShift(TObjArray *collhisto1, TObjArray *collhisto2, Float_t xmin, Float_t xmax) {
	//Check collhisto1
	if (collhisto1 == NULL) {
		if (fErr) fErrFile <<"pointer collhisto1 null in CheckShift method ..."<<endl;
		cout<<"pointer collhisto1 null in CheckShift method ..."<<endl;
		return;
	}
	//Check collhisto2
	if (collhisto2 == NULL) {
		if (fErr) fErrFile <<"pointer collhisto2 null in CheckShift method ..."<<endl;
		cout<<"pointer collhisto2 null in CheckShift method ..."<<endl;
		return;
	}
	//Check number of entries
	if (collhisto1->GetEntries() < collhisto2->GetEntries()) {
		if (fErr) fErrFile <<"collhisto1 entries > collhisto2 entries in CheckShift method ..."<<endl;
		cout<<"collhisto1 entries > collhisto2 entries in CheckShift method ..."<<endl;
		return;
	}
	
	TH1D *histo1=NULL;
	TH1D *histo2=NULL;
	for (Int_t i=0 ; i<collhisto1->GetEntries() ; i++){
		TCanvas c1("CheckShift","CheckShift");
		c1.Clear();
		histo1 = (TH1D *)collhisto1->At(i);
		histo2 = (TH1D *)collhisto2->At(i);
		
		histo1->GetXaxis()->SetRangeUser(xmin,xmax);
		histo2->GetXaxis()->SetRange();
		
		//histo1->SetLineColor(3);
		histo1->Draw();
		histo2->SetLineColor(2);
		histo2->Draw("SAME");
		
		
		c1.Update();
		c1.WaitPrimitive();
		
	}
	
	histo1 = NULL;
	histo2 = NULL;
	
}



