#include<iostream>
using namespace std;

#include<Co60.h>
#include<TH1.h>
#include<TSpectrum.h>
#include<TGraphErrors.h>
#include<TCollection.h>
#include<TFile.h>
#include<TCanvas.h>
#include<TLine.h>
#include<TSystem.h>
#include<TMath.h>

#define MINCOUNTPEAK 100

static Float_t fEnergyCo60[2] = {1173.237,1332.516};  //The 2 tabulated energies of the Co60 source
//static Float_t fIntensityCo60[NBCo60]={100,100}; // The 2 tabulated intensity of the Co60 source


Co60::Co60() : BaseCalib(){

	fHistoCalib = new TH1D("HistoCalib","Calibrated Histogram",NBCANAUX,0,SCALEENERGY);


}

Co60::~Co60(){
	if (fHistoCalib != NULL) delete fHistoCalib;
}


TH1D* Co60::GetHistoCalib() { return(fHistoCalib); }

const Char_t* Co60::DoCalib(TH1D* histo, TF1 *formula, TGraphErrors* graph, const Char_t *option, Int_t binmin, Int_t binmax){

	//Perform the calibration of the TH1 histo for a Co60 source
	//A copy of the formule is created and we work on this copy
	//   if fCheck = TRUE --> the formule is not changed
	//   if fCheck = FALSE --> formule = copy
	//Check the number of parameters in the formula (if nbpar > 2 --> par[i>1]=0.0)
	//First, search the more intense peaks in the histogram with the method Search of TSpectrum (Search(histo,sigma,option,threshold))
	//   sigma : 1 --> sigma of searched peaks, for details we refer to manual
	//   threshold: 0.1 (default=0.05)  peaks with amplitude less than threshold*highest_peak are discarded.  0<threshold<1
	//   option = ""
	//   if option is not equal to "goff" (goff is the default), then
	//   a polymarker object is created and added to the list of functions of
	//   the histogram. The histogram is drawn with the specified option and
	//   the polymarker object drawn on top of the histogram.
	//   The polymarker coordinates correspond to the npeaks peaks found in
	//   the histogram.
	//Select the 2 more intense peaks
	//Fit the 2 peaks with the Fit1Peak Method of BaseCalib (with "option" see below)
	//Add points in the resolution graph in function of energy with results of fits
	//Add the Calibration graph to the List of functions of the histogram "histo"
	//     x --> peak position (results of fits)
	//     y --> tabulated energies
	//     the calibration graph has got 2 points
	//Fit the calibration graph with the TF1 "formula" with the FitCalib method
	//The Calibrated histogram is filled (private member fHistoCalib)

	//Parameters option accepted :
	//"Gauss" : defaut option --> Gaussian Fit without background
	//"" is set to "Gauss"
	//"GaussBg" --> Gaussian Fit with background
	//"Skewed" --> Gaussian Fit first, then SkewedGaussian Fit without background substraction
	//"SkewedBg" --> Gaussian Fit first, then SkewedGaussian Fit with background substraction
	
	//if binmin=0 and binmax=0 (by default), the Search method of TSpectrum is performed in the 
	//      total range of the histo. 
	//else, Search is performed between binmin and binmax

	TString string(option);

	TSpectrum spectrum;
	Float_t position[2][2]; //Position and intensity of the 2 peaks


	//Chek the validity of the parameter "option"
	if (!string.CompareTo("")) string.Replace(0,0,"Gauss");
	if ((string.CompareTo("Gauss")) && (string.CompareTo("GaussBg"))
		&& (string.CompareTo("Skewed")) && (string.CompareTo("SkewedBg"))){
			printf("Option '%s' is not define to fit\n",string.Data());
			fNbWrong++;
			return ("Abort ...");
	}



	//Check the histogram
	if (histo == NULL) {
		if (fErr) {fErrFile <<"no histogram to calib ..."<<endl;}
		fNbWrong++;
		return ("no histogram to calib ...");
	}

	//Check the formula
	if (formula == NULL) {
		if (fErr) {fErrFile <<"no formula associated to calib ..."<<endl;}
		fNbWrong++;
		return ("no formula associated to calib ...");
	}

	//Check the graph
	if (graph == NULL) {
		if (fErr) {fErrFile <<"no graph associated to calib - pointer null ..."<<endl;}
		fNbWrong++;
		return ("no graph associated to calib - pointer null ...");
	}

	
	
	//Check the number of parameters, if NbPar > 2, param[i>2]=0.0
	//if (formula->GetNpar() > 2){
	//	for (Int_t i=2 ; i<formula->GetNpar() ; i++) formula->SetParameter(i,0.0);
	//}

	// Construction of the Canvas Principal
		
		TCanvas *cPrincipal;
		if ( ( cPrincipal = (TCanvas *)gROOT->GetListOfFunctions()->FindObject("cPrincipal") ) ) { delete cPrincipal; }
		if (gDebug >0) {
			cPrincipal = new TCanvas("cPrincipal","Co60 Calibration",50,0,1000,600);
			cPrincipal->Divide(1,2,0.0,0.0);
			cPrincipal->cd(2);
			gPad->Divide(2,1,0.0,0.0);
			cPrincipal->cd(1);
		}

	//RAZ List of functions of histogram
	((TList *)histo->GetListOfFunctions())->Clear();

	if (fLog){fLogFile<<"Analyse de l'histogramme : "<<histo->GetName()<<endl<<endl;}
	if (fOut) {fOutFile<<"Analyse de l'histogramme : "<<histo->GetName()<<endl;}

	if (fLog){
		fLogFile << "	Formula Parameters before calibration : "<<endl;
		for (Int_t i=0 ; i<formula->GetNpar() ; i++){
			fLogFile<<"		Parameter "<<formula->GetParName(i)<<" = "<<formula->GetParameter(i)<<endl;
		}
	}

	printf("Analysing histogram  %s ...\n",histo->GetName());
	printf("Searching Peaks ...\n"); 


	//Search peaks
	if ((binmin !=0) && (binmax != 0)) {
		//creation of a new TH1D only between binmin and binmax
		TH1D *histosearch = new TH1D("histosearch","histosearch",histo->GetNbinsX(),histo->GetXaxis()->GetBinLowEdge(1),histo->GetXaxis()->GetBinUpEdge(histo->GetNbinsX()));
		for (Int_t bin=0 ; bin<histo->GetNbinsX() ; bin++){
			if ((bin <= binmax) && (bin >= binmin)) histosearch->SetBinContent(bin,histo->GetBinContent(bin));
		}	 
		if (gDebug > 0) {
			spectrum.Search(histosearch,1,"",0.1);
			gPad->Update();
		} else {
			spectrum.Search(histosearch,1,"goff",0.1);
			//gPad->Update();
		}
	} else {
	
		histo->GetXaxis()->SetRange(0,0);
		if (gDebug > 0) {
			spectrum.Search(histo,1,"",0.1);
			gPad->Update();
		} else {
			spectrum.Search(histo,1,"goff",0.1);
			//gPad->Update();
		}
	}	


	if (spectrum.GetNPeaks() <= 1) {
		if (fErr) {fErrFile <<"histo : "<<histo->GetName()<<" -> number of peaks < 2  ... abort calibration for this spectrum"<<endl;}
		printf("histo : %s --> number of peaks < 2   ... abort calibration for this spectrum\n",histo->GetName());
		fNbWrong++;
		return ("");
	}
	if (spectrum.GetNPeaks() > 2) {
		printf("*** WARNING ***\n");
		printf("*  histo : %s --> number of peaks found : %d\n",histo->GetName(),spectrum.GetNPeaks());
		printf("***************\n");
	}

	if (fLog){fLogFile <<"	Number of Peak found : "<<spectrum.GetNPeaks()<<endl;}

	for (Int_t i=0 ; i < spectrum.GetNPeaks() ; i++){
		if (fLog){fLogFile<<"		peak "<<i<<" : positionX : "<<spectrum.GetPositionX()[i]<<
				    " : positionY : "<< spectrum.GetPositionY()[i]<<endl;}
	}


	//Select the Two peaks (the 2 more intense)...
	printf("Selecting the Two Peaks ...\n");
	Int_t *index = new Int_t[100];
	//TMath::Sort(spectrum.GetNPeaks(),spectrum.GetPositionY(),index);
	if (spectrum.GetPositionX()[index[0]] > spectrum.GetPositionX()[index[1]]){
		position[0][0] = spectrum.GetPositionX()[index[1]];
		position[0][1] = spectrum.GetPositionY()[index[1]];
		position[1][0] = spectrum.GetPositionX()[index[0]];
		position[1][1] = spectrum.GetPositionY()[index[0]];
	} else {
		position[0][0] = spectrum.GetPositionX()[index[0]];
		position[0][1] = spectrum.GetPositionY()[index[0]];
		position[1][0] = spectrum.GetPositionX()[index[1]];
		position[1][1] = spectrum.GetPositionY()[index[1]];
	}

	if (fLog){
		fLogFile<<"	Peak to fit : "<<endl;
		for (Int_t i =0 ; i<2 ; i++){
			fLogFile<<"		peak "<<i<<" : positionX : "<<position[i][0]<<
			    	  			  " : positionY : "<<position[i][1]<<endl;
		}
	}
	
	//Check number of counts in the peaks
	if ((position[0][1] <= MINCOUNTPEAK) || (position[1][1] <= MINCOUNTPEAK)){
		if (fErr) fErrFile << "Counts in peaks < "<<MINCOUNTPEAK<<" -> Abort calibration "<<endl;
		cout <<"Counts in peaks < "<<MINCOUNTPEAK<<endl;
		fNbWrong++;
		return("Abort calibration for this spectrum");					
	}

	printf("Fitting the Two Peaks ...\n");
	for (Int_t i=0 ; i < 2 ; i++){  //loop on the peaks
		if (gDebug >0) {cPrincipal->cd(2); gPad->cd(i+1);}
		histo->SetAxisRange(position[i][0]-RANGE,position[i][0]+RANGE);
		Fit1Peak(histo,string.Data());
		if (string.Contains("Gauss")){
			position[i][0] = fGaussian->GetParameter(1);  //Mean
		}
		if (string.Contains("Skewed")){
			position[i][0] = fSkewedGaussian->GetParameter(1);  //Mean
		}
	} //i

	graph->Set(graph->GetN()+2); //on rajoute deux points de coordonnees (0,0), les valeurs precedentes sont gardees
	graph->SetPoint(graph->GetN()-2,position[0][0],fEnergyCo60[0]);
	//graph->SetPointError(graph->GetN()-2,0.1,0.0);
	graph->SetPoint(graph->GetN()-1,position[1][0],fEnergyCo60[1]);
	//graph->SetPointError(graph->GetN()-1,0.1,0.0);

	//Now calculate coefficients
	Float_t a=0.0,b=0.0;
	if((position[1][0]-position[0][0]) != 0) {a = (fEnergyCo60[1]-fEnergyCo60[0])/(position[1][0]-position[0][0]);}
	b = fEnergyCo60[0]-a*position[0][0];

	formula->SetParameters(b,a);


	//Now do the new calibrated histo
	fHistoCalib->Reset("ICE");
	Char_t name[50];
	sprintf(name,"%s_%s",histo->GetName(),"Calib");
	fHistoCalib->SetNameTitle(name,name);
	Float_t e1;
	Int_t canal;
	for (Int_t i=1 ; i < histo->GetNbinsX() ; i++){
		e1 = histo->GetBinLowEdge(i) + 0.5*KEVCANAL;  //mean energy of the bin i
		e1 = formula->Eval(e1);  //energy after calibration
		canal = fHistoCalib->FindBin(e1);
		fHistoCalib->SetBinContent(canal,histo->GetBinContent(i));
	}

	//Draw The calibrated histogram in a new Canvas
	if (gDebug > 0){
		TCanvas *c1;
		if ( ( c1 = (TCanvas *)gROOT->GetListOfFunctions()->FindObject("CalibratedHistogram") ) ) { delete c1; }
		c1 = new TCanvas("CalibratedHistogram","Calibrated Histogram");
		c1->Divide(2,1);
		c1->cd(1);
		fHistoCalib->SetAxisRange(fEnergyCo60[0]-2*RANGE,fEnergyCo60[0]+2*RANGE);
		fHistoCalib->DrawCopy();
		TLine *line1 = new TLine(fEnergyCo60[0],0.0,fEnergyCo60[0],fHistoCalib->GetMaximum());
		line1->SetLineStyle(2); line1->SetLineColor(2);
		line1->Draw();
		gPad->Update();
		c1->cd(2);
		fHistoCalib->SetAxisRange(fEnergyCo60[1]-2*RANGE,fEnergyCo60[1]+2*RANGE);
		if (gDebug > 0) fHistoCalib->DrawCopy();
		TLine *line2 = new TLine(fEnergyCo60[1],0.0,fEnergyCo60[1],fHistoCalib->GetMaximum());
		line2->SetLineStyle(2); line2->SetLineColor(2);
		line2->Draw();
		gPad->Update();
	}


	//printf("Double-Click on Canvas to continue ...\n");
	//gPad->WaitPrimitive();
	if (gDebug > 0){
		//gPad->Modified();
		gPad->Update();
		gPad->WaitPrimitive();
		//gSystem->Sleep(3000);
	}
	

	//Delete all local pointers
	if (index != NULL) delete [] index;

	if (fLog){
		fLogFile << "	Formula Parameters after calibration : "<<endl;
		for (Int_t i=0 ; i<formula->GetNpar() ; i++){
			fLogFile<<"		Parameter "<<formula->GetParName(i)<<" = "<<formula->GetParameter(i)<<endl;
		}
		fLogFile<<endl;
	}

	fNbGood++;
	return ("Successfull Co60 Calibration for this histogram ...");

}


void Co60::DoCalib(TObjArray *collhisto,TObjArray *collformula, TObjArray *collgraph, const Char_t *option, Int_t binmin, Int_t binmax){

	//Perform the calibration Co60 source for a TObjArray of histogram with a TObjArray of formula
	//The size of collhisto must be <= size of collformula
	//Loop on all the histograms --> DoCalib(histo,formula,option) (see below for the different options)
	//	The calibrated histograms are stored in a TObjArray called fCollHistoCalib and save in the root file
	//	Save the histo with its list of functions in the root file "CalibCo60.root"
	//	The Calibrated histogram is added to the collection of calibrated histograms
	//	   The name of calibhisto is "(name of histo)_Calib"
	//End of the loop
	//Save the collection of histograms and the collection of formula in the current opened root file
	//Display the resolution graph
	//Close the files (err, out, log)

	//Parameters option accepted :
	//"Gauss" : defaut option --> Gaussian Fit without background
	//"" is set to "Gauss"
	//"GaussBg" --> Gaussian Fit with background
	//"Skewed" --> Gaussian Fit first, then SkewedGaussian Fit without background substraction
	//"SkewedBg" --> Gaussian Fit first, then SkewedGaussian Fit with background substraction
	
	

	TString string(option);

	//Test collhisto
	if (collhisto == NULL){
		if (fErr) fErrFile<<"Error collisto == NULL in DoCalib method"<<endl;
		printf("Error, collection of histo == NULL in DoCalib method ...\n");
		return;
	}

	//Test collformula
	if (collformula == NULL){
		if (fErr) fErrFile<<"Error collformula == NULL in DoCalib method"<<endl;
		printf("Error, collection of formula == NULL in DoCalib method ...\n");
		return;
	}

	//Test collgraph
	if (collgraph == NULL){
		if (fErr) fErrFile<<"Error collgraph == NULL in DoCalib method"<<endl;
		printf("Error, collgraph == NULL in DoCalib method ...\n");
		return;
	}

	//Test option
	if (!string.CompareTo("")) string.Replace(0,0,"Gauss");
	if ((string.CompareTo("Gauss")) && (string.CompareTo("GaussBg"))
		&& (string.CompareTo("Skewed")) && (string.CompareTo("SkewedBg"))){
			printf("Option '%s' is not define to fit\n",string.Data());
			return;
	}

	//Test size of the Three collections
	if ((collhisto->GetEntries() > collformula->GetEntries()) && (collhisto->GetEntries() > collgraph->GetEntries())) {
		if (fErr) {fErrFile <<"No enaugh formula or graph to calibrate ... abort Calibration"<<endl; return;}
	}
	
	fNbSpectra = collhisto->GetEntries();
	printf("Number of spectra = %d\n",fNbSpectra);

	for (Int_t i =0 ; i<collhisto->GetEntries() ; i++){ //loop on all the collection of histograms

		TH1D *histocalibcopy = new TH1D();
		printf("%s\n",DoCalib((TH1D *)collhisto->At(i),(TF1 *)collformula->At(i), (TGraphErrors *)collgraph->At(i), string.Data(), binmin, binmax));  //Perform calibration of one histogram
		//gPad->Update();

		//The Calibrated histogram is added to the collection of calibrated histograms

		fHistoCalib->Copy(*histocalibcopy);
		
		if (histocalibcopy != NULL) fCollHistoCalib->Add(histocalibcopy);

	}

	//Save the collection of histograms and the collection of formula in the root file
	collhisto->Write("FitAndHisto",TObject::kSingleKey);
	collformula->Write("FormulaCalibrated",TObject::kSingleKey);
	fCollHistoCalib->Write("HistoCalibrated",TObject::kSingleKey);
	collgraph->Write("GraphCalibrated",TObject::kSingleKey);


	//DisplayGraphResolution();
	
	cout << "Number of spectra to calibrate = "<<fNbSpectra<<endl;;
	cout << "Number of good calibrations    = "<<fNbGood<<endl;
	cout << "Number of wrong calibrations   = "<<fNbWrong<<endl;

	cout << "------- End of Calibration Co60 ---------"<<endl;

}



/*void Co60::DroiteCalib(TH1D *histo, TF1 *formula){


	if ((histo->FindObject("Gauss") == NULL) && (histo->FindObject("Skewed") == NULL)){
		if (fErr) fErrFile << "histogram "<<histo->GetName()<<" : no Gauss or Skewed functions found in DroiteCalib method"<<endl;
		return;
	}

	if (histo->FindObject("Gauss") != NULL) {

		Double_t posX[2];
		Int_t index = 0;
		TString string;

		for (Int_t i=0 ; i < histo->GetListOfFunctions()->GetSize() ; i++){
			string = histo->GetListOfFunctions()->At(i)->GetName();
			if (!string.CompareTo("Gauss")){
				posX[index] = ((TF1 *)histo->GetListOfFunctions()->At(i))->GetParameter("Mean");
				index++;
				if (index > 1) break;
			}
		}

		//Calculation of the calibration coefficients
		formula->SetParameter(1,(fEnergyCo60[1]-fEnergyCo60[0])/(posX[1]-posX[0]));
		formula->SetParameter(0,fEnergyCo60[0]-posX[0]*formula->GetParameter(1));
	}

	if (histo->FindObject("Skewed") != NULL) {

		Double_t posX[2];
		Int_t index = 0;
		TString string;

		for (Int_t i=0 ; i < histo->GetListOfFunctions()->GetSize() ; i++){
			string = histo->GetListOfFunctions()->At(i)->GetName();
			if (!string.CompareTo("Skewed")){
				posX[index] = ((TF1 *)histo->GetListOfFunctions()->At(i))->GetParameter("Mean");
				index++;
				if (index > 1) break;
			}
		}

		//Calculation of the calibration coefficients
		formula->SetParameter(1,(fEnergyCo60[1]-fEnergyCo60[0])/(posX[1]-posX[0]));
		formula->SetParameter(0,fEnergyCo60[0]-posX[0]*formula->GetParameter(1));
	}

	if (fLog){
		fLogFile << "	---------- Calibration coefficients -----------"<<endl;
		for (Int_t i=0 ; i < formula->GetNpar() ; i++){
			fLogFile << "	- parametre "<<i<<" : "<<formula->GetParameter(i)<<endl;
			fOutFile << "	"<<formula->GetParameter(i)<<endl;
		}
		fLogFile << "	-----------------------------------------------"<<endl;
	}


}*/

/*void Co60::CheckResolution(TObjArray *collhisto, TObjArray *collformula, Float_t threshold){
	//Check the resolution of a collection of histogram --> Co60 source
	//For each histogram, we search for Gaussian or SkewedGaussian function in the ListOfFunction
	//For the first peak of Co60, the resolution is set with 2*sigma
	//The default threshold = 0.1
	//The mean value is calculated for the collection
	//Loop on the collection,
	//	if (resolution-mean)/mean is grater than threshold, then the formula is set to 0 for this spectrum -->Bad resolution

	//Check the pointer collhisto
	if (collhisto == NULL) {
		if (fErr) fErrFile << "Error, NULL pointer in CheckResolution Method, No Histogram Collection ..."<<endl;
		cout <<"Error, NULL pointer in CheckResolution Method, No Histogram Collection ..."<<endl;
		return;
	}

	//Check the pointer collformula
	if (collformula == NULL) {
		if (fErr) fErrFile << "Error, NULL pointer in CheckResolution Method, No Formula Collection ..."<<endl;
		cout <<"Error, NULL pointer in CheckResolution Method, No Formula Collection ..."<<endl;
		return;
	}

	//Check the size of the histogram collection
	if (collhisto->GetEntries() == 0) {
		if (fErr) fErrFile << "Error, size of the histogram collection = 0 ..."<<endl;
		cout << "Error, size of the histogram collection = 0 ..."<<endl;
		return;
	}

	//Check the size of the two collections
	if (collhisto->GetEntries() > collformula->GetEntries()) {
		if (fErr) fErrFile << "Error, size of the histogram collection > size of the formula collection ..."<<endl;
		cout <<"Error, size of the histogram collection > size of the formula collection ..."<<endl;
		return;
	}
	

	//Check the threshold
	threshold = TMath::Abs(threshold);

	//Loop on the collection to determine the mean resolution
	Float_t mean = 0;
	TF1 locfunction;
	for (Int_t i=0 ; i<collhisto->GetEntries(); i++){
		if (((TH1D *)collhisto->At(i))->FindObject("Gauss") != NULL) {
			((TF1 *)(((TH1D *)collhisto->At(i))->FindObject("Gauss")))->Copy(locfunction);
			mean += 2 * locfunction.GetParameter("Sigma");
		}
		if (((TH1 *)collhisto->At(i))->FindObject("Skewed") != NULL) {
			((TF1 *)(((TH1D *)collhisto->At(i))->FindObject("Skewed")))->Copy(locfunction);
			mean += 2 * locfunction.GetParameter("Sigma");
		}
	} //loop on the collection
	mean = mean/(collhisto->GetEntries());

	if (mean == 0) {
		if (fErr) fErrFile << "Error : Mean resolution = 0 ..."<<endl;
		cout << "Error : Mean resolution = 0 ..."<<endl;
		return;
	}

	if (fLog) {
		fLogFile << "------ Check Resolution -----"<<endl;
		fLogFile << "Number of histograms : "<<collhisto->GetSize()<<endl;
		fLogFile << "Mean Resolution : "<<mean<<endl;
		fLogFile << "Threshold : "<<threshold<<endl<<endl;
	}

	//Check the resolution of each histogram
	Float_t value;
	for (Int_t i=0 ; i<collhisto->GetEntries(); i++){
		if (fLog) fLogFile << "Histogram : "<<((TH1D *)collhisto->At(i))->GetName()<<endl;
		if ((((TH1D *)collhisto->At(i))->FindObject("Gauss") == NULL) &&
		    (((TH1D *)collhisto->At(i))->FindObject("Skewed") == NULL)) {
			if (fErr) fErrFile << "No function 'Gauss' or 'Skewed' found ..."<<endl;
			if (fLog) fLogFile << "    No function 'Gauss' or 'Skewed' found"<<endl;
			cout << "No function 'Gauss' or 'Skewed' found ..."<<endl;
			continue; //pass this histogram

		}
		if (((TH1D *)collhisto->At(i))->FindObject("Gauss") != NULL){
			((TF1 *)((TH1D *)collhisto->At(i))->FindObject("Gauss"))->Copy(locfunction);
		}
		if (((TH1 *)collhisto->At(i))->FindObject("Skewed") != NULL){
			((TF1 *)((TH1D *)collhisto->At(i))->FindObject("Skewed"))->Copy(locfunction);
		}
		value = (2*locfunction.GetParameter("Sigma") - mean)/mean;
		//if (value > threshold) {
		//	((TF1 *)collformula->At(i))->SetParameters(0.0,0.0);
		//}
		if (fLog) {
			fLogFile << "    Resolution : "<<2*locfunction.GetParameter("Sigma")<<endl;
			fLogFile << "    (Resolution - Mean)/Mean : "<<value<<endl;
			if (value <= threshold) fLogFile <<"     *** OK - GOOD RESOLUTION ***"<<endl;
			if (value > threshold)  fLogFile <<"     *** WARNING ----> BAD RESOLUTION ***"<<endl;
		}


	} //end loop histogram collection

}*/
