#ifndef GW_RAWTOTREE_H
#define GW_RAWTOTREE_H

// root includes
#include <Rtypes.h>
#include <TObjString.h>

//! Class for Meta information on Tree
/*!
	What is stored in this object:
	- current run number on which is based the tree
	- the tree name, title and the root file in which is written the tree
	- date of creation of the run
	- Number of events read to build this tree
	- Total Number of events
	- Number of Exogam, Vamos and Diamant events 
*/
class MetaTree : public TNamed
{
public:
	Int_t	Run;		// common run number for all events in the tree	
	TObjString  Date;	// creation date of the run
		
	TObjString  TreeName;	// Current TreeName
	TObjString  TreeTitle;	// Current TreeTitle	
	TObjString  TreeFile;	// current TreeFile	

	UInt_t  EventsIn;	// total number of events read from tapes
	UInt_t  EventsOut;	// total number of events written in the tree	
	
	UInt_t  ExoEvents;	// total number of Exogam events read from tapes
	UInt_t  VamosEvents;	// total number of Exogam events read from tapes
	UInt_t  DiamantEvents;	// total number of Exogam events read from tapes
			
public:
	MetaTree() ;
	virtual ~MetaTree() {;}
	
	virtual void Clear(Option_t *opt = "");		

	ClassDef(MetaTree,1) // Some global informations about the created trees
};

//! RawToTree to read raw data from GANIL and store them in ROOT's TTree
/*!
	To run this program in a root session:
	-# run root: root LoadLib.C
	(be sure libGWPHYSICS.so, libGWGEM.so, libGWSYSTEM.so, libGWGANIL.so and libDAQGANIL.so are loaded)	
	-# start the application: .x Scan.C
	-# init the TTree attributes on the command line:     scan->AttributesTree()
	-# set the input and output file on the command line: scan->Init("/dev/st0 toto.root") or scan->Init("DAQ toto.root")
	-# now you interact with the program using the bar menu.

	How to used the bar menu.
	- Run, to run the application. If the application has not been initialized, it cannot start. To initialized properly,
	Scan->Init is mandatory (scan->AttributesTree() provides some default, so it's not mandatory)
	- Stop to stop the looping on events and Run again to restart it.
	- Status or Help could be used during the looping on events.
	- For Update and End, the application must be stopped. \n
		Update will close the current file toto.root and open a new one name toto_1.root, you can resume the looping after.
		End will close the output and input file, you are then ready to run a new application
*/
class RawToTree 
{   
private:
   class Implementation;
   Implementation *fImplementation;
          
public:
    RawToTree();
    virtual ~RawToTree();
    
    //! to modify the TTree attributes
    /*!
	the first argument is the name of the tree, the second is the title. 
		- if the title contains Exo, a branch corresponding to Exogam will be created 
		- if the title contains Diam, a branch corresponding to Diamant will be created 
		- if the title contains Vamos, a branch corresponding to Vamos will be created 	
		
	upd means the tree will be saved inside a file any upd events. \n
	maxtreesize is the maximum size of the root file. if during the looping the file size is reached, root will
	automatically close it and will open the next file one. XXX_N.root --> XXX_N+1.root
   */
    void AttributesTree(const Char_t *nametree = "RawData", const Char_t *titletree = "Exogam_Vamos_Diamant",
    			Int_t upd = 100000000, Long64_t maxtreesize = 2000000000);

    //! to init the input and the ouput files
    /*!
    	EX: \n
	scan->Init("/dev/st0 toto.root") will read a DLT tape or a regular file \n
	scan->Init("DAQ toto.root") will spy the GANIL acquisition	\n   
   */    
    void Init(Option_t *opt = "");
    void Run(Option_t *opt = "");
    
    //! tclose the current root output file an open a new one
    /*!
    	EX: \n
	scan->Init("/dev/st0 toto.root") will read a DLT tape or a regular file \n
	scan->Init("DAQ toto.root") will spy the GANIL acquisition	\n   
    
    */
    void Update(Option_t *opt = "");

    void Help(Option_t *opt = "");
    
    //! to display some informations concerning the looping on events.    
    void Status(Option_t *opt = "");

    //! to suspend the looping, Run to resume it.
    void Suspend(Option_t *opt = "");
    void End(Option_t *opt = "");
    
    //! 
    /*
    	When the looping stops because there are no more input data, a timer is started autmatically
	any 5 seconds. Use this function to stop this timer if it is running
    */
    void StopTimer();

    ClassDef(RawToTree,0); // To read raw buffers and fill a TTree
};
// RawToTree inline members ///////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////



#endif
