/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file BaseGEM.cpp compiled in libGWGEM.so */
 
#ifndef GW_BASEGEM_H
#include "BaseGEM.h"
#endif

using namespace Gw;

ClassImp(BaseGEM);

BaseGEM::BaseGEM()
 : TObject(), fLevelScheme(), fFeeding(), fRandFeeding(), fRandDown(), fRand0(), fTmp(), fEoC(), fMinFactor(10.0), fDirection(kDown)
{
	fTmp.SetOwner(kTRUE); fRandDown.SetOwner(kTRUE); fFeeding.SetOwner(kTRUE); fRandFeeding.SetOwner(kTRUE); 
}

BaseGEM::~BaseGEM()
{
	BaseGEM::Clear(""); // clear everything
}

void BaseGEM::Clear(Option_t *o)
{
	TString opt = o;
	if ( opt.Contains("all") ) fLevelScheme.Clear(); 
	
	fRand0.Clear(); fFeeding.Clear(); fRandFeeding.Clear(); fRandDown.Clear(); fTmp.Clear(); fMinFactor = 10.0, fDirection = kDown;
}

Int_t BaseGEM::Import(const Char_t *filename, Option_t *opt)
{	
	// init level scheme and test if initialisation has succeded
	Int_t ok = fLevelScheme.Import(filename,opt);  if ( ok ) InitRandom(); return ok;
}

void BaseGEM::InitRandom()
{
	if ( fLevelScheme.GetLinks().GetSize() == 0 ) return; // nothing to be done
	
	Int_t i, iend, j, jend; Link *gam1 = NULL, *gam2 = NULL;
	
	// clear the previous definition BUT NOT the level scheme which has just been read.
	BaseGEM::Clear("");

	// set each collection with the right number.
	fRandDown.Expand(fLevelScheme.GetLinks().GetSize()); 
	fRandFeeding.Expand(fLevelScheme.GetLinks().GetSize()); fFeeding.Expand(fLevelScheme.GetLinks().GetSize()); 
		
	// First it loops to check if all intensities are > 0 and it calculates the lowest positive one MIN . 
	// If at least one intensity is found <=0,
	// all the gamma-rays with an intensity <= 0 are modified with an intensity = MIN / 100
	Float_t min = 1000000; Int_t nbwrong = 0;
	iend = jend = fLevelScheme.GetLinks().GetSize();
	for (i = 0; i < iend; i++) { 
		gam1 = (Link *)fLevelScheme.GetLinks().At(i);
		if ( gam1->GetStrength().Get() <= 0 ) { 
			nbwrong++; 
		}
		else { // more than 0
			if ( gam1->GetStrength().Get() < min ) 
				min = gam1->GetStrength().Get() ;
		}
	} // iend
	if ( nbwrong > 0 ) {
		printf("\n WARNING in BaseGEM::Import \n");	
		printf("  %d links have been found with an intensity = 0 in %s \n",nbwrong,fLevelScheme.GetName());
		printf("  The intensity of those links are set to the weakest intensity %f / %f \n \n",min,fMinFactor);
		for (i = 0; i < iend; i++) {
			gam1 = (Link *)fLevelScheme.GetLinks().At(i);
			if ( gam1->GetStrength().Get() <= 0 ) { 
				gam1->GetStrength().Set(min/fMinFactor);
			} // if	
		} // i	
	} // nbwrong
	
	// now for each gamma, a generator is created to go down in the path and to set fRand0 which determine the entry point
	RandObj *down, *downfeeding; TObject *obj;
	Float_t sum_int_up_in, sum_int_up_out, sum_int_down_in, sum_int_down_out;
	
	for (i = 0; i < iend; i++) { // loop on all gammas to build the cascade generator
	
		// current link with its associated generators
		gam1 = (Link *)fLevelScheme.GetLinks().At(i); down = new RandObj(); downfeeding = new RandObj();
		
		// feeding link for gam1 
		Link *feeding = new Link(); 
		feeding->SetIL(gam1->GetIL()); feeding->SetFL(gam1->GetIL());
		
		obj = new TObject(); // gam1 is one of the link directly in coincidence with gam1
		fTmp.Add(obj); 
		obj->SetUniqueID(i);
		downfeeding->Add(obj,gam1->GetStrength().Get());	
			
		// look for all gammas up and down in coincidence with gam1
		sum_int_up_in = 0; 
		sum_int_up_out  = gam1->GetStrength().Get();
		sum_int_down_in = gam1->GetStrength().Get(); 
		sum_int_down_out = 0;

		for (j = 0; j < jend; j++) { // look for gammas in coincidence with gamma # i

			if ( i == j ) continue;

			gam2 = (Link *)fLevelScheme.GetLinks().At(j);
	
			if ( gam1->GetIL() == gam2->GetFL() ) { // gammas filling the initial level of gam1
				sum_int_up_in += gam2->GetStrength().Get();
			} 
			if ( gam1->GetIL() == gam2->GetIL() ) { // gammas desexciting the initial level of gam1
				sum_int_up_out += gam2->GetStrength().Get();
				
				obj = new TObject(); 
				fTmp.Add(obj); 
				obj->SetUniqueID(j);
				downfeeding->Add(obj,gam2->GetStrength().Get());
			}
			if ( gam1->GetFL() == gam2->GetIL() ) { // gammas desexciting the final level of gam1
				sum_int_down_out += gam2->GetStrength().Get();
				
				obj = new TObject(); 
				fTmp.Add(obj); 
				obj->SetUniqueID(j);				
				down->Add(obj,gam2->GetStrength().Get());
			}

			if ( gam1->GetFL() == gam2->GetFL() ) { // gammas filling the final level of gam1
				sum_int_down_in += gam2->GetStrength().Get();
			}
		} // jend
		
		// lateral feeding treatment
		if ( sum_int_up_out > sum_int_up_in ) { // the feeding for gam1 has a intensity greater than 0
			feeding->GetStrength().Set(sum_int_up_out - sum_int_up_in);
		}
		else { feeding->GetStrength().Set(0.0); }

		// isomeric state and end of cascade treatment
		if ( sum_int_down_in > sum_int_down_out ) { down->Add(&fEoC,sum_int_down_in - sum_int_down_out); }
	
		fRand0.Add(feeding,feeding->GetStrength().Get()); fFeeding.AddAt(feeding,i);
		fRandFeeding.AddAt(downfeeding,i); 
		fRandDown.AddAt(down,i);  
		
		if ( gDebug > 0 ) 
				printf(" There are %d gammas feeding the gamma # %d \n",downfeeding->GetSize(),i);	
		if ( gDebug > 0 )
			printf(" There are %d gammas feeding the gamma # %d \n",down->GetSize(),i);	
	} // iend
}	

Int_t BaseGEM::InitAGS(const Char_t *filename, Option_t *opt)   
{
	return Import(filename,opt);
}

/*
void BaseGEM::ls(Option_t *o) const 
{
	Link *gam1 = NULL; RandObj *robj;  Int_t iend = fRand0.GetSize();
	
	fRand0.ls(o);
	for (Int_t i = 0; i < iend; i++ ) {
		gam1 = (GammaLink *)fRand0.At(i); 
		if ( gam1 ) {
			printf("%f keV [%d] is stored with weight %f \n",gam1->GetEnergy().Get(),i,fRand0.WeightAt(i)) ;
			
			robj = (RandObj *)fRandUp.At(i); printf(" * %d gammas feed it \n",robj->GetSize()) ;
			for (Int_t j = 0; j < robj->GetSize(); j++) {
				if ( robj->At(j) == &fEoC ) printf("    Lateral feeding with weight %f \n",robj->WeightAt(j));
				else {
					printf("    Gamma #%d, feeding with weight %f \n",robj->At(j)->GetUniqueID(),robj->WeightAt(j)) ;
				}
			} // j			
			robj = (RandObj *)fRandDown.At(i); printf(" * %d gammas desexcites it \n",robj->GetSize()) ;
			for (Int_t j = 0; j < robj->GetSize(); j++) {
				if ( robj->At(j) == &fEoC ) printf("    EoC with weight %f \n",robj->WeightAt(j));
				else {
					printf("    Gamma #%d, desexcitation with weight %f \n",robj->At(j)->GetUniqueID(),robj->WeightAt(j)) ;
				}
			} // j
		} // if gam1
	} // iend
}
*/

// void BaseGEM::FillRandom(TH1 *h, Int_t ntimes) { fRand0.FillRandom(h,ntimes); }

void BaseGEM::DoCascade(TSeqCollection &cas, Int_t which_first, Option_t *o)
{
	TObject *obj; Link *link; if ( fRandFeeding.At( which_first ) == NULL ) return;

	// Ask the random generator for that feeding for the next link in the level scheme and keep on going with it	
	obj = ((RandObj *)fRandFeeding.At( which_first ))->Rand();
	while ( obj != &fEoC ) {
		link = (Link *)fLevelScheme.GetLinks().At(obj->GetUniqueID()); 
		link->DoCascade(cas,o);
		obj = ((RandObj *)fRandDown.At( obj->GetUniqueID() ))->Rand();	
	}
	
	if ( fDirection == kUp ) { // if kUp, inverse the cascade
		TObjArray tmp(30); 
		tmp.SetOwner(kFALSE); tmp.Clear("nodelete"); tmp.AddAll(&cas); cas.Clear("nodelete");
		
		for (Int_t i = tmp.GetEntries()-1; i >=0; i-- ) cas.Add(tmp.At(i));	
	} 
}

void BaseGEM::DoCascade(TSeqCollection &cas, Option_t *o)
{
	TString opt = o; Link *link;
	
	// Reset the previous cascade.
	if ( !opt.Contains("add",TString::kIgnoreCase) ) cas.Clear("nodelete"); 

	// select the 'entry' point
	Int_t which_first;	
	link = (Link *)fRand0.Rand(which_first);  DoCascade(cas,which_first,o);
} 


