#ifndef GW_EVAPLMOF_H
#define GW_EVAPLMOF_H

#ifndef ROOT_Rtypes
#include "Rtypes.h"
#endif
#ifndef GWS_BUFFER_H
#include "Buffer.h"
#endif 

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>

class TLeaf;

namespace Gw {

/** Interface to read/decode EvapOR List Mode Output Files (so-called .pax files)

	In the evapoOR input file, the PAX keyword controls the production of files in which are stored details about the full cascades
	on an event by event bases. This interface helps you to read these files. A tar file of the evapOR package is available with a
	global makefile to build it. Just ask for it at agata_@_ipnl.in2p3.fr (remove _ in the email address). It has been compiled
	successfully on Sun station (with the Sun f77 fortran compiler) and on Linux SL4 (with the Intel ifort fortran compiler). Unfortunatelly,
	it cannot be compiled with g77 or gfortran. 
	
	Four output structures are filled event by event which correspond to the four common blocks as explained in the "playback common blocks" section of
	the evapoOR documentation (section A.6). The name of the variables are identical and their signification is
	reported here.  
	
	\code 
	struct GLOBAL {
		Int_t 	AC;	// A of initial compound system
		Int_t 	ZC;	// Z of initial compound system
		Int_t	NCASC;	// total number of cascades
		Int_t 	MDIR;	// Defines coordinate system.
		Int_t 	JCASC;	// total number of casc. with weight >0
		Float_t ENER;	// E* of initial compound system
		Float_t SIGMA;	// Total CN formation cross section
		Float_t VRC;	// CN velocity
		Int_t 	MAX_P;	// number of particle decay channels
	};
	struct ID {
		Short_t MG;	// gamma multiplicity
		Float_t EE;	// entry state energy
		Float_t	JE;	// entry state spin
		Float_t L;	// CN initial angular momentum
		Short_t Z;	// final nucleus Z
		Short_t N;	// final nucleus neutron number
		Short_t NUM;	// number of emissions
		Float_t WT;	// weight of this cascade.
		
		Bool_t  FIS;	// true if the cascade ends up by fission
	};
	struct DATA {
		Short_t MODE[MAXCASCADE];	// emission type (1=n, 2=p, 3=4He, etc)
		Float_t EI[MAXCASCADE];		// excitation energy of initial state
		Float_t EP[MAXCASCADE];		// particle E in CN rest frame
		Short_t JI[MAXCASCADE];		// initial spin  (j or j-1/2)
		Short_t LP[MAXCASCADE];		// orbital ang. mom. of emitted particle
		Short_t MP[MAXCASCADE];		// projection of L on quantization axis
		Short_t MUL[MAXPARTICLES];	// multiplicity of mode j
		Float_t ES[MAXPARTICLES];	// total energy of mode j (sum of Kinetic E)
		Float_t DJ[MAXPARTICLES];	// total spin change for mode j 
	};
	struct KINMAT {
		Float_t V0[3];			// CN three momentum (k=1,3) ("lab")
		Float_t VE[MAXCASCADE][3];	// emitted particle 3 momentum ("lab")
		Float_t THETA[MAXCASCADE];	// emission angle
		Float_t PHI[MAXCASCADE];	// emission angle
		Short_t ZB;			// Z of emitting nucleus
		Short_t AB;			// A of   "         "	
		Float_t ER;			// recoil energy (CN) ("lab")
		Float_t EEJ[MAXCASCADE];	// energy of emitted particle ("lab")
		Float_t JF[MAXCASCADE];		// final angular momentum 
	}; 	
	\endcode
	
 	Several pax files could be chained (see SetPath & AddFile). Once initialized, the chain is processed with NexEvent() that returns true
	if a new event is available. Rewind allows to restart the reading while Scan just displays on the standard output the content 
	of each event. 
	 
	This interface takes care of the endian type so that you can generate your pax file on any machine and process them with EvapLMOF 
	on any machine. 
	
	@author Olivier Stezowski
 */
class EvapLMOF
{
public:
	//! Reading status
	enum Status { kEmpty, kGood, kEoE, kFail };

	//! Set verbosity level	
	static Short_t VERBOSE;
	
public:
	static const Short_t MAXCASCADE   = 200; 
	static const Short_t MAXPARTICLES =  12; // because arrays start from 0 in C++ while 1 in fortran
	
private:
	// structure corresponding to the GLOB common
	struct GLOBAL {
		Int_t 	AC;
		Int_t 	ZC;
		Int_t	NCASC;
		Int_t 	MDIR;
		Int_t 	JCASC;
		Float_t ENER;
		Float_t SIGMA;
		Float_t VRC;
		Int_t 	MAX_P;
	};
	// structure corresponding to the ID common
	struct ID {
		Short_t MG;
		Float_t EE;
		Float_t	JE;
		Float_t L;
		Short_t Z;
		Short_t N;
		Short_t NUM;
		Float_t WT;
		
		Bool_t  FIS;
	};
	// structure corresponding to the DATA common
	struct DATA {
		Short_t MODE[MAXCASCADE];	
		Float_t EI[MAXCASCADE];
		Float_t EP[MAXCASCADE];	
		Short_t JI[MAXCASCADE];	
		Short_t LP[MAXCASCADE];	
		Short_t MP[MAXCASCADE];	
		Short_t MUL[MAXPARTICLES];	
		Float_t ES[MAXPARTICLES];
		Float_t DJ[MAXPARTICLES];
	};
	// structure corresponding to the KINMAT common (always filled unless idonly is set)	
	struct KINMAT {
		Float_t V0[3];
		Float_t VE[MAXCASCADE][3];
		Float_t THETA[MAXCASCADE];	
		Float_t PHI[MAXCASCADE];	
		Short_t ZB;
		Short_t AB;	
		Float_t ER;
		Float_t EEJ[MAXCASCADE];
		Float_t JF[MAXCASCADE];	
	}; 
	
public:
	friend EvapLMOF &idonly(EvapLMOF &reader) ;
	friend EvapLMOF &full(EvapLMOF &reader) ;	

private:
	Gws::Buffer *fHeader;		// to identify the next record (Global or Id)	
	
	Gws::Buffer *fGlobalRecord;	// global record (one per run)
	Gws::Buffer *fIdRecord;		// id record (one per cascade)	
	Gws::Buffer *fDataRecord;	// full cascade
	
private:	
	struct GLOBAL	fGlobal;
	struct ID 	fId;	
	struct DATA 	fData;
	struct KINMAT 	fKin;
	
private:		
	Bool_t fIdOnly; 

	TString fError;
	TString fWarning;
		
private:	
	void SetRecord(Gws::Memory::EEndian);
		
	Bool_t ReadGlobal();
	Bool_t ReadId();	
	Bool_t ReadData();	
	void   ReadHeader();
	
private:
	FILE *fInfile;				// input stream

	Int_t fCurrent, fMaxFiles, fStatus; 	// current file number, max file number and current status

	std::vector<Double_t>     fEventsRead;	// total number of events read for each files
	std::vector<std::string>  fFileName;	// name of each file

	std::string fPath;			// path to reach input files

public:
	EvapLMOF();
	~EvapLMOF();

	//! to modified the reading
	/*!
		Usage:
		\code EvapLMOF pax; pax << idonly; \endcode
		
		The following manipulators are available:
		- idonly If you don't need the full cascade but just the global and id part.
		- full to read the full cascade 
	*/	
	EvapLMOF &operator<< (EvapLMOF &(*pf)(EvapLMOF &)) { return (*pf)(*this); }

	//! to add an input (pax) file  
	/*!
		With this function, a file is added to the list of input files.

		All the files must be in the same directory (or in a sub-directory of that directory) which is set once
		with the SetPath() member (default is current directory ./). A counter is associated with each file to
		know the number of events read. Non-existing file or non-pax file could be added without problems ... they will be
		ignored.
	*/
	void AddFile(const char *fname) { std::string s = fname; fFileName.push_back(s); Double_t zero = 0.0; fEventsRead.push_back(zero); fMaxFiles++; }
	
	//! set the directory where to find the input files
	void SetPath(const char *path ="./") { fPath = path; }
	
	//! it returns an array with the number of events read from all the input files	
	const std::vector<Double_t> &GetEventsRead() { return fEventsRead; }	

	//!  switch to next file.
	/*!
	 	if a new file is open successfully, Status is kGood. Otherwise, Status is not kGood and you cannot
		read anymore. You can switch to the next file at any moment.
	*/
	void   NextFile();

	//! To load the next event
	/*!
		It returns true in case a new event has just been read. The next parameter is not yet used. In case something goes
		wrong, it returns false and a message corresponding to the error is available (see GetError).
	*/
	Bool_t NextEvent(UInt_t next = 0);
	
	//! To restart the data stream from the beginning
	/*!
		statics are also reset and the stream is ready for reading. The verbose level and the manipulators are not modified
	*/	   
	void Rewind();
	
	//! To scan events and show them on the standard output
	/*!
		The stream is rewinded before starting the scanning so that start is the absolute position in the chain.
	*/	
	void Scan(UInt_t start = 0, UInt_t end = 10);
	
	//! To get the current global structure	
	const GLOBAL &GetGlobal() { return fGlobal;}
	//! To get the current id structure	
	const ID       &GetId() { return fId; }	
	//! To get the current data structure
	const DATA   &GetData() { return fData; }
	//! To get the current kinematic structure
	const KINMAT  &GetKin() { return fKin; }
	
	//! To get a ROOT leaf corresponding to one of the variable
	/*!
		It helps you to write general macros to fill histograms. A leaf is a pointer to one of the value
		of the global record, the id record, the data record or the kinematic record. For instance, the following part of code
		fills the gamma multiplicity histogram. To get the Z distribution of the residus, just change "MG" by "Z".
	
		\code
		EvapLMOF pax; pax.AddFile("my.pax");
		
		TLeaf *x = pax.GetLeaf("MG"), *w = pax.GetLeaf("WT");
		
		while( pax.NextEvent() ) { histo.Fill(x->GetValue(),w->GetValue()); }
		\endcode
	*/		
	TLeaf *GetLeaf(const char *leafname = "no");
		
	//! The show the content of the current event on the standard ouput	
	void ShowCurrentEvent(std::ostream &out = std::cout);
	
	//! The current status is shown on the standard ouptut
	/*!
		It gives you the number of events read in all the registered files as well as the current file that
		is being read. Additional informations that comes from the header are also displayed and the last error messages if any.
	*/	
	void ShowCurrentConditions(std::ostream &out = std::cout);
	
	//! To get the last error	
	TString &GetError() { return fError; }	
	
	//! To get the last warning
	TString &GetWarning() { return fWarning; }
};
inline EvapLMOF &idonly(EvapLMOF &reader)  { reader.fIdOnly = true ; return reader;  }
inline EvapLMOF &full(EvapLMOF &reader)     { reader.fIdOnly = false ; return reader;  }
}
#endif

