/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file GEM.h header file for a GEM (gamma-rays generator) */
 
#ifndef GW_GEM_H
#define GW_GEM_H

#ifndef GW_BASEGEM_H
#include "BaseGEM.h"
#endif

namespace Gw {

/**
Class to get randomly cascades (with lateral feeding) of gammas on the basis of a level scheme.

With BaseGEM, cascades are determined randomly in a level scheme. The produced
cascade ends with the first populated link (level) which is the entry point in the 
level scheme. It should be enought for some
experiment (beta decay experiment), but in several cases (fusion-evaporation), the
lateral feeding produces gamma-rays which can be also simulated by many different models.

This class is designed to help in this kind of situation. After the BaseGEM random
process, it calls the feeding process which is just a particular Link which has its
own DoCascade member. \n

To set the feedink link, use the SetFeeding member. Unless removed, the link belongs
to GEM and will be deleted in the destructor. In the demos directory, there is an exemple
of how to write a feeding link (see TestFeeding). 

\todo initialisation with a existing level scheme by copying it (copy constructor for level scheme)

\todo Check out what happens when a link is not bind to two levels

\author Olivier Stezowski
*/
class GEM : public BaseGEM
{
private:
     Link  *fFeeding;	// link which feeds any levels
     
protected:
     //! to be called to init the random generator 
 //    virtual void InitRandom();	
        
public:
    GEM();
    virtual ~GEM();

    //! Clear everything 
    virtual void Clear(Option_t *opt = "all");
    
    Link *SetFeeding(const char *feedingtype = "TestFeeding");
    
    //! to modify the current feeding link (general one)
    /*!
    	One 
    	It is the charge of the user to delete the previous link which is returned by this member.
	The current feeding link in deleted by the destructor so there no need to delete it explicitely.
	However it could be done by the user using: \n
	Link *l = gem.SetFeeding(NULL); delete l;
    */
    Link *SetFeeding(Link *newfeeding) { Link *tmp = fFeeding; fFeeding = newfeeding; return tmp; }

    //! starting point in the ExI plan from which starts the gamma-ray emission   
    void SetExI(Float_t E, Float_t I);
  

     //! starting point in the ExI plan from which starts the gamma-ray emission      
   // virtual Int_t Import(const Char_t *, Option_t *opt="152DY");  
       
    //! cascade simulation
    /*!
    Fill a list (Cascade) of Links with a randomly determined path.
    The cascade is cleared before running the Monte-Carlo unless the option contains add. \n
    */
    virtual void DoCascade(TSeqCollection &, Option_t *opt = "");   
//    virtual void DoCascade(TSeqCollection *cas, Option_t *opt = "") { GEM::DoCascade((*cas),opt) ; } 
        
    virtual void DoCascade(TSeqCollection &cas, Int_t from, Option_t *opt = "") { BaseGEM::DoCascade(cas, from, opt); }

    //! rootcint dictionary
    ClassDef(GEM,0); // Gamma-ray Monte-Carlo
};
// inline members
}

#endif
