#ifndef GW_GAMMAFILTER_H
#define GW_GAMMAFILTER_H

#ifndef ROOT_TH1F
#include <TH1F.h>
#endif

#ifndef GW_GAMMALINK_H
#include <GammaLink.h>
#endif

#ifndef GW_BASEFILTER_H
#include <BaseFilter.h>
#endif

namespace Gw {

//! Standard filter for gamma-rays (EUROBALL-like appproach)
/*!
Gamma-ray simulated with BaseGEM or GEM could be used as inputs for detector simulations.
The GEANT simulations give then the expected events coming from a 'real' experiment.

In order to produce events faster (probably needed to produce large amount of DATA), one
can define some global characteristics of the AGATA array and use this experimental filter
to get as much as possible realistic events. This approach is standard because 
the detector is supposed to be characterised only by its efficiency, its peak to total,
its resolution and a compton profile. It allows then some comparisons with, for
instance, EUROBALL. A more complex approach (TrackFilter) should be available soon ..  \n
The list of gamma-rays is modified by taking into account:
	-# An efficiency curve. \n
	For a given gamma-ray energy, it says if the gamma-ray is detected or not.
	-# A peak to total ratio \n
	For a given gamma-ray energy, it says if the gamma-ray is fully detected (photopeak event). If not the g-ray energy is changed
	-# A doppler broadring \n
	For a given gamma-ray energy, it gives a Gaussian distribution.
	
\callgraph

\author Olivier Stezowski \date 01/2004 modification 11/2005
*/
class GammaFilter : public BaseFilter
{
private:

protected:
	TH1F *fComptonShape;	// What is the energy after a compton scattering
	TH1F *fCompton;		// histogram that is used to do compton scattering
	TH1F *fDoppler;		// doppler broadring
	TH1F *fEfficiency;	// efficiency curve
	
public:
	GammaFilter();
	//! the simpliest filter: efficiency = cst, P/T = cst, Doppler = cst
	/*!
	\param xmin { lowest gamma-rays energy in keV }
	\param xmax { highest gamma-rays energy in keV }
	\param eff { default 1 }
	\param PT { default 0.5 - 1 means every gamma-ray give a photopeak event }
	\param dop { default 1.25 keV for any gamma-ray energy }
	Default is 1 keV per channel from xmin to xmax, which is precise enought for such quantities.
	The efficiency and the P/T must be in the range [0,1], if not true, default values are choosen.
	*/
	GammaFilter(Axis_t xmin, Axis_t xmax, Float_t eff = 1, Float_t PT = 0.5, Float_t dop = 1.25);
	virtual ~GammaFilter();

	//! check out if it has been correctly initiated
	/*!
		It means an histogram to define the efficiency curve, the doppler and the Compton 
	*/
	Bool_t IsEffective();
	
	//! to modifiy the efficency curve.
	/*!
		This is the absolute efficiency curve. Any value of the histogram should
		be smaller or equal to 1. If one value is found in the 	.. blabla
		
		
	*/
	// void SetEfficiency(TH1F &);

	
	//! to init this filter from a root file
	Bool_t  InitFilter(const char *);
	
	//! true if this gamma-ray has not given all its energy. 
	/*!
	 The second argument gives the compton scattered gamma-ray energy. The GammaLink is not modified !
	*/
	Bool_t Compton ( GammaLink *, Float_t &);

	//! return a random number assuming a gaussian at the gamma-rays position with a width given by fDoppler
	Float_t Doppler ( GammaLink * );

	//! True if this gamma ray is not detected
	bool IsRejected ( GammaLink * );

	//@{
	//! Apply this filter to the list of gamma-rays
	/*!
	It returns the number of gamma-rays after application of this experimental filter. 
	The simulated gamma-rays energies are stored in a ROOT buffer as floats from the current position.
	PS1: No need to worry about the buffer size since it is expanded if too small.
	PS2: It is the user who must 'rewind' the buffer ! 
	*/
	virtual Int_t ApplyE(const TSeqCollection &, TBuffer &);
	virtual Int_t ApplyE(const TSeqCollection &, Float_t *);
	//@}
			
	//! Root dictionnary related
	ClassDef(GammaFilter,1); // An experimental filter to real gamma-ray cascades
};
// GammaFilter inline members //////////////////////////////////////////////////////
inline bool GammaFilter::IsRejected ( GammaLink *gam )
{ return fEfficiency->GetBinContent(fEfficiency->FindBin(gam->GetEnergy().GetValue())) < fRand->Current()->Rndm();}
inline Float_t GammaFilter::Doppler ( GammaLink *gam )
{
	Float_t e = gam->GetEnergy().GetValue();
	return fRand->Current()->Gaus(e,fDoppler->GetBinContent(fDoppler->FindBin(e)));
}
// End GammaFilter inline members //////////////////////////////////////////////////

}
#endif
