#ifndef GW_GEANTLMOF_H
#define GW_GEANTLMOF_H

#ifndef ROOT_Rtypes
#include "Rtypes.h"
#endif

#ifndef GW_GAMMATRACKER_H
#include <GammaTracker.h>
#endif

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>

#define MAXCHARACTER 500

class TGeoVolume;

namespace Gw {

/** Class to read/decode Geant List Mode Output Files

	This class is an event stream to read events written in GEANT list-mode output files (see AGATA documentation). Several
	files could be chained to allow large dataset to be treated. Once a single directory is provided (see SetPath()), you can 
	add as many files as you need with the AddFile member. The file's name is relative to the path directory. For instance,
	if you need to chain two files /data/mysimu1/run1 and /data/mysimu1/run2, you should do: \n
	GeantLMOF reader; \n
	reader.SetPath("/data/mysimu1/"); reader.AddFile("run1"); reader.AddFile("run2"); \n

	Once the input files are given, the member NextEvent() loads the next event. An event is composed of a list of emmitted particles and a list of associated impacts. The members NbEmitted() and NbImpacts() returns the values for the
	current event. 
	Several additionnal members (GetXXX like) help you to get everything that is defined in the list-mode output file format.
	The Get members have about the same name as given in the AGATA-GEANT documentation.\n

	The current event could be send to a GammaTracker through the >> operator: \n
	GeantLMOF reader; \n
	GammaTracker *tracker = GammaTracker::GetTracker("TheTrackerYouWant"); \n
	while ( reader.NextEvent() ) { reader >> tracker; ... and do whatever you want with the tracker output ... }
	
	As a default, this class is as general as possible. It means the input files may contains other particles that gammas. These
particles cannot be sent to the GammaTracker. To remove the additionnal particles, the >> operator sorts the inputs before sending
them to the tracker. if you are sure that your input files contain only gamma-rays, you can call speed up the process with the 
following command: \n
	GeantLMOF reader; \n
	reader << fastout;

To toggle back on the checking: reader << checkout;

	\todo 
		- full reading of the header part
		- some reading of event by event not yet implemented (ID -101, -102 and -103)
		- position of impacts in the detector frame (mgs reference frame)
	
	@author Olivier Stezowski
 */
class GeantLMOF
{
	friend GeantLMOF &fastout(GeantLMOF &reader) ;
	friend GeantLMOF &checkout(GeantLMOF &reader) ;

private:
	bool fFastOut;

	 //! use to keep only gammas 
	GeantLMOF *fGamma;
	
protected:
	//! Fill this using another GeantLMOF by keeping only emmitted particles with fType between fTypemin & fTypemax
	UInt_t OnlyType(const GeantLMOF &input, Int_t fTypemin = 1, Int_t fTypemax = 1);
		
public:
	//! Reading status
	enum Status { kEmpty, kGood, kEoE, kFail };

private:
	Double_t fRecoilBeta;
        Double_t fRecoilSigma;
        Double_t fRecoilDx;
        Double_t fRecoilDy;
        Double_t fRecoilDz;
        Double_t fRecoilAngle;
	Double_t fTime;

	UInt_t fNbEmitted;
	UInt_t fNbImpacts;

	std::vector<Int_t>    fType;
        std::vector<Int_t>    fNdet;
        std::vector<Int_t>    fEvNum;
        std::vector<Int_t>    fNInt;
        std::vector<Int_t>    fPrimary;
	std::vector<Double_t> fEnergy;
        std::vector<Double_t> fDx;
        std::vector<Double_t> fDy;
        std::vector<Double_t> fDz;
        std::vector<Double_t> fEdep;
        std::vector<Double_t> fPx;
        std::vector<Double_t> fPy;
        std::vector<Double_t> fPz;
        std::vector<Double_t> fNseg;

	Double_t fDistFactor;
        Double_t fEnerFactor;

private:
	std::ifstream fInfile;			// input stream

	Int_t fCurrent, fMaxFiles, fStatus; 	// current file number, max file number and current status

	std::vector<Double_t>     fEventsRead;	// total number of events read for each files
	std::vector<std::string>  fFileName;	// name of each file

	std::string fPath;			// path to reach input files

	Char_t buf[MAXCHARACTER];		// buffer to get data from files

private:
	void   ReadHeader();

public:
	GeantLMOF();
	~GeantLMOF();
	
	//! to import the agata geometry from the Agata-Geant4 world
	/*!
	To define the agata geometry from the standard (Agata Geant4) input files (asolid, acluster, aeuler). If the top volume already exists in
	the GeoManager, agata is added to it.
	*/	
	static TGeoVolume *ImportAgataG4(const char *asolid = "asolid", const char *aclust = "aclust", const char *aeuler = "aeuler", const char *top = "World");
		
				
	void SetRecoil(Float_t b=0.0, Float_t dx = 0.0, Float_t dy = 0.0, Float_t dz = 1.0, Float_t angle = 0.0) {fRecoilBeta=b; fRecoilDx=dx; fRecoilDy=dy; fRecoilDz=dz; fRecoilAngle=angle;}
	
	Double_t GetDistFactor() const { return fDistFactor; } 
	Double_t GetEnerFactor() const { return fEnerFactor; }

	Double_t GetfRecoilBeta() const { return fRecoilBeta; }
	void GetRecoil(Double_t &b, Double_t &dx, Double_t &dy, Double_t &dz) const { b = fRecoilBeta; dx = fRecoilDx; dy = fRecoilDy; dz = fRecoilDz;}

	UInt_t NbEmitted() const { return fNbEmitted; }
	UInt_t NbImpacts() const { return fNbImpacts; }
		
	const std::vector<Double_t> &GetEnergy() { return fEnergy;}
	const std::vector<Double_t> &GetDx() { return fDx;}
	const std::vector<Double_t> &GetDy() { return fDy;}
	const std::vector<Double_t> &GetDz() { return fDz;}	

	const std::vector<Double_t> &GetEdep() { return fEdep;}
	const std::vector<Double_t> &GetPx()   { return fPx;}
	const std::vector<Double_t> &GetPy()   { return fPy;}
	const std::vector<Double_t> &GetPz()   { return fPz;}		
	const std::vector<Double_t> &GetNseg() { return fNseg;}	

	const std::vector<Int_t> &GetType()  { return fType;}
	const std::vector<Int_t> &GetNdet()  { return fNdet;}
	const std::vector<Int_t> &GetEvNum() { return fEvNum;}
	const std::vector<Int_t> &GetNInt()  { return fNInt;}		
	const std::vector<Int_t> &GetPrimary()  { return fPrimary;}

	const std::vector<Double_t> &GetEventsRead() { return fEventsRead; }	
				
	//! fill the tracker with the current event	
	void operator>> (GammaTracker *) ;
	
	GeantLMOF &operator<< (GeantLMOF &(*pf)(GeantLMOF &)) { return (*pf)(*this);}

	//! to add an input file  
	/*!
		With this function, a file is added to the list of input files.

		All the files must be in the same directory (or in a sub-directory of that directory) which is set once
		with the SetPath() member (default is current directory ./). A counter is associated with each file to
		know the number of events read.
	*/
	void AddFile(const char *fname) { std::string s = fname; fFileName.push_back(s); Double_t zero = 0.0; fEventsRead.push_back(zero); fMaxFiles++; }
	void SetPath(const char *path ="./") { fPath = path; }

	//!  switch to next file.
	/*!
	 	if a new file is open successfully, Status is kGood. Otherwise, Status is not kGood and you cannot
		read anymore. You can switch to any file at any moment 
	*/
	void   NextFile();

	//! To load the next event
	/*!
	It returns kFALSE if something goes wrong. The Status field gives you informations to know why the next
	event cannot be loaded (kEoE=end of events, kFail=something goes wrong during the last reading). \n
	This member function could be used to read: \n
	- formatted files (with a begin of event flag -100, if the GEANT simulation has been performed with an external file). Simply call NextEvent() to fill GeantLMOF with the next event.
	- unformatted files (no begin of event flag -100). In this case call NextEvent(mul), with mult the multiplicity of
	the cascade you would like to build.
	*/
	Bool_t NextEvent(UInt_t mult = kMaxUInt);

	void ShowCurrentEvent();
	//!  The current status is shown on the standard ouptut
	/*!
	It gives you the number of events read in all the registered files as well as the current file that
	is being read. Additional informations that comes from the header are also displayed. 
	*/	
	void ShowCurrentConditions();
};

GeantLMOF &fastout(GeantLMOF &reader)  { reader.fFastOut = true ; return reader;  }
GeantLMOF &checkout(GeantLMOF &reader) { reader.fFastOut = false ; return reader; }
		
}
#endif

