/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file BaseENSDF.cpp compiled in libGWPHYSICS.so */

#ifndef GW_BASEENSDF_H
#include "BaseENSDF.h"
#endif

#include <math.h>
#include <iostream>
#include <sstream>

using namespace std;
using namespace Gw;

string BaseENSDF::BLANK = "                                                                                \n";
    
BaseENSDF::DataSetKey::DataSetKey(const DataSetKey &key)
{
	Nuclide = key.Nuclide;
	Dsid 	= key.Dsid;
	Dsref 	= key.Dsref;
	Pub 	= key.Pub;
	Date 	= key.Date;
	Position.first  = key.Position.first; 
	Position.second = key.Position.second;
	
	CurrentPosition = key.CurrentPosition;
 }

void BaseENSDF::DataSetKey::Clear()
{
	Nuclide = "     ";
	Dsid 	= "                             ";
	Dsref 	= "                         ";
	Pub 	= "        ";
	Date 	= "     ";
	Position.first = Position.second = CurrentPosition = 0;
}

void BaseENSDF::DataSetKey::Print(std::ostream &out) const 
{
	out << "Nuclide " << Nuclide << " with id " << Dsid 
	    << " at position " << Position.first << " " << Position.second << endl;
} 	
	
BaseENSDF::BaseENSDF()
{
}

BaseENSDF::~BaseENSDF() { Close(); }

bool BaseENSDF::IsRecord(const char *r, const char *t) const
{
	string id, type, record; bool ok;

	ok = false; 
	for (int i = 0; i < RSIZE; i++ ) record += r[i]; // cannot use just record = r ... it doesn't work !
	
	id = record.substr(5,4); type = t;
	
	if ( type == "IDENT" && (id == "    " || id == "1   ") ) { 
		if ( record != BLANK ) ok = true; 
	}	
	
	if ( type == "Q-VALUE" && id == "  Q " ) {  ok = true; }	
	if ( type == "LEVEL"   && id == "  L " ) {  ok = true; }
	if ( type == "BETA"    && id == "  B " ) {  ok = true; }	
	if ( type == "EC"      && id == "  E " ) {  ok = true; }	
	if ( type == "ALPHA"   && id == "  A " ) {  ok = true; }
	if ( type == "GAMMA"   && id == "  G " ) {  ok = true; }
	
	if ( type == "END"     && record == BLANK )  {  ok = true; } 
	return ok;
}

string BaseENSDF::Nuclide(const char *r) const
{
	string tmp; 
	for (int i = 0; i < 5; i++ ) tmp += r[i]; return tmp;
}
string BaseENSDF::Dsid(const char *r) const
{
	string tmp; 
	for (int i = 9; i < 39; i++ ) tmp += r[i]; return tmp;
}
string BaseENSDF::Dsref(const char *r) const
{
	string tmp;
	for (int i = 39; i < 65; i++ ) tmp += r[i]; return tmp;	
}    
string BaseENSDF::Pub(const char *r) const
{
	string tmp;
	for (int i = 65; i < 74; i++ ) tmp += r[i]; return tmp;	
}

string BaseENSDF::Date(const char *r) const
{
	string tmp;    
	for (int i = 74; i < 80; i++ ) tmp += r[i]; return tmp;	 
}

unsigned int BaseENSDF::IsDataSet(const char *nuclide, const char *datasetid) const
{
	unsigned int which = 0;
	for ( unsigned int i = 0; i < fDataSets.size(); i++ ) {
		if ( fDataSets[i].Dsid.find(datasetid) != string::npos
			&& fDataSets[i].Nuclide.find(nuclide) != string::npos  ) which = i + 1;
	}
	return which;
}

bool BaseENSDF::NextRecord(unsigned int whichdataset, char *record)
{
	bool ok = true; unsigned int realwhich;
	
	Input.clear(); // 
	if (  (0 < whichdataset) && (whichdataset < (fDataSets.size()+1)) ) {
	
		realwhich = whichdataset - 1;
		
		fDataSets[realwhich].CurrentPosition++; 
		if ( fDataSets[realwhich].CurrentPosition >= fDataSets[realwhich].Position.second ) 
		  { ok = false; }
		else { 
		  Input.seekg(fDataSets[realwhich].CurrentPosition * RSIZE,ios::beg);
		  // Input.read(record,RSIZE); 
		  for (int i = 0; i < RSIZE; i++) Input.get(record[i]);
		  ok = Input.good();
		}
	}
	else { ok = false; } 
	
	return ok;
}

bool BaseENSDF::FirstRecord(unsigned int whichdataset, char *record)
{
	bool ok = true; unsigned int realwhich;
	
	Input.clear(); //
	if (  (0 < whichdataset) && (whichdataset < (fDataSets.size()+1)) ) {

		realwhich = whichdataset - 1;
		
		fDataSets[realwhich].CurrentPosition = fDataSets[realwhich].Position.first; 
		Input.seekg(fDataSets[realwhich].CurrentPosition * RSIZE,ios::beg); 
		
		// Input.read(record,RSIZE); 
		for (int i = 0; i < RSIZE; i++) Input.get(record[i]);		
		ok = Input.good();
	}
	else { ok = false; } 
	
	return ok;
}
              
bool BaseENSDF::Open(const char *filename)
{
	DataSetKey tmpkey;
	char record[RSIZE]; int read_record, expected_record; bool anew;
	
	if ( IsOpen() ) { Close() ; } 
	
	Input.open(filename); 
	
	anew = false; read_record = 0;
	if ( Input.is_open() ) { // the file is open, read it and creat the list of data sets.
			
		// determines the number of expected records
		Input.seekg(0,ios::end); expected_record = (int)Input.tellg(); // OK .. ENSDF files cannot be larged ...
		if ( (expected_record % RSIZE) != 0 ) { 
			cerr << " BaseENSDF::Open(): " << filename << " is not an ENSDF file \n" << endl; expected_record = 0;
		}  
		else expected_record /= RSIZE;	
		
		Input.seekg(0,ios::beg);	
				
		while( read_record != expected_record ) { // 
		
			// cannot used read because of a bug on solarisCC !
			for (int i = 0; i < RSIZE; i++) Input.get(record[i]);
			
			if ( Input ) { 

				if ( IsRecord(record,"IDENT") && anew == false ) {

					tmpkey.Nuclide = Nuclide(record);
					tmpkey.Dsid    = Dsid(record);
					tmpkey.Dsref   = Dsref(record);
					tmpkey.Pub     = Pub(record);
					tmpkey.Date    = Date(record);
					tmpkey.Position.first = tmpkey.CurrentPosition = read_record; 
				
					// tmpkey.Print(cout);
					anew = true;	
				}
				if ( IsRecord(record,"END") && anew == true ) { 
			
					tmpkey.Position.second = read_record; fDataSets.push_back(tmpkey); 
				
					tmpkey.Clear(); anew = false;
				}
				read_record++;	
			}
			else { read_record = expected_record; } // something goes wrong during the reading
		} // while reading
	} // is_open
	cout << fDataSets.size() << " datasets have been found in the ENSDF file " << filename << endl;
	
	return Input.is_open();

	/* 
	old version buggy because solarisCC ifstream.read doesn't work properly !!
	
	DataSetKey tmpkey;
	char record[RSIZE]; int nb_record; bool anew;
	
	if ( IsOpen() ) { Close() ; } 
	
	Input.open(filename); 
	
	anew = false; nb_record = 0;
	if ( Input.is_open() ) { // the file is open, read it and creat the list of data sets.
		// read the file and determine the number of identiciation records.
		
		Input.seekg(0,ios::beg);	
				
		while( Input.read(record,RSIZE) ) {

			if ( IsRecord(record,"IDENT") && anew == false ) {
			
				tmpkey.Nuclide = Nuclide(record);
				tmpkey.Dsid    = Dsid(record);
				tmpkey.Dsref   = Dsref(record);
				tmpkey.Pub     = Pub(record);
				tmpkey.Date    = Date(record);
				tmpkey.Position.first = tmpkey.CurrentPosition = nb_record; 
				
				anew = true;	
			}
			if ( IsRecord(record,"END") && anew == true ) { 
			
				tmpkey.Position.second = nb_record; fDataSets.push_back(tmpkey); 
				
				tmpkey.Clear(); 
				anew = false;
			}
			nb_record++;
		}
	}
	cout << fDataSets.size() << " datasets have been found in the ENSDF file " << filename << endl;
	
	return Input.is_open(); */
}

void BaseENSDF::Close()
{
	fDataSets.clear(); Input.close(); Input.clear();
}

void BaseENSDF::ls(std::ostream &out) const
{

// out << " Nuclide |" << " Dataset identification |" << " Date | " << "[first,last] line " << " Current Record " << endl;
	for (unsigned int i = 0; i < fDataSets.size(); i++) {
	 out << fDataSets[i].Nuclide << " " 
		<< fDataSets[i].Dsid << " "
//		<< fDataSets[i].Date << ""
		<< "from line " << fDataSets[i].Position.first << " to " << fDataSets[i].Position.second << " Current: " 
		<< fDataSets[i].CurrentPosition << endl;
	}
} 

float BaseENSDF::GetCC(const char *r)
{
	string tmp;
	
	tmp = "";
	for (int i = 55; i < 62; i++ ) 
		if ( r[i] != ' ' ) tmp += r[i]; // copy only the interesting part 
		
	float c;
	if (tmp.size() > 0 ) { istringstream estring(tmp); estring >> c; }
	else c = 0.0; 
	
	return c;
}

float BaseENSDF::GetDCC(const char *r)
{
	float dvalue, precision; unsigned int i; string tmp1, tmp2;
	
	// copy the cc string in tm1 to look for the precision
	tmp1 = "";
	for (i = 55; i < 62; i++) 
		if ( r[i] != ' ' ) tmp1 += r[i]; 
	
	if ( tmp1.size() > 0 )	
		precision = GetPrecision(tmp1);
	else 
		precision = 0.0;	
		
	// copy the dcc string in tmp2 
	tmp2 = "";	
	for (i = 62; i < 64; i++) 
		if ( r[i] != ' ' ) tmp2 += r[i];  

	if ( tmp2.size() == 0 || 
	     tmp2 == "LT" || tmp2 == "GT" || tmp2 == "LE" ||
	     tmp2 == "GE" || tmp2 == "AP" || tmp2 == "CA" || tmp2 == "SY" ) dvalue = 0;
	else {
		istringstream estring(tmp2); 
		estring >> dvalue; dvalue *= precision;
	}
	return dvalue;
}

float BaseENSDF::GetMR(const char *r)
{
	string tmp;
	
	tmp = "";
	for (int i = 41; i < 49; i++ ) 
		if ( r[i] != ' ' ) tmp += r[i]; // copy only the interesting part 
	
	float c;
	if (tmp.size() > 0 ) { istringstream estring(tmp); estring >> c; }
	else c = 0.0; 
	
	return c;
}

// aym a faire
float BaseENSDF::GetDMR(const char *r)
{
	float dvalue, precision; unsigned int i; size_t l; string tmp1, tmp2;
	
	// MR to get the precision 
	tmp1 = "";
	for (i = 41; i < 49; i++ ) 
		if ( r[i] != ' ' ) tmp1 += r[i]; // copy only the interesting part 

	if ( tmp1.size() > 0 )	
		precision = GetPrecision(tmp1);
	else 
		precision = 0.0;	
	
	tmp2 = "";
	for (i = 49; i < 55; i++ ) 
		if ( r[i] != ' ' ) tmp2 += r[i]; // copy only the interesting part 	
		
	if ( (l = tmp2.find('+', 0 ) ) != string::npos 
		&& (l = tmp2.find('-', 0 ) != string::npos ) ) { // aymetric error
		dvalue = 0.0; //for the moment ..
	}
	else { 
		if ( tmp2.size() > 0 ) {
			istringstream estring(tmp2); estring >> dvalue; 
		}
		else dvalue = 0.0;
	}
	return dvalue*precision;
}


float BaseENSDF::GetRI(const char *r, std::string &s)
{
	float value; unsigned int i; size_t l; string tmp;
	
	// copy the good part string in tmp
	tmp = "";
	for (i = 21; i < 29; i++) 
		if ( r[i] != ' ' ) tmp += r[i]; // copy only the interesting part 
	
	// tentative value
	if ( (l = tmp.find('(', 0 ) ) != string::npos ) 
		{ s.insert(s.begin(),'?'); tmp.erase(l,1); }
	if ( (l = tmp.find(')', 0 ) ) != string::npos ) 
		{ s.insert(s.begin(),'?'); tmp.erase(l,1); }
	
	if ( tmp.size() > 0 ) {	
		istringstream vstring(tmp); vstring >> value; 
	}
	else value = 0;
	
	return value;
}

float BaseENSDF::GetDRI(const char *r)
{
	float dvalue, precision; unsigned int i; string tmp1, tmp2;
	
	// copy the cc string in tm1 to look for the precision
	tmp1 = "";
	for (i = 21; i < 29; i++) 
		if ( r[i] != ' ' && r[i] != '(' && r[i] != ')' ) tmp1 += r[i]; 

	if ( tmp1.size() > 0 )	
		precision = GetPrecision(tmp1);
	else 
		precision = 0.0;	
	
	// copy the dcc string in tmp2 
	tmp2 = "";
	for (i = 29; i < 31; i++) 
		if ( r[i] != ' ' ) tmp2 += r[i];  
	
	if ( tmp2.size() == 0 || 
	     tmp2 == "LT" || tmp2 == "GT" || tmp2 == "LE" ||
	     tmp2 == "GE" || tmp2 == "AP" || tmp2 == "CA" || tmp2 == "SY" ) dvalue = 0;
	else {
		istringstream estring(tmp2); 	
		estring >> dvalue; dvalue *= precision;
	}
	return dvalue;
} 

float BaseENSDF::GetTI(const char *r, std::string &s)
{
	float value; unsigned int i; size_t l; string tmp;
	
	// copy the good part string in tmp
	tmp = "";
	for (i = 64; i < 74; i++) 
		if ( r[i] != ' ' ) tmp += r[i]; // copy only the interesting part 
	
	// tentative value
	if ( (l = tmp.find('(', 0 ) ) != string::npos ) 
		{ s.insert(s.begin(),'?'); tmp.erase(l,1); }
	if ( (l = tmp.find(')', 0 ) ) != string::npos ) 
		{ s.insert(s.begin(),'?'); tmp.erase(l,1); }

	if ( tmp.size() > 0 ) {	
		istringstream vstring(tmp); vstring >> value; 
	}
	else value = 0;		
	
	return value;
}

float BaseENSDF::GetDTI(const char *r)
{
	float dvalue, precision; unsigned int i; string tmp1, tmp2;
	
	// copy the cc string in tm1 to look for the precision
	tmp1 = "";
	for (i = 64; i < 74; i++) 
		if ( r[i] != ' ' && r[i] != '(' && r[i] != ')' ) tmp1 += r[i]; 

	if ( tmp1.size() > 0 )	
		precision = GetPrecision(tmp1);
	else 
		precision = 0.0;
		
	// copy the dcc string in tmp2 
	tmp2 = "";
	for (i = 74; i < 76; i++) 
		if ( r[i] != ' ' ) tmp2 += r[i];  
	
	if ( tmp2.size() == 0 || 
	     tmp2 == "LT" || tmp2 == "GT" || tmp2 == "LE" ||
	     tmp2 == "GE" || tmp2 == "AP" || tmp2 == "CA" || tmp2 == "SY" ) dvalue = 0;
	else {
		istringstream estring(tmp2); 	
		estring >> dvalue; dvalue *= precision;
	}
	return dvalue;
} 

float BaseENSDF::GetE(const char *r, std::string &offset)
{
	static char whichoffset[] = "XYZUVWABCDEFGHIJKLMNOPQRST";
	
	float energy; unsigned int i; size_t l; string tmp;
	
	// copy the E string in tmp
	tmp = "";
	for (i = 9; i < 19; i++) tmp += r[i]; // copy only the interesting part 

	offset = "  ";
	// check out if there is an offset
	for (i = 0; i < 26; i++) { 
		if ( (l = tmp.find(whichoffset[i],0 ) ) != string::npos ) 
			{ offset.insert(offset.begin(),whichoffset[i]); tmp.erase(l,1); break; }
	} 	
	// check out if it has been deduced or taken from other experiment
	if ( (l = tmp.find('(', 0 ) ) != string::npos ) 
		{ offset.insert(offset.begin()+1,'?'); tmp.erase(l,1); }
	if ( (l = tmp.find(')', 0 ) ) != string::npos ) 
		{ offset.insert(offset.begin()+1,'?'); tmp.erase(l,1); }
	// remove + in case of offset
	if ( (l = tmp.find('+', 0 ) ) != string::npos ) { tmp.erase(l,1); }		
	
	// to get the energy
	if ( tmp.size() > 0 ) {
		istringstream estring(tmp); estring >> energy; 
		// check out if it has been correctly read. if bad returns 0 
		if ( estring.fail() ) energy = 0.0;	
	}
	else energy = 0.0;

	return energy;
}

float BaseENSDF::GetPrecision(std::string st)
{
	float precision = 1.0; int expo; size_t l1,l2;
	
	l1 = st.find('.'); l2 = st.find('E'); if ( l2 == string::npos ) l2 = st.find('e');
	if ( l1 == string::npos ) { // no point
		if ( l2 == string::npos ) // no exponant
			precision = 1.0;
		else { // exponant
			st.erase(0,l2+1); 
			istringstream out(st); out >> expo; precision = pow(10.0,expo);
		}
	} else { // one point
		if ( l2 == string::npos ) { // no exponant
			int e = - (st.size() - l1 - 1); precision = pow(10.0,e);
		}
		else { // exponant
			st.erase(0,l2+1); 
			istringstream out(st); out >> expo;	
			
			int e = - (l2 - l1 - 1) + expo; precision = pow(10.0,e);	
		}
	}
	if ( precision < 0 ) precision = -1.0 * precision;
	
	return precision;	
}

float BaseENSDF::GetDE(const char *r)
{
	static char whichoffset[] = "XYZUVWABCDEFGHIJKLMNOPQRST";
	
	float denergy, precision; unsigned int i; size_t l; string tmp1, tmp2;
	
	// copy the E string in tmpe to look for the precision
	tmp1 = "";
	for (i = 9; i < 19; i++) 
		if ( r[i] != ' ' ) tmp1 += r[i]; 

	// check out if there is an offset
	for (i = 0; i < 26; i++) { 
		if ( (l = tmp1.find(whichoffset[i],0 ) ) != string::npos ) 
			{ tmp1.erase(l,1); break; }
	} 	
	// check out if it has been deduced or taken from other experiment
	if ( (l = tmp1.find('(', 0 ) ) != string::npos ) 
		{ tmp1.erase(l,1); }
	if ( (l = tmp1.find(')', 0 ) ) != string::npos ) 
		{ tmp1.erase(l,1); }
	// remove + in case of offset
	if ( (l = tmp1.find('+', 0 ) ) != string::npos ) { tmp1.erase(l,1); }	
				
	if ( tmp1.size() > 0 )	
		precision = GetPrecision(tmp1);
	else 
		precision = 0.0;	
			
	// copy the DE string in tmpde  
	tmp2 = "";
	for (i = 19; i < 21; i++) 
		if ( r[i] != ' ' ) tmp2 += r[i];  
	
	if ( tmp2.size() == 0 || 
	     tmp2 == "LT" || tmp2 == "GT" || tmp2 == "LE" ||
	     tmp2 == "GE" || tmp2 == "AP" || tmp2 == "CA" || tmp2 == "SY" ) denergy = 0;
	else {
		istringstream estring(tmp2); 
		estring >> denergy; denergy *= precision;
	}
	return denergy;
}

float BaseENSDF::GetT(const char *r, float &unit, std::string &option)
{	
	float t; unsigned int i; size_t l; string tmp;
	
	t = 0; 
	tmp = "";
	for (i = 39; i < 49; i++) tmp += r[i]; 
	
	option = "";
	if ( (l = tmp.find('?', 0 ) ) != string::npos ) { tmp.erase(l,1); option = "?"; }
		
	if ( tmp.find("STABLE") != string::npos ) { t = 3.2E23; unit = 1; }
	else {
		if ( tmp.find("Y")  != string::npos ) { unit = 31622400; }
		if ( tmp.find("D")  != string::npos ) { unit = 86400; }
		if ( tmp.find("H")  != string::npos ) { unit = 3600; }
		if ( tmp.find("M")  != string::npos ) { unit = 60; }
		if ( tmp.find("S")  != string::npos ) { unit = 1; }
		if ( tmp.find("MS") != string::npos ) { unit = 1E-3; }
		if ( tmp.find("US") != string::npos ) { unit = 1E-6; }	
		if ( tmp.find("NS") != string::npos ) { unit = 1E-9; }
		if ( tmp.find("PS") != string::npos ) { unit = 1E-12; }
		if ( tmp.find("FS") != string::npos ) { unit = 1E-15; }	
		if ( tmp.find("AS") != string::npos ) { unit = 1E-18; }
		
		if ( tmp.size() > 0 ) { 
			istringstream tmpstring(tmp); 
			tmpstring >> t; t *= unit;
			if ( tmpstring.fail() ) { option = "?"; t = 0.0; unit = 0.0; }				
		}
		else { t = 0.0; unit = 0.0; }
	}

	return t;
} 

float BaseENSDF::GetDT(const char *r)
{
	float dt, precision, unit; unsigned int i; size_t l; string tmp1, tmp2;
	
	// copy the T string in tmp1 to look for the precision
	unit = 1; tmp1 = "";	
	if ( BaseENSDF::GetT(r,unit,tmp1) == 3.2E23 ) return 0.0; // stable nucleus
	
	tmp1 = "";
	for (i = 39; i < 49; i++) {
		if ( r[i] != ' ' ) tmp1 += r[i]; 
	}
	if ( ( l = tmp1.find("MS")) != string::npos ) { tmp1.erase(l,2); }
	if ( ( l = tmp1.find("US")) != string::npos ) { tmp1.erase(l,2); }	
	if ( ( l = tmp1.find("NS")) != string::npos ) { tmp1.erase(l,2); }
	if ( ( l = tmp1.find("PS")) != string::npos ) { tmp1.erase(l,2); }
	if ( ( l = tmp1.find("FS")) != string::npos ) { tmp1.erase(l,2); }	
	if ( ( l = tmp1.find("AS")) != string::npos ) { tmp1.erase(l,2); }
	if ( ( l = tmp1.find("Y") ) != string::npos ) { tmp1.erase(l,1); }
	if ( ( l = tmp1.find("D") ) != string::npos ) { tmp1.erase(l,1); }
	if ( ( l = tmp1.find("H") ) != string::npos ) { tmp1.erase(l,1); }
	if ( ( l = tmp1.find("M") ) != string::npos ) { tmp1.erase(l,1); }
	if ( ( l = tmp1.find("S") ) != string::npos ) { tmp1.erase(l,1); }

	if ( (l = tmp1.find('?', 0 ) ) != string::npos ) { tmp1.erase(l,1); }

	if ( tmp1.size() > 0 )	
		precision = GetPrecision(tmp1);
	else 
		precision = 0.0;

		
	// copy the DT string in tmp2 
	tmp2 = ""; 
	for (i = 49; i < 55; i++) 
		if ( r[i] != ' ' ) tmp2 += r[i];  
	
	if ( tmp2.size() == 0 || 
	     tmp2 == "LT" || tmp2 == "GT" || tmp2 == "LE" ||
	     tmp2 == "GE" || tmp2 == "AP" || tmp2 == "CA" || tmp2 == "SY" ) dt = 0;
	else {
		istringstream estring(tmp2); estring >> dt; dt = dt * precision * unit; 
	}
	return dt;
}    

string BaseENSDF::GetJPI(const char *r)
{
	string tmp;
	for (int i = 21; i < 39; i++) 
		if ( r[i] != ' ' ) tmp += r[i];
		
	return tmp;	
} 







