/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file BaseENSDF.h base classe to interface ENSDF (Evaluated Nuclear Structure Data File) */  
 
#ifndef GW_BASEENSDF_H
#define GW_BASEENSDF_H

#include <vector>
#include <iostream>
#include <string>
#include <fstream>

namespace Gw {

/** General interface to ENSDF files.

Utility class to deal with ENSDF files. A ENSDF file is viewed as a collection of datasets. Each dataset is composed
of records, a record being a buffer of a fixed length (81 characters including a return character at the end).

Once a valid ENDSF file is open, a list of dataset keys is created for each data set. 
This keys described the data set using the fields coming from the identification record (see ENSDF doc):
It is possible then to check out if a given data set exists by using the IsDataSet member.
If it exists, any of its records could be read with BlankRecord, FirstRecord and NextRecord.

Some members allow to decode some records like:

IsRecord GetCC GetPrecision GetE GetDE GetT GetDT GetJPI ...
 
	\todo - For the moment, asymetric errors are not proposed for DT, DMR etc .. and when
	given in an ENSDF file, GetDT returns a wrong error.
	
	
@author Olivier Stezowski
*/
class BaseENSDF 
{   
private:
    std::ifstream Input;
    
protected:
    static std::string BLANK;
    static const int  RSIZE = 81;	// a record size is 81 bytes (80 data + \n)

    // structure that describes a dataset
    class DataSetKey {
    public:
	std::string Nuclide;
	std::string Dsid;
	std::string Dsref;
	std::string Pub;
	std::string Date;
	int CurrentPosition;
	std::pair<int,int> Position;
	
	DataSetKey()  { Clear() ; }
	DataSetKey(const DataSetKey &)  ; 
	~DataSetKey() { Clear() ; }	

	void Clear();
	void Print(std::ostream &) const  ; 	
    }; 
    std::vector<DataSetKey> fDataSets;
        
    std::string Dsid(const char *) const; 
    std::string Dsref(const char *) const;    
    std::string Pub(const char *) const;
    std::string Date(const char *) const;             
    std::string Nuclide(const char *) const;
        
public:
    BaseENSDF();
    virtual ~BaseENSDF();
    
    //!
    /*!
       To open a new BaseENSDF file. The previous open file is lost even if this fails.
    */
    virtual bool Open(const char *);
    virtual void Close();
    virtual bool IsOpen() { return Input.is_open(); }
    
    //! returns a proper empty record. It is the charge of the user to delete it
    /*!
    	It is strongly recommended to use this member to set a proper buffer length.
	Because many functions map the record to get the physical values, it is important to
	be sure that this class methods will not try to read out of the real buffer length. 
    */
    static char *BlankRecord() { return new char[RSIZE] ; }
    
    //! to get dataset's records
    /*!
       After being open, each dataset contained in this file can be read indepandently record by record.
       All readings are initialised on the first record. To rewind, use FirstRecord.
       As for FirstRecord, this member uses a number to identify which dataset is read. This unique number
       must be obtained with IsDataSet. It returns false if there is nothing to be read. 
    */     
    virtual bool NextRecord(unsigned int, char *);   
    
    //! to init the records reading
    /*!
      To get the first record of a dataset. The dataset number has been returned by IsDataSet. 
      You should used BlankRecord to properly initiate the buffer. It returns false if the record cannot be reached 
    */     
    virtual bool FirstRecord(unsigned int which, char *record);   
        
    //! look for a given data set
    /*!
       
       if the dataset doesn't exist in this file, it returns 0, otherwise it returns a number that identifies
       uniquely the dataset. This number is used to read the records that composed this data set with the NextRecord member. 
    */           
    virtual unsigned int IsDataSet(const char *nuclide, const char *dsid = "ADOPTED LEVELS, GAMMAS") const;
    
    //! check record's type.
    /*!
    	To test if a record (first parameter) is of a given type (second parameter)
    	Record's type are defined by names as summary in ENSDF documentation (p31). Recognized record's types are then:
	 - IDENT
	 - Q-VALUE
	 - LEVEL
	 - BETA
	 - EC
	 - ALPHA
	 - GAMMA
	 - END
	 
	 Ex: IsRecord(record, "LEVEL") returns true if the record is a level record
    */
    bool IsRecord(const char *, const char *) const;

    //! on a gamma record, to get the total conversion coefficient... check the record before with IsRecord!!
    float GetCC(const char *);     
    float GetDCC(const char *);  
       
    //! to get the relative intensity field of a gamma record
    /*!
	if the value is deduced or taken from another experiment, a question marked is returned
	in the second argument.
    */
    float GetRI(const char *, std::string &);  
    float GetDRI(const char *);          

    //! to get the relative total transition intensity field of a gamma record
    /*!
	if the value is deduced or taken gfrom another experiment, a question marked is returned
	in the second argument.
    */
    float GetTI(const char *, std::string &);  
    float GetDTI(const char *);       
        
    //! extract the precision for a given ENSDF data
    /*!
    	In ENSDF files, the uncertainty is a two characters field given in the standard form.
	'The standard numeric uncertainty denotes an uncertainty in the last significant figures'.
	
	Ex (from ENSDF doc):
	- NR=0.873 DNR=11 means 0.873[+/-]0.011
	- QP=2.3E6 DQP=10 means (2.3[+/-]1.0)E6
	
	Given the quantity (NR or QP ...) as a string, this member returns the coefficient that must 
	multiply the uncertainty field to get the true uncertainty.
    */
    float GetPrecision(std::string);
    
    //! to get the energy field of a record
    /*!
    	As specified in the ENSDF documentation, the energy could be defined with an offset which is a character in
	the following order: X, Y, Z, U, V, W, A, B ....
	If the energy field in the ENSDF file contains such a character, it is copied in the second argument at
	the first place in the string. If there is no shift, the first character is set to blank.
	If there are parentheses around this field (tentative value), 
	it is also written in the second argument as a question mark in the second character.
	
	\warning the case of resonance energies is not yet treated !
    */
    float GetE(const char *, std::string &);  
    float GetDE(const char *);  
    
    //! to get the T (half-life) field of a record.
    /*!
    	The unit is always second .
	ENSDF files provides T with a numerical value and an associated unit given as a string. This function
	returns T in second. The second arguments returns the ENSDF units in second. This is needed
	to set the proper DT value. As for the enrgy field, a question mark in the third argument means
	a tentative value.
	
	\todo - There is no translation for the moment if the unit given is eV
	
	PS: STABLE means 10E15 years
    */   
    float GetT(const char *, float &, std::string &);  
    float GetDT(const char *);      

    //! to get the multipolarity of a gamma level
    /*!
    */           
    // std::string GetM(const char *);  
 
    //! to get the mixing ratio of a gamma level
    /*!
    */           
    float GetMR(const char *);   
    float GetDMR(const char *);    
              
    //! to get the spin-parity field JPI: J
    /*!
    	Because the spin-parity field could have many forms, this member only returns a copy
	of the spin-parity field. Concret implementation that uses the informations must have
	there own functions to decode the field.
    */           
    std::string GetJPI(const char *);  
     
    virtual void ls(std::ostream &) const ; 
};

} // namespace
#endif
