/***************************************************************************
 *   Copyright (C) 2004-2006 by Olivier Stezowski & Christian Finck        *
 *   stezow(AT)ipnl.in2p3.fr, cfinck(AT)ires.in2p3.fr                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file BaseLevelSchemeReader.cpp compiled in libGWPHYSICS.so */

#ifndef Gw_BaseLevelSchemeReader
#include "BaseLevelSchemeReader.h"
#endif

#ifndef GW_NUCLEARLEVEL_H
#include "NuclearLevel.h"
#endif

#ifndef GW_XGAMMALINK_H
#include "XGammaLink.h"
#endif


using namespace Gw;

//__________________________________________________________
BaseLevelSchemeReader::BaseLevelSchemeReader(const char* filename, Option_t* opt) 
  : TObject(),
    fFileName(filename),
    fOptions(opt),
    fIsChecked(false),
    fLog("BaseLevelSchemeReader")
{ 
  //Default Constructor
  fBand.SetOwner(kTRUE); fLevel.SetOwner(kTRUE); 
  fLink.SetOwner(kTRUE); fText.SetOwner(kTRUE);
}

//__________________________________________________________
BaseLevelSchemeReader::~BaseLevelSchemeReader()
{
// default destructor
}

//__________________________________________________________
Int_t BaseLevelSchemeReader::Import(LevelScheme& levelScheme)
{
  // check before filling
  Int_t ok = 1;
  if ( !fIsChecked )
    ok = Check();
  
  // if check failed return
  if ( !ok )
    return 0;
  
  Fill(levelScheme);
  FillLabels(levelScheme);
  
  return ok;
}


//__________________________________________________________
void BaseLevelSchemeReader::Fill(LevelScheme& levelScheme)
{
  
  fLog.SetProcessMethod("Fill(LevelScheme& )");
  
  if ( !fIsChecked) {
    fLog << error <<   Form("File %s has not been read/checked before", fFileName.Data()) << dolog;
    return;
  }
  
  TObject *obj; 
  Cascade *cascade; GammaLink *gamma;
  TList cl; cl.SetOwner(kFALSE);	
  Int_t j, j1;
  // everything is right .. so now previous definition is cleared
  levelScheme.Clear();
  
  // now remove levels, links etc from temporary and put them in fLevels, fLinks ..
  fLog << debug << Form(" Deleting temporary levels") << nline;	
  for(j = 0; j < fLevel.GetEntries() ; j++ ){ //levels
    levelScheme.AddLevel((Level*)fLevel.At(j));
  }
  fLevel.SetOwner(kFALSE); fLevel.Clear(); // kFALSE cos fBand is owner
  
  fLog << debug << Form(" Deleting temporary links") << nline;	
  for(j = 0; j < fLink.GetEntries() ; j++ ){ // links
    levelScheme.AddLink((Link*)fLink.At(j));
  }
  fLink.SetOwner(kFALSE); fLink.Clear(); 
  
  fLog << debug << Form(" Deleting temporary texts") << nline;	  
  for( j = 0; j < fText.GetEntries() ; j++ ){ // texts
    levelScheme.AddText((TLatex*)fText.At(j));
  }
  fText.SetOwner(kFALSE); fText.Clear(); 
  
  fLog << debug << Form(" Deleting temporary cascades") << nline;	       
  for(j = 0; j < fBand.GetEntries(); j++ ){ // cascades
    
    // move all levels temporary stored in cascade into another collection
    cascade = (Cascade *)fBand.At(j);	

    for(j1 = 0; j1 < cascade->GetSize(); j1++) {
      obj = cascade->At(j1); cl.Add(obj); 
    }
    cascade->Clear();
    
    // now loop over all links and keep in the current cascade only those 
    // having their initial and final level in the corresponding band    
    for(j1 = 0; j1 < levelScheme.GetLinks().GetSize(); j1++){
      gamma = (GammaLink *)levelScheme.GetLinks().At(j1);
      if ( cl.FindObject(gamma->GetIL()) == NULL ) continue;
      if ( cl.FindObject(gamma->GetFL()) == NULL ) continue;
      cascade->Add(gamma);
    }
    cl.Clear();
    // add this cascade to the list of cascades
    levelScheme.AddCascade(cascade);
  }
  fBand.SetOwner(kFALSE); fBand.Clear();    
	
  fLog << dolog; 

}

