/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file Cascade.h header file for a Cascade */
 
#ifndef Gw_Cascade
#define Gw_Cascade
#define GW_CASCADE_H Gw_Cascade

#ifndef ROOT_TList
#include <TList.h>
#endif

#ifndef ROOT_TLatex
#include <TLatex.h>
#endif

#include "GwLogMessage.h"

class TGraphErrors;

namespace Gw {


/**
A cascade is a list of links

A cascade represents a list of links. It has a name and a type.
By default, the links doesn't belong to this cascade, so they are not deleted in the destructor.
It can be changed by calling the TList member SetOwner().

@author Olivier Stezowski
*/
class Cascade : public TList
{
  
public:
  
    enum EDefaultStyle { kKnown, kTentative };
    enum EDefaultColor { kDefault, kSelected };
    
    static Style_t GetDefaultStyle(EDefaultStyle d) { return fgDefaultStyle[d]; }
    static Color_t GetDefaultColor(EDefaultColor d) { return fgDefaultColor[d]; }
	static Float_t GetDefaultLabelSize()            { return fgDefaultLabelSize; }

    static void SetStylePalette(Style_t* pal)       { ::memcpy(fgDefaultStyle, pal, fgkMaxStyle*sizeof(Style_t)); }
    static void SetColorPalette(Color_t* pal)       { ::memcpy(fgDefaultColor, pal, fgkMaxColor*sizeof(Color_t)); }
	static void SetDefaultLabelSize(Float_t size)   { fgDefaultLabelSize = size; }

protected:
    TLatex fName;	// cascade's name
    TLatex fType;	// cascade's type
  
    EDefaultStyle fDefaultStyle; // line style
    EDefaultColor fDefaultColor; // line color
  
    Bool_t fIsSelected;          // graphically selected 
    Bool_t fIsVisible;           // flag to know whether the level is visible

    mutable LogMessage fLog;     //! log message
    
private:
  
    static const Int_t fgkMaxColor;
    static const Int_t fgkMaxStyle;
    
	static Float_t fgDefaultLabelSize;  // default text size
    static Style_t fgDefaultStyle[];    // 0 for known, 1 for tentative  
    static Color_t fgDefaultColor[];    // 0 for default, 1 for selected
  
private:
   
   Double_t fXName; // position in X for name
   Double_t fYName; // position in Y for name
   
public:
    Cascade();
    virtual ~Cascade();
    
    //! to get and modify the displayed properties of name and type
    virtual TLatex& GetLatex(Option_t *opt = "name");
    
    //! to get only the cascade's name
    /*!
    	Name is a TLatex object. To get more characteritics (like position), see the GetLatex member
    */
   
    virtual const Char_t* GetName() const ;
    //! to set only the cascade's name 
    /*!
    	Name is a TLatex object. To modify more characteritics (like position), see the GetLatex member
    */      
    virtual void SetName(const Char_t* name);
   
    // interface for using SetText from TLatex in graphical context 
    //! set cascade label
    virtual void SetLabels(const Char_t* name, const Char_t* type = "ND", Double_t x = 0, Double_t y = 0);  //*MENU* *ARGS={name=>fName,type=>fType,x=>fXName,y=>fYName}
   
    //! Get X for name
    virtual Double_t GetXName() const { return fName.GetX(); }
   
    //! Get Y for name
    virtual Double_t GetYName() const { return fName.GetY(); }

    //! to get only the cascade's type
    /*!
    	Type is a TLatex object. To get more characteritics (like position), see the GetLatex member
    */    
    virtual const Char_t* GetType() const ;
    
    //! to set only the cascade's type 
    /*!
    	Type is a TLatex object. To modify more characteritics (like position), see the GetLatex member
    */         
    virtual void SetType(const Char_t *);
    
    //! Draw/Paint methods
    virtual void Paint(Option_t* opt ="");
    virtual void Draw(Option_t* opt ="");
    
    //! Compute distance to object
    virtual Int_t DistancetoPrimitive(Int_t px, Int_t py);
  
    //! to execute an event for this cascade
    virtual void ExecuteEvent(Int_t event, Int_t px, Int_t py);
  
    //! Print out
    void ls(Option_t* opt = "") const; //*MENU*
  
    //! Selected link
    virtual void Selected() {fIsSelected = true;}
    
    //! unselected link
    virtual void DeSelected() {fIsSelected = false;}
    
    //! return selection
    virtual Bool_t IsSelected() {return fIsSelected;}
  
    //! Set cascade (un)visible, do not delete the objects, still present, just not drawn !
    virtual void SetVisible(Bool_t visu); //*TOGGLE* *ARGS={visu=>fIsVisible} *GETTER=IsVisible
    
    //! return visible
    virtual Bool_t IsVisible() {return fIsVisible;}
  
    //! return default style
    virtual EDefaultStyle GetDefaultStyle() {return fDefaultStyle;}
    
    //! return default color
    virtual EDefaultColor GetDefaultColor() {return fDefaultColor;}
    
    //! Set color label
    virtual void SetLabelColor(EDefaultColor color) 
                              { fName.SetTextColor(GetDefaultColor(color) ); 
                                fType.SetTextColor(GetDefaultColor(color) ); } 

    //! Compute J2
    TGraphErrors* DoJ2() const;
  
    //! Compute J1
    TGraphErrors* DoJ1() const;
 
    //! rootcint dictionary
    ClassDef(Cascade,3); // a cascade is a list of links
};
// inline members
inline void Cascade::SetName(const Char_t *name) { fName.SetTitle(name); }
inline void Cascade::SetType(const Char_t *type) { fType.SetTitle(type); }
inline const Char_t *Cascade::GetName() const { return fName.GetTitle(); }
inline const Char_t *Cascade::GetType() const { return fType.GetTitle(); }

}
#endif
