/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   Copyright (C) 2010 by Christian Finck                                 * 
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file CorrelatedSpace.h header file for all type of CorrelatedSpace */  

#ifndef Gw_CorrelatedSpace
#define Gw_CorrelatedSpace

#include "GwLogMessage.h"

#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

class TBrowser;
class TList;
class TH1D;

namespace Gw {

class CSFilter;

/** CorrelatedSpace base class that contains correlated informations 
 
*/
class CorrelatedSpace : public TNamed
{

protected:
  
  Int_t fDimension;           // dimension of the correlated space
  
  TList* fHistoTotList;       // list of histograms for total projection
  
  TList* fHistoList;          // list of histograms for projection
  
  TList* fHistoBkgList;       // list of histograms for bkg projection
  
  Int_t fNofCells;            // number of cells
    
  Int_t fStatus;              // status of the space
  
  mutable LogMessage fLog;    //! log message
  
private:
  
  static const Char_t* fgkDefaultName; //! name of the correlated space
  static const TString fgkFolderName;  //! name of the CS folder
  static        TList* fgCorrSpaces;   //! list of Correlated spaces

protected:
  
  virtual Int_t GetAxisNumber(Option_t* name) const = 0;

public:
  
  enum  {kDefined, kUndefined};
  
  //! Get default name for CS
  static const Char_t* GetDefaultName() { return fgkDefaultName; }
  
  //! Get folder name
  static const TString GetFolderName()  { return fgkFolderName; }
  
  //! Get correlated space lsit
  static TList* GetCSList() { return fgCorrSpaces; }
  
public:
  CorrelatedSpace();
  CorrelatedSpace(const char* name, const char* title);
  virtual ~CorrelatedSpace();
  
  //! fill correlated space
  virtual void Fill(const Double_t* /*Xn*/, const Double_t /*weight*/ = 1) = 0;
  
  //! project
  virtual void Project(CSFilter* /*filter*/, Option_t* /*axis*/)  = 0;
  
  //! project
  virtual void Project(Option_t* axis, Option_t* gateName = "Gate0");
  
  //! Set dimension
  virtual void SetDimension(Int_t dimension) = 0;
  
  //! Get dimension
  virtual Int_t GetDimension() const = 0;
  
  //! Get Bin
  virtual Int_t GetNbin(Int_t /*axis*/) const  = 0;
  
  //! Get cell content
  virtual Double_t GetCellContent(Int_t /*iCell*/)  = 0;
  
  //! Set cell content
  virtual void SetCellContent(Int_t /*iCell*/, Double_t /*weight*/)  = 0;
   
  //! Browser method
  virtual void Browse(TBrowser* b) = 0;
  
  //! Set bin for each axis
  virtual void SetBins(Int_t bin, Double_t min, Double_t max, Int_t axis = 0);
  
  //! Get number of cells
  Int_t GetNofCells() { return fNofCells; }
  
  //! Set number of cells
  void SetNofCells(Int_t nCells) { fNofCells = nCells; }
  
  //! Get histo
  virtual TH1D* GetHisto(Option_t* axis = "x");
  
  //! Get bkg histo
  virtual TH1D* GetHistoBkg(Option_t* axis = "x");

  //! Get histo total projection
  virtual TH1D* GetHistoTot(Option_t* axis = "x");
  
  //! Get histo total projection
  virtual TH1D* GetHistoTot(Option_t* axis = "x") const;

  //! Import another CS
  void Import(CorrelatedSpace* cs);
  
  ClassDef(CorrelatedSpace,3)  // Correlated Space
};
  
} // end namespace

#endif
