/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   Copyright (C) 2010 by Christian Finck                                 * 
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file CorrelatedSpaceLightTree3.h header file for all type of CorrelatedSpaceLightTree3 */  


#ifndef Gw_CorrelatedSpaceLightTree3
#define Gw_CorrelatedSpaceLightTree3

#include "GwLogMessage.h"

#ifndef Gw_SymCorrelatedSpace
#include "SymCorrelatedSpace.h"
#endif

#ifndef ROOT_TObject
#include "TObject.h"
#endif

#ifndef ROOT_TString
#include "TString.h"
#endif

class TTree;
class TFile;
class TH3;
class TH1;
class TBranch;

namespace Gw {
	
class CSFilter;

/** CorrelatedSpaceLightTree3 class that manages a 3D symmetrical correlated space stored in a tree
    Here we use a simple tree structure to store the 3 dimensionnal array in a short word (16 bits max size).
    This space is suited for large dimension of axis up tp 16k, but with maximum number of events around 1 million.
    Otherwise the performance starts to decreasse.
*/
class CorrelatedSpaceLightTree3 : public SymCorrelatedSpace
{

protected:
  
  TTree* fTree;              //! tree that contains the matrix
  TString fBranchName;       //  name of branch where the matrix is stored  
  Short_t fX[3];             //! array to store axis
  
  mutable LogMessage fLog;   //! log message

protected:

   void ProjectSimple(CSFilter* filter, Int_t idx);
   void ProjectCombined(CSFilter* filter, Int_t idx);
   
public:
  CorrelatedSpaceLightTree3(const char* name = CorrelatedSpace::GetDefaultName(), const char* title = "Tree");
  virtual ~CorrelatedSpaceLightTree3();
  
  //! Set address
  virtual TBranch* SetAddress();
  
  //! Set Branch
  virtual void SetBranch();
  
   //! Check if h3 already in tree
 // virtual void CheckH3InTree(TH3* h3);
    
  //! fill correlated space
  void Fill(const Double_t* Xn, const Double_t weight = 1.);
  
  //! fill random
  void FillRandom(TH1* h1, Int_t nEvents);
  
  //! fill correlated space from H3
  void FillFromH3(TH3* h3);
  
  //! Project
  void Project(CSFilter* filter, Option_t* axis = "X");

  //! Project
  virtual void Project(Option_t* axis, Option_t* gateName = "Gate0");
  
  //! Browser method
   virtual void Browse(TBrowser* b);
   
   //! Set branch name
  void SetBranchName(TString name = "cube") { fBranchName = name; }
  
  //! Get branch name
  const Char_t* GetBranchName() { return fBranchName.Data(); }
  
  //! Get tree
  void AssertTree();
  
  //! Get Tree name
  const Char_t* GetTreeName() const;
  
  //! Get Tree title
  const Char_t* GetTreeTitle() const;

  //! Reset Tree 
  void DeleteTree(Option_t* opt = "");
  
  //! Get Bin
  Int_t GetNbin(Int_t axis) const;
  
  //! Set cell content
  void SetCellContent(Int_t iCell, Double_t weight);
  
  //! Get cell content
  Double_t GetCellContent(Int_t iCell);  
  
  ClassDef(CorrelatedSpaceLightTree3,1)  // Correlated Space Tree in 3 dim
};
	

  
} // end namespace

#endif
