/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   Copyright (C) 2010 by Christian Finck                                 * 
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file CorrelatedSpaceTree2.h header file for all type of CorrelatedSpaceTree2 */  


#ifndef Gw_CorrelatedSpaceTree2
#define Gw_CorrelatedSpaceTree2

#include "GwLogMessage.h"

#ifndef Gw_SymCorrelatedSpace
#include "SymCorrelatedSpace.h"
#endif

#ifndef ROOT_TObject
#include "TObject.h"
#endif

#ifndef ROOT_TString
#include "TString.h"
#endif

class TTree;
class TFile;
class TH2;
class TH1;
class TBranch;

namespace Gw {
	
class CSFilter;

/** CorrelatedSpaceTree2 class that manages a 2D symmetrical correlated space stored in a tree
    In the entries of the tree are stored the dimension of the matrix. To retrieve the counts of a 
    given (i,j) point of the matrix, tree->GetEntry(i*dim + j), the value is store in member fCounts.
    This correlated space is suited for big space or while using more than one space since it do not 
    load the full space in memory. For small space (typically a 4k matrix of float) and big number of 
    events better use the correlated space using a TH2 root object (Gw::CorrelatedSpaceTH2).
*/
class CorrelatedSpaceTree2 : public SymCorrelatedSpace
{

protected:
  
  TTree* fTree;               //! tree that contains the matrix
  TString fBranchName;        // name of branch where the matrix is stored  
  
  mutable LogMessage fLog;    //! log message

public:
  CorrelatedSpaceTree2(const char* name = CorrelatedSpace::GetDefaultName(), const char* title = "Tree");
  virtual ~CorrelatedSpaceTree2();
  
  //! Set address
  virtual TBranch* SetAddress() = 0;
  
  //! Set Branch
  virtual void SetBranch() = 0;
  
  //! Fill histo
  virtual void FillHisto(Int_t j, TH1* h) = 0;
  
  //! fill counts from h2
  virtual void FillCounts(Int_t i, Int_t j, TH2* h2) = 0;
  
  //! fill counts from vector
  virtual void FillCounts(Int_t i, Int_t j, Int_t dimX, const Double_t* Xn, const Double_t weight = 1.) = 0;
  
  //! Set counts
  virtual void SetCounts(Double_t weight) = 0;
  
  //! Get counts
  virtual Double_t GetCounts() = 0;
  
  //! fill counts random
  virtual void FillCounts(Int_t j) = 0;
  
  //! Check if h2 already in tree
  virtual void CheckH2InTree(TH2* h2) = 0;
    
  //! fill correlated space
  void Fill(const Double_t* Xn, const Double_t weight = 1.);
  
  //! fill random
  void FillRandom(Int_t dim);
  
  //! fill correlated space from H2
  void FillFromH2(TH2* h2);
  
  //! Project
  void Project(CSFilter* filter, Option_t* axis = "X");

  //! Project
  virtual void Project(Option_t* axis, Option_t* gateName = "Gate0");
  
  //! Browser method
   virtual void Browse(TBrowser* b);
   
  //! Set branch name
  void SetBranchName(TString name = "mat1") { fBranchName = name; }
  
  //! Get branch name
  const Char_t* GetBranchName() { return fBranchName.Data(); }
  
  //! Get tree
  void AssertTree();
  
  //! Get Tree name
  const Char_t* GetTreeName() const;
  
  //! Get Tree title
  const Char_t* GetTreeTitle() const;

  //! Reset Tree 
  void DeleteTree(Option_t* opt = "");
  
  //! Get Bin
  Int_t GetNbin(Int_t axis) const;
  
  //! Set cell content
  void SetCellContent(Int_t iCell, Double_t weight);
  
  //! Get content
  Double_t GetContent(Int_t i, Int_t j);
  
  //! Get cell content
  Double_t GetCellContent(Int_t iCell);  
  
  ClassDef(CorrelatedSpaceTree2,1)  // Correlated Space Tree in 2 dim
};
	

  
} // end namespace

#endif
