/***************************************************************************
 *   Copyright (C) 2010 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef Gw_GCondition
#define Gw_GCondition


#include "TNamed.h"
#include "TObjArray.h"

#include <vector>
#include <utility>

class TCut;
class TCutG;

namespace Gw {

class Peak1D;
class PeakCreator;

//! Base for a Graphical condition on a spectrum (1D, 2D)
/*!
 
 A graphical condition is an helper class to use GW/ROOT Peak/TGcut to be used to apply conditions on a spectra.
 If Disabled (use Enable), the test is always passed while enable really test the input.
 As well it could be usd as a condition or an anti-condition using UseAsAnti. Default is condition
 
 Import/Export methods allow to Import/Export conditions in different formats.
  
*/
class GCondition : public TNamed 
{
public:
	static const Style_t gCoincStyle = 3004;
	static const Style_t gAnticStyle = 3005;
	
protected:
	TObjArray fListOfConditions; // list of TCut or TGCut
	
protected:
	Bool_t fEnable;         // switch on/off the cut
	Bool_t fExpectedResult; // used for coincidence/anti coincidence treatment
	//
	mutable Bool_t fHasPassed;                  // true if last test has been passed
	mutable std::vector<Bool_t> fHasPassed_;    // individual gate
	
public:	
	GCondition();
	GCondition(const char * name, const char * title);
    GCondition(const GCondition& p);
	virtual ~GCondition();
    
    //! to get one by one the individual list of conditions
    const TObjArray &GetListOfConditions()
    {
        return fListOfConditions;
    }
    
    // Root copy this to object
    virtual void Copy(TObject &o) const
    {
        ((GCondition &)o) = (*this);
    }
	
	//! Set var name
	virtual void SetVarName(const Char_t * /*var_name*/, const Char_t * /*which_var*/ = "X")
		{;}
	
	//! Browser means Draw
	virtual void Browse(TBrowser* /*b*/) 
	{
		Draw();
	}
	
	//! delete all object from fListOfConditions
	virtual void Delete(Option_t* option = ""); //*MENU*
    
    //! returns true if the condition is passed with success
	virtual Bool_t IsPassed(Double_t, Double_t = 0.0) const
	{
		return !fExpectedResult;
	}
	
	//! return true if last condition has been passed. Set to false at the beginning of IsPassed
	virtual Bool_t HasPassed() const
	{
		return fHasPassed;
	}
	virtual Bool_t HasPassed(UInt_t i) const
	{
		if ( i < fHasPassed_.size() )
			return fHasPassed_[i];
		
		return false;
	}
	
	//! toggle on/off the cut. if off returns always passed 
	virtual void Enable();  //*TOGGLE* *GETTER=IsEnable
	virtual Bool_t IsEnable() const
    {
        return fEnable;
    }
	//! to change gate into an anti gate
	virtual void UseAsAnti(Bool_t r = false);  //*TOGGLE* *GETTER=IsAnti
	virtual Bool_t IsAnti() const
	{ 
		return fExpectedResult == false; 
	} 
	
	
	virtual Short_t Import(const char * /*what*/ = "*", const char * /*from*/ = "gpad", Option_t * /*opt*/ = "+") 
	{
		return 0;
	}
	
	virtual Short_t Export(const char * /*what*/ = "*", const char * /*to*/ = "MyCuts", Option_t * /*opt*/ = "") const
	{
		return 0;
	}
	
    //! overwritte print method ... should be to gw log system
    virtual void	Print(Option_t* option = "") const; //*MENU*
    
	ClassDef(GCondition,1)  // a graphical cut (1d,2d)
};	

//! Base for a Graphical condition 1D
/*!

	a GCond1 is a serie of conditions [min,max]. The real Test is an OR of all individual tests
 
	The real implementation is a list of Peak1D that are used as gates.
	In case of Export to ROOT world, it is converted in TCut 
 
*/
class GCond1D : public GCondition
{
private:
	TString fVarX;
	
protected:
    // methods to convert peak to/from TCut
	TCut *PeakToCut(const Gw::Peak1D *) const;
    Gw::Peak1D * CutToPeak(TCut *);
    
public:
    // static method to convert a cut to a gate (gw::peak). Peak name is cut name while title is the variable on which this cut applys
    static Gw::Peak1D *TCutToPeak(TCut *);
   
public:
	GCond1D();
	GCond1D(const char * name, const char * title);
    GCond1D(const GCond1D &p);
	virtual ~GCond1D();
	
    // Root copy, this to o
    virtual void Copy(TObject &o) const
    {
        ((GCond1D &)o) = (*this);
    }
    
	//! Set var name
	virtual void SetVarName(const Char_t * var_name, const Char_t * /*which_var*/ = "X")
	{
		fVarX = var_name;
	}

	void Draw(Option_t* option = "");

	//! need to be redefined since Peak does not inherits from TAttFill
	virtual void UseAsAnti(Bool_t r = false);  //*TOGGLE* *GETTER=IsAnti
	
	//! Add a window to the list of conditions
	Gw::Peak1D *AddWindow(Double_t, Double_t, Option_t *opt = "+"); //*MENU*

	//! returns true if the condition is passed with success
	virtual Bool_t IsPassed(Double_t, Double_t  = 0.0) const;
	
	//! Import 1D conditions 
	/*!
		From gpad, look for Peak1D and import them 
		From gdir, look for TCut/Gw::Peak1D and import them
	 
		if what = "*", import all of them that are not already in the collection
		if opt = "+", does not reset before import (add mode)
	 
		return the number of imported conditions.
		
	*/
	virtual Short_t Import(const char *what = "*", const char *from = "gpad", Option_t *opt = "+"); //*MENU*
	
	//! Export 1D conditions
	/*!
		to gdir, export in gdirectory. option is empty or overwrite
		to 
	 */
	virtual Short_t Export(const char *what = "*", const char *to = "MyCuts", Option_t *opt = "") const; //*MENU*

	
	//! add to the current canvas the way to add graphically 1D gates (type h after in the canvas to get list of commands) 
	Gw::PeakCreator *WindowGCreator(const char *c_name = ""); //*MENU*
	
	ClassDef(GCond1D,1)  // a graphical cut 1d
	
};
		
//! Base for a Graphical condition 2D
/*!
 
	List of TCutG. they are not owned by this and thus not deleted, but can be removed of this.
	As any graphical ROOT cut, they are stored in the ListOfSpecials objects. 
	For security, it is better to try and avoid having two different cuts with the same name. 
 
 */
class GCond2D : public GCondition
{
private:
	TString fVarX;
	TString fVarY;
	
public:
	GCond2D();
	GCond2D(const char * name, const char * title);
    GCond2D(const GCond2D &p);
	virtual ~GCond2D();
    
    // Root copy
    virtual void Copy(TObject &o) const
    {
        ((GCond2D &)o) = (*this);
    }
    
	void Draw(Option_t* option = "");

	//! Set var name
	virtual void SetVarName(const Char_t * var_name, const Char_t * which_var = "X")
	{
		TString which(which_var); 
		if ( which == "X" )
			fVarX = var_name;
		if ( which == "Y" )
			fVarY = var_name;
	}
	
	//! Add a window to the list of conditions
	TCutG *AddWindow(Double_t xmin, Double_t ymin, Double_t xmax, Double_t ymax, Option_t *opt = "+"); //*MENU*
	
	//! returns true if the condition is passed with success
	virtual Bool_t IsPassed(Double_t, Double_t) const;

	//! Import 2D conditions 
	/*!
	 From pad, look for TCutG  
	 From gDirectory, look for TCut and import them
	 
	 if what = "*", import all of them that are not already in the collection, or the one named what
	 if opt = "+", does not reset before import (add mode)
	 
	 return the number of imported conditions  
	 
	 */	
	virtual Short_t Import(const char *what = "*", const char *from = "gpad", Option_t *opt = "+"); //*MENU*

	virtual Short_t Export(const char *what = "*", const char *to = "MyCuts", Option_t *opt = "") const; //*MENU*

	//! prepare current canvas to draw a TCutG.
	virtual void WindowGCreator(const char *c_name = ""); //*MENU*

	
	ClassDef(GCond2D,1)  // a graphical cut 2d
	
};
	
	
//! Link between the leaf, the condition and the histogram
/*!
	They are not owned by the user. It is then the taks of the user to delete them
*/
class GCondHandler
{
protected:
	GCondition *fCond;
public:
	GCondHandler(GCondition *cond) : fCond(cond) 
		{;}
	virtual ~GCondHandler()
		{;}		
	
	//! to get back the condition
	GCondition *GetGCondition()
	{
		return fCond;
	}
	
	virtual Bool_t Check()
	{
		return false;
	}
	virtual void Fill()
	{
		;
	}
};

	
} // namespace

#endif