#ifndef Gw_GLSPlayer
#define Gw_GLSPlayer

#ifndef Gw_CorrelatedSpace
#include "CorrelatedSpace.h"
#endif

#ifndef ROOT_TString
#include "TString.h"
#endif

#ifndef Gw_BaseLSPlayer
#include "BaseLSPlayer.h"
#endif

#ifndef Gw_NuclearLevel
#include "NuclearLevel.h"
#endif

#ifndef Gw_GammaLink
#include "GammaLink.h"
#endif

#ifndef Gw_Measure
#include "Measure.h"
#endif

namespace Gw {
  
class LevelScheme;
class CSManager;
/** Most of the methods relies on graphical approach for level scheme
 
 A level scheme could be read from a file (radware or ensdf) or a level could be built from scratch.
 
 Setting attributes as color style labels for cascade, level and link are available via menus. 
 
 Dedicated actions could be performed:
 Cascade::DoJ1(2): selected a cascade and the moment is calculated and plot onto a canvas.
 Link::DoGate: from selected links, draw gates onto a existing spectrum.
 
 Direct actions are possible via the LS menu:
 AddGroundLevel: put the ground level in an empty level scheme for a given (Z,A),
 AddLink: this method put a link on the top by selecting a link or insert a link by selecting two levels,
 AddLevel: this method put a level on the top of a selected level,
 AddFirstLevel: add the first level on a new cascade.
 
 The labels inherited from TLatex class are all editable.
 The objects as link, level and cascade have menus to set individually their attributes (colors, styles, names, etc...) by clicking.
 By default these objects could not be moved, SetMovable as true for moving in menu.
 
*/
class  GLSPlayer : public BaseLSPlayer {
  
public:
  GLSPlayer(LevelScheme* levelScheme);
  virtual ~ GLSPlayer();

  //! To draw this level scheme 
  virtual void Draw(Option_t *opt = "");
    
  //! Register correlated space
  virtual void RegisterCS(const Char_t* fileName, const Char_t* name = CorrelatedSpace::GetDefaultName() ); //*MENU={Hierarchy="CS/Register"}* *ARGS={name=>fCSDefaultName}*
  
  //! Show gate on spectrum
  virtual void ShowGate(Option_t* opt = "gate");  //*MENU={Hierarchy="CS/ShowGate"}*
  
  //! Collect gate on a histogram and add them to filter
  virtual Int_t CollectGate(Option_t* opt = "reset");  //*MENU={Hierarchy="CS/CollectGate"}*

  //! Draw projection
  virtual void DrawProj(Option_t* opt = "");  //*MENU={Hierarchy="CS/DrawProj"}*
  
  //! Draw projection total
  virtual void DrawProjTot(Option_t* opt = "");  //*MENU={Hierarchy="CS/DrawProjTot"}*

  //! Project with CSManger
  virtual void Project(Option_t* axis = "x");  //*MENU={Hierarchy="CS/Project"}*
  
  //! Make gate, project and drawing
  virtual void DoProject(Option_t* axis = "x", Option_t* gateName = "Gate", 
						 Option_t* optFilter = "reset", Option_t* optDraw = "");  //*MENU={Hierarchy="CS/DoProject"}*
  
  
  //! Select all objects
  virtual void Select(); //*MENU={Hierarchy="LS/Select"}*
  
  //! Deselect all objects
  virtual void DeSelect(); //*MENU={Hierarchy="LS/DeSelect"}*
  
  //! Add link with selected link/level
  virtual void AddLink(Float_t strength = 100,  Float_t errStrength = 1, 
                       Float_t energy   = 300, Float_t errEnergy   = 1,
                       const Char_t* mult = "E2" ); // *MENU={Hierarchy="LS/AddLink"}*

  //! Save level scheme
  virtual void WriteLS(const char* name, Int_t write = TObject::kWriteDelete); //*MENU={Hierarchy="LS/Write"}* *GETTER=GetLSName* *ARGS={name=>fLSName}*
  
  //! Add first level to a new cascade beside the current cascade
  virtual void AddFirstLevel(Float_t energy = 300, Float_t errEnergy = 1, const Char_t* jp = "0+", Option_t* opt = "left");  // *MENU={Hierarchy="LS/AddFirstLevel"}* 
    
  //! Add new level to the current cascade
  virtual void AddLevel(Float_t energy = 300, Float_t errEnergy = 1, const Char_t* dJp = "0+");  // *MENU={Hierarchy="LS/AddLevel"}* 
  
  //! Add ground state level
  virtual void AddGroundLevel(Int_t z = 66, Int_t a = 152, Option_t* opt = "0+", Float_t energy = 0, Float_t errEnergy = 1); //*MENU={Hierarchy="LS/AddGroundLevel"}*
  
  //! Add arrows to the LS between two selected levels
  virtual void AddArrow(Option_t* opt = "|>");  // *MENU={Hierarchy="LS/AddArrow"}* 
  
  //! Add text to the LS at last mouse position
  virtual void AddText(const Char_t* msg);  // *MENU={Hierarchy="LS/AddText"}* 
  
  
  //! Make Cascade from selected links
  virtual void MakeCascade(const Char_t* label = "New");  //*MENU={Hierarchy="Cascade/MakeCascade"}*
  
  //! Do J1
  virtual void DoJ1(Option_t* opt = "AP") const;  //*MENU={Hierarchy="Cascade/DoJ1"}*
    
  //! Do J2
  virtual void DoJ2(Option_t* opt = "AP") const;  //*MENU={Hierarchy="Cascade/DoJ2"}*
  
  //! set color palette for level
  virtual void SetCascadeColorPalette(Color_t def, Color_t selected);  //*MENU={Hierarchy="Cascade/SetColorPalette"}*  *ARGS={def=>fCascadeDefColor,selected=>fCascadeSelColor}
  
  //! Align selected cascade's levels/links respect to the first selected in X 
  virtual void AlignCascade();  //*MENU={Hierarchy="Cascade/Align"}*
  
  //! Add existing link to cascade
  virtual void AddLinkToCascade();  //*MENU={Hierarchy="Cascade/AddLink"}*
  
  //! Move all levels/link (extra line) from a casacde (excluded last level if lastLevel = false)
  virtual void MoveCascade(Float_t mv = 200, Bool_t lastLevel = true);  //*MENU={Hierarchy="Cascade/Move"}*
  

  //! Align selected links on level posistions in X
  virtual void AlignLinks();  //*MENU={Hierarchy="Link/Align"}*
  
  //! set color palette for link
  virtual void SetLinkDefaultColor(Color_t col);  //*MENU={Hierarchy="Link/SetLinkDefaultColor"}* *ARGS={col>fLinkColor}

  //! Disable/Enable move of links
  virtual void   SetLinkMove(Bool_t move) { move ? Link::EnableMove() :  Link::DisableMove();} //*MENU={Hierarchy="Link/SetMovable"}* *TOGGLE* *GETTER=IsLinkMovable
  virtual Bool_t IsLinkMovable() { return Link::IsMovable(); }
  
  //! Set vis label for levels
  virtual void SetLinkVisLabel(const char* s = "000");  //*MENU={Hierarchy="Link/SetVisLabel"}* 
  
  //! Set label size for levels
  virtual void SetLinkLabelsSize(Float_t size);  //*MENU={Hierarchy="Link/SetLabelSize"}*
  
  //! Set label color for levels
  virtual void SetLinkLabelsColor(Color_t color);  //*MENU={Hierarchy="Link/SetLabelColor"}*
  
  //! Delete current link
  virtual void DeleteLink(); //*MENU={Hierarchy="Link/Delete"}*
  
  
  //! set color palette for level
  virtual void SetLevelColorPalette(Color_t plus, Color_t minus, 
                                    Color_t unknown, Color_t selected);  //*MENU={Hierarchy="Level/SetColorPalette"}*  *ARGS={plus=>fLevelPlusColor,minus=>fLevelMinusColor,unknown=>fLevelUnknownColor,selected=>fLevelSelColor}
  
  //! set style palette for level
  virtual void SetLevelStylePalette(Color_t known, Color_t tentative);  //*MENU={Hierarchy="Level/SetStylePalette"}*  *ARGS={known=>fLevelKnownStyle,tentative=>fLevelTentStyle}
  
  //! Disable/Enable move of levels
  virtual void   SetLevelMove(Bool_t move) { move ? Level::EnableMove() :  Level::DisableMove();} //*MENU={Hierarchy="Level/SetMovable"}* *TOGGLE* *GETTER=IsLevelMovable
  virtual Bool_t IsLevelMovable() { return Level::IsMovable(); }
  
  //! Set vis label for levels
  virtual void SetLevelVisLabel(const char* s = "0000");  //*MENU={Hierarchy="Level/SetVisLabel"}* 
  
  //! Set label size for levels
  virtual void SetLevelLabelsSize(Float_t size);  //*MENU={Hierarchy="Level/SetLabelSize"}*
  
  //! Set label color for levels
  virtual void SetLevelLabelsColor(Color_t color);  //*MENU={Hierarchy="Level/SetLabelColor"}*
  
  //! Align selected levels respect to the first selected in X (does not affect the associated links)
  virtual void AlignLevels();  //*MENU={Hierarchy="Level/Align"}*
  
  //! Delete current level
  virtual void DeleteLevel(); //*MENU={Hierarchy="Level/Delete"}*
  

  //! Popup DoProject menu
  virtual void PopupDoProject(); 
  
  //! Popup AddLink menu
  virtual void PopupAddLink(); 
  
  //! Popup AddLevel menu
  virtual void PopupAddLevel(); 
  
  //! Popup AddFirstLevel menu
  virtual void PopupAddFirstLevel();
   
  //! Popup AddGroundLevel menu
  virtual void PopupAddGroundLevel();
  
  //! Popup write LS menu
  virtual void PopupWriteLS();
  
  //! Popup AddArrow menu
  virtual void PopupAddArrow();
  
  //! Popup AddText menu
  virtual void PopupAddText();
  
  //! Add gamma link on the top of a band by  selected a link or a link
  virtual void AddLinkOnTop(Measure<Float_t> strength, Measure<Float_t> energy);
  
  //! insert a gamma link between two seleted levels
  virtual void InsertLink(Measure<Float_t> strength);
  
  Color_t GetCascadeDefColor()   { return Cascade::GetDefaultColor(Cascade::kDefault); }
  Color_t GetCascadeSelColor()   { return Cascade::GetDefaultColor(Cascade::kSelected); }

  Color_t GetLevelKnownStyle()   { return Level::GetDefaultStyle(Level::kKnown); }
  Color_t GetLevelTentStyle()    { return Level::GetDefaultStyle(Level::kTentative); }

  Color_t GetLevelPlusColor()    { return Level::GetDefaultColor(Level::kPlus); }
  Color_t GetLevelMinusColor()   { return Level::GetDefaultColor(Level::kMinus); }
  Color_t GetLevelUnknownColor() { return Level::GetDefaultColor(Level::kUnknown); }
  Color_t GetLevelSelColor()     { return Level::GetDefaultColor(Level::kSelected); }
  
  Float_t GetLevelLabelsSize()   { return Level::GetDefaultLabelSize(); }
  Float_t GetLevelLabelsColor()  { return Level::GetDefaultLabelColor(); }
  
  Color_t GetLinkColor()         { return Link::GetDefaultColor(); }

  const char* GetLSName()        { return fLevelScheme->GetName(); }
  
  const char* GetCSDefaultName() { return CorrelatedSpace::GetDefaultName(); }
  
protected:
  
  //! Add gamma link on the top of any selected link (under progress) 
  virtual void AddLinkBetween(Float_t strength = 10, Float_t energy = 300, 
                              Float_t errStrength = 1, Float_t errEnergy = 1); 
  
  //! Add gamma link on the top of the selected link
  virtual void AddLinkOnTop(Float_t strength = 10, Float_t energy = 300, 
                            Float_t errStrength = 1, Float_t errEnergy = 1,
                            const Char_t* mult = "E1"); 
  
  //! insert a gamma link between two seleted levels
  virtual void InsertLink(Float_t strength = 10, Float_t errStrength = 1, const Char_t* mult = "E1"); 
  
  //! update drawing options color/style
  virtual void UpdateDrawingOptions();
  
  //! Check multi-polarity
  virtual Bool_t CheckPolarity(GammaLink& link);
  
private:
  
  Color_t fCascadeDefColor;
  Color_t fCascadeSelColor;
  
  Style_t fLevelKnownStyle;
  Style_t fLevelTentStyle;
  
  Color_t fLevelPlusColor;
  Color_t fLevelMinusColor;
  Color_t fLevelUnknownColor;
  Color_t fLevelSelColor;
  
  Float_t fLevelLabelsSize;
  Color_t fLevelLabelsColor;
  
  Color_t fLinkColor;
 
  TString fLSName;        // level scheme name
  
  Char_t* fCSDefaultName; //  default CS name
  
  CSManager* fCSManager;  // Correlated space manager
   
  ClassDef( GLSPlayer, 2);  // some standard/graphical approach
};

} // end namespace
#endif
