/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file GammaLink.h header file for a GammaLink */

#ifndef Gw_GammaLink
#define Gw_GammaLink
#define GW_GAMMALINK_H Gw_GammaLink

#ifndef Gw_link
#include "Link.h"
#endif
#ifndef GW_DATA_H
#include "Data.h"
#endif
#ifndef GW_MEASURE_H
#include "Measure.h"
#endif

namespace Gw {


/**
A GammaLink binds two nuclear levels

A GammaLink gives the informations concerning the gamma-ray that binds two nuclear levels. 
The Link strength is the gamma-ray intensity with the associated intensity. It has also an energy (Measure with its error) and a multipolarity.  
Some additionnal informations are available like a mixing ratio and a conversion coefficient.

@author Olivier Stezowski
*/
class GammaLink : public Link
{

private:
    static Int_t fgMaxAttributes; // maximum number of attributes for a gamma link

private:
    TString fAttr0;             // attribute for energy
    TString fAttr1;             // attribute for multipolarity
    TString fAttr2;             // attribute for mixing
    TString fAttr3;             // attribute for conversion
    TString fAttr4;             // attribute for strength
    TString fAttr5;             // attribute for tau
    TString fAttrLabel;	        // To know the attribute for each label 

protected:
    Data<Char_t>   fEM;		// Electric/Magnetic transition
    Data<UShort_t> fLambda;	// numeric part of the multipolarity
        
    Measure<Float_t> fEnergy;	// gamma ray energy
    Measure<Float_t> fMixing;	// mixing ratio
    Measure<Float_t> fConv;	// conversion

public:
    GammaLink();
    GammaLink(const GammaLink &);
    GammaLink& operator=(const GammaLink &);
    virtual ~GammaLink();

    //! to get/modify the mixing ratio and its error 
    virtual Measure<Float_t> & GetMixing() { return fMixing; }

    //! to get/modify the gamma-ray energy and its error 
    virtual Measure<Float_t> & GetEnergy() { return fEnergy; }

    //! to get/modify the conversion coefficient and its error
    virtual Measure<Float_t> & GetConversion() { return fConv; }
    
    //! to set the multipolarity 
    /*!
      The input string is decomposed to init fEM and fLambda.
      Ex: E2 will give fEM = E, fLambda = 2. 
      Default is E2.
      If something is wrong, the Unknown flag will be set for the wrong part(s)
    */
    virtual void SetEM(const char *em= "E2"); 
    
    //! true if this transition is electric 
    virtual bool IsE() const { return fEM.Get() == 'E'; }
    //! true if this transition is magnetic   
    virtual bool IsM() const { return fEM.Get() == 'M'; }
    //! returns the numeric part of the multipolarity    
    virtual UShort_t GetLambda() const { return fLambda.Get() ; }
    
    //! it adds this gamma to the given cascade   
    //virtual void DoCascade(Cascade &cas, Option_t *opt = "") ;
    virtual void DoCascade(TSeqCollection &col, Option_t *opt = "");

    virtual void ls(Option_t* option = "") const ;      
      
    void SetGamma(const char* energy,   const char* multipolarity,
                  const char* mixing,   const char* conversion,
                  const char* strength, const char* tau); //*MENU* *ARGS={energy=>fAttr0,multipolarity=>fAttr1,mixing=>fAttr2,conversion=>fAttr3,strength=>fAttr4,tau=>fAttr5}

    const char* GetAttr(Int_t idx);  
    const char* GetAttr0() { return Form("%7.1f %4.1f", fEnergy.Get(), fEnergy.GetError()); }   
    const char* GetAttr1() { return Form("%c%d",        fEM.Get(), fLambda.Get()); }   
    const char* GetAttr2() { return Form("%7.1f %4.1f", fMixing.Get(), fMixing.GetError()); }   
    const char* GetAttr3() { return Form("%7.1f %4.1f", fConv.Get(), fConv.GetError()); }   
    const char* GetAttr4() { return Form("%7.1f %4.1f", GetStrength().Get(), GetStrength().GetError()); }   
    const char* GetAttr5() { return Form("%7.1f %4.1f", GetTau().Get(), GetTau().GetError()); }   

    //! Used to determine the attribute for each label   
    /*!    
     
     The number are respectively for: 0->Energy, 1->EM, 2->Mixing, 3->Conv, 4->Strenght, 4->Tau.
     There are 3 labels associated with this link. The string given as argument helps determining which attributes belongs to which label
     For instance, "045" (which is the default value) sets label 0 to the energy, label 1 to strength and label 2 to conversion value.
     */    
    virtual void  SetAttrLabel(const char *);
  
    //!  return attribute pattern
    virtual const char * GetAttrLabel() { return fAttrLabel.Data(); }
    
    //! Get labels title
    virtual const char* GetLabel0();  
    virtual const char* GetLabel1();   
    virtual const char* GetLabel2();  
  
    //! rootcint dictionary
    ClassDef(GammaLink,1); // A GammaLink binds two nuclear levels 
};
// inline members

}
#endif
