/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** \file Level.h header file for a general Level */  
 
#ifndef Gw_Level
#define Gw_Level
#define GW_LEVEL_H Gw_Level

#ifndef ROOT_TLine
#include <TLine.h>
#endif

#ifndef ROOT_TLatex
#include <TLatex.h>
#endif

#include "GwLogMessage.h"

namespace Gw {

/**
Base class describing a general level. 

A level has a graphical representation which is just a 2D line (TLine) \n
It also defines 4 differents labels (ROOT TLatex), one for each corner \n

The labels could be hidden by using the SetVisLabel method which is available in the root graphical menu.
It consists of a string of four characters. If the first character is equal to 0 then the first label is hidden 
otherwise it is shown. 

\image html level_label.gif
 
@author Olivier Stezowski
*/
class Level : public TLine 
{
  
  
public:
  enum EDefaultStyle { kKnown, kTentative };
  enum EDefaultColor { kPlus, kMinus, kUnknown, kSelected };
  
  static Style_t GetDefaultStyle(EDefaultStyle d) { return fgDefaultStyle[d]; }
  static Color_t GetDefaultColor(EDefaultColor d) { return fgDefaultColor[d]; }
  static Float_t GetDefaultLabelSize()  { return fgDefaultLabelSize; }
  static Color_t GetDefaultLabelColor() { return fgDefaultLabelColor; }

  static void SetStylePalette(Style_t* pal) { ::memcpy(fgDefaultStyle, pal, fgkMaxStyle*sizeof(Style_t)); }
  static void SetColorPalette(Color_t* pal) { ::memcpy(fgDefaultColor, pal, fgkMaxColor*sizeof(Color_t)); }
  static void SetDefaultLabelSize(Float_t size)   { fgDefaultLabelSize = size; }
  static void SetDefaultLabelColor(Color_t color) { fgDefaultLabelColor = color; }
  static void EnableMove()  { fgMovable = true; }
  static void DisableMove() { fgMovable = false; }
  static Bool_t IsMovable() { return fgMovable == true; }
  
private:
    static const Short_t fgkMaxLevelLabel;
    static const Int_t fgkMaxColor;
    static const Int_t fgkMaxStyle;

    static Style_t fgDefaultStyle[];    // 0 for known, 1 for tentative  
    static Color_t fgDefaultColor[];    // 0 for positive parity, 1 for negative and 2 for parity unknown , 3 for selected
    static Color_t fgDefaultLabelColor; // default text color
    static Float_t fgDefaultLabelSize;  // default text size
    static Bool_t  fgMovable;           // graphically movable

private:
    TString fVisLabel;	// To know if a label should be drawn ok not
    TString fPosLabel;	// To know the position of label 0:ul; 1:ur; 2:lr; 3:ll

    TLatex  fLabel0;	// Some labels associated with this level    
    TLatex  fLabel1;	// Some labels associated with this level    
    TLatex  fLabel2;	// Some labels associated with this level    
    TLatex  fLabel3;	// Some labels associated with this level    
    
    TLine   fExtra;	    // Extra line for links far away from this level
    Bool_t  fHasExtra;  // flag to know whether the level has an extra line 
        
    EDefaultStyle fDefaultStyle; // line style
    EDefaultColor fDefaultColor; // line color
  
    mutable LogMessage fLog;    //! log message

protected:
  
    Bool_t fIsSelected;  // graphically selected 
    Bool_t fIsVisible;   // flag to know whether the level is visible
	Bool_t fIsModified;  // flag to know whether the link has been modified
	Float_t fDY;         // relative displacement when modified

public:
    Level();
    virtual ~Level();
    
    //! extra line 
    /*!
    	An additional line is drawn for links connected to this level but far away from this level in the level scheme.
    */
    void SetExtraLine(Double_t, Double_t);
    void SetExtraLineX1(Double_t);
    void SetExtraLineX2(Double_t);
    
    //! Used to determine which label should be drawn
    /*!    
    	There are four labels associated with this level. The string given as argument helps determining if a label is drawn.
	For instance, "0000" (which is the default value) correspond to no label while for "1010" only the first and the third
	ones are drawn.
	This method is available through the graphical menu (right click) in a pad.
    */    
    virtual void SetVisLabel(const char *); //*MENU*
        
    //! set label titles
    virtual void SetLabels(const char *l0, const char *l1, const char *l2, const char *l3) ; // *MENU* *ARGS={l0=>fLabel0,l1=>fLabel1,l2=>fLabel2,l3=>fLabel3}

    //! Set extraline drawning
    virtual void ExtraLine(Bool_t extra = true); //*TOGGLE* *GETTER=HasExtraLine

    //! set all labels size
    virtual void SetLabelsSize(Float_t size);
  
    //! set all labels color
    virtual void SetLabelsColor(Color_t color);
    
    //! set position for levels
    virtual void SetPosLabel(const char *);
  
    //! Used to determine which label should be drawn    
    virtual const char * GetVisLabel() { return fVisLabel.Data(); }
            
    //! Used to determine the position of labels   
    virtual const char * GetPosLabel() { return fPosLabel.Data(); }

    //! All labels are shifted when moving a level to another position     
    virtual TLatex& GetLabel(Int_t);
  
    //! Get extra line
    virtual TLine& GetExtraLine() {return fExtra;}

    //! Get labels title
    virtual const char* GetLabel0() { return fLabel0.GetTitle(); }   
    virtual const char* GetLabel1() { return fLabel1.GetTitle(); }   
    virtual const char* GetLabel2() { return fLabel2.GetTitle(); }   
    virtual const char* GetLabel3() { return fLabel3.GetTitle(); }       
        
    //! to set labels         
    /*!
    
    */
    virtual void SetLabel(const char *, Int_t);
  
    //! Set default labels
    virtual void SetDefaultLabels();
  
    //! All labels are shifted when moving a level to another position 
    virtual void SetX1(Double_t);
    //! All labels are shifted when moving a level to another position    
    virtual void SetX2(Double_t);
    //! All labels are shifted when moving a level to another position       
    virtual void SetY1(Double_t);
    //! All labels are shifted when moving a level to another position           
    virtual void SetY2(Double_t);
    
    //! used to sort a list of levels
    /*!
      The comparison is done using the vertical position of this level \n
      Because Level overwrites TLine::SetY1 and TLine::SetY2, Y1 is always equal to Y2.
    
    \todo Warnings during compilation
    */
    virtual Int_t Compare(const TObject *) const;  
  
    //! Set Sortable true 
    virtual Bool_t IsSortable() const { return true; }  

    //! used to paint a level
    /*!
    	The labels associated with this level are painted only if there visible. To hide/show labels, see SetVisLabel
    */
    virtual void Paint(Option_t* option = "");
    virtual void Draw(Option_t* option = "");
    
    //! to execute an event for this level
    virtual void ExecuteEvent(Int_t event, Int_t px, Int_t py);
    
    //! Selected link
    virtual void Selected() {fIsSelected = true;}
    
    //! unselected link
    virtual void DeSelected() {fIsSelected = false;}
    
    //! return selection
    virtual Bool_t IsSelected() {return fIsSelected;}
  
    //! Set link visible
    virtual void SetVisible() {fIsVisible = true;}
    
    //! Set link unvisble
    virtual void SetUnvisible() {fIsVisible = false;}
    
    //! return visible
    virtual Bool_t IsVisible() {return fIsVisible;}

	//! Set link modified
	virtual void SetModified() {fIsModified = true;}
	
	//! Set link unvisble
	virtual void SetUnmdified() {fIsModified = false;}
	
	//! return modified
	virtual Bool_t IsModified() {return fIsModified;}
  
	// Set DY
	virtual void SetDY(Float_t dy) { fDY = dy; }
  
	// Get DY
	virtual Float_t GetDY() { return fDY; }
  
    //! return default style
    virtual EDefaultStyle GetDefaultStyle() {return fDefaultStyle;}
   
    //! return default color
    virtual EDefaultColor GetDefaultColor() {return fDefaultColor;}
        
    //! Set style
    virtual void SetStyle(EDefaultStyle style);
  
     //! Set color
    virtual void SetColor(EDefaultColor color);
  
    //! check if extra line
    virtual Bool_t HasExtraLine() { return fHasExtra; }
  
    //! Delete, redifined to avoid appearence in Menu
    void Delete(const Option_t* /*opt*/ = "") { return; }
    
    //! Root dictionary related
    ClassDef(Level, 4); // base class for a level
};

}

#endif
