/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file LevelScheme.h header file for a LevelScheme */
 
#ifndef GW_LEVELSCHEME_H
#define GW_LEVELSCHEME_H
#define Gw_LevelScheme GW_LEVELSCHEME_H

#ifndef ROOT_TBits
#include "TBits.h"
#endif

#ifndef ROOT_TObject
#include <TObject.h>
#endif

#ifndef ROOT_TList
#include <TList.h>
#endif

#ifndef ROOT_TLatex
#include <TLatex.h>
#endif

#include "GwLogMessage.h"

#ifndef GW_LEVEL_H
#include "Level.h"
#endif

#ifndef GW_LINK_H
#include "Link.h"
#endif

#ifndef GW_CASCADE_H
#include "Cascade.h"
#endif

#include <map>

class TArrow;
namespace Gw {

  class BaseLevelSchemeReader;
  class BaseLSPlayer;
/**
A level Scheme.

A level sheme is more general and complex that a Cascade. It's composed of a list of levels, a list of links and a list of cascades.
The LevelScheme is the true owner of its links, levels and cascades. They will be deleted
by the destructor.\n
A level scheme has a name, a type and a reference to keep some informations
about the origin of the level scheme. An additional list of TLatex is available to display some comments.
 
A Cascade is a particular path in the level scheme. With the level sheme the notion of current
cascade exists. When adding new links, they are added to the current cascade unless it
is null. 
 
\todo copy constructor - some functions to modify an Level Scheme

@author Olivier Stezowski
*/
  
class LevelScheme : public TObject
{
  friend class BaseLevelSchemeReader;
  friend class BaseLSPlayer;
    
private:
    
    Link*     fCLink;    //! pointer to current link
    Level*    fCLevel;   //! pointer to current level
    Cascade*  fCCascade; //! pointer to current cascade
  
    std::map <Int_t, TBits*> fCoincMap; // map of gamma coincidence

  
protected:
    TList fLevels;	   // list of levels
    TList fLinks;	   // list of links
    TList fCascades;   // list of cascades
    
    TLatex fName;	    // LevelScheme's name
    TLatex fReference;	// some information concerning the origin of this level scheme
	TList  fTexts;	    // a list of additional TLatex text to add some comments on the level scheme
    TList  fArrows;	    // a list of additional TArrow  to add some arrows on the level scheme

    BaseLSPlayer* fPlayer; //! level scheme player pointer
  
protected:
  
  //! Add level
  void AddLevel(Level* level);

  //! Add link
  void AddLink(Link* link);
  
  //! Add cascade
  void AddCascade(Cascade* Cascade);
  
  //! Add text
  void AddText(TLatex* latex);
  
  //! Add arrow
  void AddArrow(TArrow* arrow);

  //! to get the list of levels
  TList& GetLevelList()   { return fLevels; }
  //! to get the list of links
  TList& GetLinkList()    { return fLinks; }
  //! to get the list of cascades
  TList& GetCascadeList() { return fCascades; }
  //! to get the list of additional comments
  TList& GetTextList()   { return fTexts; }
  //! get current link
  Link* GetCurrentLink() { return fCLink; } 
  //! get current level
  Level* GetCurrentLevel() { return fCLevel; } 
  //! get current cascade
  Cascade* GetCurrentCascade() { return fCCascade; } 
  //! to get Name latex object
  TLatex& GetNameLatex() { return fName; }
  //! to get ref latex object
  TLatex& GetRefLatex() { return fReference; }

  //! Fill map of coincidence
  virtual void FillMap();
  
public:
    LevelScheme();
    LevelScheme(const Char_t *, Option_t *);
    virtual ~LevelScheme();
    
    virtual void Browse(TBrowser *b);
    
    //! set level scheme name.
    virtual void SetName(const Char_t* name); //*MENU* *ARGS={name=>fName}
    virtual const char* GetName() const; 
    //! set level scheme reference. 
    virtual void SetReference(const Char_t *);
    virtual const Char_t * GetReference() const;
  
    //! create new level
    Level* NewLevel(const char* name);
    
    //! create new level
    Link* NewLink(const char* name);
    
    //! create new cascade
    Cascade* NewCascade(const char* name);
    
    //! create new text
    TLatex* NewText(const char* msg);
  
	//! create new arrow
	TArrow* NewArrow(Float_t x1, Float_t y1, Float_t x2, Float_t y2);
  
    //! to get the list of levels
    const TList & GetLevels();
    //! to get the list of links
    const TList & GetLinks();
    //! to get the list of cascades
    const TList & GetCascades();
    //! to get the list of additional comments
    const TList & GetTexts();
     
    //! set current link
    void SetCLink(Link* link) { fCLink = link; }
    
    //! get current link
    const Link* GetCLink() { return fCLink; } 
    
    //! set current level
    void SetCLevel(Level* level) { fCLevel = level; }
    
    //! get current level
    const Level* GetCLevel() { return fCLevel; } 
    
    //! set current cascade
    void SetCCascade(Cascade* cascade) { fCCascade = cascade; }
    
    //! get current cascade
    const Cascade* GetCCascade() { return fCCascade; } 
  
     //! Get player
	BaseLSPlayer* GetPlayer(Option_t* opt="Graphical");
  
    //! BETTER TO USE Import - To init this level scheme with a Radford AGS file 
    /*! 
      Init this LevelSheme with an ascii gls Level Sheme File (.ags). The second argument (optional) is the nuclide
      to which this level scheme belongs. It is better to give it since the ags file's content does not have this information.  
      
      It returns 0 if something is wrong and 1 if the initialisation is ok.
      In case of success, the previous data members are lost.
      In case of error, the previous data members are still available.
      
      \warning The Radware defines bands and not Cascades. They are differents since
      a band is a collection of levels while a cascade is a collection of links.
      In this function, bands are converted into Cascades. A link is in the cascade if
      the initial and final level are in the corresponding band.
    */
    virtual Int_t InitAGS(const Char_t *filename, Option_t *opt = "152Dy");
    
    //! BETTER TO USE Import - To init this level scheme with a ENSDF file
    /*! 
      Init this LevelSheme with an ENSDF File (.ens). Because ENSDF files are collections of datasets, the
      dataset to be used must be given in the second argument. 
      
      	Example: InitENSDF("toto.ens","152DY ADOPTED LEVELS, GAMMAS") or InitENSDF("toto.ens","152DY")
	
      the first 5 characters of the second arguments must be a valid nuclide and the following characters correspond to the 
      dataset id. If it is not given, the assumed value is ADOPTED LEVELS, GAMMAS.
     
      It returns 0 if something is wrong and 1 if the initialisation is ok.
      In case of success, the previous data members are lost.
      In case of error, the previous data members are still available.
    */    
    virtual Int_t InitENSDF(const Char_t *, Option_t *);   
    
    //! to init this level scheme from an existing formatted file (ENSDF, Radware ..)
    /*! 
      It is better to use this member instead of directly InitAGS or InitENSDF since it proposes the same
      interface for all initialisations. The choise of the Init function is based on the filename's extention:
      - .ens --> InitENSDF
      - .ags --> InitAGS
      - ... could be extented 
      
      if your file has not the correct extention, change it or use directly the suitable member which does not check the extention.
      The second argument is, at least, the nuclide to which this nuclide belongs. 
      See the suitable members for a full description of the arguments required. 
      
      It returns 0 if something is wrong and 1 if the initialisation is ok.
      In case of success, the previous data members are lost.
      In case of error, the previous data members are still available.
    */              
    virtual Int_t Import(const Char_t *, Option_t *);    
        
    virtual void Clear(Option_t *opt = ""); 
                         
     //! to draw this level scheme 
    virtual void Draw(Option_t *opt = "");
  
    //! list levels/links
    virtual void ls(Option_t* option = "all") const ;

    //! Compute distance to object
    virtual Int_t DistancetoPrimitive(Int_t px, Int_t py);
  
    //! to execute an event for this level scheme
    virtual void ExecuteEvent(Int_t event, Int_t px, Int_t py);
  
    //! Find the cascades that own that link
    virtual TList* FindLinkInCascade(Link* link) const;
  
    //! Find the cascades that own the currengt link
    virtual TList* FindCLinkInCascade() const;

    //! Find links attached to level
    virtual const TList* FindLink(Level* level) const;
  
    //! Sort levels level
    virtual void SortLevels(Bool_t order = kSortAscending) { fLevels.Sort(order); }
  
    //! get bit map coincidence above a given link
    virtual TBits* GetBitMap(Link* link);
  
protected:
  
  mutable LogMessage fLog; //! log message
  

    //! rootcint dictionary
    ClassDef(LevelScheme,4); // a level scheme is a list of levels, links and cascades
};
// inline members
inline const TList & LevelScheme::GetLevels() { return fLevels;}
inline const TList & LevelScheme::GetLinks()  { return fLinks; }
inline const TList & LevelScheme::GetCascades() { return fCascades; }
inline const TList & LevelScheme::GetTexts() { return fTexts; }
inline void  LevelScheme::SetName(const Char_t *name) { fName.SetTitle(name); }
inline void  LevelScheme::SetReference(const Char_t *ref) { fReference.SetTitle(ref); }
inline const Char_t * LevelScheme::GetName() const { return fName.GetTitle(); }
inline const Char_t * LevelScheme::GetReference() const { return fReference.GetTitle(); }
}
#endif
