/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file Link.cpp compiled in libGWPHYSICS.so */

// A link is basically a TPolyLine with 8 points :
//
//       0=8  1   2
//        |   |   |
//        *********
//        *       *
//        *       *
//        *       *
//   6  7 *       * 3  4
//   |   \*       */   |
//   ******       ******
//    *               *
//     *             *
//      *           *
//       *         *
//        *       *
//         *     *
//          * 5 *
//           *|*
//            *



#ifndef Gw_Link
#include "Link.h"
#endif

#ifndef ROOT_TROOT_H
#include "TROOT.h"
#endif

#ifndef ROOT_TVirtualPad_H
#include "TVirtualPad.h"
#endif

using namespace Gw;

ClassImp(Link);

	  Color_t Link::fgDefaultColor      = 1;  
      Bool_t  Link::fgMovable           = false;
const Short_t Link::fgkMaxLinkLabel     = 3;
      Float_t Link::fgDefaultLabelSize  = 0.03; 
      Color_t Link::fgDefaultLabelColor = 1; 
//__________________________________________________________
Link::Link()
 : TPolyLine(),
   fVisLabel("000"),
   fLabel0(),
   fLabel1(),
   fLabel2(),
   fBox1(),
   fLog("Link"),
   fInitial(0x0),
   fFinal(0x0),
   fIsSelected(false),
   fIsVisible(true),
   fIsModified(false),
   fStrength(),
   fTau()
{
    SetFillColor(fgDefaultColor); SetLineColor(fgDefaultColor); SetOption("F");
    
    fLabel0.SetX(0); fLabel0.SetY(0); fLabel0.SetTextAlign(32); fLabel0.SetBit(TObject::kCannotPick);
    fLabel1.SetX(0); fLabel1.SetY(0); fLabel1.SetTextAlign(22); fLabel1.SetBit(TObject::kCannotPick);
    fLabel2.SetX(0); fLabel2.SetY(0); fLabel2.SetTextAlign(12); fLabel2.SetBit(TObject::kCannotPick);
    fBox1.SetFillColor(0); fBox1.SetLineColor(0);
}

//__________________________________________________________
Link::Link(const Link &link)
  : TPolyLine(link),
    fLog("Link"),
    fInitial(link.fInitial),
    fFinal(link.fFinal),
    fIsSelected(link.fIsSelected),
    fStrength(link.fStrength),
    fTau(link.fTau)
{
}
    
//__________________________________________________________
Link& Link::operator=(const Link &link) 
{
  if (this == &link) return *this;

  fInitial    = link.fInitial;
  fFinal      = link.fFinal;
  fIsSelected = link.fIsSelected;
  fStrength   = link.fStrength;
  fTau        = link.fTau;
  
  
  return *this;
}

//__________________________________________________________
Link::~Link()
{
}

//__________________________________________________________
void Link::SetLabelsSize(Float_t size)
{
  fLabel0.SetTextSize(size);
  fLabel1.SetTextSize(size);
  fLabel2.SetTextSize(size);
}

//__________________________________________________________
void Link::SetLabelsColor(Color_t color)
{
  fLabel0.SetTextColor(color);
  fLabel1.SetTextColor(color);
  fLabel2.SetTextColor(color);
}

//__________________________________________________________
void Link::SetVisLabel(const char *s)
{
  TString st = s; Short_t end = (st.Length() < fgkMaxLinkLabel) ? st.Length(): fgkMaxLinkLabel;
  
  for( Short_t i = 0; i < end; i++ ) {
    if ( st[i] == '0' ) {
      fVisLabel[i] = '0';
    }
    else {
      fVisLabel[i] = '1';
    }
  }
}


//__________________________________________________________
void Link::SetLabel(const char *label, Int_t which)
{
  if (which < 0 || which >= fgkMaxLinkLabel) return;
  switch (which) {
    case 0:
      fLabel0.SetText(fLabel0.GetX(),fLabel0.GetY(),label);
      break;
    case 1:
      fLabel1.SetText(fLabel1.GetX(),fLabel1.GetY(),label);
      break;
    case 2:
      fLabel2.SetText(fLabel2.GetX(),fLabel2.GetY(),label);
      break;
  }
}

//__________________________________________________________
void Link::SetLabels(const char *l0, const char *l1, const char *l2) 
{
  fLabel0.SetText(fLabel0.GetX(),fLabel0.GetY(),l0);
  fLabel1.SetText(fLabel1.GetX(),fLabel1.GetY(),l1);
  fLabel2.SetText(fLabel2.GetX(),fLabel2.GetY(),l2);
}

//__________________________________________________________
void Link::SetDefaultLabels()
{
  SetLabelsSize(Link::GetDefaultLabelSize());
  SetLabelsColor(Link::GetDefaultLabelColor());
  SetLabels(GetLabel0(), GetLabel1(), GetLabel2());
}

//__________________________________________________________
TLatex& Link::GetLabel(Int_t which)
{
  if ( which <= 0  ) return fLabel0;   
  
  if ( which == 1 )  return fLabel1;   
  
  return fLabel2;
}


//__________________________________________________________
void Link::SetPoints()
{
  if ( Link::GetIL() && Link::GetFL() ) {
	if (GetIL()->HasExtraLine())
	  SetPoints((GetFL()->GetX1()+GetFL()->GetX2())/2.0,
				(GetFL()->GetX1()+GetFL()->GetX2())/2.0 );
	
	else if (GetFL()->HasExtraLine())
	  SetPoints((GetIL()->GetX1()+GetIL()->GetX2())/2.0,
				(GetIL()->GetX1()+GetIL()->GetX2())/2.0 );
	else 
	  SetPoints((GetIL()->GetX1()+GetIL()->GetX2())/2.0,
				(GetFL()->GetX1()+GetFL()->GetX2())/2.0 );
  }	
}

//__________________________________________________________
void Link::SetPoints(Double_t xi, Double_t xf)
{
	Double_t x = 0, y = 0;
	Double_t offset1, offset2, prop, height, offset3;
	
	prop = 3.0/4.0;
	
	offset1 = TMath::Max(TMath::Abs(GetIL()->GetX2() - GetIL()->GetX1())/30.,
						 GetStrength().Get()/2.0 );
	
	offset2 = TMath::Min(2.5*offset1,TMath::Abs(GetIL()->GetX2() - GetIL()->GetX1())/2.);
	
	offset3 = TMath::Abs(GetIL()->GetX2() - GetIL()->GetX1())*0.25;
	
	height = GetFL()->GetY1() + TMath::Abs(GetIL()->GetY1() - GetFL()->GetY1())*(0.6);
  
					        
	if ( Link::GetIL() && Link::GetFL() ) {

		x = xi; y = GetIL()->GetY1();
		SetPoint(0,x-offset1,y);
		SetPoint(1,x,y);
		SetPoint(2,x+offset1,y);	
		SetPoint(8,x-offset1,y);
			
		x = xi + prop*(xf-xi); y = GetIL()->GetY1() - prop*(GetIL()->GetY1()-GetFL()->GetY1());
		SetPoint(3,x+offset1,y);
		SetPoint(4,x+offset2,y);   
		SetPoint(6,x-offset2,y);
		SetPoint(7,x-offset1,y);	
		
		x = xf; y = GetFL()->GetY1(); 
		SetPoint(5,x,y);
	}
  
        for (Int_t i = 0; i < fgkMaxLinkLabel; ++i) {
          GetLabel(i).SetY(height);
        }

        GetLabel(0).SetX(GetX()[1] - offset3);
        GetLabel(1).SetX(GetX()[1]);
        GetLabel(2).SetX(GetX()[1] + offset3);
}


//_____________________________________________________________________________
Int_t Link::Compare(const TObject* obj) const
{
  const Link* link = static_cast<const Link*>(obj);
  
  if ( GetIL()->GetY1() > link->GetIL()->GetY1() ) {
    return 1;
  }  else if (  GetIL()->GetY1() < link->GetIL()->GetY1() ) {
    return -1;
  } else {
	return 0;
  }
}

//__________________________________________________________
void Link::UpdatePoints()
{
  Double_t y = 0;
  Double_t prop, height;
  
  prop = 3.0/4.0;  
  height = GetFL()->GetY1() + TMath::Abs(GetIL()->GetY1() - GetFL()->GetY1())*(0.6);
  
  if ( Link::GetIL() && Link::GetFL() ) {
	
	y = GetIL()->GetY1();
	SetPoint(0,GetX()[0],y);
	SetPoint(1,GetX()[1],y);
	SetPoint(2,GetX()[2],y);	
	SetPoint(8,GetX()[8],y);
	
	y = GetIL()->GetY1() - prop*(GetIL()->GetY1()-GetFL()->GetY1());
	SetPoint(3,GetX()[3],y);
	SetPoint(4,GetX()[4],y);   
	SetPoint(6,GetX()[6],y);
	SetPoint(7,GetX()[7],y);	
	
	y = GetFL()->GetY1(); 
	SetPoint(5,GetX()[5],y);
  }
  
  for (Int_t i = 0; i < fgkMaxLinkLabel; ++i) {
	GetLabel(i).SetY(height);
  }
}

//__________________________________________________________
void Link::Paint(Option_t* option)
{
  if ( !IsVisible() ) return;
  
  TPolyLine::Paint(option);
  
  if ( fVisLabel[0] == '1' ) {
    GetLabel(0).Paint(option);
  }
  if ( fVisLabel[1] == '1' ) { 
    fBox1.SetX1(fLabel1.GetX() - fLabel1.GetXsize()/2.6); 
    fBox1.SetX2(fLabel1.GetX() + fLabel1.GetXsize()/2.5); 
    fBox1.SetY1(fLabel1.GetY() - fLabel1.GetYsize()/1.6); 
    fBox1.SetY2(fLabel1.GetY() + fLabel1.GetYsize()/2.0);
    fBox1.Paint(option);
    GetLabel(1).Paint(option);
  }
  if ( fVisLabel[2] == '1' ) {
    GetLabel(2).Paint(option);
  }
  
}


//__________________________________________________________
void Link::Draw(Option_t* option)
{
  if ( !IsVisible() ) return;
  
  TPolyLine::Draw(option);
  
   if ( fVisLabel[0] == '1' ) {
    GetLabel(0).Draw(option);
  }
  if ( fVisLabel[1] == '1' ) { 
    fBox1.Draw(option);
    GetLabel(1).Draw(option);
  }
  if ( fVisLabel[2] == '1' ) {
    GetLabel(2).Draw(option);
  }
  
}

//____________________________________________________________________________________________
void Link::ExecuteEvent(Int_t event, Int_t px, Int_t py)
{
   // Execute action corresponding to one event.
	//
   //  This member function is called when a polyline is clicked with the locator
	//
   //  If Left button clicked on one of the line end points, this point
   //     follows the cursor until button is released.
	//
   //  if Middle button clicked, the line is moved parallel to itself
   //     until the button is released.

        if ( !IsMovable() ) return;

	Int_t i, d;
	Double_t xmin, xmax, ymin, ymax, dx, dy, dxr, dyr;
	const Int_t kMaxDiff = 30;
	static Bool_t middle;
	static Int_t ipoint, pxp, pyp;
	static Int_t px1,px2,py1,py2;
	static Int_t pxold, pyold, px1old, py1old, px2old, py2old;
	static Int_t dpx, dpy;
	static Int_t *x=0, *y=0;
	static Double_t a=0.0,b=0.0;
	static Int_t xprime=0,t1=0,t2=0;

	if (!gPad->IsEditable()) return;

	Int_t np = Size();

	Double_t vx0 = fLabel0.GetX() - fLabel1.GetX();
	Double_t vx1 = 0.;
	Double_t vx2 = fLabel2.GetX() - fLabel1.GetX();
	Double_t vy  = fLabel2.GetY() - GetY()[5];
  
	switch (event) {

		case kButton1Down:
			gVirtualX->SetLineColor(-1);
			TAttLine::Modify();  //Change line attributes only if necessary
			px1 = gPad->XtoAbsPixel(gPad->GetX1());
			py1 = gPad->YtoAbsPixel(gPad->GetY1());
			px2 = gPad->XtoAbsPixel(gPad->GetX2());
			py2 = gPad->YtoAbsPixel(gPad->GetY2());
			ipoint = -1;


			if (x || y) break;
			x = new Int_t[np+1];
			y = new Int_t[np+1];
			for (i=0;i<np;i++) {
				pxp = gPad->XtoAbsPixel(gPad->XtoPad(fX[i]));
				pyp = gPad->YtoAbsPixel(gPad->YtoPad(fY[i]));
				x[i] = pxp;
				y[i] = pyp;
				d   = TMath::Abs(pxp-px) + TMath::Abs(pyp-py);
				if ( (d < kMaxDiff) && ((i==1) || (i==5))) {
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
					ipoint =i;
				}
			}
			dpx = 0;
			dpy = 0;
			pxold = px;
			pyold = py;
			if (ipoint < 0) return;
			if (ipoint == 0) {
				px1old = 0;
				py1old = 0;
				px2old = gPad->XtoAbsPixel(fX[1]);
				py2old = gPad->YtoAbsPixel(fY[1]);
			} else if (ipoint == fN-1) {
				px1old = gPad->XtoAbsPixel(gPad->XtoPad(fX[fN-2]));
				py1old = gPad->YtoAbsPixel(gPad->YtoPad(fY[fN-2]));
				px2old = 0;
				py2old = 0;
			} else {
				px1old = gPad->XtoAbsPixel(gPad->XtoPad(fX[ipoint-1]));
				py1old = gPad->YtoAbsPixel(gPad->YtoPad(fY[ipoint-1]));
				px2old = gPad->XtoAbsPixel(gPad->XtoPad(fX[ipoint+1]));
				py2old = gPad->YtoAbsPixel(gPad->YtoPad(fY[ipoint+1]));
			}
			pxold = gPad->XtoAbsPixel(gPad->XtoPad(fX[ipoint]));
			pyold = gPad->YtoAbsPixel(gPad->YtoPad(fY[ipoint]));

			break;


		case kMouseMotion:

			middle = kTRUE;
			for (i=0;i<np;i++) {
				pxp = gPad->XtoAbsPixel(gPad->XtoPad(fX[i]));
				pyp = gPad->YtoAbsPixel(gPad->YtoPad(fY[i]));
				d   = TMath::Abs(pxp-px) + TMath::Abs(pyp-py);
				if ( (d < kMaxDiff) && ((i==1) || (i==5))) middle = kFALSE;
			}


   // check if point is close to an axis
			if (middle) gPad->SetCursor(kMove);
			else gPad->SetCursor(kHand);
			break;

		case kButton1Motion:
			if (middle) {
				for(i=0;i<np-1;i++) {
					gVirtualX->DrawLine(x[i]+dpx, y[i]+dpy, x[i+1]+dpx, y[i+1]+dpy);
					pxp = x[i]+dpx;
					pyp = y[i];
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				}
				pxp = x[np-1]+dpx;
				pyp = y[np-1];
				gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
				gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
				gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
				gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				dpx += px - pxold;
				//dpy += py - pyold;
				pxold = px;
				pyold = py;
				for(i=0;i<np-1;i++) {
					gVirtualX->DrawLine(x[i]+dpx, y[i]+dpy, x[i+1]+dpx, y[i+1]+dpy);
					pxp = x[i]+dpx;
					pyp = y[i];
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				}
				pxp = x[np-1]+dpx;
				pyp = y[np-1];
				gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
				gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
				gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
				gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
			} else {
				if (ipoint==1) {
					if ((x[5]-x[1]-dpx != 0) && (y[5]-y[1] != 0)) {
						a = (y[5]-y[1])/((x[5]-x[1]-dpx)*1.0);
						b = y[5] - a * x[5];
						xprime = (Int_t)((y[3] - b) / a);
					}  else {
						xprime = x[5];
					}
					t1 = (x[3] - x[7]) / 2;
					t2 = (x[4] - x[6]) / 2;
					//printf("x1 = %d; x5 = %d, dpx = %d xprime = %f, t1 = %d, t2=%d, y1=%d , y5=%d , y3=%d, a=%f, b=%f\n",x[1],x[5],dpx,xprime,t1,t2,y[1],y[5],y[3],a,b);
					gVirtualX->DrawLine(x[0]+dpx, y[0], x[2]+dpx, y[2]);
					gVirtualX->DrawLine(x[0]+dpx, y[0], xprime-t1, y[7]);
					gVirtualX->DrawLine(x[2]+dpx, y[2], xprime+t1, y[3]);
					gVirtualX->DrawLine(xprime-t1, y[7], xprime-t2, y[6]);
					gVirtualX->DrawLine(xprime+t1, y[3], xprime+t2, y[4]);
					gVirtualX->DrawLine(xprime-t2, y[6], x[5], y[5]);
					gVirtualX->DrawLine(xprime+t2, y[4], x[5], y[5]);
					pxp = x[1]+dpx;
					pyp = y[1];
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				  
				  // draw extra line if needed
				  Int_t x1IL = gPad->XtoAbsPixel(GetIL()->GetX1());
				  Int_t x2IL = gPad->XtoAbsPixel(GetIL()->GetX2());
				  if (pxp > x2IL) {
					gVirtualX->DrawLine(x2IL, pyp, pxp,  pyp);
				  } else if (pxp < x1IL) {
					gVirtualX->DrawLine(pxp, pyp, x1IL,  pyp);
				  }
				}

				if (ipoint==5) {
					if ((x[1]-x[5]-dpx != 0) && (y[1]-y[5] != 0)) {
						a = (y[1]-y[5])/((x[1]-x[5]-dpx)*1.0);
						b = y[1] - a * x[1];
						xprime = (Int_t)((y[3] - b) / a);
					}  else {
						xprime = x[1];
					}
					t1 = (x[3] - x[7]) / 2;
					t2 = (x[4] - x[6]) / 2;
					gVirtualX->DrawLine(xprime-t1, y[7], x[0], y[0]);
					gVirtualX->DrawLine(xprime+t1, y[3], x[2], y[2]);
					gVirtualX->DrawLine(xprime-t1, y[7], xprime-t2, y[6]);
					gVirtualX->DrawLine(xprime+t2, y[4], xprime+t1, y[3]);
					gVirtualX->DrawLine(x[5]+dpx, y[5], xprime-t2, y[6]);
					gVirtualX->DrawLine(x[5]+dpx, y[5], xprime+t2, y[4]);
					pxp = x[5]+dpx;
					pyp = y[5];
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				   
				    // draw extra line if needed
				    Int_t x1FL = gPad->XtoAbsPixel(GetFL()->GetX1());
				    Int_t x2FL = gPad->XtoAbsPixel(GetFL()->GetX2());
				    if (pxp > x2FL) {
					   gVirtualX->DrawLine(x2FL, pyp, pxp,  pyp);
				    } else if (pxp < x1FL) {
					   gVirtualX->DrawLine(pxp, pyp, x1FL,  pyp);
				    }
				}

				dpx += px - pxold;
				
				pxold = px;
				pxold = TMath::Max(pxold, px1);
				pxold = TMath::Min(pxold, px2);

				if (ipoint==1) {
					if ((x[5]-x[1]-dpx != 0) && (y[5]-y[1] != 0)) {
						a = (y[5]-y[1])/((x[5]-x[1]-dpx)*1.0);
						b = y[5] - a * x[5];
						xprime = (Int_t)((y[3] - b) / a);
					}  else {
						xprime = x[5];
					}
					t1 = (x[3] - x[7]) / 2;
					t2 = (x[4] - x[6]) / 2;
					//printf("x1 = %d; x5 = %d, dpx = %d xprime = %f, t1 = %d, t2=%d, y1=%d , y5=%d , y3=%d, a=%f, b=%f\n",x[1],x[5],dpx,xprime,t1,t2,y[1],y[5],y[3],a,b);
					gVirtualX->DrawLine(x[0]+dpx, y[0], x[2]+dpx, y[2]);
					gVirtualX->DrawLine(x[0]+dpx, y[0], xprime-t1, y[7]);
					gVirtualX->DrawLine(x[2]+dpx, y[2], xprime+t1, y[3]);
					gVirtualX->DrawLine(xprime-t1, y[7], xprime-t2, y[6]);
					gVirtualX->DrawLine(xprime+t1, y[3], xprime+t2, y[4]);
					gVirtualX->DrawLine(xprime-t2, y[6], x[5], y[5]);
					gVirtualX->DrawLine(xprime+t2, y[4], x[5], y[5]);
					pxp = x[1]+dpx;
					pyp = y[1];
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				}

				if (ipoint==5) {
					if ((x[1]-x[5]-dpx != 0) && (y[1]-y[5] != 0)) {
						a = (y[1]-y[5])/((x[1]-x[5]-dpx)*1.0);
						b = y[1] - a * x[1];
						xprime = (Int_t)((y[3] - b) / a);
					}  else {
						xprime = x[1];
					}
					t1 = (x[3] - x[7]) / 2;
					t2 = (x[4] - x[6]) / 2;
					gVirtualX->DrawLine(xprime-t1, y[7], x[0], y[0]);
					gVirtualX->DrawLine(xprime+t1, y[3], x[2], y[2]);
					gVirtualX->DrawLine(xprime-t1, y[7], xprime-t2, y[6]);
					gVirtualX->DrawLine(xprime+t2, y[4], xprime+t1, y[3]);
					gVirtualX->DrawLine(x[5]+dpx, y[5], xprime-t2, y[6]);
					gVirtualX->DrawLine(x[5]+dpx, y[5], xprime+t2, y[4]);
					pxp = x[5]+dpx;
					pyp = y[5];
					gVirtualX->DrawLine(pxp-4, pyp-4, pxp+4,  pyp-4);
					gVirtualX->DrawLine(pxp+4, pyp-4, pxp+4,  pyp+4);
					gVirtualX->DrawLine(pxp+4, pyp+4, pxp-4,  pyp+4);
					gVirtualX->DrawLine(pxp-4, pyp+4, pxp-4,  pyp-4);
				}
			}
			
			break;

		case kButton1Up:

   // Compute x,y range
			xmin = gPad->GetUxmin();
			xmax = gPad->GetUxmax();
			ymin = gPad->GetUymin();
			ymax = gPad->GetUymax();
			dx   = xmax-xmin;
			dy   = ymax-ymin;
			//dxr  = dx/(1 - gPad->GetLeftMargin() - gPad->GetRightMargin());
			//dyr  = dy/(1 - gPad->GetBottomMargin() - gPad->GetTopMargin());
			dxr = 0;
			dyr = 0;

   // Range() could change the size of the pad pixmap and therefore should
   // be called before the other paint routines
			gPad->Range(xmin - dxr*gPad->GetLeftMargin(),
				    ymin - dyr*gPad->GetBottomMargin(),
				    xmax + dxr*gPad->GetRightMargin(),
				    ymax + dyr*gPad->GetTopMargin());
			gPad->RangeAxis(xmin, ymin, xmax, ymax);

			if (middle) {
				for(i=0;i<np;i++) {
					fX[i] = gPad->PadtoX(gPad->AbsPixeltoX(x[i]+dpx));
					//fY[i] = gPad->PadtoY(gPad->AbsPixeltoY(y[i]+dpy));
				}
			} else {
				if (ipoint == 1) {
					if ((x[5]-x[1]-dpx != 0) && (y[5]-y[1] != 0)) {
						a = (y[5]-y[1])/((x[5]-x[1]-dpx)*1.0);
						b = y[5] - a * x[5];
						xprime = (Int_t)((y[3] - b) / a);
					}  else {
						xprime = x[5];
					}
					t1 = (x[3] - x[7]) / 2;
					t2 = (x[4] - x[6]) / 2;
					fX[1] = gPad->PadtoX(gPad->AbsPixeltoX(x[1]+dpx));
				        fX[0] = gPad->PadtoX(gPad->AbsPixeltoX(x[0]+dpx));
				        fX[2] = gPad->PadtoX(gPad->AbsPixeltoX(x[2]+dpx));
					fX[8] = gPad->PadtoX(gPad->AbsPixeltoX(x[8]+dpx));
					fX[3] = gPad->PadtoX(gPad->AbsPixeltoX(xprime+t1));
					fX[4] = gPad->PadtoX(gPad->AbsPixeltoX(xprime+t2));
					fX[6] = gPad->PadtoX(gPad->AbsPixeltoX(xprime-t2));
					fX[7] = gPad->PadtoX(gPad->AbsPixeltoX(xprime-t1));
				  
				  // add extra line if needed
				  if (fX[1] > GetIL()->GetX2() ) {
					GetIL()->SetExtraLineX2(fX[1]);
				  } else if (fX[1] < GetIL()->GetX1()) {
					GetIL()->SetExtraLineX1(fX[1]);
				  } else {
					GetIL()->GetExtraLine().SetBit(TObject::kInvalidObject);
				  }
				}
				if (ipoint == 5) {
					if ((x[1]-x[5]-dpx != 0) && (y[1]-y[5] != 0)) {
						a = (y[1]-y[5])/((x[1]-x[5]-dpx)*1.0);
						b = y[1] - a * x[1];
						xprime = (Int_t)((y[3] - b) / a);
					}  else {
						xprime = x[1];
					}
					t1 = (x[3] - x[7]) / 2;
					t2 = (x[4] - x[6]) / 2;
					fX[5] = gPad->PadtoX(gPad->AbsPixeltoX(x[5]+dpx));
					fX[3] = gPad->PadtoX(gPad->AbsPixeltoX(xprime+t1));
					fX[4] = gPad->PadtoX(gPad->AbsPixeltoX(xprime+t2));
					fX[6] = gPad->PadtoX(gPad->AbsPixeltoX(xprime-t2));
					fX[7] = gPad->PadtoX(gPad->AbsPixeltoX(xprime-t1));
				   
				    // add extra line if needed
				    if (fX[5] > GetFL()->GetX2() ) {
					   GetFL()->SetExtraLineX2(fX[5]);
					} else if (fX[5] < GetFL()->GetX1()) {
					   GetFL()->SetExtraLineX1(fX[5]);
				    } else {
					   GetFL()->GetExtraLine().SetBit(TObject::kInvalidObject);
					}

			 	}
				//fY[ipoint] = gPad->PadtoY(gPad->AbsPixeltoY(pyold));
			}
		
			fLabel0.SetX((GetX()[1] + GetX()[5])/2. + vx0); fLabel0.SetY(fY[5]+vy);
			fLabel1.SetX((GetX()[1] + GetX()[5])/2. + vx1); fLabel0.SetY(fY[5]+vy);
			fLabel2.SetX((GetX()[1] + GetX()[5])/2. + vx2); fLabel0.SetY(fY[5]+vy);
		
			delete [] x; x = 0;
			delete [] y; y = 0;
			gPad->Modified(kTRUE);
			gVirtualX->SetLineColor(-1);
	}
}




