/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
 /** \file Link.h header file for a Link */
 
#ifndef Gw_Link
#define Gw_Link
#define GW_LINK_H Gw_Link

#ifndef ROOT_TBox
#include <TBox.h>
#endif
#ifndef ROOT_TPolyLine
#include <TPolyLine.h>
#endif
#ifndef ROOT_TSeqCollection
#include <TSeqCollection.h>
#endif

#include "GwLogMessage.h"

#ifndef GW_LEVEL_H
#include "Level.h"
#endif

#ifndef GW_MEASURE_H
#include "Measure.h"
#endif

#define MAX_LINK_LABEL 2

namespace Gw {

/**
A link between two levels.

A link binds two levels with a given strength and a characteritic time to 'go' from one level to the other one.
\warning The initial and final level don't belong to this link. They will not be deleted in the destructor.

A link could be a single path from one level to the other one but it could be also more complex. \n
For instance, a converted gamma-ray gives several X-rays with a given discret distribution. It
could also be a continuous distribution like for a link between a SD level and a ND level or a
feeding link. A link has a DoCascade member to really determine a unique path from the inital
level to the final level. 

@author Olivier Stezowski
*/
class Link : public TPolyLine
{
  
public:

  static Color_t GetDefaultColor()                { return fgDefaultColor; } 
  static Float_t GetDefaultLabelSize()            { return fgDefaultLabelSize; }
  static Color_t GetDefaultLabelColor()           { return fgDefaultLabelColor; }
  
  static void SetDefaultColor(Color_t color)      { fgDefaultColor = color; }
  static void SetDefaultLabelSize(Float_t size)   { fgDefaultLabelSize = size; }
  static void SetDefaultLabelColor(Color_t color) { fgDefaultLabelColor = color; }
  
  static void   EnableMove()  { fgMovable = true; }
  static void   DisableMove() { fgMovable = false; }
  static Bool_t IsMovable()   { return fgMovable == true; }

private:
  static Color_t fgDefaultColor;      // default color (black)
  static Bool_t  fgMovable;           // graphically movable
  
  static Color_t fgDefaultLabelColor; // default text color
  static Float_t fgDefaultLabelSize;  // default text size
  
private:
    Double_t fgMinWidth;	//!
    Double_t fgMaxArrow;	//!
    Double_t fgProArrow;	//!
      

    TString fVisLabel;	// To know if a label should be drawn ok not
    TString fPosLabel;	// To know the position of label 0:left; 1:center; 2:right
    
    TLatex  fLabel0;	// Some labels associated with this link  
    TLatex  fLabel1;	// Some labels associated with this link    
    TLatex  fLabel2;	// Some labels associated with this link     
    TBox    fBox1;      // box underlying label2 onto the link
  
protected:
  
    mutable LogMessage fLog; //! log message

    Level *fInitial;		// initial level
    Level *fFinal;		    // final level
  
    Bool_t fIsSelected;     // graphically selected 
	Bool_t fIsVisible;      // flag to know whether the link is visible
	Bool_t fIsModified;     // flag to know whether the link has been modified

    static const Short_t fgkMaxLinkLabel; // max number of labels

protected:
    Measure<Float_t> fStrength;	// strength that binds the two levels
    Measure<Float_t> fTau;	    // characteritic time to ge from one level to the other one

public:
    Link();
    Link(const Link &);
    Link& operator=(const Link &);
    virtual ~Link();
    
    //! to set points from the levels and the Strength 
    virtual void SetPoints(); 
    virtual void SetPoints(Double_t, Double_t);
  
	//! update points respect to a modified level (for future application)
    virtual void UpdatePoints();
  
    //! to get the characteritic time and modify it if it is needed 
    virtual Measure<Float_t> & GetTau() { return fTau; }
    //! to get the strength and modify it if it is needed 
    virtual Measure<Float_t> & GetStrength() { return fStrength; }
    
    //! to change the final level - return the previous one 
    virtual Level * SetFL(Level *final) { Level *tmp = fFinal; fFinal = final; return tmp; }
    //! to change the final level - return the previous one
    virtual Level * SetIL(Level *initial) { Level *tmp = fInitial; fInitial = initial; return tmp; }
    //! to get the initial level 
    virtual Level * GetIL() { return fInitial; }  
    //! to get the final level 
    virtual Level * GetFL() { return fFinal; }
  
	virtual Level * GetIL() const { return fInitial; }  
	//! to get the final level 
	virtual Level * GetFL() const { return fFinal; }  
    
    //! to add this link to a cascade
    /*! 
     A cascade is a path from one level to another one. This member adds this links to a cascade. \n
     If the link involves some random distribution then a unique path is determine and all the single links are added to the cascade.
    */        
    virtual void DoCascade(TSeqCollection &/*col*/, Option_t */*opt*/ = "") {;}

    //! to execute an event for this link
    virtual void ExecuteEvent(Int_t event, Int_t px, Int_t py);
  
    
    //! Used to determine which label should be drawn
    /*!    
     There are four labels associated with this level. The string given as argument helps determining if a label is drawn.
     For instance, "000" (which is the default value) correspond to no label while for "101" only the first and the third
     ones are drawn.
     The first label is set to the left of the link while the second and the third are repectively set to the middle and 
     right of the link.
     */    
    virtual void SetVisLabel(const char *); //*MENU*
    
    //! set label titles
    virtual void SetLabels(const char *l0, const char *l1, const char *l2) ; // *MENU* *ARGS={l0=>fLabel0,l1=>fLabel1,l2=>fLabel2}
    
    //! set all labels size
    virtual void SetLabelsSize(Float_t size);
    
    //! set all labels color
    virtual void SetLabelsColor(Color_t color);
    
    
    //! Used to determine which label should be drawn    
    virtual const char * GetVisLabel() { return fVisLabel.Data(); }
    
    //! All labels are shifted when moving a level to another position     
    virtual TLatex& GetLabel(Int_t);
    
    //! Get labels title
    virtual const char* GetLabel0() { return fLabel0.GetTitle(); }   
    virtual const char* GetLabel1() { return fLabel1.GetTitle(); }   
    virtual const char* GetLabel2() { return fLabel2.GetTitle(); }   
    
    //! to set labels         
    virtual void SetLabel(const char *, Int_t);
    
    //! Set default labels
    virtual void SetDefaultLabels();
  
    //! Selected link
    virtual void Selected() {fIsSelected = true;}
    
    //! unselected link
    virtual void DeSelected() {fIsSelected = false;}
  
    //! return selection
    virtual Bool_t IsSelected() {return fIsSelected;}
   
    //! Set link visible
    virtual void SetVisible() {fIsVisible = true;}
    
    //! Set link unvisble
    virtual void SetUnvisible() {fIsVisible = false;}
    
	//! return visible
	virtual Bool_t IsVisible() {return fIsVisible;}
	
	//! Set link modified
	virtual void SetModified() {fIsModified = true;}
	
	//! Set link unvisble
	virtual void SetUnmdified() {fIsModified = false;}
	
	//! return modified
	virtual Bool_t IsModified() {return fIsModified;}
  
    //! draw
    virtual void Draw(Option_t* option = "");
    virtual void Paint(Option_t* option = "");

    //! ls on link
    virtual void ls(Option_t* /*option*/ = "") const { return; }      

    //! Delete, redifined to avoid appearence in Menu
    void Delete(const Option_t* /*opt*/ = "") { return; }
  
	/// Return true as the function Compare is implemented
	Bool_t IsSortable() const { return kTRUE; }
	
	virtual Int_t Compare(const TObject* obj) const;
  
    //! rootcint dictionary
    ClassDef(Link,3); // a Link binds two levels
};
// inline members
}

#endif
