/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file NuclearLevel.h header file for a NuclearLevel */  

#ifndef Gw_NuclearLevel
#define Gw_NuclearLevel
#define GW_NUCLEARLEVEL_H Gw_NuclearLevel

#ifndef GW_LEVEL_H
#include "Level.h"
#endif

#ifndef GW_BASENUCLEUS_H
#include <BaseNucleus.h>
#endif

#ifndef GW_PARITY_H
#include "Parity.h"
#endif

#ifndef GW_SPIN_H
#include "Spin.h"
#endif

#ifndef GW_MEASURE_H
#include <Measure.h>
#endif

namespace Gw {

/**
A nuclear level 

A nuclear level is a particular level. It belongs to a nucleus (BaseNucleus), it has some quantum numbers,
like a Spin and a Parity,  

By default, the first level label is used to store the spin value, 
the second one the parity, the third one the level lifetime (for isomers) and last one the level energy.
The spin can contain also the parity (radware like, e.g.: 43/2+), if parity is in spin, parity label is set to 
this value (e.g.: 43/2+ -> parity set to kPlus).
Default wise energy and lifetime labels include the associated errors.

\todo a configuration should be implemented with additional quantum numbers, like K,
     with some informations concerning the configuration label, the deformation ...etc 

@author Olivier Stezowski
*/
class NuclearLevel : public Level {
	
protected:
    BaseNucleus fNucleus;	  // to which nucleus this level belongs
    
    Measure<Float_t> fEnergy; // energy and its error
    Measure<Float_t> fT;	  // half-life of this level
    Parity fParity;		      // parity of this level
    Spin fSpin;			      // spin of this level
     
public:
    NuclearLevel();
    virtual ~NuclearLevel();

    //! set labels and modify the attributes of this nuclear level
    virtual void SetLabels(const char *l0, const char *l1, const char *l2, const char *l3) ; // *MENU* *ARGS={l0=>fLabel0,l1=>fLabel1,l2=>fLabel2,l3=>fLabel3}
  
    //! to get the spin and modify it if it is needed
    Spin&   GetSpin()   { return fSpin;   }
    
    //! get the parity and modify it if it is needed 
    Parity& GetParity() { return fParity; }
    
    //! to get the energy and modify it if it is needed
    Measure<Float_t> & GetE()      { return fEnergy; }
    Measure<Float_t> & GetEnergy() { return fEnergy; }

    // to get the half-life time and modify it if it is needed 
    Measure<Float_t> & GetT()      { return fT; }
        
    virtual void ls(Option_t* option = "") const ;
    
    //! overload function for specified used in this class.  
    const char* GetLabel0();
    const char* GetLabel1() { return fParity.IsParity(Parity::kPlus) ? "+" : "-"; }   
    const char* GetLabel2() { return Form("%7.1f %4.1f", fT.Get(), fT.GetError()); }   
    const char* GetLabel3() { return Form("%7.1f %4.1f", fEnergy.Get(), fEnergy.GetError()); }   
    
    //! rootcint dictionary
    ClassDef(NuclearLevel,1); // a NuclearLevel level
};
// inline members

}

#endif
