/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file XGammaLink.h header file for a XGammaLink */  

#ifndef GW_XGAMMALINK_H
#define GW_XGAMMALINK_H
#define Gw_XGammaLink GW_XGAMMALINK_H

#ifndef ROOT_TList
#include <TList.h>
#endif

#ifndef GW_GAMMALINK_H
#include "GammaLink.h"
#endif

#ifndef GW_RANDOBJ_H
#include "RandObj.h"
#endif

namespace Gw {

/**
A XGammaLink is a converted gamma-ray link.

When converted, a gamma-ray produces an electron and a X-ray. \n
Compared to a GammaLink, a XGammaLink owns a list of associated Xrays. Those
Xrays themself are GammaLink. The major difference with a GammaLink is then when
the DoCascade members is called in order to simulate a path from the initial to the final level.
If the fIsX flag is set (the associated list of X-Rays must not be empty), then a random selection is done. 

Illustration: see simu1X() in firstdemo.C

@author Olivier Stezowski
*/
class XGammaLink : public GammaLink
{
private:
    Bool_t  fIsX;	// true if the conversion must be done during DoCascade
    
protected:
    TList   fX;		// list of X-rays associated with this gamma
    RandObj fRandX;	// Random generator to select one link
    
public:
    XGammaLink();
    virtual ~XGammaLink();

    virtual Bool_t IsX() { return fIsX == kTRUE; }
    virtual void DoX(Bool_t x = kTRUE ) { fIsX = x ; }    
        
    //! it initialises the list of X-rays associated with this link  
    /*!
        The first argument is the number of Xrays, the second is a vector of energies and the third one the associated intensities. \n
	Because it sets all X-rays, the previous definition is lost. \n
	It creats a list of X-rays (as GammaLink) with a associated random generator (RandObj).
	All GammaLinks are from the same initial and final level.
	
	\todo a member function that initialises the X-rays list by consulting a DataCenter.
	\todo URGENT: set the correct x-rays properties (conversion,multipolarity ..). For the moment only copy of this
    */    
    void SetX(Int_t, Float_t *, Float_t *);
    
    //! it adds this gamma to the given cascade   
    /*!
    	A first random selection is performed to determine if this gamma is converted
	- if not converted, the GammaLink is added to the cascade.
	- if converted, it is the X-ray which is added to the cascade.

	\todo if needed, the electron must be added to the cascade 
    */
    //virtual void DoCascade(Cascade &cas, Option_t *opt = "") ;  
    virtual void DoCascade(TSeqCollection &col, Option_t *opt = "");    
    
    //virtual void ls(Option_t* option = "") const ;      
    
    //! rootcint dictionary
    ClassDef(XGammaLink,1); // a XGammaLink in a converted GammaLink
};
// inline members

}
#endif
