
#ifndef Gw_AgataGeometryTransformer
#define Gw_AgataGeometryTransformer

#include "GwLogMessage.h"

#ifndef ROOT_TArrayF
#include "TArrayF.h"
#endif

#ifndef ROOT_TObject
#include "TObject.h"
#endif

#ifndef ROOT_TString
#include "TString.h"
#endif

class TGeoCombiTrans;
class TObjArray;
class TGeoVolume;
class TGeoMaterial;

namespace Gw {
   
   /** AgataGeometryTransformer class that manage geometry transformations
	
	*/
   
class AgataGeometryTransformer : public TObject
{
	  
public:
   
   AgataGeometryTransformer();
   virtual  ~AgataGeometryTransformer();
   
   
   //! read transformations file
   Bool_t  ReadTransformations(const TString fileName);
   
   //! fill list of matrices
   TGeoCombiTrans* FillTransMatrix(Double_t* trans, Double_t* rot);
   
   // Transformation methods 
   
   //! Transform point from the global reference frame
   //! to the local reference frame of the detection id
   void Global2Local(Int_t detID,
					 Double_t xg, Double_t yg, Double_t zg, 
					 Double_t& xl, Double_t& yl, Double_t& zl) const;
   
   void Global2Local(Int_t detID,
					 Float_t xg, Float_t yg, Float_t zg, 
					 Float_t& xl, Float_t& yl, Float_t& zl) const;
   
   //! Transform point from the local reference frame
   //! of the detection id to the global reference frame 
   void Local2Global(Int_t detID,
					 Double_t xl, Double_t yl, Double_t zl, 
					 Double_t& xg, Double_t& yg, Double_t& zg) const;
   
   void Local2Global(Int_t detID,
					 Float_t xl, Float_t yl, Float_t zl, 
					 Float_t& xg, Float_t& yg, Float_t& zg) const;
   
   
   //! Get sector depth from z position
   Int_t GetDepthId(Double_t z);
   
   //! Get sector id from (x,y) position
   Int_t GetSectorId(Double_t x, Double_t y);
   
public:
   
   //! to build agata crytals from a formatted text-file (asolid)
   /*!
	To build Germanium crytal as defined in the asolid file (see http://agata.pd.infn.it/documents/simulations/docs/GeometryDescription.html
	or in the sources of the Agata simulation code).\n
	
	The created crystals are TGeoVolumes that are added to the TGeoManager. They have the same base name with a varying indice. So if
	basecrysname is gePoly, this function extracts the different shapes defined in asolid and creats volume's names gePoly0, gePoly1 ... gePolyi ..
	with i being the integer used in asolid to reference different shapes (first column in the file).
	
	It returns the number of crystals really built.
	
	The implementation comes mainly from Enrico's and Joa's codes. 
	
	\todo for the moment, the shapes are created using TGeoXtru shapes that intersect TGeoPcon shapes. The irregular polygon corresponding to the front face is the same that in Geant4. However the back ones is deduced from the front ones just by applying a scaling factor. The scaling factor is calculated so that the real polygon at the back is contained into the calculated ones giving a volume slightly bigger in ROOT that in Geant4.  
	*/
   static int BuildAgataCrystals(const char *filename = "asolid", const char *basecrysname = "gePoly");
   
   //! to build agata clusters from a formatted text-file (aclust)
   /*!
	To build the different clusters as defined in the file (aeuler-like file, see http://agata.pd.infn.it/documents/simulations/docs/GeometryDescription.html) and had them to the TGeoManager with, as base name, baseclusname.
	Each cluster is an assembly of crystals that HAVE to be already defined in the TGeoManager (BuildAgataCrystals must be called successfully first). 
	
	It returns the number of clusters really built.
	
	The implementation comes mainly from Enrico's and Joa's codes. 
	*/
   static int BuildAgataClusters(const char *filename = "aclust", const char *baseclusname = "agataCluster", const char *basecrysname = "gePoly");
   
   //! to build agata from a formatted text-file (aeuler)
   /*!
	To define the agata geometry from the standard (Agata Geant4, see http://agata.pd.infn.it/documents/simulations/docs/GeometryDescription.html) input files (aeuler). It is constructed with clusters that must already be defined (for instance BuildAgataClusters has already been called). 
	
	It returns a TGeoVolume corresponding to the geometry. This volumes could be added to a world (top) volume with a possible additionnal translation or rotation. The name of the volume is agata.
	
	The implementation comes mainly from Enrico's and Joa's codes. 	
	*/	
   static TGeoVolume *BuildAgataDetector(const char *filename = "aeuler", const char *baseclusname = "agataCluster", const char *agataName = "Agata");
   
   //! to import the agata geometry from the Agata world
   /*!
	To define the agata geometry from the standard (Agata Geant4) input files (asolid, acluster, aeuler). If the top volume already exists in
	the GeoManager, agata is added to it.
	*/	
   static TGeoVolume *ImportAgata(const char *asolid = "asolid", const char *aclust = "aclust", const char *aeuler = "aeuler", 
								  const char *baseclusname = "agataCluster", const char* agataName = "Agata");

   //! Add Dante module geometry to the Agata-Geant4 world
   static TGeoVolume* AddDanteModule(const Float_t dz= 0., const Float_t rotz1 = 0., const Float_t roty = 0., const Float_t rotz2 = 0, 
									 const Float_t rotx = 0., const char* basemoduleName = "Module", const char *danteName = "Dante");
   //! Add beam pipe
   static TGeoVolume* AddBeamPipe(const Float_t dZ = 18, const Float_t minR = 2., const Float_t maxR = 2.3, const char *pipeName = "Pipe");
   
   //! Add beam pipe
   static TGeoVolume* AddTarget(const Float_t dx = 0.8, const Float_t dy = 0.8, const Float_t dz = 0.05, 
								const char *targetName = "Target", TGeoMaterial* mat = 0x0);
   //! Add floor
   static TGeoVolume* AddFloor(Float_t dx = 20, const Float_t dy = 0.1, const Float_t dz = 25, const char *floorName = "Floor");
   
   //! Add MCP
   static TGeoVolume* AddPrismaMCP(const char *prismaName = "PrismaMCP", Float_t dx = 10., Float_t dy = 10., Float_t dz = 1., 
								   Float_t phi = 135., TGeoMaterial* mat = 0x0);
   //! Add Prisma Qpole
   static TGeoVolume* AddPrismaQpole(const char *prismaName = "PrismaQpole", Float_t dz = 21., Float_t rmin = 15., Float_t rmax = 15.7);
   
   //! Add Prisma Dipole
   static TGeoVolume* AddPrismaDipoleOld(const char *prismaName = "PrismaDipole", Float_t dz = 10., Float_t rmin = 70., 
										 Float_t rmax = 170., Float_t phi1 = -20, Float_t phi2 = 65);
   //! Add Prisma Dipole
   static TGeoVolume* AddPrismaDipole(const char *prismaName = "PrismaDipole", Float_t dx = 50., Float_t dy = 10.,Float_t dz = 1.);
   
   //! Add Multi Wire Proportional Chamber (MWPC)
   static TGeoVolume* AddPrismaMWPC(const char *prismaName = "PrismaMWPC", Float_t dzFcIc = 72., Float_t dx = 75., Float_t dy = 10., 
									Float_t dz = 65., TGeoMaterial* mat = 0x0);
   
   //! For uniformity of names overload ImportAgata method
   static TGeoVolume* BuildAgata(const char* agataPath = "./", const char *asolid = "asolid", const char *aclust = "aclust", const char *aeuler = "aeuler", 
								 const char *baseclusname = "agataCluster", const char* agataName = "Agata");
   
   //! Build Dante
   static TGeoVolume* BuildDante(const char* danteConf = "Dante.conf", const char* basemoduleName = "Module", const char *danteName = "Dante");
   
   //! Build Prisma
   static TGeoVolume* BuildPrisma(const char* prismaConf = "solver.conf", Float_t dzMCP = 25., Float_t dzQpole = 54., Float_t dzDipole = 160., Float_t dzDipMWPC = 328.5, const char* nameMCP = "PrismaMCP", 
								  const char* nameQPole = "PrismaQpole", const char* nameDipole = "PrismaDipole", const char* nameMWPC  = "PrismaMWPC",  const char* prismaName = "Prisma");
   //! Set Default tranparency
   static void SetDefaultTransparency(Char_t transparency) { fgDefaultTransp = transparency; }

//private:
   
   //! Read prisma configuration file
   static Bool_t ReadPrismaConf(Float_t& dzMCP, Float_t& dzQpole, Float_t& dzDipole, Float_t& dzDipMWPC, const char* prismaConf = "solver.conf");
   
private:
   
   TObjArray* fMatrixList;   //! list of transformation matrices  (rotation+translation for each crystal)
   
   mutable LogMessage fLog;  //! log message
   
private: 
   
   static const Int_t   fgkMaxDetectorID; //! Default number of crystal
   static const Int_t   fgkMaxDepthID;    //! Default number of slices (depth)
   static const Float_t fgkDepth[];       //! array of depths of sectors in mm 
   static const Char_t* fgkGeomName;      //! name of GeoManager
   static const Char_t* fgkGeomTitle;     //! title of GeoManager
   static       Char_t  fgDefaultTransp; // default transparency value

};
} // end namespace

#endif 







