/********************************************************************************************/
#ifndef GW_GPSICONVERTER_H
#define GW_GPSICONVERTER_H

#ifndef GW_HISTODB_H
#include "HistoDB.h"
#endif

#define MAX_SP 254

/*
Base class to read gpsi files.

Currently only reading of gpsi files is implemented to be able to get old spectra. The type
for this format is gpsi. So to open a gpsi file via HistoDB: \n
	- HistoDB hdb; hdb.Open("onegpsi","gpsi:toto.s");
	
You can browse it with hdb.ls() to get a full list of spectra in this file. You can load
single spectrum one by one or use the THStack to load several spectra in one operation. To
do this, just give a name to the THStack that is a pattern of the spectra you want. Suppose
you want to load only the spectra with a name starting with P1, then use HistoDB as following:
	- THstack hs; hs.SetName("P1*"); hdb >> hs; 

The THStack in not cleared by HistoDB, to be able to use it in adding mode. \n
The converter try to guess on which operating system the file has been created 
(little or big endian). So in principle, there no need to worry about where the file 
has been created. It also somehow checks if the given file is likely to be a gpsi file.
Those operations may produce a error when trying to read a spectrum.
	
\author Christophe Theisen & Olivier Stezowski	
*/

using namespace Gw;

class GPSIConverter: public HistoConverter 
{ 
private:
    std::ifstream fStreamIN;	  // stream used to read the gpsi file
    
    //! header file describing the number of spectra in the file 
    struct Header {
	short fNbSpectra;	// number of spectrum in xxx.s
	short fFileSize;	// filesize ( unit 512 Ko)
	short fWhere[MAX_SP];   // position to read/write spectrum
    } fHeader ;

    //! structure that defines a spectrum characteritics     
    struct Label {
	char  fSpName[12]; // name of the spectrum
	int   fNx;         // dimension x
	int   fNy;         // dimension y
    } fLabel[MAX_SP] ;
    
    //! it return false if something is wrong in the header part (not a gpsi file)    
    bool ReadHeaders();
      
    
public:
    GPSIConverter();
    GPSIConverter(const char *name);
    virtual ~GPSIConverter();

    virtual TH1 *Get(int);   
    virtual TH1 *Get(const char *);
    
    virtual bool Read(TH1 &);
    virtual bool Write(const TH1 &);

    virtual bool Read(TH2 &);
    virtual bool Write(const TH2 &);

    virtual bool Read(THStack &);
    virtual bool Write(const THStack &);

    virtual HistoConverter *NewDB(const char *) const;

    virtual const char *GetType() const { return fType.Data(); }
    
    virtual void ls(Option_t *) const ;   
     
    virtual void CheckDirectory(const char *name);
};


#endif
