/***************************************************************************
 *   Copyright (C) 2005-2006 by Christophe Theisen & Olivier Stezowski     *
 *   christophe.theisen(AT)cea.fr, stezow(AT)ipnl.in2p3.fr                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file HistoConverter.h header file for the HistoDB facility */
 
#ifndef GW_HISTOCONVERTER_H
#define GW_HISTOCONVERTER_H

#include "GwLogMessage.h"

// root's includes
#ifndef ROOT_TH1
#include <TH1.h>
#endif

#ifndef ROOT_TH2
#include <TH2.h>
#endif

#ifndef ROOT_THStack
#include <THStack.h>
#endif

#ifndef ROOT_TString
#include <TString.h>
#endif

// std's includes
#include <map>
#include <string>
#include <fstream>

// to correct a CC (solaris) bug
//	The fstream::read method does not work properly on CC solaris (at least on my machine ..)
//	Because it wasn't easy to change things that work properly on other
//	systems,  a wrapper is used and the fstream::get method which works is used instead.
//	
//	\warning if you use this wrapper, be careful that fstream::gcount does not
//	give you the correct answer .. better to use fstream::good to know if the reading was correct
//	
//	Much better to use the c or the system implementation in the futur ! 
//
/*
namespace {
class CCBug : public ifstream  
{
public:
       explicit CCBug();
	
       virtual istream &CCBug::read(char *, int);
 }; 
   
 CCBug::CCBug() : ifstream() { ; }
 
 istream & CCBug::read(char *s, int n) 
 {
	for ( int i = 0; i < n; i++ ) ifstream::get(s[i]);
	return (*this);
 }
 }
*/ 

namespace Gw {

/**
Base class that interfaces all other histograms DB systems.

An HistoConverter has a name and a type. The name is in most cases a path 
to a directory or a file. The type is a static value that is used to charaterise a DB system.\n

As for ROOT, there are two modes for reading/writings. The first one (default) is kVersion and
the second one is kOverwrite. In the first case, if one tries to write a spectrum SPNAME on
a DB system in which a spectrum with the same name already exists, the ROOT histogram will
be saved with the following name SPNAME;cycle (cycle being an integer corresponding to the 
number of times the spectrum has been saved, �la VMS). Because in many case the spectrum name 
is a file name and to avoid problems with the underlying file system, the base class HistoConverter
defines a character called fCycle that can be changed for instance by the '_' character. In this
case the name of the written spectrum will be SPNAME_cycle. In the second mode (kOverwrite),
the previous spectrum with the same name will be overwritten and then lost.

To determine if one operation has failed or not, a status flag exists and can take two 
values kFail if the last operation failed or kGood if it succeded. If case of failure,
a string is provided to explain the problem. 

It provides basic methods to read/write ROOT histograms that must be overwritten by the real
implementation. 

\author Christophe Theisen & Olivier Stezowski
*/

#define MAXCYCLE 999
#ifndef ROOT_TString
#include "TString.h"
#endif

class HistoConverter 
{
private:
    static std::map <const char *, HistoConverter *> *gPrototypes; 
    static Int_t gAllocated;

public:
    //! current status of the converter
    /*!
       good means the converter is working properly so far \n
       fail means the last operation has failed \n
       zombie means the converter has not been properly build
    */
    enum EStatus { kGood, kFail, kZombie };
    
     //! current working mode 
    /*!
       version means no spectrum will be overwritten and a versionning system is used. \n
       overwrite means if a spectrum is about to be saved and a spectrum with the same name
       already exists in the DB system, it will be overwritten and then lost.\n
    */   
    enum EMode   { kVersion, kOverwrite };

    static int fVerboseLevel; // 0 for no message when writing / reading spectra
    		              // 1 for message when writing / reading spectra. Default mode

protected:
    int fMode ;		// open mode: kVersion or kOverwrite
    TString fName;	// name of the histoconverter system
    
    int fStatus ;	// current status of the stream    
    TString fError;	// string to explain the last error in case the last operation failed

    char fCycle;	// character used in a versionning mode 

    TString fType;   	// character used for type definition (e.g.: eg, rad, gpsi, etc...) 
    
    TString fFileExt;  	// character used for file extension (e.g.: asc, spe, s, etc...) 

    bool fIsSwapping;	// true if swapping is needed to read the data
    
    TString fOptions; // option for the differents converter
  
    mutable LogMessage fLog; // log message
  
  
protected:    
    short Short(char *buf); 
    int   Int(char *buf);
    long  Long(char *buf);           
    float Float(char *buf);	

    unsigned short UShort(char *buf);	  
    unsigned int   UInt(char *buf);     
    unsigned long  ULong(char *buf);       
    
    //! check if the given directory can be read/written
    /*!
    	if the directory can not be accessed, it set status to zombie and fill
	the error string with an appropriate message. Useful for systems in which 
	one spectrum is stored per file (see RadConverter::RadConverter(const char *))
	
	It also expands to set environment variables like $(HOME), 
	and change the full path to the correct path depending on the operating system.
    */
    virtual void CheckDirectory(const char *name);
    
    TString CheckFileVersion(TString filename);
  
    void AddMineTypes();

           
public:
    HistoConverter();
    HistoConverter(const char *name);
    virtual ~HistoConverter();
    
      //! to add a new type of database service 
    static void AddPrototype(const char *, HistoConverter *) ;
    
      //! Get prototype map
    std::map <const char *, HistoConverter *>*  GetMapPrototype() {return gPrototypes;} 
    
     // ! set verbose level. 0 for no I/O messages. 1 (default) for I/O messages.
     static void SetVerboseLevel(const int);
    // ! return verbose level
     static int GetVerboseLevel();
    
    //! 
    int  GetMode() const { return fMode; }
    void SetMode(HistoConverter::EMode mode) { fMode = mode; }

    int  GetStatus() const { return fStatus; }
    const char *GetError() const { return fName.Data() ; }
    
    //! to set/reset the status of this converter
   void SetError(HistoConverter::EStatus status = kGood, const char *msgerror = "");
	
    //! to set/reset the status of this converter	
    void SetCycle(char c) { fCycle = c; }
    
     //! to set the type of the spectra/histo	
    void SetType(TString type) { fType = type; }
    
    //! to set/get file extension of the spectra/histo	
    void SetFileExt(TString ext) { fFileExt = ext; }
    TString GetFileExt() const { return fFileExt; }

    //! set options
    void SetOption(Option_t* opt, const char* mode = "reset");
    Option_t* GetOption() const {return fOptions.Data();}
    
        //! to get histo number i in the list
    /*!
    	returns NULL if it is not possible. It is the charge of the user to delete the returned histo 
    */
   virtual TH1 *Get(int /*i*/);
    //! to get an histo by its name  
    /*!
    	It returns NULL if it doesn't exist. 
	It is the charge of the user to delete the returned histogram 
    */     
   virtual TH1 *Get(const char */*name*/);
    
    //! to read an histogram 
    /*!
    	It returns false and set the status to kFail. The error message could be checked with GetError 
    */    
    virtual bool Read(TH1 &);  
    //! to write an histogram 
    /*!
    	It returns false and set the status to kFail. The error message could be checked with GetError 
    */        
    virtual bool Write(const TH1 &);
    //! to read an histogram 
    /*!
    	Only if supported. It returns kFail if it is not supported. \n
    	It returns false and set the status to kFail. The error message could be checked with GetError 
    */        
    virtual bool Read(TH2 &/*h*/) { return false; }     
    //! to write an histogram 
    /*!
    	Only if supported. It returns kFail if it is not supported. \n
    	It returns false and set the status to kFail. The error message could be checked with GetError 
    */            
    virtual bool Write(const TH2 &/*h*/) { return false;}
    
    //! to read several histograms 
    /*!
    	The stack is not cleared so the new histograms are added to the end. The name of the
	THStack could be used as a pattern to read only subset of the spectra stored in the DB system.
	Ex: \n
		HistoDB hdb; hdb.Open("mydir","rad:$(PWD)") \n
		THStack hs; hs.SetName("P*.spe"); hdb >> hs; // will read only spectra with filename P*.spe
    */            
    virtual bool Read(THStack &/*hs*/) { return false; }
    //! to write several histograms in one operation
    virtual bool Write(const THStack &/*hs*/) { return false; }  
        
    //! to creat a new DB system
    virtual HistoConverter *NewDB(const char *) const;
    
    //! to know what type of HistoConverter system it is   
    virtual const char *GetType() const { return "null" ; }
    
    //! to show the spectra in this DB system on the standard output     
    virtual void ls(Option_t *) const ;

    //! check mine types
    void CheckMineTypes();  
    
      //! draw spectrum for mine types
    static void DrawSpectrum(const char* spectrum);

    //! rootcint dictionary
    ClassDef(HistoConverter,0); // base class to read/write histograms  
};
}
#endif
