/***************************************************************************
 *   Copyright (C) 2005-2006 by Christophe Theisen & Olivier Stezowski     *
 *   christophe.theisen(AT)cea.fr, stezow(AT)ipnl.in2p3.fr                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/** \file HistoDB.h header file for the HistoDB facility */
 
#ifndef GW_HISTODB_H
#define GW_HISTODB_H

// root's includes
#ifndef ROOT_TH1
#include <TH1.h>
#endif

#ifndef ROOT_TH2
#include <TH2.h>
#endif

#ifndef ROOT_THStack
#include <THStack.h>
#endif

#ifndef ROOT_TString
#include <TString.h>
#endif

// std's includes
#include <map>
#include <string>
#include <fstream>

#ifndef GW_HISTOCONVERTER_H
#include "HistoConverter.h"
#endif

namespace Gw {

/**
Service to read/write ROOT histograms from/to different DB systems

This class is an interface to help reading/writing/converting ROOT histograms from/to
other existing or future way to store spectra. \n
The interface is design as a growing system to help adding new DB types without
having to recompile HistoDB. If you want to add your own new format, you will have to write
a new class that inherits from HistoConverter. To help you in this task, HistoDB provides
the method MakeConverter that built for you the basic and needed structure of the new class.
This class inherits from HistoConverter. DO NOT DELETE anything from this file.

Current supported convertion format are:
	- null   : black hole
	- rad    : Radware for one and two dimensionnal spectra [RADConverter]
	- gpsi   : reading mode only [GPSIConverter]
	- eg     : reading mode only [EGConverter]
	- ascii  : reading mode only [ASCConverter]

If you want to add your own format, see RADConverter as a example.

Some way to use this interface (see also DemoHistoDB.C):

	TH1F h;
	HistoDB hdb;
	
	hdb.Open("in","gpsi:/where/i/have/my/spectra.s");
	hdb.Open("out","rad:/where/i/want/to/convert");
	
	// check out that they are both open.
	hdb.ls("all");
	
	h.SetName("the_name_of_the_specrum_i_want_to_convert");
	hdb << "in" >> h << "out" << h; // to read a gpsi spectrum and convert it into a rad format
	
\author Christophe Theisen & Olivier Stezowski
*/
class HistoDB
{    
    
private:
    std::map <const char *, HistoConverter *> fConnections; // list of open connections to DB services
    
    HistoConverter *fCurrentDB;	// current DB system to read/write ROOT histograms
    
public:
    enum EStatus { kGood, kFail };
        
protected:
    int fStatus ; // current status of the last operation when using << and >>

public:
    HistoDB();
    virtual ~HistoDB();

    //! to get the current status flag for the HistoDB stream           
    virtual bool IsStatus(HistoDB::EStatus status)  { return fStatus == status ; } 
    //! modify the status flag for the HistoDB stream       
    virtual void SetStatus(HistoDB::EStatus status = kGood) { fStatus = status ; }
    
    // ! set verbose level. 0 for no I/O messages. 1 (default) for I/O messages.
     void SetVerboseLevel(const int);
    // ! return verbose level
     int GetVerboseLevel();
    
    
    //! to change the mode (overwrite/version) of the current DB service   
    virtual void SetMode(Option_t *) ;
    //! to change the mode (overwrite/version) of one of the DB service       
    virtual void SetMode(const char *, Option_t *) ;
    
    //! to get histo number i of the current DB
    /*!
    	returns NULL if it is not possible. It is the charge of the user to delete the returned histo 
    */
    virtual TH1 *Get(int i);  
    //! to get an histo by its name  
    /*!
    	It returns NULL if it doesn't exist. 
	It is the charge of the user to delete the returned histogram 
    */     
    virtual TH1 *Get(const char *name);
    
    
    //! to write a TH1 histogram in the current Histo DB service
    virtual bool Write(const TH1 &);
    //! to write a TH2 histogram in the current Histo DB service
    virtual bool Write(const TH2 &);

    //! to write an histogram in the current Histo DB service
    virtual HistoDB & operator << (const TH1 &);
    //! to read an histogram from the current Histo DB service  
    virtual HistoDB & operator >> (TH1 &);   
 
    //! to write a matrix in the current Histo DB service
    virtual HistoDB & operator << (const TH2 &);
    //! to read a matrix from the current Histo DB service  
    virtual HistoDB & operator >> (TH2 &);   
       
    //! to change the current DB service
    virtual HistoDB & operator << (const char *name) { SetCurrent(name) ; return (*this); }     

    //! to write several histograms in the current Histo DB service
    virtual HistoDB & operator << (const THStack &);
    //! to read several histograms from the current Histo DB service  
    virtual HistoDB & operator >> (THStack &);      

    //! change the current service. It uses the alias defined in open
    virtual void SetCurrent(const char *);     
       
    //! to open a new DB service 
    /*!
    	It returns NULL if:
	- the service is not known
	- the connection cannot be established 
	- the converter has been open but in a zombie status
	
	If it fails, it is not added to the list of DB services.
	
	The returned HistoConverter provides a interface to any implementation. However
	it could not be plainty used in the root interpreter because the real implementation
	are not known by root (no dictionnary). 
    */  
    virtual HistoConverter *Open(const char *, const char *);        

    //! to list the current configuration
    /*!
       The following options are available:
	- proto: list the type of HistoConverter this DB service knows
	- cdb:   to get informations on the current service
	- all:   to get informations on all opened services
    */
    virtual void ls(Option_t *opt = "cdb") ;

    //! Use it if you want to extend the DB types handled
    /*! 
    	it creates two files classname.h and classname.C - 
	DO NOT erase anything from those files.
     */  
    void MakeConverter(const char *classname) const ;
    
    //! add file extension
    void AddFileExt(const char* type, const char* ext);
    
    //! add option
    void AddOption(const char* type, Option_t* opt, const char* mode = "reset");
    
    //! rootcint dictionary
    ClassDef(HistoDB,0); // Service to read/write ROOT histograms to others DB systems
};
// inline members

} // namespace

#endif
