//
//
// MatPlayer Version 0.1
// MatPlayer is used to analyse a 2D histogram by putting gates interactively.
//
// A MatPlayer object must be connected to a TCanvas and a TH2D histogram. It is done by using SetCanvas & SetMatrix.
// Once a canvas and a matrix has been linked, any key pressed on the defined canvas gives a gate (TBox) that can be
// modify graphically. The gates on the current canvas are use to  
//
//	Olivier Stezowski - 
//
// 
//

#ifndef GW_MATPLAYER_H
#define GW_MATPLAYER_H

#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

#ifndef ROOT_TBox
#include "TBox.h"
#endif

#ifndef ROOT_TList
#include "TList.h"
#endif

#ifndef ROOT_TFormula
#include "TFormula.h"
#endif

#ifndef ROOT_TH1
#include "TH1.h"
#endif

#ifndef ROOT_TH2
#include "TH2.h"
#endif

#include <iostream> 
#include <utility> 

#include <RQ_OBJECT.h>

namespace Gw {

/**
	MatPlayer is a tool to play with 2D matrices. It is mainly designed in the spirit of gamma-gamma matrices but could be used
	also for general purposes.

	@author Olivier Stezowski
 */
class MatPlayer : public TNamed
{
	RQ_OBJECT("MatPlayer")
	
private:
	static Int_t fLastX, fLastY;	//! last (x,y) position of the moose
	static TList *fgPlayers;	//! list of all players opened .. in case a player is called from the command line
		
private:
	//! Box characteristics 
	TBox    fBoxX;	// Use to draw X-gates on the current pad
	TBox    fBoxY;	// Use to draw Y-gates on the current pad
	TBox    fBoxBX;	// Use to draw BX-gates on the current pad	
	TBox    fBoxBY;	// Use to draw BY-gates on the current pad	
	
	//! Default with for gate X and Y
	std::pair<TFormula *, TFormula *> fWidthGate; 
   	
	//! Default stepping for gate X and Y in case of 
	std::pair<Float_t,Float_t> fStep; 	
   
	TList   fLastGates;	// to keep a copy of the last gates list
	// TList   fGatesY;	// to keep a current copy the gates list
	// TList   fGatesBX;	// to keep a current copy the gates list
	// TList   fGatesBY;	// to keep a current copy the gates list
	
private:
	TH2       *fMat;     	//! 2D matrix to work with
   
	TH1D      *fPx;		//! Current projection on X
	TH1D      *fPy;		//! Current protection on Y
	TH1D      *fBx;		//! Current background on X
	TH1D      *fBy;		//! Current background on Y
	TH1D      *fTotx;	//! Current total protection on X
	TH1D      *fToty;	//! Current total protection on Y
		
private:
	//! Matrix to be analysed (see SetMatrix(const char *))
	void SetMatrix(TH2 *);
	//! Reset histograms .. called when this is initialised to work on a new matrix
	virtual void Clear(const Option_t *opt="");
	
	//! true if the first box shares the same properties of the second one
	bool IsBox(const TBox *, const TBox *) const;
	
	static void SubstractBG_RAD(TH2 *);
	static void SubstractBG_SNIP(TH2 *);

protected:	
	//! set a gate using the last mouse position 
	/*!
		Four options are available:
		- Gx: gate on the first axis of the matrix
		- Gy: gate on the second axis of the matrix
		- Bx, By:  background gate on the first and second axis of the matrix
	*/
	void SetGateX(Option_t *opt = "Gx");	
	Int_t InitGating(Option_t *opt = "Gx");
	
public:
	MatPlayer();
	virtual ~MatPlayer();
	
	//! To get the list of all open MatPlayers
	static TList *GetListOfPlayers() { return fgPlayers; }
   
	//! Generic method to substract background of the current matrix 
	static void SubstractBG(const char *, Option_t *);
  
	//!  set current canvas as the mother of the current pad
	void SetCanvas(); 
	
	//! Matrix to be analysed
	/*!
		Load the matrix (if needed). In the matrix directory, several 
		spectra are created (if not yet done) corresponding to different projections:
		- TH1D      *fPx;		//! Current protection on Y		
		- TH1D      *fPy;		//! Current protection on Y
		- TH1D      *fBx;		//! Current background on X
		- TH1D      *fBy;		//! Current background on Y
		- TH1D      *fTotx;		//! Current total protection on X
		- TH1D      *fToty;		//! Current total protection on Y	
	*/
	void SetMatrix(const char *);
   
 	//! to reset the conditions of gating in the current pad
	void Reset();  
	//! to show the gates used for the last gating 
	void ShowGates(Option_t *o = "BGx");	 
		
//  void SetGate(const Char_t *filename ="MatPlayer.Gx", Option_t *opt = "Gx"); // to set gates for a formated file
//	void SetGate2D(Double_t, Double_t, Option_t *);	// to set a gate a position x1 and X2
	
	//! get width for gate at position pos
	Double_t GetWidthGate(Double_t pos, Option_t *opt = "Gx");
	//! set default width for gates
	void SetWidthGate(Double_t wx = 2.5, Double_t wy = 2.5) { fWidthGate.first->SetParameter(0,wx); fWidthGate.second->SetParameter(0,wy); } 
	
	//! to project
	void ProjectionY(Option_t *option);
     
  	//! slots
   	void XEventAction(Int_t event, Int_t px, Int_t py, TObject *obj);
	
	//! rootcint dictionary
	ClassDef(MatPlayer,0); // MatPlayer  	
};

} // namespace

#endif
