/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file Buffer.h header file for Buffer.cpp */

#ifndef GWS_BUFFER_H
#define GWS_BUFFER_H

#ifndef GWS_MEMORY_H
#include "Memory.h"
#endif

#include <iostream>
using namespace std;

namespace Gws {

//! A buffer is used to read/write raw data buffers from/on files
/*!
	It is an array of fSize bytes with a cursor on it. Stream operators are defined for some basic types. \n
	By default, the size of the buffer is 32 kBYTE. For some reason, the allocation could failed in the constructor.
	In this case the kBad flags is set. Be sure your buffer is good after creating it before any readings/writings.
	
	@author Olivier Stezowski
*/
class Buffer : protected Memory
{
public:
    enum EStatus { kBad = BIT(1), kGood = BIT(2), kEoB = BIT(3), kFail = BIT(4), kCorrupt = BIT(5) } ;
    //! Status for this buffer
    /*!
    	-  kBad:	true if this buffer is not allocated	   
    	 - kGood:	everything is all right
	 - kEoB:	the end of the buffer has been reached
	 - kFail:	the last operation failed
	 - kCorrupt:	More bytes than the buffer size has been added.
	 
	 The real allocated buffer has a size fSize+2 and the last two slots are initialised to specific values ( E & B for End of Block).
	 If those last values are not what is expected, the buffer status is set to kCorrupt.  
	 
	 \todo - the copy constructor must ne written as fast as possible !
    */
private:
    static const Char_t fC1 = 'E', fC2 = 'B';
    
protected:    
    Int_t 	fStatus;
    Int_t	fEndian;
    
    Char_t * 	fBuffer;
    UInt_t 	fSize;
    UInt_t 	fCurrent;
    
    void DoAlloc(UInt_t);
  
public:
    //! Default size for a Buffer is 32 KBYTE.
    /*!
    */
    explicit Buffer(UInt_t s = 32 * KBYTE);
    virtual ~Buffer();
    
    //! copy n bytes from one buffer to another one
    /*!
	Conventions: copy from the first buffer (current position) to the second one n bytes. \n
	if the destination buffer is to 
    */
    // static void Copy(const Buffer &, Buffer &, Int_t);
    //! return the right buffer taking Int_to account the endian argument compared to the underlying endian system type. 
    /*!

    */    
    static Buffer * New(Memory::EEndian e, UInt_t s = 32 * KBYTE);

    bool IsStatus(Buffer::EStatus);
    void SetStatus(Buffer::EStatus);
    
    bool IsBytes(Memory::EEndian) const;
    void SetBytes(Memory::EEndian);
    
    //! Pointer to the underlying array of chars.
    /*!
	\warning DO NOT delete it !!
    */
    Char_t *GetBuffer();
    //! it returns the maximum number of bytes in this buffer   
    UInt_t Size() const;
    //! it returns the number of free bytes to the end
    UInt_t FreeSize() const;   
     //! it returns the current position in the buffer   
    UInt_t Offset() const;
    
    //! change the current position.
     /*!
     	detail
     */ 
    UInt_t SetOffset(UInt_t off = 0);
    
    //! Reset means set all elements to 0 and the current position is 0
    /*!
    	to change also the size of this buffer, use Reset(UInt_t)
    */
    virtual void Reset(); 
    virtual void Reset(UInt_t); 
    
    
    //! Expand the actual size of the buffer
    /*!
    	the content of the buffer is not change, the adding part is set to 0.
	
	It returns false if something goes wrong; true otherwise. In case of problems, the Buffer is unchanged
    */   
    virtual Bool_t Expand(UInt_t);
    
    
    virtual Buffer &operator << (Char_t);
    virtual Buffer &operator << (UChar_t);
    virtual Buffer &operator << (Short_t);
    virtual Buffer &operator << (UShort_t);
    virtual Buffer &operator << (Int_t);
    virtual Buffer &operator << (UInt_t);
    virtual Buffer &operator << (Float_t);
    virtual Buffer &operator << (Double_t);    
//    virtual Buffer &operator << (long);
//    virtual Buffer &operator << (unsigned long);
    
    virtual Buffer &operator>> (Char_t &);
    virtual Buffer &operator>> (UChar_t &);
    virtual Buffer &operator>> (Short_t &);
    virtual Buffer &operator>> (UShort_t &);
    virtual Buffer &operator>> (Int_t &);
    virtual Buffer &operator>> (UInt_t &);
    virtual Buffer &operator>> (Float_t &);
    virtual Buffer &operator>> (Double_t &); 
//    virtual Buffer &operator>> (long &);
//    virtual Buffer &operator>> (unsigned long &);
};
// inline Buffer members
inline bool Buffer::IsStatus(Buffer::EStatus s) { 
	if ( s == Buffer::kCorrupt )
		if ( fBuffer[fSize] != fC1 && fBuffer[fSize+1] != fC2 )  fStatus |= Buffer::kCorrupt;
	return fStatus & s; 
}
inline void Buffer::SetStatus(Buffer::EStatus s) { fStatus = s; }
inline bool Buffer::IsBytes(Memory::EEndian e) const { return fEndian == e; }
inline void Buffer::SetBytes(Memory::EEndian e)  { fEndian = e; }
inline Char_t * Buffer::GetBuffer() { return fBuffer ; }
inline UInt_t Buffer::Size() const { return fSize; }
inline UInt_t Buffer::Offset() const { return fCurrent ; }
inline UInt_t Buffer::FreeSize() const { return fSize - fCurrent; }       
inline Buffer &Buffer::operator << (Char_t data) { 
	if (FreeSize()>=sizeof(Char_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(Char_t)); fCurrent += sizeof(Char_t); }
	else fStatus = Buffer::kFail ;
	return (*this);
}
inline Buffer &Buffer::operator << (UChar_t data){ 
	if (FreeSize()>=sizeof(Char_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(UChar_t)); fCurrent += sizeof(Char_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);
}
inline Buffer &Buffer::operator << (Short_t data) {
	if (FreeSize()>=sizeof(Short_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(Short_t)); fCurrent += sizeof(Short_t); }
	else fStatus = Buffer::kFail ;
	return (*this);
}
inline Buffer &Buffer::operator << (UShort_t data){
	if (FreeSize()>=sizeof(Short_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(UShort_t)); fCurrent += sizeof(UShort_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &Buffer::operator << (Int_t data) {
	if (FreeSize()>=sizeof(Int_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(Int_t)); fCurrent += sizeof(Int_t); }
	else fStatus = Buffer::kFail ;
	return (*this);
}
inline Buffer &Buffer::operator << (UInt_t data){
	if (FreeSize()>=sizeof(Int_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(UInt_t)); fCurrent += sizeof(UInt_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &Buffer::operator << (Float_t data) { 
	if (FreeSize()>=sizeof(Float_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(Float_t)); fCurrent += sizeof(Float_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &Buffer::operator << (Double_t data) {
	if (FreeSize()>=sizeof(Double_t)) { ::memcpy((fBuffer+fCurrent),&data,sizeof(Double_t));fCurrent += sizeof(Double_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
/*
inline Buffer &Buffer::operator << (long data){ 
	if (FreeSize()>=sizeof(long)) { Memory::Copy((Char_t*)&data,(fBuffer+fCurrent),sizeof(long));fCurrent += sizeof(long); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &Buffer::operator << (unsigned long data) {
	if (FreeSize()>=sizeof(long)) { Memory::Copy((Char_t*)&data,(fBuffer+fCurrent),sizeof(long));fCurrent += sizeof(long); }
	else fStatus = Buffer::kFail ;// inline Buffer members
	return (*this);
}*/

inline Buffer &Buffer::operator >> (Char_t &data){
	if (FreeSize()>=sizeof(Char_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(Char_t));fCurrent += sizeof(Char_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &Buffer::operator >> (UChar_t &data){
	if (FreeSize()>=sizeof(Char_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(UChar_t));fCurrent += sizeof(UChar_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &Buffer::operator >> (Short_t &data){
	if (FreeSize()>=sizeof(Short_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(Short_t));fCurrent += sizeof(Short_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &Buffer::operator >> (UShort_t &data){
	if (FreeSize()>=sizeof(Short_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(UShort_t));fCurrent += sizeof(UShort_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer & Buffer::operator >> (Int_t &data){
	if (FreeSize()>=sizeof(Int_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(Int_t));fCurrent += sizeof(Int_t); }
	else fStatus = Buffer::kFail ;	
	return (*this) ;
}
inline Buffer &Buffer::operator >> (UInt_t &data){
	if (FreeSize()>=sizeof(Int_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(UInt_t));fCurrent += sizeof(UInt_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &Buffer::operator >> (Float_t &data){
	if (FreeSize()>=sizeof(Float_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(Float_t));fCurrent += sizeof(Float_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &Buffer::operator >> (Double_t &data){
	if (FreeSize()>=sizeof(Double_t)) { ::memcpy(&data,fBuffer+fCurrent,sizeof(Double_t));fCurrent += sizeof(Double_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}

// inline Buffer members //////////////////////////

//! A Endianbuffer is used to read/write raw data buffers from/on files
/*!
	Should never be used directly. Use Buffer::New instead
	
	@author Olivier Stezowski
*/
class EndianBuffer : public Buffer
{  
public:
    EndianBuffer(UInt_t s = 32 * KBYTE);
    virtual ~EndianBuffer();

    virtual Buffer &operator << (Char_t) ;
    virtual Buffer &operator << (UChar_t);
    virtual Buffer &operator << (Short_t);
    virtual Buffer &operator << (UShort_t);
    virtual Buffer &operator << (Int_t);
    virtual Buffer &operator << (UInt_t);
    virtual Buffer &operator << (Float_t);
    virtual Buffer &operator << (Double_t); 
//    virtual Buffer &operator << (long);
//     virtual Buffer &operator << (unsigned long);   

    virtual Buffer &operator >> (Char_t &);
    virtual Buffer &operator >> (UChar_t &);    
    virtual Buffer &operator >> (Short_t &);
    virtual Buffer &operator >> (UShort_t &);
    virtual Buffer &operator >> (Int_t &);
    virtual Buffer &operator >> (UInt_t &);
    virtual Buffer &operator >> (Float_t &);
    virtual Buffer &operator >> (Double_t &);
//     virtual Buffer &operator >> (long &);
//     virtual Buffer &operator >> (unsigned long &);
};
// inline EndianBuffer members
inline Buffer &EndianBuffer::operator << (Char_t data)  { return Buffer::operator << (data) ;}
inline Buffer &EndianBuffer::operator << (UChar_t data) { return Buffer::operator << (data); }    
inline Buffer &EndianBuffer::operator << (Short_t data) {
	if (FreeSize()>=sizeof(Short_t)) { Memory::Swap(data,fBuffer+fCurrent);fCurrent += sizeof(Short_t); }
	else fStatus = Buffer::kFail ;
	return (*this);
}
inline Buffer &EndianBuffer::operator << (UShort_t data){
	if (FreeSize()>=sizeof(Short_t)) { Memory::Swap(data,fBuffer+fCurrent); fCurrent += sizeof(Short_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &EndianBuffer::operator << (Int_t data) {
	if (FreeSize()>=sizeof(Int_t)) { Memory::Swap(data,fBuffer+fCurrent);fCurrent += sizeof(Int_t); }
	else fStatus = Buffer::kFail ;
	return (*this);
}
inline Buffer &EndianBuffer::operator << (UInt_t data){
	if (FreeSize()>=sizeof(Int_t)) { Memory::Swap(data,fBuffer+fCurrent);fCurrent += sizeof(Int_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}

inline Buffer &EndianBuffer::operator << (Float_t data) { 
	if (FreeSize()>=sizeof(Float_t)) { Memory::Swap(data,fBuffer+fCurrent); fCurrent += sizeof(Float_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &EndianBuffer::operator << (Double_t data) {
	if (FreeSize()>=sizeof(Double_t)){  Memory::Swap(data,fBuffer+fCurrent);fCurrent += sizeof(Double_t); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
/*
inline Buffer &EndianBuffer::operator << (long data){ 
	if (FreeSize()>=sizeof(long)) { Memory::Swap(data,fBuffer+fCurrent);fCurrent += sizeof(long); }
	else fStatus = Buffer::kFail ;
	return (*this);	
}
inline Buffer &EndianBuffer::operator << (unsigned long data) {
	if (FreeSize()>=sizeof(long)){  Memory::Swap(data,fBuffer+fCurrent);fCurrent += sizeof(long); }
	else fStatus = Buffer::kFail ;
	return (*this);
}
*/
inline Buffer &EndianBuffer::operator >> (Char_t &data)  { return Buffer::operator >> (data) ;}
inline Buffer &EndianBuffer::operator >> (UChar_t &data) { return Buffer::operator >> (data); }    
inline Buffer &EndianBuffer::operator >> (Short_t &data){
	if (FreeSize()>=sizeof(Short_t)){ Memory::Swap(fBuffer+fCurrent,data); fCurrent += sizeof(Short_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &EndianBuffer::operator >> (UShort_t &data){
	if (FreeSize()>=sizeof(Short_t)){  Memory::Swap(fBuffer+fCurrent,data);fCurrent += sizeof(Short_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &EndianBuffer::operator >> (Int_t &data){
	if (FreeSize()>=sizeof(Int_t)){ Memory::Swap(fBuffer+fCurrent,data);  fCurrent += sizeof(Int_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &EndianBuffer::operator >> (UInt_t &data){
	if (FreeSize()>=sizeof(Int_t)){  Memory::Swap(fBuffer+fCurrent,data);fCurrent += sizeof(Int_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
/*
inline Buffer &EndianBuffer::operator >> (long &data){
	if (FreeSize()>=sizeof(long)){  Memory::Swap(fBuffer+fCurrent,data); fCurrent += sizeof(long); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &EndianBuffer::operator >> (unsigned long &data){
	if (FreeSize()>=sizeof(long)){  Memory::Swap(fBuffer+fCurrent,data); fCurrent += sizeof(long); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
*/
inline Buffer &EndianBuffer::operator >> (Float_t &data){
	if (FreeSize()>=sizeof(Float_t)){  Memory::Swap(fBuffer+fCurrent,data); fCurrent += sizeof(Float_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
inline Buffer &EndianBuffer::operator >> (Double_t &data){
	if (FreeSize()>=sizeof(Double_t)){  Memory::Swap(fBuffer+fCurrent,data); fCurrent += sizeof(Double_t); }
	else fStatus = Buffer::kFail ;	
	return (*this);	
}
}; // gws

#endif
