/***************************************************************************
 *   Copyright (C) 2004 by Olivier Stezowski                               *
 *   stezow(AT)ipnl.in2p3.fr                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   aLong_t with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GWS_MEMORY_H
#define GWS_MEMORY_H

#ifndef GWS_BASE_H
#include "Base.h"
#endif

#ifndef ROOT_Bytes
#include <Bytes.h>
#endif


namespace Gws
{ 

const Int_t  BYTE = 8;
const Int_t KBYTE = 1024;
const Int_t MBYTE = KBYTE*KBYTE;
const Int_t GBYTE = KBYTE*MBYTE;

//! Interface to the memory
/*!
	It defines a set of useful functions to deal with buffers of basic types. It also determined the endian
	type of the system you are working with. 
	This class is used by Buffer and EndianBuffer to get buffers that can be read whatever their origin 
	as soos as the endian type of the source of data is known. 
*/
class Memory : public Base 
{
public:
    /**
       The adjectives big-endian and little-endian refer to which bytes are most significant in multi-byte data 
       types and describe the order in which a sequence of bytes is stored in a computer’s memory.
       In a big-endian system, the most significant value in the sequence is stored at the lowest storage address (i.e., first). 
       In a little-endian system, the least significant value in the sequence is stored first. 
       
       \todo check byte swapping for Long_t, Float_t & Double_t
    */
    enum EEndian { kLittle = 0, kBig = 1 } ;
    
private:
    //! Agent that determines the endian type at starting time
    /*!
    	The endian type is determined in the constructor (many thanks to Fons for the tip) which is called at least once when allocating gAgent
    */
    class Agent {
    	Int_t fEndian; 
    public:
    	Agent();
	Int_t GetEndian() { return fEndian; }
    };
    static Agent *gAgent;
    
public:
    Memory();
    virtual ~Memory();
    
    //! check out the endian type of the running system
    /*!
    
    */
    static bool IsBytes(Memory::EEndian e) { return gAgent->GetEndian() == e; }
    static Int_t GetEndian() { return gAgent->GetEndian() ; }
       
    //! To allocate a buffer that is set to 0 (some compilers don't do it)
    /*!
  	if New has been used to allocate, you must use the member Delete to delete
	or delete it with [] since new [] is used even if just one value is allocated.  	
    */
    static Char_t   *New(Char_t   *p, Int_t nb = 32*KBYTE);
    static Short_t  *New(Short_t  *p, Int_t nb = 32*KBYTE);
    static Int_t    *New(Int_t    *p, Int_t nb = 32*KBYTE);
    static UChar_t  *New(UChar_t  *p, Int_t nb = 32*KBYTE);
    static UShort_t *New(UShort_t *p, Int_t nb = 32*KBYTE);
    static UInt_t   *New(UInt_t   *p, Int_t nb = 32*KBYTE);
    static Float_t  * New(Float_t *p, Int_t nb = 32*KBYTE);
    static Double_t * New(Double_t *p, Int_t nb = 32*KBYTE);
//    static Long_t   *New(Long_t   *p, Int_t nb = 32*KBYTE);
//    static ULong_t  *New(ULong_t  *p, Int_t nb = 32*KBYTE);
   
    static Char_t   *New(Char_t   *p, UInt_t nb = 32*KBYTE);
    static Short_t  *New(Short_t  *p, UInt_t nb = 32*KBYTE);
    static Int_t    *New(Int_t    *p, UInt_t nb = 32*KBYTE);
    static Long_t   *New(Long_t   *p, UInt_t nb = 32*KBYTE);
    static UChar_t  *New(UChar_t  *p, UInt_t nb = 32*KBYTE);
    static UShort_t *New(UShort_t *p, UInt_t nb = 32*KBYTE);
    static UInt_t   *New(UInt_t   *p, UInt_t nb = 32*KBYTE);
    static Float_t  * New(Float_t *p, UInt_t nb = 32*KBYTE);
    static Double_t * New(Double_t *p, UInt_t nb = 32*KBYTE);   
     
//    static Long_t   *New(Long_t   *p, UInt_t nb = 32*KBYTE);
//    static ULong_t  *New(ULong_t  *p, UInt_t nb = 32*KBYTE);

    static void Delete(Char_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(Short_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(Int_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(UChar_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(UShort_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(UInt_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(Float_t *p)	{ ::delete [] p; p = NULL; }
    static void Delete(Double_t *p)	{ ::delete [] p; p = NULL; }

//    static void Delete(Long_t *p)	{ ::delete [] p; p = NULL; }
//    static void Delete(ULong_t *p)	{ ::delete [] p; p = NULL; }	
    
    //! fast initialization of the buffer at 0
    /*!
	It overwrites the memset function 
    */
    static void Zero(Char_t   *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(Short_t  *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Short_t)*nb); }
    static void Zero(Int_t    *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Int_t)*nb);   }
    static void Zero(UChar_t  *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(UShort_t *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(UShort_t)*nb); }
    static void Zero(UInt_t   *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(UInt_t)*nb);   }
    static void Zero(Float_t  *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Float_t)*nb); }
    static void Zero(Double_t *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Double_t)*nb); }
        
//    static void Zero(Long_t   *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Long_t)*nb);  }
//    static void Zero(ULong_t  *p, Int_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(ULong_t)*nb);  }
    
    static void Zero(Char_t   *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(Short_t  *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Short_t)*nb); }
    static void Zero(Int_t    *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Int_t)*nb);   }
    static void Zero(UChar_t  *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,nb); }
    static void Zero(UShort_t *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(UShort_t)*nb); }
    static void Zero(UInt_t   *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(UInt_t)*nb);   }
    static void Zero(Float_t  *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Float_t)*nb); }
    static void Zero(Double_t *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Double_t)*nb); } 
      
//    static void Zero(Long_t   *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(Long_t)*nb);  }    
//    static void Zero(ULong_t  *p, UInt_t nb) { if ( nb < 1 ) return; ::memset(p,0,sizeof(ULong_t)*nb);  }
    
    //! copy the content from -> to
    /*!
    	detail
    */
    static void Copy(const Char_t *from, Char_t *to, Int_t size) { ::memcpy(to,from,size); }
    
    //! the Swap members
    /*!
    
    	Swap(T, Char_t *buf) writes T in the buffer pointed by buf 
	This members used the ROOT macros defined in Bytes.h which take into account byte swapping optimizations from the some systems.
    */
    static void Swap(Short_t, Char_t *);
    static void Swap(UShort_t, Char_t *); 
    static void Swap(Int_t, Char_t *);
    static void Swap(UInt_t, Char_t *);
    static void Swap(Float_t, Char_t *);
    static void Swap(Double_t, Char_t *);
//    static void Swap(Long_t, Char_t *);
//    static void Swap(ULong_t, Char_t *);   
    
    static void Swap(Char_t *,Short_t &);
    static void Swap(Char_t *,UShort_t &); 
    static void Swap(Char_t *,Int_t &);
    static void Swap(Char_t *,UInt_t &);   
    static void Swap(Char_t *,Float_t &);
    static void Swap(Char_t *,Double_t &);    
//    static void Swap(Char_t *,Long_t &);
//    static void Swap(Char_t *,ULong_t &);  
};

// Memory inline members //////////////////////////////////////////////////////
inline Char_t   * Memory::New(Char_t   *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new Char_t[nb];   if ( p ) Zero(p,nb); return p; }
inline Short_t  * Memory::New(Short_t  *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new Short_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Int_t    * Memory::New(Int_t    *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new Int_t[nb];    if ( p ) Zero(p,nb); return p; }
inline UChar_t  * Memory::New(UChar_t  *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new UChar_t[nb];  if ( p ) Zero(p,nb); return p; }
inline UShort_t * Memory::New(UShort_t *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new UShort_t[nb]; if ( p ) Zero(p,nb); return p; }
inline UInt_t   * Memory::New(UInt_t   *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new UInt_t[nb];   if ( p ) Zero(p,nb); return p; }
inline Float_t  * Memory::New(Float_t  *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new Float_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Double_t * Memory::New (Double_t * p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new Double_t[nb]; if ( p ) Zero(p,nb); return p; }
//inline Long_t   * Memory::New(Long_t   *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new Long_t[nb];   if ( p ) Zero(p,nb); return p; }
//inline ULong_t  * Memory::New(ULong_t  *p,Int_t nb) { if ( nb < 1 ) return NULL; p = ::new ULong_t[nb];  if ( p ) Zero(p,nb); return p; }

inline Char_t   * Memory::New(Char_t   *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new Char_t[nb];   if ( p ) Zero(p,nb); return p; }
inline Short_t  * Memory::New(Short_t  *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new Short_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Int_t    * Memory::New(Int_t    *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new Int_t[nb];    if ( p ) Zero(p,nb); return p; }
inline UChar_t  * Memory::New(UChar_t  *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new UChar_t[nb];  if ( p ) Zero(p,nb); return p; }
inline UShort_t * Memory::New(UShort_t *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new UShort_t[nb]; if ( p ) Zero(p,nb); return p; }
inline UInt_t   * Memory::New(UInt_t   *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new UInt_t[nb];   if ( p ) Zero(p,nb); return p; }
inline Float_t  * Memory::New(Float_t  *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new Float_t[nb];  if ( p ) Zero(p,nb); return p; }
inline Double_t * Memory::New (Double_t * p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new Double_t[nb]; if ( p ) Zero(p,nb); return p; }
//inline Long_t   * Memory::New(Long_t   *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new Long_t[nb];   if ( p ) Zero(p,nb); return p; }
//inline ULong_t  * Memory::New(ULong_t  *p,UInt_t nb) { if ( nb < 1 ) return NULL; p = ::new ULong_t[nb];  if ( p ) Zero(p,nb); return p; }


inline void Memory::Swap(UShort_t x, Char_t *buf)
{
# if defined(__linux) && defined(__i386__)
   *((UShort_t *)buf) = Rbswap_16(x);
# else
   // To work around a stupid optimization bug in MSVC++ 6.0
   const UShort_t *intermediary = &x;
   char *sw = (char *) intermediary;
   buf[0] = sw[1];
   buf[1] = sw[0];
# endif
}
inline void Memory::Swap(Short_t x, Char_t *buf) { Memory::Swap((UShort_t) x,buf) ;}

inline void Memory::Swap(UInt_t x, Char_t *buf)
{
# if defined(__linux) && defined(__i386__)
   *((UInt_t *)buf) = Rbswap_32(x);
# else
   // To work around a stupid optimization bug in MSVC++ 6.0
   const UInt_t *intermediary = &x;
   char *sw = (char *)intermediary;
   buf[0] = sw[3];
   buf[1] = sw[2];
   buf[2] = sw[1];
   buf[3] = sw[0];
# endif
}
inline void Memory::Swap(Int_t x, Char_t *buf) { Memory::Swap((UInt_t) x,buf);}

inline void Memory::Swap(Float_t x, Char_t *buf)
{
# if defined(__linux) && defined(__i386__) && \
     defined(__GNUC__) && __GNUC__ >= 2
   *((UInt_t *)buf) = Rbswap_32(*((UInt_t *)&x));
# elif defined(R__KCC)
   // Use an union to prevent over-zealous optimization by KCC
   // related to aliasing float.
   // + Use a volatile here to work around error in KCC optimizer
   union {
     volatile char    c[4];
     volatile Float_t f;
   } u;
   u.f = x;
   buf[0] = u.c[3];
   buf[1] = u.c[2];
   buf[2] = u.c[1];
   buf[3] = u.c[0];
# else
   char *sw = (char *)&x;
   buf[0] = sw[3];
   buf[1] = sw[2];
   buf[2] = sw[1];
   buf[3] = sw[0];
# endif
}

inline void Memory::Swap(Double_t x, Char_t *buf)
{
# if defined(__EXTENSIONS__) && defined(__linux) && defined(__i386__) && \
     defined(__GNUC__) && __GNUC__ >= 2
   *((ULong64_t *)buf) = Rbswap_64(*((ULong64_t *)&x));
# elif defined(R__KCC)
   // Use an union to prevent over-zealous optimization by KCC
   // related to aliasing double.
   // + Use a volatile here to work around error in KCC optimizer
   union {
     volatile char     c[8];
     volatile Double_t d;
   } u;
   u.d = x;
   buf[0] = u.c[7];
   buf[1] = u.c[6];
   buf[2] = u.c[5];
   buf[3] = u.c[4];
   buf[4] = u.c[3];
   buf[5] = u.c[2];
   buf[6] = u.c[1];
   buf[7] = u.c[0];
# else
   char *sw = (char *)&x;
   buf[0] = sw[7];
   buf[1] = sw[6];
   buf[2] = sw[5];
   buf[3] = sw[4];
   buf[4] = sw[3];
   buf[5] = sw[2];
   buf[6] = sw[1];
   buf[7] = sw[0];
# endif
}

inline void Memory::Swap(Char_t *buf, UShort_t &x)
{
# if defined(__linux) && defined(__i386__)
   x = Rbswap_16(*((UShort_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[1];
   sw[1] = buf[0];
# endif
}
inline void Memory::Swap(Char_t *buf, Short_t &x)
{
# if defined(__linux) && defined(__i386__)
   x = Rbswap_16(*((Short_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[1];
   sw[1] = buf[0];
# endif
}

inline void Memory::Swap(Char_t *buf, UInt_t &x)
{
# if defined(__linux) && defined(__i386__)
   x = Rbswap_32(*((UInt_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[3];
   sw[1] = buf[2];
   sw[2] = buf[1];
   sw[3] = buf[0];
# endif
}

inline void Memory::Swap(Char_t *buf, Int_t &x)
{
# if defined(__linux) && defined(__i386__)
   x = Rbswap_32(*((Int_t *)buf));
# else
   char *sw = (char *)&x;
   sw[0] = buf[3];
   sw[1] = buf[2];
   sw[2] = buf[1];
   sw[3] = buf[0];
# endif
}
inline void Memory::Swap(Char_t *buf, Float_t &x)
{
# if defined(__linux) && defined(__i386__) && \
     defined(__GNUC__) && __GNUC__ >= 2
   // Use a union to allow strict-aliasing
   union {
      volatile UInt_t  i;
      volatile Float_t f;
   } u;
   u.i = Rbswap_32(*((UInt_t *)buf));
   x = u.f;
# elif defined(R__KCC)
   // Use an union to prevent over-zealous optimization by KCC
   // related to aliasing float.
   // + Use a volatile here to work around error in KCC optimizer
   union {
      volatile char    c[4];
      volatile Float_t f;
   } u;
   u.c[0] = buf[3];
   u.c[1] = buf[2];
   u.c[2] = buf[1];
   u.c[3] = buf[0];
   x = u.f;
# else
   char *sw = (char *)&x;
   sw[0] = buf[3];
   sw[1] = buf[2];
   sw[2] = buf[1];
   sw[3] = buf[0];
# endif
}

inline void Memory::Swap(Char_t *buf, Double_t &x)
{
# if defined(__EXTENSIONS__) && defined(__linux) && defined(__i386__) && \
     defined(__GNUC__) && __GNUC__ >= 2
   // Use a union to allow strict-aliasing
   union {
      volatile ULong64_t l;
      volatile Double_t  d;
   } u;
   u.l = Rbswap_64(*((ULong64_t *)buf));
   x = u.d;
# elif defined(R__KCC)
   // Use an union to prevent over-zealous optimization by KCC
   // related to aliasing double.
   // + Use a volatile here to work around error in KCC optimizer
   union {
      volatile char     c[8];
      volatile Double_t d;
   } u;
   u.c[0] = buf[7];
   u.c[1] = buf[6];
   u.c[2] = buf[5];
   u.c[3] = buf[4];
   u.c[4] = buf[3];
   u.c[5] = buf[2];
   u.c[6] = buf[1];
   u.c[7] = buf[0];
   x = u.d;
# else
   char *sw = (char *)&x;
   sw[0] = buf[7];
   sw[1] = buf[6];
   sw[2] = buf[5];
   sw[3] = buf[4];
   sw[4] = buf[3];
   sw[5] = buf[2];
   sw[6] = buf[1];
   sw[7] = buf[0];
# endif
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

} // namespace gws



#endif





